
#include "DominoDevice.h"

#include <QDebug>
#include <QString>
#include <QVector>

#include "OneWire/OneWireArrayModule.h"
#include "WaveBlc/WaveBlcModule.h"
#include "base/Globals.h"
#include "cal/CalResult.h"
#include "device-modules/AdcSpiBaseModule.h"
#include "DeserializersControl/DeserializersControlModule.h"
#include "TrigInfo/TrigInfoModule.h"
#include "TrigCtrlStat/TrigCtrlStatModule.h"

namespace {
const int SEND_STATUS_TIMEOUT = 100; // in ms
} // namespace

DominoDevice::DominoDevice(const DeviceIndex &index, QObject *parent)
    : ModularMregDevice(index, status, parent)
    , status(index)
    , oneWireArrayModule(createModule<OneWireArrayModule>())
{
    // register types this class emits
    qRegisterMetaType<AdcStatus>();
}

void DominoDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();

    dropConfigCache();

    if (config_key.isValid()) {
        writeConfigWithPause();
    }
}

void DominoDevice::on_deviceDisconnected()
{
    ModularMregDevice::on_deviceDisconnected();
    dropConfigCache();
}

bool DominoDevice::readStatus()
{
    return
        ModularMregDevice::readStatus() &&
        readAdc64Status();
}

bool DominoDevice::readAdc64Status()
{
    if (!isConnected())
        return false;

    if (oneWireArrayModule) {
        auto bmStatus = oneWireArrayModule->readStatus();
        status.bmStatus = bmStatus;
    }

    /*
    if(!settingsCacheValid)
        writeSettings();
*/
    if (!lastStatusTimestamp.isValid() || lastStatusTimestamp.elapsed() > SEND_STATUS_TIMEOUT) {
        lastStatusTimestamp.start();
        // status.deviceAttr.fpgaTemps.insert(status.getBMTemps()); // TODO

        if (isConfigCacheValid())
            status.setup_key = config_key;
        else
            status.setup_key.reset();
        emit adcStatusUpdated(status);
    }

    return true;
}

bool DominoDevice::writeConfig()
{
    bool ok = true;
    if (!config.isRun()) {
        ok &= writeSoftRun(false);
    }
    ok &= ModularMregDevice::writeConfig();

    if(Globals::getDebugMode()) {
        if(!isConfigCacheValid() || config.adcSystemCore != configCache.adcSystemCore) {
            auto spiMod = getAdcSpiModule();
            if(spiMod) {
                ok &= spiMod->setCustomPattern(config.adcSystemCore.pattern);
                ok &= spiMod->setCustomPatternMode(config.adcSystemCore.usePattern);
            }
        }
    }

    if (config.isRun()) {
        ok &= writeSoftRun(true);
    }
    if (!ok) {
        qWarning().noquote()
            << QString("[%1] Error writing config").arg(getIdent());
    }
    return ok;
}

bool DominoDevice::writeConfigWithPause()
{
    if (!isConnected())
        return {};

    writeSoftRun(false);
    syncConfigDeferred();
//    clearStatistics();

    return true;
}

std::optional<CalResult> DominoDevice::performCalibration()
{
    if (!isConnected())
        return {};
    auto mod = getFirstModule<WaveBlcModule>();
    if(!mod)
        return {};

    QElapsedTimer t;
    t.start();
    auto st = mod->readStatus();
    if(!st)
        return {};

    CalResult res;
    res.index = getDeviceIndex();
    res.nch = st->channelCount;

    auto ok = mod->readBaseline(&res.baseline_x1, 1000);
    if (!ok)
        return {};

    qInfo() << "calibrateBaseline:" << t.elapsed() << "ms";
    res.temps = readTemp();

    return res;
}

void DominoDevice::startCalibration(const ConfigKey &cal_key)
{
    const auto &calResult = performCalibration();
    if (!calResult.has_value()) {
        qWarning() << "Calibration failed";
        emit calibrateFinished(getDeviceIndex(), ConfigKey());
        return;
    }
    auto res = calResult.value();
    emit calResultChanged(res);
    emit calibrateFinished(getDeviceIndex(),
                           isConnected() ? cal_key : ConfigKey());
}

void DominoDevice::preRunReset(const QVariant &startEvNum, const RcRunId &)
{
    preRunActions(startEvNum);
}

void DominoDevice::enforceConfigConstraints(ModularDeviceConfig *config)
{
    config->tqdcCore.makeWaveRecConfig(&config->waveRec, getDeviceId());
    config->tqdcCore.makeWaveTrigConfig(&config->waveTrig);


    auto &adcStaticCompensation = config->adcMpdTrig.adcStaticCompensation;
    adcStaticCompensation.clear();
    const auto &cal = config->calResult.baseline_x1;
    for (auto it = cal.begin(); it != cal.end(); ++it) {
        auto ch = it.key();
        auto channel = config->waveBlc.getChannel(ch);
        // channel.staticOffset = -it->mean;
        // config->waveBlc.setChannel(ch, channel);
        if(adcStaticCompensation.size()<=it.key())
            adcStaticCompensation.resize(it.key()+1);
        adcStaticCompensation[it.key()] = -it->mean;
    }
}
