/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2014-2020 Ilia Slepnev
**    Copyright 2014-2015 Nikita Tarasov
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DreMainWindow.h"
#include "ui_DreMainWindow.h"

#include <limits>
#include <math.h>

#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QSettings>
#include <QStatusBar>
#include <QTimer>

#include "base/DeviceId.h"
#include "device-discover/DiscoverDialog.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"
#include "RegOpModel.h"
#include "sdb/SdbTypes.h"

#include "DreConfigConverter.h"

namespace {
enum { COL_ID, COL_SERIAL, COL_STATE, COL_FIRMWARE, COL_TEMP, COL_NUM };
}

DreMainWindow::DreMainWindow(QWidget *parent)
    : BaseMainWindow(parent)
    , ui(new Ui::DreMainWindow)
    , deviceDispatcher(new DeviceDispatcher(this))
    , discoverDialog(new DiscoverDialog(this))
    , regOpModel(new RegOpModel(this))
{
    qRegisterMetaType<DeviceIndex>();
    AbstractConfigConverter<DreConfig>::registerTypes();

    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, nullptr);
    connect(ui->pushButtonChange, &QPushButton::clicked, discoverDialog, &DiscoverDialog::show);
    connect(ui->pushButtonAdd, &QAbstractButton::clicked, this, &DreMainWindow::on_pushButtonAdd_clicked, Qt::UniqueConnection);
    connect(ui->pushButtonRemove, &QAbstractButton::clicked, this, &DreMainWindow::on_pushButtonRemove_clicked, Qt::UniqueConnection);

    ui->pushButtonChange->setShortcut(QKeySequence::Find);

    ui->mainToolBar->hide();
    ui->regOpTableView->setModel(regOpModel);
    connect(regOpModel, &RegOpModel::configChanged,
            this, &DreMainWindow::setConfig);
    connect(this, &DreMainWindow::regResultsChanged,
            regOpModel, &RegOpModel::setRegResults);


    config.loadSettings();
    regOpModel->setConfig(config);

    ui->comboBoxRegType->addItem("16-bit", DreConfig::REG_TYPE_16);
    ui->comboBoxRegType->addItem("32-bit", DreConfig::REG_TYPE_32);
    ui->comboBoxRegType->addItem("64-bit", DreConfig::REG_TYPE_64);

    ui->comboBoxRegAddr->addItem("MLink::CTRL 16-bit", 0x40);
    ui->comboBoxRegAddr->addItem("MLink::Temp 16-bit", 0x54);
    ui->comboBoxRegAddr->addItem("MLink::Temp2 16-bit", 0x4b);
    ui->comboBoxRegAddr->addItem("MLink::DevId 16-bit", 0x42);
    ui->comboBoxRegAddr->addItem("MLink::OneWireID 64-bit", 0x48);

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.subscribe(deviceDispatcher);
    lldpDiscovery.subscribe(discoverDialog);

    deviceDispatcher->subscribeDevices(this);
    deviceDispatcher->subscribeDevices(discoverDialog);
    // deviceDispatcher->subscribeHostAddressUpdates(dev);

    DeviceIndexEnabledMap map;
    if (!config.deviceIndex.isNull()) {
        map[config.deviceIndex] = DeviceIndexEnabled(config.deviceIndex, true);
        emit deviceListChanged(map);
    }
    restoreGui();
    onDeviceStateChange();
}

DreMainWindow::~DreMainWindow()
{
    saveConfig();
    delete ui;
}

void DreMainWindow::on_comboBoxRegAddr_currentIndexChanged(int)
{
    ui->radioButtonPreDefReg->setChecked(true);
}

void DreMainWindow::on_lineEditRegAddr_textEdited(const QString)
{
    ui->radioButtonCustomReg->setChecked(true);
}

void DreMainWindow::on_comboBoxRegType_currentIndexChanged(int)
{
    updateLineEditWriteLen();
}

void DreMainWindow::on_pushButtonRead_clicked()
{
    if (!saveConfig(false)) {
        qWarning() << "Can't read. Some field has incorrect value";
        return;
    }

    if (!dev || !dev->isConnected()) {
        if (dev)
            qWarning() << "Can't read. Device disconnected";
        else
            qWarning() << "Can't read. Device undefined";
        return;
    }

    RegOpVector r;
    bool ok = false;
    quint64 readResult;
    switch (config.regType) {
    case DreConfig::REG_TYPE_16:
        quint16 val16;
        r.RegRead16(config.regAddr, &val16);
        if ((ok = dev->regOpExec(r)))
            readResult = val16;
        break;
    case DreConfig::REG_TYPE_32:
        quint32 val32;
        r.RegRead32(config.regAddr, &val32);
        if ((ok = dev->regOpExec(r)))
            readResult = val32;
        break;
    case DreConfig::REG_TYPE_64:
        quint64 val64;
        r.RegRead64(config.regAddr, &val64);
        if ((ok = dev->regOpExec(r)))
            readResult = val64;
        break;
    }
    if (ok) {
        QString readValStr;
        readValStr = QString("%L1\n0x%2")
                         .arg(readResult)
                         .arg(readResult, config.regType*4, 16, QChar('0'));
        ui->labelRegVal->setText(readValStr);
    } else {
        ui->labelRegVal->clear();
    }
    ui->lineEditResult->setText(
        QString("Read [%1] %2").arg(config.regAddr, 8, 16, QChar('0')).arg(ok ? "Ok" : "Error"));
}

void DreMainWindow::on_pushButtonWrite_clicked()
{
    if (!saveConfig()) {
        qWarning() << "Can't read. Some field has incorrect value";
        return;
    }

    if (!dev || !dev->isConnected()) {
        return;
    }
    RegOpVector r;
    bool ok = false;
    switch (config.regType) {
    case DreConfig::REG_TYPE_16:
        r.RegWrite16(config.regAddr, static_cast<quint16>(config.regWriteVal));
        ok = dev->regOpExec(r);
        break;
    case DreConfig::REG_TYPE_32:
        r.RegWrite32(config.regAddr, static_cast<quint32>(config.regWriteVal));
        ok = dev->regOpExec(r);
        break;
    case DreConfig::REG_TYPE_64:
        r.RegWrite64(config.regAddr, config.regWriteVal);
        ok = dev->regOpExec(r);
        break;
    }
    ui->lineEditResult->setText(
        QString("Write [%1] %2").arg(config.regAddr, 8, 16, QChar('0')).arg(ok ? "Ok" : "Error"));
}

void DreMainWindow::deviceConnected()
{
    onDeviceStateChange();
    updateSdbComboBox();
}

void DreMainWindow::deviceDisconnected()
{
    onDeviceStateChange();
}

void DreMainWindow::setConfig(const DreConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;
    saveConfig();
}

void DreMainWindow::onDeviceStateChange()
{
    bool connected = dev && dev->isConnected();
    ui->pushButtonRead->setEnabled(connected);
    ui->pushButtonWrite->setEnabled(connected);
    if (!connected) {
        ui->labelRegVal->clear();
        ui->lineEditResult->clear();
    }
    if (connected) {
        auto message = QString("%1\nConnected to %2")
                           .arg(dev ? dev->getIdent() + ", f/w "+ dev->getFirmwareVersionStr() : config.deviceIndex.getIdent())
                           .arg(dev ? dev->getAddressStr() : "");
        ui->labelDev->setText(message);
    } else {
        if (config.deviceIndex.isNull())
            ui->labelDev->clear();
        else {
            auto message = QString("%1\nNot connected")
                               .arg(dev ? dev->getIdent() : config.deviceIndex.getIdent());
            ui->labelDev->setText(message);
        }
    }
}

void DreMainWindow::updateSdbComboBox()
{
    if(!dev || !dev->isConnected()) {
        return;
    }
    sdb_t sdb = dev->getSdb();
    QMap<QString, QPair<quint32, quint32>> sdbMap;
    for(auto sdbId : sdb.getDeviceIds()) {
        int seq = 0;
        auto devices = sdb.filterDevices({sdbId});
        for(auto d : devices) {
            QString name = d.component.product.name;
            if(devices.size() > 1)
                name += QString(" #%1").arg(seq);

            sdbMap[name] = {d.component.product.device_id, seq++};
        }
    }

    ui->regOpTableView->setSdbMap(sdbMap);
}

QtMregDevice *DreMainWindow::createDevice(const DeviceIndex &index)
{
    auto device = new QtMregDevice(index, this);
    deviceDispatcher->subscribeHostAddressUpdates(device->getDeviceIndex(), device);
    device->setLiveMagicCheck(false);

    connect(device, &QtMregDevice::deviceConnected, this, &DreMainWindow::deviceConnected);
    connect(device, &QtMregDevice::deviceDisconnected, this, &DreMainWindow::deviceDisconnected);

    return device;
}

void DreMainWindow::addDevice(const DeviceIndex &index)
{
    dev = QSharedPointer<QtMregDevice>(createDevice(index));
    config.deviceIndex = index;
    config.saveSettings();
    onDeviceStateChange();
}

void DreMainWindow::removeDevice(const DeviceIndex &index)
{
    dev = nullptr;
    config.deviceIndex = DeviceIndex();
    onDeviceStateChange();
}

DeviceIndexSet getDeviceIndexSetFromMap(const DeviceIndexEnabledMap &devices)
{
    DeviceIndexSet set;
    for (auto i : devices)
        if (i.enabled)
            set.insert(i.index);
    return set;
}

void DreMainWindow::setDeviceList(const DeviceIndexEnabledMap &devices)
{
    static DeviceIndexEnabledMap deviceListCache;
    if (devices == deviceListCache)
        return;
    const auto old_set = getDeviceIndexSetFromMap(deviceListCache);
    const auto new_set = getDeviceIndexSetFromMap(devices);
    deviceListCache = devices;
    const auto toRemove = old_set - new_set;
    const auto toAdd = new_set - old_set;
    for (auto i : toRemove)
        removeDevice(i);
    for (auto i : toAdd)
        addDevice(i);
}

bool DreMainWindow::saveConfig(bool checkWriteVal)
{
    bool addrOk, valOk;
    config.isCustom = ui->radioButtonCustomReg->isChecked();

    if (config.isCustom) {
        int regAddr = ui->lineEditRegAddr->text().toInt(&addrOk);
        if (!addrOk) {
            regAddr = ui->lineEditRegAddr->text().toInt(&addrOk, 16);
        }

        if (addrOk) {
            addrOk = regAddr >= 0 && regAddr < std::numeric_limits<quint16>::max();
            config.regAddr = addrOk ? static_cast<quint16>(regAddr) : 0;
        }

        if (!addrOk)
            qWarning() << "Bad custom reg address:" << ui->lineEditRegAddr->text();
    } else {
        config.regAddr = static_cast<quint16>(ui->comboBoxRegAddr->currentData().toUInt(&addrOk));
        addrOk &= config.regAddr != 0;
    }

    config.regType = static_cast<DreConfig::REG_TYPE>(ui->comboBoxRegType->currentData().toInt());

    quint64 val = ui->lineEditRegVal->text().toULongLong(&valOk);
    if (valOk) {
        config.regWriteVal = val;
    } else {
        val = ui->lineEditRegVal->text().toULongLong(&valOk, 16);
        if (valOk) {
            config.regWriteVal = val;
            ui->lineEditRegVal->setText(QString("0x%1").arg(config.regWriteVal, 0, 16));
        } else if (checkWriteVal)
            qWarning() << "Bad writting value:" << ui->lineEditRegVal->text();
    }

    config.saveSettings();
    return addrOk & (valOk || !checkWriteVal);
}

void DreMainWindow::restoreGui()
{
    (config.isCustom ? ui->radioButtonCustomReg : ui->radioButtonPreDefReg)->setChecked(true);

    ui->lineEditRegAddr->setText(QString("0x%1").arg(config.regAddr, 0, 16));
    ui->lineEditRegVal->setText(QString("0x%1").arg(config.regWriteVal, 0, 16));
    auto index = ui->comboBoxRegType->findData(config.regType);
    ui->comboBoxRegType->setCurrentIndex(index);
    updateLineEditWriteLen();
}

void DreMainWindow::updateLineEditWriteLen()
{
    bool ok;
    int len = ui->comboBoxRegType->currentData().toInt(&ok);
    //    ui->lineEditRegVal->setMaxLength(2+len*4);
}

void DreMainWindow::on_pushButtonAdd_clicked()
{
    config.regOpActions.push_back({});
    regOpModel->setConfig(config);
    config.saveSettings();
}

void DreMainWindow::on_pushButtonRemove_clicked()
{
    auto i = ui->regOpTableView->currentIndex();
    if(!i.isValid())
        return;
    auto r = i.row();
    config.regOpActions.removeAt(r);
    regOpModel->setConfig(config);
    config.saveSettings();
}

void DreMainWindow::on_pushButtonExec_clicked()
{
    if (!dev || !dev->isConnected()) {
        return;
    }

    QVector<quint64> results;
    sdb_t sdb = dev->getSdb();
    for(auto a : config.regOpActions) {
        quint64 res = 0;
        mlink::RegOpVector r;
        bool ok = false;
        quint16 addr = a.regOffset;
        if(a.sdbId)
            addr += sdb.getSdbBaseAddr(static_cast<AFI_SDB_ID>(a.sdbId), a.sdbSeq);
        if(a.isWrite) {
            auto writeVal = a.getWriteVal();
            if(!writeVal) {
                qWarning() << __PRETTY_FUNCTION__ << "Invalid write val: " << a.writeVal;
                results.push_back(0);
                continue;
            }

            switch (a.regType) {
            case REG_OP_ACTION_16:
                r.RegWrite16(addr, *writeVal, &ok);
                break;
            case REG_OP_ACTION_32:
                r.RegWrite32(addr, *writeVal, &ok);
                break;
            case REG_OP_ACTION_64:
                r.RegWrite64(addr, *writeVal, &ok);
                break;
            }
        } else {
            switch (a.regType) {
            case REG_OP_ACTION_16:
                r.RegRead16(addr, reinterpret_cast<quint16*>(&res), &ok);
                break;
            case REG_OP_ACTION_32:
                r.RegRead32(addr, reinterpret_cast<quint32*>(&res), &ok);
                break;
            case REG_OP_ACTION_64:
                r.RegRead64(addr, &res, &ok);
                break;
            }
        }

        if (!dev->regOpExec(r) || !ok) {
            qWarning() << __PRETTY_FUNCTION__ << "Err on " << results.size();
            break;
        }
        results.push_back(res);
    }
    emit regResultsChanged(results);
}
