#include "GenServer.h"

#include <QTcpServer>
#include <QtNetwork>
#include <unistd.h>
#include <stdexcept>

#include "base/DeviceId.h"
#include "base/Globals.h"

//const int TCP_SERVER_BUFFER_SIZE = 10000;
// EventSize(s)=40+4*4*s; EventSize(200)=3240; 10Mb for 3236 Events
namespace {
const int TCP_SERVER_BUFFER_SIZE = 10*1024*1024;
const int SOCKET_PENDING_DATA_LIMIT = TCP_SERVER_BUFFER_SIZE;
const int SEND_TIMEOUT_MS = 10;
}

GenServer::GenServer(int _id, const QString &programIndexBase, int port)
    : QObject(nullptr)
    , id(_id)
    , fsm(new GenServerFsm(id, this))
    , genProgIndexStr(QString("%1_%2").arg(programIndexBase).arg(id))
    , tcpPort(port)
    , readPos(TCP_SERVER_BUFFER_SIZE-1)
    , qTcpServer(new QTcpServer(this))
    , freeSpace(TCP_SERVER_BUFFER_SIZE)
    , sendTimer(new QTimer(this))
    , workTimer(new QTimer(this))
{
    status.id = id;

    progDescr.setClientIndex({Globals::getDaqModule(), ProgramIndex::fromString(genProgIndexStr)});
    connect(workTimer, &QTimer::timeout, this, &GenServer::work);
    connect(fsm, &GenServerFsm::stateChanged, this, &GenServer::sendFsmState);
    connect(fsm, &GenServerFsm::stateChanged, this, &GenServer::onFsmStateChange);
    workTimer->setSingleShot(true);

    mpdHdr.sync = MPD_SYNC_NORM_TLV;
    device_hdr.deviceId = DEVICE_ID_SOFTWARE;
    device_hdr.deviceSerial = id;

    QTimer::singleShot(0, this, &GenServer::init);
}

GenServer::~GenServer()
{
    emit closeProgram(progDescr);

    qint64 bytes = getDataSize();
    if (tcpSocket)
        bytes += tcpSocket->bytesToWrite();
    if (bytes)
        qWarning() << genProgIndexStr + QString("TcpServer not empty and destroyed, %1 bytes lost").arg(bytes);
}

void GenServer::init()
{
    while(!qTcpServer->listen(QHostAddress::Any, tcpPort)) {
        QAbstractSocket::SocketError errNum = qTcpServer->serverError();
        if(errNum==QAbstractSocket::AddressInUseError){
            qWarning() << genProgIndexStr + tr("Unable to start server on port %1: %2, exiting...").arg(tcpPort).arg(qTcpServer->errorString());
            // sleep(5);
            exit(-1);
            continue;
        }
        QString s = tr("Unable to start server on port %1: %2").arg(tcpPort).arg(qTcpServer->errorString());
        qWarning() << genProgIndexStr + s;
        throw std::runtime_error(s.toStdString());
    }
    tcpPort = qTcpServer->serverPort();
    status.tcpPort = tcpPort;
    connect(sendTimer, &QTimer::timeout, this, &GenServer::on_sendTimer_timeout);
    sendTimer->setSingleShot(true);
    sendTimer->setInterval(SEND_TIMEOUT_MS);
    tcpData.resize(TCP_SERVER_BUFFER_SIZE);

    connect(qTcpServer, &QTcpServer::newConnection, this, &GenServer::updateConnections);
    qInfo() << genProgIndexStr + QString("TcpServer started on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
    emit updateProgramDescription(progDescr);
    // QTimer::singleShot(0, this, &GenServer::updateStatus);
    // QTimer::singleShot(0, this, &GenServer::updateConnectionState);
    updateStatus();
    updateConnectionState();
    sendFsmState();
}

void GenServer::startGen(const RcRunId &rcId, const GenServerConfig &cfg)
{
    if(rcRunId.isValid()) {
        qCritical() << "Got" << __FUNCTION__ << " while previous run didn't complemed";
        return;
    }
    rcRunId = rcId;
    runStartTime = QDateTime::currentDateTime().addMSecs(cfg.preRunWait);

    config = cfg;
    adoptPayloadDataToConfig();
    upExpectedEvSize();

    fsm->setAllDone(false);
    fsm->setRequestState(FsmState::Run);
    requestFinish = false;

    status.workLoop = 0;
    workTimer->start();
    qInfo() << __PRETTY_FUNCTION__ << id << rcRunId;
}

void GenServer::handleDisconnect()
{
    if (!m_connectionName.isEmpty())
        qInfo() << genProgIndexStr + QString("Closed connection from %1").arg(m_connectionName);
    deleteSocket();
    rcRunId = {};

    updateStatus();
}

void GenServer::checkPendingData()
{
    if(!requestFinish)
        return;
    if((getDataSize()==0) && tcpSocket && (tcpSocket->bytesToWrite()==0)) {
        fsm->setAllDone(true);
        disconnect(tcpSocket, &QTcpSocket::bytesWritten,
                   this, &GenServer::checkPendingData);
    }
}

void GenServer::updateStatus()
{
    status.total_events = evNum;
    emit statusUpdated(status);
}

bool GenServer::isConnected() const
{
    return tcpSocket && tcpSocket->isValid() && tcpSocket->state() == QTcpSocket::ConnectedState;
}

void GenServer::deleteSocket()
{
    tcpSocket->deleteLater();
    tcpSocket = nullptr;
    status.peerHost.clear();
    status.peerPort = 0;
    status.isFree = isFree();
    status.numberBytesSent = 0;
    status.numberBytesDropped = 0;
    m_connectionName.clear();

    updateConnectionState();
}

bool GenServer::isEnoughSpace(int size) const
{
    int freeWriteSize = getFreeSize();
    if(freeWriteSize < size)
        return false;

    double freeSizeRate = 1.*freeWriteSize/TCP_SERVER_BUFFER_SIZE;
    if(freeSizeRate<0.5){
        // Enable porosity
        double chance = 1.*qrand()/RAND_MAX;
        if(chance>freeSizeRate)
            //Skipp this event
            return false;
    }
    return true;
}

double GenServer::getOccupancy() const
{
    return 1.*getDataSize()/TCP_SERVER_BUFFER_SIZE;
}

void GenServer::deleteGen(int i)
{
    if(id == i)
        deleteLater();
}

QString GenServer::connectionName() const
{
    Q_ASSERT(tcpSocket);
    return connectionName(*tcpSocket);
}

QString GenServer::connectionName(const QTcpSocket &socket) const
{
    return QString("%1:%2").arg(socket.peerAddress().toString()).arg(socket.peerPort());
}

void GenServer::on_sendTimer_timeout()
{
    sendData();
    if (tcpSocket)
        sendTimer->start(SEND_TIMEOUT_MS);
}

void GenServer::sendData()
{
    //wait data available
    int sentSize = 0;
    int totalSize = getDataSize();
    if(totalSize==0){
        writerWC.wakeAll();
        return;
    }
    //Check connetion
    if (!tcpSocket) {
        return;
    }

    if(tcpSocket->bytesToWrite()>SOCKET_PENDING_DATA_LIMIT){
        if(!slowConnection){
            qDebug() << genProgIndexStr + QString("Connection with %1 is slow.").arg(connectionName());
            slowConnection = true;
        }
        return;
    }
    if(slowConnection){
        qDebug() << genProgIndexStr + QString("Connection with %1 recovered.").arg(connectionName());
        slowConnection = false;
    }
    if(isConnected()) {
        int tailSize = TCP_SERVER_BUFFER_SIZE-readPos-1;
        if(tailSize && tailSize < totalSize){
            sentSize = tcpSocket->write(tcpData.constData()+readPos+1, tailSize);
            Q_ASSERT(sentSize != -1);
            if(sentSize > 0){
                readPos += tailSize; // TCP_SERVER_BUFFER_SIZE-1
           } else {
                qWarning() << genProgIndexStr + "Can't sent data to socket:"  << tcpSocket->errorString();
                return;
            }
        }
        if(readPos == TCP_SERVER_BUFFER_SIZE-1)
            readPos = -1;

        tcpSocket->write(tcpData.constData()+readPos+1, totalSize-sentSize);
        readPos+=totalSize-sentSize;
        sentSize+=totalSize-sentSize;
        writerWC.wakeAll();
        status.numberBytesSent += sentSize;
        emit bytesSent(status.numberBytesSent);
//        qDebug()<<QTime::currentTime().toString("hh:mm:ss.zzz")<<" Data sent "<< sentSize;
    } else {
        deleteSocket();
        updateStatus();
        sendTimer->stop();
    }

    if(requestFinish)
        checkPendingData();
}

void GenServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()){
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)) {
            tcpSocket->abort();
            deleteSocket();
        }
        if(tcpSocket==nullptr) {
            tcpSocket = pendingConnection;
            status.peerHost = tcpSocket->peerAddress();
            status.peerPort = tcpSocket->peerPort();
            status.isFree = isFree();
            if(requestFinish)
                connect(tcpSocket, &QTcpSocket::bytesWritten,
                        this, &GenServer::checkPendingData, Qt::UniqueConnection);
            connect(tcpSocket, SIGNAL(disconnected()), SLOT(handleDisconnect()));
            m_connectionName = connectionName();
            qInfo() << genProgIndexStr + QString("Accepted connection from %1").arg(m_connectionName);
            sendTimer->start(0);
            updateConnectionState();
        } else {
            qInfo() << genProgIndexStr + QString("Refused connection from %1").arg(connectionName(*pendingConnection));
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
//    if(tcpSocket==nullptr){
//        qDebug()<<"No connections";
//    } else {
//        qDebug() << QString("Current connection: %1").arg(connectionName());
//    }
}


bool GenServer::addData(const char *event, int len)
{
//    qDebug() << QTime::currentTime().toString("hh:mm:ss.zzz") << QString("addEvent: %1 bytes").arg(len);
    if(dropDataOnFreeSocket && tcpSocket==nullptr){
        status.numberBytesDropped += len;
        return true;
    }

    int writtenSize=0;
    Q_ASSERT(len < TCP_SERVER_BUFFER_SIZE); //todo complex partial writing
    int freeWriteSize = getFreeSize();

    while(freeWriteSize<len){
        //wait client read
        qWarning() << genProgIndexStr + "Buffer is full, waiting client";
        writerMutex.lock();
        writerWC.wait(&writerMutex);
        freeWriteSize = getFreeSize();
        writerMutex.unlock();
    }
//    if(freeSpace.tryAcquire(len)){
//        qWarning()<<"Buffer is full, waiting client";
//        freeSpace.acquire(len);
//    }

    if(writePos+len > TCP_SERVER_BUFFER_SIZE){
        int tailSize = TCP_SERVER_BUFFER_SIZE-writePos;
        memcpy(tcpData.data()+writePos, event, tailSize);
        writtenSize+=tailSize;
        writePos+=tailSize;
        if(writePos==TCP_SERVER_BUFFER_SIZE)
            writePos=0;
    }
    memcpy(tcpData.data()+writePos, event+writtenSize, len-writtenSize);
    writePos += len-writtenSize;
    writtenSize+=len-writtenSize;
    if(writePos==TCP_SERVER_BUFFER_SIZE)
        writePos=0;
    return true;
}

void GenServer::work()
{
    switch (fsm->getState()) {
    case FsmState::Run:
    case FsmState::Stopping:
        //send events
        break;
    case FsmState::Ready:
        // break work loop
        return;
    default:
        workTimer->start(100);
        return;
    }
    ++status.workLoop;

    // qInfo() << __PRETTY_FUNCTION__
    //         << "id:" << id
    //         << "loop:" << status.workLoop
    //         << "total_events:" << status.total_events;

    auto evToSend = getNumEvToSend();

    bool someEvWasSent = false;
    auto expectedFreeSize = getFreeSize();
    auto initialFreeSize = expectedFreeSize;
    QList<quint32> evSizes;
    while((evNum < evToSend) && (expectedFreeSize >= expectedEvSize)) {
        evSizes << expectedEvSize;
        auto lastFreeSize = getFreeSize();
        expectedFreeSize -= expectedEvSize;
        sendEvent();
        someEvWasSent = true;
    }

    updateStatus();
    quint32 runEvToSend = config.spill_count*config.getEvCntPerSpill();
    if(evNum == runEvToSend) {
        if(someEvWasSent) {
            qInfo() << __PRETTY_FUNCTION__ << id << "all was sent";
        }
    } else {
        workTimer->start(100);
    }
}

void GenServer::requestWorkCompleted()
{
    fsm->setRequestState(FsmState::Stopping);
    requestFinish = true;
    if(tcpSocket)
        connect(tcpSocket, &QTcpSocket::bytesWritten,
                this, &GenServer::checkPendingData, Qt::UniqueConnection);

    checkPendingData();
}

void GenServer::setEvbGenServerConfig(const GenServerConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;

    adoptPayloadDataToConfig();
    upExpectedEvSize();
}

void GenServer::onFsmStateChange(const FsmState &state)
{
    switch (state) {
    case FsmState::Stopping:
        fsm->setRequestState(FsmState::Ready);
        break;
    case FsmState::Ready:
        rcRunId = {};
        evNum = 0;
        status.total_events = 0;
        status.total_size = 0;
        workTimer->stop();
        break;
    default:;
    }
}

int GenServer::getFreeSize() const
{
//    return freeSpace.available();
    //If buffer fully written we use all-1 bytes and positions are equal
    //If buffer empty positions differs on 1
    if(readPos == writePos)
        return 0;

    int size=0;
    if(readPos < writePos)
        size = TCP_SERVER_BUFFER_SIZE-writePos+readPos;
    else
        size = readPos-writePos;

    return size;
}

int GenServer::getDataSize() const
{
//    return TCP_SERVER_BUFFER_SIZE-getFreeSize();
    //If buffer fully written we use all-1 bytes and positions are equal
    if(readPos==writePos)
        return TCP_SERVER_BUFFER_SIZE;

    int size=0;
    if(readPos<writePos)
        size = writePos-readPos-1;
    else
        size = (TCP_SERVER_BUFFER_SIZE-readPos-1)+writePos;

    return size;
}

void GenServer::updateConnectionState()
{
    bool conn = isConnected();
    fsm->setConnected(conn);
    bool changed = progDescr.updateProgramInterface(
        ProgramInterfaceDataFlow, 0,
        status.tcpPort, true, !conn,
        status.peerHost, status.peerPort);
    // if(changed)
    {
        emit updateProgramDescription(progDescr);
        updateStatus();
    }
}

void GenServer::sendFsmState()
{
    emit fsmStateChanged(id, fsm->getState());
}

void GenServer::sendEvent()
{
    ++evNum;
    ++status.total_events;

    quint32 len = sizeof(mpdHdr);
    len += sizeof(evNum);
    len += sizeof(MpdDeviceHdr);
    len += device_hdr.length;
    addData((char *)&mpdHdr, sizeof(mpdHdr));
    addData((char *)&evNum, sizeof(evNum));
    addData((char *)&device_hdr, sizeof(MpdDeviceHdr));
    if(device_hdr.length)
        addData(eventPayloadData, device_hdr.length);

    Q_ASSERT(len == expectedEvSize);

    status.total_size += sizeof(MpdHdr) + mpdHdr.length;
    if(config.useMax)
        upExpectedEvSize();
}

void GenServer::adoptPayloadDataToConfig()
{
    eventPayloadData.resize(config.useMax ? config.eventLengthMax : config.event_length);
}

quint32 GenServer::upExpectedEvSize()
{
    const constexpr static quint32 baseSize = sizeof(MpdHdr) + sizeof(quint32) + sizeof(MpdDeviceHdr);
    if(!config.useMax || config.event_length == config.eventLengthMax) {
        expectedEvSize = baseSize + config.event_length;
    } else {
        quint32 width =  config.eventLengthMax - config.event_length;
        expectedEvSize = baseSize + config.event_length + std::rand() % width;
    }
    updateLenInHdr();
    return expectedEvSize;
}

void GenServer::updateLenInHdr()
{
    const constexpr static quint32 baseSize = sizeof(MpdHdr) + sizeof(quint32) + sizeof(MpdDeviceHdr);
    device_hdr.length = expectedEvSize-baseSize;
    mpdHdr.length = sizeof(evNum) + sizeof(device_hdr) + device_hdr.length;
}

quint32 GenServer::getNumEvToSend() const
{
    auto curTime = QDateTime::currentDateTime();
    auto dt = runStartTime.msecsTo(curTime);
    // int spillNum = std::ceil(dt/(config.spillMs + config.interspillMs));
    // spillNum = std::min(spillNum, config.spill_count);
    // quint32 evToSend = spillNum*config.spill_events;
    int fullSpillNum = std::floor(dt/(config.spillMs + config.interspillMs));
    fullSpillNum = std::min(fullSpillNum, config.spill_count);
    auto evPerSpill = config.getEvCntPerSpill();
    quint32 evToSend = fullSpillNum*evPerSpill;
    if(fullSpillNum <config.spill_count) {
        dt -= fullSpillNum*(config.spillMs + config.interspillMs);
        if(dt>0) {
            evToSend += 1.*evPerSpill*dt/config.spillMs;
        }
    }
    return evToSend;
}
























