#ifndef GENSERVER_H
#define GENSERVER_H

#include <QObject>
#include <QByteArray>
#include <QHostAddress>
#include <QVector>
#include <QMutex>
#include <QWaitCondition>
#include <QSemaphore>
#include <QDateTime>

#include "base/RcRunId.h"
#include "pnp-server/PNPMessage.h"
#include "lib-common/MpdRawTypes.h"
#include "fsm/FsmState.h"

#include "GenServerConfig.h"
#include "GenServerStatus.h"
#include "GenServerFsm.h"

class QTcpServer;
class QTcpSocket;
class QTimer;

class GenServer : public QObject
{
    Q_OBJECT

public:
    GenServer(int id, const QString &programIndexBase,
              int port = defaultPort());
    virtual ~GenServer();
    bool isEnoughSpace(int size) const; // threadsafe
    bool isFree() const { return tcpSocket==nullptr; } // deprecated
    void setDropDataOnFreeSocket(bool v) { dropDataOnFreeSocket = v; }
    double getOccupancy() const;

signals:
    void bytesSent(quint64 value) const;
    void statusUpdated(const GenServerStatus &) const;
    void closeProgram(const ProgramDescription &) const;
    void updateProgramDescription(const ProgramDescription &) const;
    void fsmStateChanged(int id, const FsmState &cur) const;

public slots:
    void deleteGen(int i);
    void init();
    void startGen(const RcRunId &rcId, const GenServerConfig &cfg);
    bool addData(const char *event, int len);
    void requestWorkCompleted();
    void setEvbGenServerConfig(const GenServerConfig &newConfig);

private slots:
    void onFsmStateChange(const FsmState &state);
    void work();

    void sendData();
    void updateConnections();
    void on_sendTimer_timeout();
    void deleteSocket();
    void handleDisconnect();
    void checkPendingData();
    void updateStatus();
    void updateConnectionState();
    void sendFsmState();

private:
    bool isConnected() const;
    QString connectionName() const;
    QString connectionName(const QTcpSocket &socket) const;
    static int defaultPort() { return 0; } // 33400
    int getFreeSize() const;  // threadsafe
    int getDataSize() const;  // threadsafe

private:
    void sendEvent();
    void adoptPayloadDataToConfig();
    quint32 upExpectedEvSize();
    void updateLenInHdr();
    quint32 getNumEvToSend() const;

    const int id=-1;
    GenServerFsm * const fsm;
    QString genProgIndexStr;
    std::uint16_t tcpPort;
    std::atomic_int writePos = 0; // data position, where we start writing
    std::atomic_int readPos; // last transfered data position
    QTcpServer * const qTcpServer;
    QByteArray tcpData;
    QTcpSocket *tcpSocket = nullptr;
    QWaitCondition writerWC;
    QSemaphore freeSpace;
    QMutex writerMutex;
//    QMutex readerMutex;
    QTimer * const sendTimer;
    bool slowConnection = false;
    QString m_connectionName;
    bool dropDataOnFreeSocket = false;
    bool requestFinish = false;

    GenServerStatus status;

    MpdHdr mpdHdr;
    quint32 evNum;
    MpdDeviceHdr device_hdr;
    QByteArray eventPayloadData;
    quint32 expectedEvSize = 0;

    QDateTime workTime;
    QTimer * const workTimer;

    ProgramDescription progDescr;
    RcRunId rcRunId;
    GenServerConfig config;
    QDateTime runStartTime;
};

#endif // GENSERVER_H
