#include "EvbClient.h"
#include "params.h"
#include "Analizer.h"
#include "RingBuffer.h"
#include "TcpClient.h"
#include "base/threadQuit.h"
#include "base/RcRunId.h"

#include <QCoreApplication>
#include <QElapsedTimer>
#include <QThread>
#include <QRegularExpression>

EvbClient::EvbClient(const EvBClientConfig &cc,
                             int evNumIndex,
                             QObject *parent) :
    EvbClientFsm(parent),
    index(cc.index),
    enabled(cc.enabled),
    dataRingBuff(new RingBuffer<char>(0)),
    evRingBuff(new RingBuffer<EventMetaInfo>(EVENTS_IN_BUFER)),
    tcpClientThread(new QThread(this)),
    analizerThread(new QThread(this)),
    tcpClient(new TcpClient(index, evNumIndex, dataRingBuff, cc.bufSize)),
    analizer(new Analizer(index, dataRingBuff, evRingBuff))
{
    auto shortIndexStr = QString("%1#%2")
                             .arg(index.getProgramType().toString().remove(QRegularExpression{"[a-z]"}))
                             .arg(index.getProgramIndex().toString());
    setObjectName(__func__+(":"+shortIndexStr));


    tcpClientThread->setObjectName(QString("TCl_%1").arg(shortIndexStr));
    analizerThread->setObjectName(QString("A_%1").arg(shortIndexStr));
    tcpClient->moveToThread(tcpClientThread);
    tcpClientThread->start();

    analizerThread->start();
    analizer->moveToThread(analizerThread);

    connect(this, &BaseFsm::stateChanged, this, &EvbClient::onFsmStateChanged);
    connect(this, &EvbClient::readOutEnChanged, tcpClient, &TcpClient::setReadOutEn);

    connect(this, &EvbClient::requestState, analizer, &Analizer::setRequestState);
    connect(this, &EvbClient::requestState, tcpClient, &TcpClient::setRequestState);

    connect(analizer, &Analizer::stateChanged, this, &EvbClientFsm::setAnalizerState);
    connect(tcpClient, &BaseFsm::stateChanged, this, &EvbClientFsm::setTcpState);

    connect(this, &EvbClient::destroyingClient, analizer, &Analizer::deleteLater);
    connect(this, &EvbClient::destroyingClient, tcpClient, &TcpClient::deleteLater);

    connect(analizer, &Analizer::destroyed, analizerThread, &QThread::quit);
    connect(tcpClient, &TcpClient::destroyed, tcpClientThread, &QThread::quit);

    connect(analizer, &Analizer::statUpdated, this, &EvbClient::statUpdated);
    connect(analizer, &Analizer::evSizePeriodUpdated, this, &EvbClient::evSizePeriodUpdated);

    connect(this, &EvbClient::reqPrintFsmInput, this, &BaseFsm::printFsmInput);
    connect(this, &EvbClient::reqPrintFsmInput, tcpClient, &BaseFsm::printFsmInput);
    connect(this, &EvbClient::reqPrintFsmInput, analizer, &BaseFsm::printFsmInput);
}

EvbClient::~EvbClient()
{
    emit destroyingClient();
    QElapsedTimer timer;
    timer.start();

    if(!threadQuiting(analizerThread, 1000, &timer))
        qWarning() << objectName() << "Analizer's thread isn't stopped";

    if(!threadQuiting(tcpClientThread, 1000, &timer))
        qWarning() << objectName() << "TcpClient's thread isn't stopped";

    delete dataRingBuff;
    delete evRingBuff;
}

void EvbClient::destroyClient(const ClientIndex &i)
{
    if(!i.isEmpty() && index!=i)
        return;
    deleteLater();
}

void EvbClient::updateClientReadout(const ClientIndex &i, bool readout)
{
    if(index != i)
        return;
    enabled = readout;
    emit readOutEnChanged(enabled);
    if(getState()!=FsmState::Error && getState()!=FsmState::Ready)
        setState(FsmState::Error);
}

void EvbClient::setClientRequestState(const FsmState &st)
{
    if(!enabled)
        return;
    setRequestState(st);
}

void EvbClient::onFsmStateChanged(const FsmState &state)
{
    emit clientStateUpdated(index, state);
    if(state == FsmState::Error)
        doReqStateActions(FsmState::Error);
}

void EvbClient::doReqStateActions(const FsmState &st)
{
    if(performedStateAction == st)
        return;
    performedStateAction = st;

    qDebug() << objectName() << __PRETTY_FUNCTION__ << st;
    emit requestState(st);
}
