﻿//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "EvbMainWindow.h"
#include "ui_EvbMainWindow.h"

#include "AdjustBufferDialog.h"
#include "AdjustBufferModel.h"
#include "EvbClient.h"
#include "Combiner.h"
#include "EvBConfigDispatcher.h"
#include "EvbFileNameDialog.h"
#include "EvbProgramDescriptionHelper.h"
#include "EvBStartupOptions.h"
#include "EvbSignalRouter.h"
#include "FileSplitDialog.h"
#include "FileWriterStats.h"
#include "StatisticDialog.h"
#include "SizeHelper.h"
#include "EvbClientManagerFsm.h"

#include "add-pnp-dialog/AddPnpDialog.h"
#include "app/AppVersion.h"
#include "base/AbstractConfigConverter.h"
#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "base/Globals.h"
#include "base/ThreadController.h"
#include "base/threadQuit.h"
#include "base/RunDocument.h"
#include "base/RunDocumentConverter.h"
#include "configuration-manager/ConfigurationManager.h"
#include "pnp-server/PnpDiscovery.h"
#include "widgets/LogViewWidget.h"
#include "metric-sender/MetricSender.h"
#include "metric-sender/MetricTypes.h"
#include "remote-control-server/RemoteControlServer.h"
#include "TelnetRemoteControl/TelnetRemoteControl.h"

#include <QActionGroup>
#include <QCheckBox>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QJsonDocument>
#include <QJsonObject>
#include <QHostAddress>
#include <QInputDialog>
#include <QMessageBox>
#include <QMetaType>
#include <QThread>
#include <QTimer>
#include <QSemaphore>
#include <QSettings>
#include <QScreen>
#include <time.h>
#include <QTime>
#include <iostream>

namespace {
QString getTcpName(int tcpId) {
    return tcpId ? QString("Tcp #%1").arg(tcpId-1) : QString("Mon");
}
const DaqModuleSet EVB_SEARCH_TARGETS {
    DaqModule(DaqModule::Type::EventBuilder),
    DaqModule(DaqModule::Type::Generator),
    DaqModule(DaqModule::Type::GenericClient),
    DaqModule(DaqModule::Type::MStream),
    DaqModule(DaqModule::Type::VmeDaq),
    DaqModule(DaqModule::Type::Land)
};

static constexpr const char* prop_outType = "outType";
static constexpr const char* prop_id = "id";

void updateFsmLabel(QLabel *label, const QString &prefix, FsmState st)
{
    label->setText(prefix + ": " + fsmStateToString(st));
    auto p = label->palette();
    auto b = p.brush(label->foregroundRole());
    switch (st) {
    case FsmState::Error:
        b.setColor(Qt::red);
        break;
    case FsmState::Ready:
        b.setColor(QColor());
        break;
    case FsmState::Run:
        b.setColor(Qt::darkGreen);
        break;
    default:
        b.setColor(Qt::darkYellow);
    }
    p.setBrush(label->foregroundRole(), b);
    label->setPalette(p);
}
}

using namespace std;

Q_DECLARE_METATYPE(Qt::GlobalColor)

EvbMainWindow::EvbMainWindow(const EvBStartupOptions &so, QWidget *parent)
    : BaseMainWindow(parent)
    , ui(new Ui::EvbMainWindow)
    , configuration_manager (new ConfigurationManager(this))
    , configDb(new EvBConfigDb(this))
    , configDispatcher(new EvBConfigDispatcher(this))
    , evb_fsm(new EventBuilderFsm(this))
    , client_manager_fsm(new EvbClientManagerFsm(this))
    , combiner(new Combiner())
    , combinerThread(new QThread(this))
    , remoteControl(new RemoteControlServer(this, 0))
    , labelFileErrSize(new QLabel(this))
    , statisticDialog(new StatisticDialog(this))
    , labelEvbFsmState(new QLabel(this))
    , labelCombinerFsmState(new QLabel(this))
    , labelClientManagerFsmState(new QLabel(this))
    , labelCombinerStatus(new QLabel(this))
    , fileSplitDialog(new FileSplitDialog(this))
    , fileNameDialog(new EvbFileNameDialog(this))
    , adjustBufferDialog(new AdjustBufferDialog(this))
    , adjustBufferModel(new AdjustBufferModel(this))
    , applyCacheTimer(new QTimer(this))
    , telnetRc(so.telnetPort ? new TelnetRemoteControl(*so.telnetPort) : nullptr)
{
    registerTypes();

    ui->setupUi(this);
    ui->mainToolBar->hide();
    labelEvbFsmState->setMinimumWidth(120);
    labelCombinerFsmState->setMinimumWidth(150);
    labelClientManagerFsmState->setMinimumWidth(180);
    ui->statusBar->addWidget(labelEvbFsmState);
    ui->statusBar->addWidget(labelCombinerFsmState);
    ui->statusBar->addWidget(labelClientManagerFsmState);
    ui->statusBar->addWidget(labelCombinerStatus);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);

    ui->splitter->addWidget(new LogViewWidget(ui->splitter));
    adjustBufferDialog->setModel(adjustBufferModel);

    auto evbSignalRouter = &EvbSignalRouter::instance();
    evbSignalRouter->setConfigurationManager(configuration_manager);
    evbSignalRouter->setEvBConfigDb(configDb);
    evbSignalRouter->setEvBConfigDispatcher(configDispatcher);
    evbSignalRouter->setMainWindow(this);
    evbSignalRouter->setRemoteControlServer(remoteControl);
    evbSignalRouter->setWidgetRunControl(ui->widgetRunControl);
    evbSignalRouter->setEvbFsm(evb_fsm);
    evbSignalRouter->setClientManagerFsm(client_manager_fsm);
    evbSignalRouter->setAdjustBufferModel(adjustBufferModel);
    evbSignalRouter->setStatisticDialog(statisticDialog);
    if(telnetRc) {
        evbSignalRouter->setTelnetRemoteControl(telnetRc);
        new ThreadController(telnetRc, this);
    }

    evbSignalRouter->connectAll();

    QThread::currentThread()->setObjectName("GUI thread");
    combinerThread->setObjectName("combinerThread");

    QActionGroup *dataOutActionGroup = new QActionGroup(this);
    dataOutActionGroup->addAction(ui->actionOutToFile);
    dataOutActionGroup->addAction(ui->actionOutToTcp);
    dataOutActionGroup->addAction(ui->actionCombinerIdle);
    connect(dataOutActionGroup, &QActionGroup::triggered,
            this, &EvbMainWindow::data_out_changed);
    connect(ui->radioButtonOutFile, &QRadioButton::toggled, this, &EvbMainWindow::dataOutChangedRB);
    connect(ui->radioButtonOutIdle, &QRadioButton::toggled, this, &EvbMainWindow::dataOutChangedRB);
    connect(ui->radioButtonOutTcp, &QRadioButton::toggled, this, &EvbMainWindow::dataOutChangedRB);

    ui->radioButtonOutFile->setProperty(prop_outType, COMBINER_FILE);
    ui->radioButtonOutTcp->setProperty(prop_outType, COMBINER_TCP);
    ui->radioButtonOutIdle->setProperty(prop_outType, COMBINER_IDLE);

    connect(ui->actionOutput_file_name, &QAction::triggered,
            fileNameDialog, &EvbFileNameDialog::show);

    new ThreadController(remoteControl, this);

    configDispatcher->subscribeCS(this);
    configDispatcher->subscribeCS(configDb);
    configDispatcher->subscribeCS(configuration_manager);
    configDispatcher->subscribeEvbConfig(this);
    configDispatcher->subscribeEvbConfig(configDb);
    configDispatcher->subscribeClientList(this);
    configDispatcher->subscribeFileSplitterConfig(fileSplitDialog);
    configDispatcher->subscribeWriterConfig(fileNameDialog);
    configDispatcher->subscribeFileSplitterConfig(combiner);
    configDispatcher->subscribeClientList(client_manager_fsm);
    configDispatcher->subscribeClientList(ui->tableView);

    configDispatcher->subscribeEvbConfig(adjustBufferDialog);
    configDispatcher->subscribeEvbConfig(adjustBufferModel);
    configDispatcher->subscribeWriterConfig(this);
    configDispatcher->subscribeEvNumberConfig(this);

    configuration_manager->setDefaultConfig(EvBConfig());

    applyCacheTimer->setSingleShot(true);
    applyCacheTimer->setInterval(500);
    connect(applyCacheTimer, &QTimer::timeout, this, &EvbMainWindow::applyCache);
    applyCacheTimer->start();

    init();
}

void EvbMainWindow::init()
{
    ui->actionEnable_inf_output->setVisible(Globals::getDebugMode());
    ui->statusBar->addWidget(labelFileErrSize);
//    statusBar()->addPermanentWidget(labelFileErrSize);
    statusBar()->setToolTip("reg:\tNumber of regular events\n"
                            "skip:\tNumber of skipped events from clients\n"
                            "eob:\tNumber of 'end of burst' events\n"
                            "clients:\tClient's ids from last combined event\n"
                            "nextEv:\tEvent number of next expected event\n"
                            "monitor:\tNumber of sent events to monitor");
    labelFileErrSize->hide();

    initCombiner();
    updateOutTcpLabelList();
    setup_clients();
    updateGui();
}

void EvbMainWindow::setConfig(const EvBConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;

    applyConfig();
}

void EvbMainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void EvbMainWindow::applyCache()
{
    updatePendingProgramDescription();
    applyCacheTimer->start();
}

void EvbMainWindow::doReqStateAction(const FsmState &reqSt)
{
    if(requestedState == reqSt)
        return;
    requestedState = reqSt;

    qDebug() << __PRETTY_FUNCTION__ << reqSt;

    QVariant requestArg;
    switch (reqSt) {
    case FsmState::Run:
        requestArg = QVariant::fromValue(loc_runId);
        break;
    case FsmState::PrepareStopping:
        requestArg = QVariant::fromValue(stopEvNum.value_or(0));
        break;
    default:;
    }

    switch (reqSt) {
    case FsmState::Configure:
    case FsmState::Run:
    case FsmState::PrepareStopping:
    case FsmState::Ready:
        emit requestState(reqSt, requestArg);
        if(reqSt == FsmState::Run){
            sendRunDocument();
        }
        break;
    default:;
    }
}

void EvbMainWindow::onEvbFsmStateChanged(const FsmState &cur, const FsmState &prev)
{
    updateFsmLabel(labelEvbFsmState, "EvbFsm", cur);
    switch (cur) {
    case FsmState::Error:
    {
        if(prev != FsmState::Init && prev != FsmState::Ready) {
            emit printFsmInput();
            qDebug() << __PRETTY_FUNCTION__ << "send stopping signal on Error state";
            doReqStateAction(FsmState::Ready);
        }
        break;
    }
    default:;
    }

    updateGuiEnableStates();
}

void EvbMainWindow::updateCombinerFsmLabel(const FsmState &cur_state)
{
    updateFsmLabel(labelCombinerFsmState, "CombinerFsm", cur_state);
}

void EvbMainWindow::updateClientFsmLabel(const FsmState &cur_state)
{
    updateFsmLabel(labelClientManagerFsmState, "ClientManagerFsm", cur_state);
}

void EvbMainWindow::startLocalRun()
{
    if(getFsmState() == FsmState::Ready)
        evb_fsm->setRequestState(FsmState::Starting);
}

void EvbMainWindow::stopLocalRun()
{
    if(getFsmState() == FsmState::Run) {
        stopEvNum.reset();
        evb_fsm->setLookForStopEvNum(stopEvNum.has_value());
        evb_fsm->setRequestState(FsmState::Stopping);
    }
}

void EvbMainWindow::configSelectorChanged(const ConfigIndexName& conf)
{
    BaseMainWindow::configSelectorChanged(conf);

    // Initialize MetricSender
    MetricSender::setBaseTag("index", conf.getIndex().toString());

    initProgramDescription(conf);
}

EvbMainWindow::~EvbMainWindow()
{
    QElapsedTimer timer;
    timer.start();

    auto allClients = clients.keys();
    ClientIndexSet toDel{allClients.cbegin(), allClients.cend()};
    deleteClients(toDel);

    disconnect(remoteControl, &RemoteControlServer::startRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stopRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::connectionStatusUpdated, nullptr, nullptr);

    emit closeProgram(progDesc);

//    Q_ASSERT(ret);

    emit destroyAll();
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"destroyClient(0);";
    //todo: wait all clients are destroyed

    timer.start();

    //Give time to proceed destroyed signal from Combiner
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"threadQuiting(combinerThread";
    if(!threadQuiting(combinerThread, 5000, &timer)){
        qWarning() << "Combiner's thread isn't stopped";
//        qInfo()<<"[MainWindow] "<<timer.elapsed()<<"Combiner's thread isn't stopped";
    }

    delete ui;
}

EvbClientStatusTableView * const EvbMainWindow::getStatusTable() const
{
    return ui->tableView;
}

void EvbMainWindow::on_pushButtonAdd_clicked()
{
    QVector<ProgramDescription> freeProgs = getFreeProgs();
    if(!freeProgs.size())
        return;

    AddPnpDialog dialog(progDescrsCache, true, this);
    PnpDiscovery::subscribeCacheUpdaed(&dialog);

    dialog.set_function_filter(std::bind(&EvbMainWindow::isFreeProg, this, std::placeholders::_1));

    auto clients_map = config.getClientIndexEnabledMap();
    if(dialog.exec() == QDialog::Accepted) {
        for(auto pd : dialog.getProgDescr()) {
            auto index = pd.getClientIndex();
            clients_map[index] = true;
        }
        emit clientListChanged(clients_map);
    }
}

void EvbMainWindow::on_pushButtonRemove_clicked()
{
    auto cis = ui->tableView->getSelectedClients();
    if(cis.isEmpty())
        return;

    QStringList list;
    for(auto c : cis)
        list << c.toString();
    QString message = QString("Are you shure to remove client(s):\n%1?")
                          .arg(list.join('\n'));

    int ret = QMessageBox::warning(this, tr("Confirm removing"), message,
                                   QMessageBox::Yes | QMessageBox::No,
                                   QMessageBox::No);
    if(ret == QMessageBox::No)
        return;

    auto clients_map = config.getClientIndexEnabledMap();
    for(auto index : cis) {
        if(index.isEmpty() || !clients_map.contains(index)) {
            qWarning() << "Don't know what client was selected; removing canceled";
            return;
        }

        clients_map.remove(index);
        qInfo() << QString("Client '%1' deleted").arg(index.toString());
    }
    emit clientListChanged(clients_map);
}

void EvbMainWindow::updateClientStatus(const TcpClientStatus &status)
{
    auto index = status.index;
    if(!config.clients.contains(index))
        return;
    if(status.state!=TcpClientState::Connected) {
        if(isInRun() && config.clients.contains(index) && config.clients[index].enabled){
            QString errStr = QString("%1 disconnected").arg(index.toString());
            qWarning() << errStr;
            // todo handle mstream restart period
            if(lastCombinerStats.hasPendingEvent && lastCombinerStats.regularEvents)
                deviceError(index, errStr);
        }
    }

    lastTcpClientStatus[status.index] = status;
}

void EvbMainWindow::setup_clients()
{
    ClientIndexSet toDel;
    for(ClientIndex index : clients.keys()) {
        if(!config.clients.contains(index))
            toDel << index;
    }
    deleteClients(toDel);

    for(ClientIndex index : config.clients.keys()) {
        auto cc = config.clients[index];
        if(!clients.contains(index)) {
            addClient(cc);
        } else {
            emit updateClientReadout(index, cc.enabled);
        }
    }

    updateAnalizerList();
}

void EvbMainWindow::deleteClients(const ClientIndexSet &set)
{
    for(ClientIndex index : set) {
        delete clients.take(index);
    }
}

bool EvbMainWindow::addClient(const EvBClientConfig &client_config)
{
    Q_ASSERT(getFsmState()==FsmState::Init ||
             getFsmState()==FsmState::Ready ||
             getFsmState()==FsmState::Error);

    ClientIndex index = client_config.index;
    if (index.isEmpty())
        return false;
    if (clients.contains(index))
        return false;

    auto client = new EvbClient(client_config, config.evNumCfg.getEvNumIndex(), this);
    EvbSignalRouter::instance().connect_Client(client);
    clients[index] = client;
    emit updateClientReadout(index, client_config.enabled);
    qDebug() << QString("Added client %1").arg(index.toString());
    return true;
}

void EvbMainWindow::updateGui() const
{
    ui->actionAdvanced_counters->setChecked(config.advancedCounters);
    ui->tableView->setShowAdvancedCounters(config.advancedCounters);
    ui->labelDebugInfo->setVisible(config.advancedCounters);

    ui->actionAllow_Remote_Control->setChecked(config.isRemoteEnabled);
    ui->actionEnable_client_time_distribution->setChecked(config.combinerCfg.timeDistrEn);
    ui->actionEnable_inf_output->setChecked(config.combinerCfg.infOutputEn);

    updateGuiCombinerOut();

    ui->actionFix_TLV_format->setChecked(config.combinerCfg.fixTlvFormat);
    ui->actionProtect_data_file_while_writing->setChecked(config.writerCfg.protectFile);
    ui->labelEvNumIndex->setVisible(config.evNumCfg.step != 1);
    ui->labelEvNumIndex->setText(QString("EvNum index: %1").arg(config.evNumCfg.getEvNumIndex()));
    updateSplitGui();
}

void EvbMainWindow::updateGuiCombinerOut() const
{
    ui->actionOutToFile->setChecked(config.isOutFile());
    ui->actionOutToTcp->setChecked(config.isOutTcp());
    ui->actionCombinerIdle->setChecked(config.isOutIdle());

    ui->radioButtonOutFile->setChecked(config.isOutFile());
    ui->radioButtonOutTcp->setChecked(config.isOutTcp());
    ui->radioButtonOutIdle->setChecked(config.isOutIdle());
}

void EvbMainWindow::updateSplitGui() const
{
    const auto &cfg = config.fileSplit;
    ui->groupBoxFileSplit->setVisible(config.isOutFile() && cfg.hasAnyEn());
    auto label = ui->labelSplitEv;
    label->setVisible(cfg.evNumEn);
    label->setText(QString("EvNum:\n%L1").arg(cfg.evNumFileSplit));
    label = ui->labelSplitSize;
    label->setVisible(cfg.fileSizeEn);
    label->setText(QString("File size: %1").arg(humanReadableSize(cfg.fileSizeSplit)));
    label = ui->labelSplitTime;
    label->setVisible(cfg.timeEn);
    label->setText(QString("Time: %L1s").arg(cfg.timeFileSplit));
}

void EvbMainWindow::updateGuiEnableStates() const
{
    bool hasSelectedClients = !ui->tableView->getSelectedClients().isEmpty();

    bool isStopped = (getFsmState()==FsmState::Ready) || (getFsmState()== FsmState::Error);
    bool rmEn = hasSelectedClients && isStopped;
    ui->pushButtonRemove->setEnabled(rmEn);
    updateButtonRemoveEnableState();
    ui->tableView->setControlEnabled(isStopped);

    updateButtonAddEnableStates();
    ui->actionSet_EvB_index->setEnabled(isStopped);
    ui->actionOutToFile->setEnabled(isStopped);
    ui->actionOutToTcp->setEnabled(isStopped);
    ui->groupBoxOutput->setEnabled(isStopped);
    ui->actionCombinerIdle->setEnabled(isStopped);
    ui->actionChange_storage_dir->setEnabled(isStopped);
    ui->actionReset_to_defaults->setEnabled(isStopped);
    ui->actionFix_TLV_format->setEnabled(isStopped);
}

void EvbMainWindow::updateButtonAddEnableStates() const
{
    bool isStopped = (getFsmState()==FsmState::Ready) || (getFsmState()== FsmState::Error);
    ui->pushButtonAdd->setEnabled(isStopped && hasFreeProgs());
}

void EvbMainWindow::updateAnalizerList()
{
    if(evb_fsm->getState() != FsmState::Error && evb_fsm->getState() != FsmState::Ready) {
        qCritical("Trying to '%s' while FSM not in Ready or Error state", __func__);
        evb_fsm->setRequestState(FsmState::Error);
    }
    QVector<Analizer *> analizers;
    ClientIndexSet clients_;
    nonEvBasedClients.clear();
    for(auto cfg : config.clients) {
        if(!cfg.enabled)
            continue;
        Q_ASSERT(clients.contains(cfg.index));
        if(clients.contains(cfg.index))
            analizers.push_back(clients[cfg.index]->getAnalizer());
        clients_.insert(cfg.index);
        if(isNonEventBaseClient(cfg.index))
            nonEvBasedClients << cfg.index;
    }

    enabledClientNum = static_cast<uint>(analizers.size());
    emit analizerListUpdated(clients_, nonEvBasedClients, analizers);
    refreshProgramDescription();
}

void EvbMainWindow::initProgramDescription(const ConfigIndexName& conf)
{
    progDesc.setClientIndex({DaqModule(DaqModule::Type::EventBuilder),
                             conf.getIndex()});

    progDesc.ver_date = getAppDateStr();
    progDesc.ver_hash = getAppVersion();
}

void EvbMainWindow::refreshProgramDescription()
{
    bool hasRcInterface = progDesc.containsProgramInterface(ProgramInterfaceRemoteControl, 0);
    ProgramInterface rcInterface;
    if(hasRcInterface)
        rcInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    progDesc.clear();
    if(hasRcInterface)
        progDesc.interfaces.push_back(rcInterface);

    for (int i=0; i<lastCombinerStats.tcpStatus.size(); ++i) {
        const OutputTcpStatus &tcpStatus = lastCombinerStats.tcpStatus[i];
        if(!tcpStatus.listening)
            continue;
        ProgramInterface tcpInterface;
        tcpInterface.setType(i==0 ? ProgramInterfaceMonitorData : ProgramInterfaceDataFlow);
        tcpInterface.enabled = true;
        tcpInterface.id = tcpStatus.index ? tcpStatus.index-1 : 0;
        tcpInterface.port = *tcpStatus.localPort;
        tcpInterface.isFree = !tcpStatus.connected;
        if(tcpStatus.connected) {
            tcpInterface.peers << ProgramInterfacePeer(tcpStatus.peer.host, tcpStatus.peer.port);
        }
        progDesc.interfaces.push_back(tcpInterface);
    }

    int clientCnt=0;
    bool nonEvBasedMode = true;
    for (auto cfg : config.clients) {
        if(cfg.enabled) {
            ++clientCnt;
            if(nonEvBasedMode)
                nonEvBasedMode = isNonEventBaseClient(cfg.index);
        }
    }
    if(clientCnt)
        progDesc.options["Clients"] = QString::number(clientCnt);
    if(nonEvBasedMode)
        markProgramAsNonEventBase(progDesc);

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif
    progDesc.options["fsm"] = fsmStateToString(getFsmState());
    progDesc.options["runIndex"] = loc_runId.runIndex;
    progDesc.options["runNumber"] = QString::number(loc_runId.runNumber);
    progDesc.options["output"] = config.getOutputStr();

    emit updateProgramDescription(progDesc);
}

void EvbMainWindow::updateTcpPortLabels()
{
    for (const auto &st : lastCombinerStats.tcpStatus) {
        QLabel *l = getTcpLabel(st.index);
        Q_ASSERT(l);

        QString text = QString("%1: ").arg(getTcpName(st.index));
        if(st.listening) {
            text += QString("%1").arg(*st.localPort);
            if(st.connected) {
                text += "->" + st.peer.toString();
            } else {
                text += " (free)";
            }
        } else {
            text += QString("disabled");
        }
        text += QString(" (%1%)").arg((int)(st.dataPart));
        l->setText(text);

        l->setStyleSheet(QString("QLabel { color : %1; }")
                             .arg(!st.connected ? (st.index ? "red" : "black") : "green"));
    }
}

void EvbMainWindow::setOutTcpServerNumber(int tcpNum)
{
    if(tcpNum < 1)
        tcpNum = 1;
    if(config.combinerCfg.combinerOutTcpNum != tcpNum){
        config.combinerCfg.combinerOutTcpNum = tcpNum;
        emit configChanged(config);
    }

    updateOutTcpLabelList();

    emit configChanged(config);
}

QVector<ProgramDescription> EvbMainWindow::getFreeProgs() const
{
    QVector<ProgramDescription> freeProgs;
    for (const ProgramDescription &descr : progDescrsCache) {
        if(isFreeProg(descr))
            freeProgs.push_back(descr);
    }
    return freeProgs;
}

bool EvbMainWindow::hasFreeProgs() const
{
    for (const ProgramDescription &descr : progDescrsCache) {
        if(isFreeProg(descr))
            return true;
    }
    return false;
}

bool EvbMainWindow::isFreeProg(const ProgramDescription &pd) const
{
    if(clients.contains(pd.getClientIndex()))
        return false;

    int interface_id = config.evNumCfg.getEvNumIndex();
    for (const ProgramInterface &i : pd.interfaces) {
#ifdef QT_DEBUG
        if((i.type == ProgramInterfaceDataFlow ||
             i.type == ProgramInterfaceMStream)
#else
        if(i.type == ProgramInterfaceDataFlow
#endif
            && i.enabled && i.isFree && i.id == interface_id) {
            return true;
        }
    }

    return false;
}

void EvbMainWindow::combinerOutUpdated()
{
    if(config.isOutTcp() && config.combinerCfg.combinerOutTcpNum==0){
        setOutTcpServerNumber(1);
        emit configChanged(config);
    }

    ui->actionSet_number_of_outputTcpServers->setEnabled(config.isOutTcp());
    ui->labelDataFileName->setVisible(config.isOutFile());
    ui->labelCombinerIdle->setVisible(config.isOutIdle());
    updateOutTcpLabelList();
    updateTcpLabelVisibility();
    emit configChanged(config);
}

QLabel *EvbMainWindow::getTcpLabel(int index) const
{
    bool ok;
    for(int i=0; i<ui->verticalLayoutOutTcp->count(); ++i) {
        QLayoutItem *item = ui->verticalLayoutOutTcp->itemAt(i);
        if(item) {
            QWidget *w = item->widget();
            if(w) {
                QLabel *l = qobject_cast<QLabel *>(w);
                if(l) {
                    if(l->property(prop_id).toInt(&ok)==index && ok)
                        return l;
                }
            }
        }
    }
    return nullptr;
}

void EvbMainWindow::updateTcpLabelVisibility() const
{
    for(int id=1; id<=config.combinerCfg.combinerOutTcpNum; ++id) {
        auto l = getTcpLabel(id);
        Q_ASSERT(l);
        if(l)
            l->setVisible(config.isOutTcp());
    }
}

void EvbMainWindow::registerTypes()
{
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<AnalyzerStat>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<ClientIndex>();
    qRegisterMetaType<ClientIndexEnabledMap>();
    qRegisterMetaType<ClientIndexSet>();
    qRegisterMetaType<EvbFileSplitterConfig>();
    qRegisterMetaType<EvbCombinerConfig>();
    qRegisterMetaType<EvbEvNumberConfig>();
    qRegisterMetaType<EventSizePeriod>();
    AbstractConfigConverter<ClientIndex>::registerTypes();
    AbstractConfigConverter<RunDocument>::registerTypes();
}

void EvbMainWindow::initCombiner()
{
    combiner->initOutputBuffers(config.combinerCfg);

    connect(combiner, &Combiner::destroyed, combinerThread, &QThread::quit);
    combiner->moveToThread(combinerThread);
    combinerThread->start();
}

void EvbMainWindow::applyConfig()
{
    updateGui();

    //TODO: implement check that all parts are corresponding to Config
    emit allowRemoteControl(config.isRemoteEnabled);
    setup_clients();
    update_combiner_to_config();
}

void EvbMainWindow::update_combiner_to_config()
{
    updateAnalizerList();
    combinerOutUpdated();
    setOutTcpServerNumber(config.combinerCfg.combinerOutTcpNum);

    emit evNumberConfigChanged(config.evNumCfg);
    if(config.monitorPort)
        emit setMonitorPortNum(config.monitorPort);
}

void EvbMainWindow::updateOutTcpLabelList()
{
    QLayout *layout = ui->verticalLayoutOutTcp;

    const auto &tcpNum = config.combinerCfg.combinerOutTcpNum;
    for(int index=0; index < layout->count(); ) {
        bool ok;
        QWidget *w = layout->itemAt(index)->widget();
        if(w->property(prop_id).toInt(&ok) > tcpNum && ok) {
            layout->removeWidget(w);
            delete w;
        } else {
            ++index;
        }
    }

    for(int tcpId=0; tcpId <= tcpNum; ++tcpId) {
        if(getTcpLabel(tcpId)) continue;
        QLabel *label = new QLabel(QString("%1: creating...").arg(getTcpName(tcpId)), ui->widgetOutTcp);
//        if(tcpId == 0) label->setVisible(false);
        label->setProperty(prop_id, tcpId);
        layout->addWidget(label);
        label->setAlignment(Qt::AlignHCenter);
        if(tcpId)
            label->setVisible(config.isOutTcp());
    }
}

void EvbMainWindow::checkTrashedEvents(const CombinerStats &cs,
                                       const CombinerStats &prevCs) const
{
    for(auto it=cs.clientsStat.cbegin(); it!=cs.clientsStat.cend(); ++it){
        auto clSt = *it;
        ClientIndex index = it.key();
        if(index.isEmpty())
            continue;
        if(clSt.trashedEvents) {
            if(prevCs.clientsStat[index].trashedEvents != clSt.trashedEvents) {
                static QElapsedTimer trashedEventsTimer;
                if(!trashedEventsTimer.isValid() || trashedEventsTimer.elapsed() > 1500) {
                    trashedEventsTimer.start();
                    auto mess = QString("Trashed events from %1: %L2")
                                    .arg(index.toString()).arg(cs.partedEvents);
                    emit sendRunWarning(loc_runId, mess);
                }
            }
        }
    }
}

void EvbMainWindow::updateCombinerStatusLabel(const CombinerStats &cs) const
{
    auto str = QString("reg:%1;skip:%2;eob:%3;nextEv:%5  monitor:%6")
                   .arg(cs.regularEvents)
                   .arg(cs.skipped)
                   .arg(cs.eob)
                   .arg(cs.expectedEvNum)
                   .arg(cs.monitorEvCount);

    labelCombinerStatus->setText(str);
}

void EvbMainWindow::updateFileErrLabel(const CombinerStats &cs) const
{
    labelFileErrSize->hide();
    labelFileErrSize->setText(QString("fileErr:%1B").arg(cs.missedFileSize));
    if(cs.missedFileSize){
        labelFileErrSize->show();
        labelFileErrSize->setStyleSheet("QLabel { color : red;}");
    }
    labelFileErrSize->setStyleSheet(QString("QLabel { color : %1;}")
                                        .arg(cs.missedFileSize?"red" : "black"));
}

void EvbMainWindow::sendCombinerMetrics(const CombinerStats &cs) const
{
    // metrics
    for(auto it=cs.clientsStat.cbegin(); it!=cs.clientsStat.cend(); ++it) {
        const QMap<QString, QString> tags({{"client", it.key().toString()}});
        MetricFields fields{{"drop", it->trashedEvents},
                            {"pass_event", it->eventBlocks},
                            {"pass_json", it->jsonBlocks},
                            {"pass_stat", it->statBlocks}};
        MetricSender::send("evb.in.subevents", fields, tags);
    }
}

void EvbMainWindow::checkBackPreasure(const CombinerStats &cs)
{
    if(backPreasure) {
        if(cs.getMaxOutputDataPart() < 50) {
            backPreasure = false;
            qInfo() << "BackPressure disabled";
            emit sendBackPreasure(loc_runId, false);
        }
    } else {
        if(cs.getMaxOutputDataPart() > 90) {
            backPreasure = true;
            QString mess = "Enable backPressure";
            mess += QString(" fwDataPart:%1%").arg(cs.fwDataPart,0,'f',1);
            for(auto i=cs.tcpStatus.cbegin();i!=cs.tcpStatus.cend();++i) {
                if(i.key())
                    mess += QString(" outTcpDataPart_%1:%2%")
                                .arg(i.key())
                                .arg(i->dataPart,0,'f',1);
            }
            qWarning() << mess;
            emit sendRunWarning(loc_runId, mess);
            emit sendBackPreasure(loc_runId, true);
        }
    }
}

void EvbMainWindow::combinerError(const QString &errStr)
{
    qCritical() << "Got error from combiner" << errStr;
    if(isInRun())
        emit sendRunError(loc_runId, errStr);
}

void EvbMainWindow::deviceError(const ClientIndex &index, const QString &errStr)
{
    QString msg = QString("Got error from %1").arg(index.toString());
    qCritical() << msg << errStr;
    if(isInRun())
        emit sendRunError(loc_runId, errStr);
}

void EvbMainWindow::setCombinerStats(const CombinerStats &cs)
{
    checkBackPreasure(cs);

    bool outDataInterfaceChanged = false;
    bool outDataPartsChanged = false;
    for (int id : cs.tcpStatus.keys()) {
        if(!lastCombinerStats.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
        const OutputTcpStatus &prevS = lastCombinerStats.tcpStatus[id];
        const OutputTcpStatus &nextS = cs.tcpStatus[id];
        outDataInterfaceChanged |= prevS.listening!=nextS.listening
                || prevS.connected!=nextS.connected
                || prevS.localPort!=nextS.localPort
                || prevS.peer.hostName!=nextS.peer.hostName;
        outDataPartsChanged |= !qFuzzyCompare(prevS.dataPart, nextS.dataPart);
    }
    for (int id : lastCombinerStats.tcpStatus.keys()) {
        if(!cs.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
    }
    if(cs.partedEvents && lastCombinerStats.partedEvents != cs.partedEvents) {
        static QElapsedTimer partedEventsTimer;
        if(!partedEventsTimer.isValid() || partedEventsTimer.elapsed() > 1500) {
            partedEventsTimer.start();
            auto mess = QString("Incompleted events: %L1").arg(cs.partedEvents);
            emit sendRunWarning(loc_runId, mess);
        }
    }
    const auto prevCs = lastCombinerStats; // todo: use lastCombinerStats instead
    ui->labelDebugInfo->updateDebugInfoLabel(lastTcpClientStatus, cs);
    lastCombinerStats = cs; // todo: move updating lastCombinerStats after table update
//    fsm->setExpectedEvNum(ss.expectedEvNum);

    if(outDataInterfaceChanged)
        refreshProgramDescription();
    if(outDataInterfaceChanged || outDataPartsChanged)
        updateTcpPortLabels();

    if(cs.clientsStat.isEmpty()) {
        return;
    }

    checkTrashedEvents(cs, prevCs);

    sendCombinerMetrics(cs);
    updateCombinerStatusLabel(cs);
    updateFileErrLabel(cs);
}

void EvbMainWindow::updateButtonRemoveEnableState() const
{
    bool isStopped = (getFsmState()==FsmState::Ready) || (getFsmState()== FsmState::Error);
    bool hasSelectedClients = !ui->tableView->getSelectedClients().isEmpty();
    bool rmEn = hasSelectedClients && isStopped;
    ui->pushButtonRemove->setEnabled(rmEn);
}

void EvbMainWindow::fileWriterStatsChanged(const FileWriterStats &fwStat)
{
    QString dataFileNameText;
    if(!fwStat.fileName.isEmpty()){
        QString suf;
        if(fwStat.writtenFileSize != fwStat.writtenTotalSize)
            suf = humanReadableSize(fwStat.writtenTotalSize).prepend('/');
        dataFileNameText = QString("%1 (%2%3) (%4%)")
                .arg(fwStat.fileName)
                .arg(humanReadableSize(fwStat.writtenFileSize))
                .arg(suf).arg(fwStat.dataPart*100, 0, 'f', 1);

        if(fwStat.missedFileSize)
            dataFileNameText +=QString(" (missed: %1)")
                    .arg(humanReadableSize(fwStat.missedFileSize));
    } else {
        dataFileNameText = QString(fwStat.predefStorage).append("...");
    }
    ui->labelDataFileName->setText(dataFileNameText);

    auto p = ui->labelDataFileName->palette();
    p.setColor(ui->labelDataFileName->foregroundRole(), fwStat.dataPart>0.8 ? Qt::darkYellow : Qt::black);
    ui->labelDataFileName->setPalette(p);
}

void EvbMainWindow::remoteStartRequest(const RcRunId &rcRunId, const QVariant &startEvNumber)
{
    auto msg = QString("Got remote request to start %1").arg(rcRunId.toString());
    if(startEvNumber.isValid()) {
        startEvNum = startEvNumber.toULongLong();
        msg += QString(" (StartEvNum=%L1)").arg(*startEvNum);
    } else {
        startEvNum.reset();
    }

    qInfo() << msg;
    emit configChanged(config);

    static bool warn_start_event_number = true;
    if(startEvNum && warn_start_event_number) {
        qWarning() << "Remote start with start_event_number not implemented yet."
                   << "startEvNumber:" << *startEvNum;
        warn_start_event_number = false;
    }

    if(loc_runId.runIndex == rcRunId.runIndex &&
            loc_runId.runNumber > rcRunId.runNumber) {
        const auto &msg = QString("Previous runNumber(%1) is greater that remote control runNumber(%2)")
                              .arg(loc_runId.runNumber)
                              .arg(rcRunId.runNumber);
        qWarning() << msg;
        emit sendRunWarning(loc_runId, msg);
        return;
    }

    if(getFsmState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Ready))
                              .arg(fsmStateToString(getFsmState()));
        qWarning() << msg;
        emit sendRunWarning(rcRunId, msg);
        return;
    }

    loc_runId = rcRunId;
    evb_fsm->setRequestState(FsmState::Starting);
}

void EvbMainWindow::remoteStopRequest(const RcRunId &rcRunId, const QVariant &runStopEvNum)
{
    auto msg = QString("Got remote request to stop %1").arg(rcRunId.toString());
    if(runStopEvNum.isValid()) {
        stopEvNum = runStopEvNum.toULongLong();
        msg += QString(" (StopEvNum=%L1)").arg(*stopEvNum);
    } else {
        stopEvNum.reset();
    }
    qInfo() << msg;

    if(loc_runId.runIndex != rcRunId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                     .arg(loc_runId.runIndex)
                     .arg(rcRunId.runIndex);
        return;
    }
    if(loc_runId.runNumber != rcRunId.runNumber) {
        qWarning()<< QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                     .arg(loc_runId.runNumber)
                     .arg(rcRunId.runNumber);
        return;
    }

    if(getFsmState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(getFsmState()));
        qWarning() << msg;
        emit sendRunWarning(loc_runId, msg);
        if(getFsmState() != FsmState::Ready && getFsmState() != FsmState::Error)
            evb_fsm->setRequestState(FsmState::Error);
        return;
    }

    evb_fsm->setLookForStopEvNum(stopEvNum.has_value());
    evb_fsm->setRequestState(FsmState::Stopping);
}

void EvbMainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &rcSerStat)
{
    bool pdChanged = false;
    if(!progDesc.containsProgramInterface(ProgramInterfaceRemoteControl, 0)){
        ProgramInterface remoteControlInterface;
        remoteControlInterface.setType(ProgramInterfaceRemoteControl);
        remoteControlInterface.id = 0;
        progDesc.interfaces.push_back(remoteControlInterface);
        pdChanged = true;
    }

    ProgramInterface &remoteInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    pdChanged |= remoteInterface.port != rcSerStat.tcpPort ||
            remoteInterface.enabled != rcSerStat.listening ||
            remoteInterface.isFree != !rcSerStat.portInUse;
    remoteInterface.port = rcSerStat.tcpPort;
    remoteInterface.enabled = rcSerStat.listening;
    remoteInterface.isFree = !rcSerStat.portInUse;

    if(rcSerStat.portInUse) {
        ProgramInterfacePeer p{rcSerStat.peerHost, rcSerStat.peerPort};
        if(!remoteInterface.peers.contains(p)) {
            remoteInterface.setPeer(p);
            pdChanged = true;
        }
        qInfo() << "Remote control connected.";
    } else {
        qWarning() << "Remote control disconnected.";
        loc_runId = RcRunId();

        if(getFsmState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            stopEvNum.reset();
            evb_fsm->setLookForStopEvNum(stopEvNum.has_value());
            evb_fsm->setRequestState(getFsmState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }

    if(ui->widgetRunControl->isEnabled() == rcSerStat.portInUse){
        ui->widgetRunControl->setEnabled(!rcSerStat.portInUse);
    }
    if(pdChanged)
        refreshProgramDescription();
}

void EvbMainWindow::on_actionChange_storage_dir_triggered()
{
//    QFileDialog::QFileDialog ( QWidget * parent = 0, const QString & caption = QString(), const QString & directory = QString(), const QString & filter = QString() )
//    QFileDialog dialog(this, "Set predefined storage for mpd data file", "", "*.data");
//    dialog.setOption(QFileDialog::ShowDirsOnly);
//    dialog.exec();
//    qDebug() << dialog.selectedFiles();

    auto &storageDir = config.writerCfg.storageDir;
    QString dir = QFileDialog::getExistingDirectory( this, "Choose storage directory...",
                                                     storageDir, QFileDialog::ShowDirsOnly );
    if(!dir.isEmpty() && storageDir != dir){
        storageDir = dir;
        emit writerConfigChanged(config.writerCfg);
    }
}

void EvbMainWindow::on_actionReset_to_defaults_triggered()
{
    emit configChanged({});
}

void EvbMainWindow::on_actionAdvanced_counters_toggled(bool showColumns)
{
    if(config.advancedCounters != showColumns) {
        config.advancedCounters = showColumns;
        emit configChanged(config);
    }

    ui->tableView->setShowAdvancedCounters(showColumns);
    ui->labelDebugInfo->setVisible(showColumns);
}

void EvbMainWindow::on_actionAllow_Remote_Control_toggled(bool checked)
{
    if(config.isRemoteEnabled != checked) {
        config.isRemoteEnabled = checked;
        emit configChanged(config);
    }
    emit allowRemoteControl(config.isRemoteEnabled);
}

void EvbMainWindow::data_out_changed(QAction *a)
{
    auto &out = config.combinerCfg.combinerOut;
    if(a == ui->actionCombinerIdle) out = COMBINER_IDLE;
    else if(a == ui->actionOutToTcp) out = COMBINER_TCP;
    else if(a == ui->actionOutToFile) out = COMBINER_FILE;
    else {
        qCritical() << "Data out was switch to unknown state.";
    }

    combinerOutUpdated();
    emit configChanged(config);
    refreshProgramDescription();
    updateGuiCombinerOut();
    updateSplitGui();
}

void EvbMainWindow::dataOutChangedRB()
{
    auto rb = qobject_cast<QRadioButton*>(sender());
    if(!rb->isChecked())
        return;

    auto out = rb->property(prop_outType).value<COMBINER_OUT>();
    if(out == config.combinerCfg.combinerOut)
        return;
    config.combinerCfg.combinerOut = static_cast<COMBINER_OUT>(out);

    combinerOutUpdated();
    emit configChanged(config);
    refreshProgramDescription();
    updateGuiCombinerOut();
    updateSplitGui();
}

void EvbMainWindow::on_actionSet_number_of_outputTcpServers_triggered()
{
    bool ok;
    const int tcpNum = QInputDialog::getInt(this, tr("TCP servers for output"),
                                 tr("Set number of TCP servers for output:"), config.combinerCfg.combinerOutTcpNum, 1, 20, 1, &ok);
    if (ok && config.combinerCfg.combinerOutTcpNum != tcpNum)
        setOutTcpServerNumber(tcpNum);
}

void EvbMainWindow::on_actionSet_EvB_index_triggered()
{
    bool ok;
    quint32 startNumber = QInputDialog::getInt(this, tr("Event builder index"),
                                 tr("Set 1st event number (EvB index):"), config.evNumCfg.startEvNum, 0, 19, 1, &ok);
    if(!ok)
        startNumber = config.evNumCfg.startEvNum;
    quint32 stepNumber = QInputDialog::getInt(this, tr("EvB event numbet step"),
                                 tr("Set event number step:"), config.evNumCfg.step, 1, 20, 1, &ok);
    if(!ok)
        startNumber = config.evNumCfg.step;

    if (config.evNumCfg.startEvNum != startNumber || config.evNumCfg.step != stepNumber) {
        int evbIndexOld = config.evNumCfg.getEvNumIndex();
        EvbEvNumberConfig cfg{startNumber, stepNumber};
        emit evNumberConfigChanged(cfg);
        if(evbIndexOld != cfg.getEvNumIndex()){
            QMessageBox msgBox;
            msgBox.setText("Event Builder index was changed, configuration was saved.\n"
                           "Needed manual restart\n"
                           "Closing...");
            msgBox.exec();
            close();
        }
    }
}

void EvbMainWindow::on_actionSet_ring_buffer_sizes_triggered()
{
    adjustBufferDialog->show();
}

void EvbMainWindow::on_actionSet_file_split_options_triggered()
{
    fileSplitDialog->show();
}

void EvbMainWindow::on_actionShow_statistics_triggered()
{
    statisticDialog->show();
}

void EvbMainWindow::on_actionEnable_client_time_distribution_toggled(bool en)
{
    if(config.combinerCfg.timeDistrEn != en) {
        config.combinerCfg.timeDistrEn = en;
        emit configChanged(config);
    }

    ui->tableView->setShowTimeDistrCol(en);
}

void EvbMainWindow::on_actionSet_monitor_port_number_triggered()
{
    bool ok;
    const int MIN_PORT_NUMBER = 0;
    const int MAX_PORT_NUMBER = 32999;
    const int portNum = QInputDialog::getInt(this, tr("Set monitor port number"),
                                             QString("Set monitor port number (in range %1-%2):")
                                             .arg(MIN_PORT_NUMBER).arg(MAX_PORT_NUMBER),
                                             config.monitorPort ? config.monitorPort : MAX_PORT_NUMBER, MIN_PORT_NUMBER, MAX_PORT_NUMBER, 1, &ok);
    if (ok && config.monitorPort != portNum) {
        config.monitorPort = portNum;
        emit configChanged(config);
        emit setMonitorPortNum(static_cast<quint16>(portNum));
//        QMessageBox msgBox;
//        msgBox.setText("Monitor port number was changed, configuration was saved.\n"
//                       "Needed manual restart\n"
//                       "Closing...");
//        msgBox.exec();
//        close();
    }
}

void EvbMainWindow::on_actionPrint_FSM_input_triggered()
{
    emit printFsmInput();
}

void EvbMainWindow::on_actionFix_TLV_format_toggled(bool en)
{
    if(config.combinerCfg.fixTlvFormat == en)
        return;
    config.combinerCfg.fixTlvFormat = en;
    emit configChanged(config);
}

void EvbMainWindow::on_actionProtect_data_file_while_writing_toggled(bool protect)
{
    if(config.writerCfg.protectFile == protect)
        return;
    config.writerCfg.protectFile = protect;
    emit writerConfigChanged(config.writerCfg);
    emit configChanged(config);
}

void EvbMainWindow::on_actionEnable_inf_output_toggled(bool en)
{
    if(config.combinerCfg.infOutputEn == en)
        return;
    config.combinerCfg.infOutputEn = en;
    emit configChanged(config);
}

void EvbMainWindow::pnpCacheUpdated(const PnpInfoCache &cache)
{
    progDescrsCache = cache;
    hasNewPnpCache = true;
}

bool EvbMainWindow::isNonEventBaseClient(const ClientIndex &index) const
{
    for(auto pd : progDescrsCache) {
        if(pd.getClientIndex() == index)
            return isNonEventBaseProgram(pd);
    }
    return false;
}

FsmState EvbMainWindow::getFsmState() const
{
    return evb_fsm->getState();
}

bool EvbMainWindow::isInRun() const
{
    return getFsmState() != FsmState::Ready && getFsmState() != FsmState::Error && getFsmState() != FsmState::Error;
}

void EvbMainWindow::updatePendingProgramDescription()
{
    if(!hasNewPnpCache)
        return;
    hasNewPnpCache = false;

    for(auto pd : progDescrsCache) {
        if(!EVB_SEARCH_TARGETS.contains(pd.getDaqModule()))
            continue;

        auto index = pd.getClientIndex();
        if(!clients.contains(index))
            continue;

        if(config.clients.value(index).enabled &&
            isNonEventBaseProgram(pd) != nonEvBasedClients.contains(index)) {
            //what if in run?!
            updateAnalizerList();
        }
    }

    updateButtonAddEnableStates();
}

void EvbMainWindow::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(loc_runId, configDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}
