#include "OutputTcpServer.h"

#include <QCoreApplication>
#include <QHostInfo>
#include <QTcpServer>
#include <QTcpSocket>
#include <QTimer>
#include <QElapsedTimer>
#include <stdexcept>

#include "params.h"
#include "metric-sender/MetricSender.h"
#include "TcpServerStats.h"
#include "SizeHelper.h"


#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

namespace {
const int MINIMAL_PACKET_SIZE = 100e3; // 100 KB
const int MAXIMAL_PACKET_SIZE = 100e6; // 100 MB
const int TIMEOUT_SEND = 10;
const int METRICS_SEND_PERIOD_MS = 1000;
const int FLUSH_TIMEOUT_MS = 1000;
}

OutputTcpServer::OutputTcpServer(int _index)
    : QObject(nullptr)
    , index(_index)
    , stats(new TcpServerStats())
    , dnsLookup(new DnsLookup(this))
    , workLoopTimer(new QTimer(this))
    , statusTimer(new QTimer(this))
    , ringBuf(new RingBuffer<char>(TCP_SERVER_BUFFER_SIZE))
{
    if(index)
        setObjectName(__func__+QString::number(index-1));
    else
        setObjectName("MonitorTcpServer");

    connect(workLoopTimer, &QTimer::timeout, this, &OutputTcpServer::workLoop);
    workLoopTimer->setInterval(0);
    workLoopTimer->setSingleShot(true);
    connect(statusTimer, &QTimer::timeout, this, &OutputTcpServer::sendStatus);
    statusTimer->start(500);

    connect(dnsLookup, &DnsLookup::reverseEntryUpdated,
            this, &OutputTcpServer::reverseEntryUpdated);

    QTimer * sendMetricsTimer = new QTimer(this);
    connect(sendMetricsTimer, &QTimer::timeout, this, &OutputTcpServer::sendMetrics);
    sendMetricsTimer->start(METRICS_SEND_PERIOD_MS);
}

OutputTcpServer::~OutputTcpServer()
{
    qDebug() << objectName() << "deletting...";
    disconnectFromHost();
    closeServer();

    delete stats;
}

void OutputTcpServer::startListen(int i, bool listen)
{
    if(index != i)
        return;
    if(listen) {
        serverFd = socket(AF_INET, SOCK_STREAM, 0);
        if(*serverFd < 0) {
            qFatal("[%s] socket creation failed: %s",
                   objectName().toStdString().c_str(), strerror(errno));
        }

        int opt = 1;
        if (setsockopt(*serverFd, SOL_SOCKET,
                       SO_REUSEADDR | SO_REUSEPORT, &opt,
                       sizeof(opt))) {
            qFatal("[%s] setsockopt failed: %s",
                   objectName().toStdString().c_str(), strerror(errno));
        }

        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = INADDR_ANY;
        addr.sin_port = htons(preferedPortNum);

        if(bind(*serverFd, (struct sockaddr*)&addr, sizeof(addr))< 0) {
            qFatal("[%s] bind failed: %s",
                   objectName().toStdString().c_str(), strerror(errno));
        }
        if(::listen(serverFd.value(), 1) < 0) {
            qFatal("[%s] listen error: %s",
                   objectName().toStdString().c_str(), strerror(errno));
        }
        status.updateLocalPort(*serverFd);

        Q_ASSERT(!preferedPortNum || (*status.localPort == preferedPortNum));
        sendStatus();
        qInfo() << QString("%1 listen on port #%2")
                   .arg(objectName()).arg(*status.localPort);
        workLoopTimer->start();
    } else {
        disconnectFromHost();
        closeServer();
        qInfo() << objectName() << "Stopped";
    }
}

void OutputTcpServer::deleteTcp(int i)
{
    if(i==index || i==0) {
        deleteLater();
    }
}

void OutputTcpServer::writeData()
{
    if(!socketFd)
        return;
    writeLatencyElapsedTimer.restart();
    int writeLoops = 0;
    while(dataSizeCache) {
        int flags = MSG_DONTWAIT | MSG_NOSIGNAL;
        auto ret = ::send(*socketFd, ringBuf->getReadPointer(), dataSizeCache, flags);
        if(ret < 0) {
            if(errno == EAGAIN)
                return;
            qCritical() << QString("%1 %2: %3")
                           .arg(objectName(), __PRETTY_FUNCTION__, strerror(errno));
            disconnectFromHost();
            return;
        }
        if(ret == 0)
            break;

        ++writeLoops;
        stats->bytesSent += ret;
        status.bytes += ret;
        if(!infOutput) {
            dataSizeCache -= ret;
            ringBuf->trashData(ret);
            ringBuf->freeData(ret);
        }
    }
    if(writeLoops) {
        double elapsedAver = writeLatencyElapsedTimer.nsecsElapsed();
        elapsedAver /= writeLoops;
        do {
            stats->addLatencyValue(elapsedAver);
        } while(--writeLoops);
    }
    checkWorkEnd();
}

void OutputTcpServer::checkDisconnect()
{
    char buf[1] = {0};

    int bytes = recv(*socketFd, buf, sizeof(buf), 0);
    if (bytes > 0) {
        static bool warn = true;
        if(warn) {
            qCritical() << objectName()
                        << " recieved some data from peer";
            warn = false;
        }
    } else if (bytes == 0) {
        // remote has close the connection
        disconnectFromHost();
    } else {
        qCritical() << objectName()
                    << " recieved some data from peer" << strerror(errno);
        disconnectFromHost();
    }
}


void OutputTcpServer::workLoop()
{
    static const int LOOP_TIMEOUT = 100; // in ms
    if(!serverFd)
        return;
    struct timeval timeout = {0, LOOP_TIMEOUT*1000};

    int maxFd = *serverFd;
    fd_set readFds, writeFds;
    FD_ZERO(&readFds);
    FD_ZERO(&writeFds);

    FD_SET(*serverFd, &readFds);

    if(socketFd) {
        FD_SET(*socketFd, &readFds);

        if(!dataSizeCache) {
            if(!updateDataSize()) {
                ringBuf->waitDataSpace(MINIMAL_PACKET_SIZE, LOOP_TIMEOUT);
                updateDataSize();
            }
        }
        if(dataSizeCache) {
            FD_SET(*socketFd, &writeFds);
            maxFd = std::max(maxFd, socketFd.value_or(0));
        } else {
            // in case connected peer and no data we should wait in ringBub, not in select
            timeout.tv_usec = 0;
        }
    }

    maxFd = std::max(maxFd, socketFd.value_or(0));
    int activity = select(maxFd+1, &readFds, &writeFds, nullptr, &timeout);
    if(activity == -1) {
        qCritical("[%s] select failed: %s",
               objectName().toStdString().c_str(), strerror(errno));
        disconnectFromHost();
        return;
    }
    if(activity==0) {
        workLoopTimer->start();
        return;
    }
    int activityOld = activity;

    if (FD_ISSET(*serverFd, &readFds)) {
        acceptConnection();
        --activity;
    }

    int oldFd = socketFd.value_or(-1);
    if(socketFd && FD_ISSET(*socketFd, &writeFds)) {
        writeData();
        --activity;
    }

    if(oldFd!=-1 && FD_ISSET(oldFd, &readFds)) {
        if(socketFd)
            checkDisconnect();
        --activity;
    }

    if(activity != 0) {
        qWarning() << __PRETTY_FUNCTION__ <<  " activityOld:" << activityOld <<  " activity:" << activity;
        Q_UNREACHABLE();
    }

    workLoopTimer->start();
}

void OutputTcpServer::checkWorkEnd() const
{
    if(!finishId)
        return;

    if(ringBuf->getDataSize()==0) {
        emit workFinished(index, finishId);
    }
}

void OutputTcpServer::clear()
{
    ringBuf->clear();
    sendMetrics();
    stats->clear();
    status.clear();
    updateDataSize();
}

void OutputTcpServer::sendMetrics()
{
    const QMap<QString, QString> tags({{"output", QString::number(index)}});
    MetricFields fields{{"free", ringBuf->getFreePart()},
                        {"data", ringBuf->getDataPart()},
                        {"trash", ringBuf->getTrashPart()}};
    MetricSender::send("evb.out.tcp.bytes", stats->bytesSent, tags);
    MetricSender::send("evb.out.tcp.buffer", fields, tags);
    MetricSender::send("evb.out.tcp.latency.mean", stats->getMeanLatency(), tags);
}

void OutputTcpServer::reverseEntryUpdated(const DnsReverseEntry &ent)
{
    if(!status.connected)
        return;

    bool ok;
    Q_ASSERT(!status.peer.host.isNull());
    const auto ip = status.peer.host.toIPv4Address(&ok);
    Q_ASSERT(ok);
    if(ent.first == ip && status.peer.hostName != ent.second) {
        status.peer.hostName = ent.second;
        sendStatus();
    }
}

void OutputTcpServer::disconnectFromHost()
{
    if(!socketFd)
        return;
    close(*socketFd);
    connectionCooldown.setRemainingTime(1500);

    const QString errStr = objectName() + " disconnected from " + status.peer.toString();
    socketFd.reset();
    status.connected = false;
    status.peer = {};

    (index ? qWarning() : qInfo()) << errStr;
    // We can't use clear at this moment, todo: Combiner should stop using our RB
    // // Drop all elapsed data in case of monitor TCP
    // if(!index) {
    //     ringBuf->clear();
    // }
    // Send error only in Run state and if tcp out not a monitor one
    if(index && runEnabled)
        emit error(index, errStr);
    sendStatus();
}

void OutputTcpServer::closeServer()
{
    disconnectFromHost();

    if(serverFd) {
        shutdown(*serverFd, SHUT_RDWR);
        serverFd.reset();
        status.localPort.reset();
        sendStatus();
    }
}

int OutputTcpServer::updateDataSize()
{
    return dataSizeCache = ringBuf->getContinuosDataSize(0);
}

void OutputTcpServer::acceptConnection()
{
    if(!connectionCooldown.hasExpired()) {
        return;
    }

    // Drop all elapsed data from previous connection in case of monitor TCP
    if(!index) {
        ringBuf->clear();
        updateDataSize();
    }

    int fd = accept(*serverFd, nullptr, nullptr);
    if(fd == -1) {
        qCritical() << QString("%1 failed to accept pending connection: %2")
                           .arg(objectName(), strerror(errno));
    } else {
        if(socketFd) {
            SocketAddress peer;
            peer.updateAddr(fd);
            close(fd);
            qWarning() << QString("[%1] Reject incoming second connection: %2")
                              .arg(objectName(), peer.toString());
            return;
        }

        socketFd = fd;
        status.peer.updateAddr(fd);
        if(!status.peer.host.isNull())
            dnsLookup->reverseLookup(status.peer.host.toString());
        status.connected = true;

        qInfo() << QString("[%1] Current connection: %2")
                       .arg(objectName(), status.peer.toString());
    }
}

void OutputTcpServer::sendStatus() const
{
    status.listening = serverFd.has_value();
    status.updateRingBuffet(ringBuf);
    status.updateSocketOptions(serverFd);
    status.updateSpeed();
    emit statusUpdated(status);
}

void OutputTcpServer::startWork()
{
    finishId = 0;
    runEnabled = true;
}

void OutputTcpServer::finishWork(uint fid)
{
    runEnabled = false;
    if(fid==0) {
        qWarning() << objectName()
                   << QString("Got signal to finishWork with incorrect fid=%1").arg(fid);
        return;
    }
    if(finishId)
        qWarning() << objectName()
                   << QString("Got signal to finishWork with fid=%1, but i'm already in finishing state with finishId=%2")
                      .arg(fid).arg(finishId);
    finishId = fid;
    checkWorkEnd();
}

void OutputTcpServer::setPreferedPortNum(quint16 portNum)
{
    if(preferedPortNum == portNum)
        return;

    preferedPortNum = portNum;
    if(serverFd) {
        startListen(index, false);
        startListen(index, true);
    } else {
        if(!index)
            qInfo() << objectName() << " set prefered port to:" << preferedPortNum;
    }
}

void OutputTcpServer::setInfOutput(bool en)
{
    if(infOutput == en)
        return;
    infOutput = en;
    if(infOutput) {
        auto s = ringBuf->getContinuosDataSize(0);
        if(s<1e7) {
            ringBuf->trashData(s);
            ringBuf->freeData(s);
            s = ringBuf->getContinuosFreeSize();
            ringBuf->writeData(s);
        }
    }
}
