#include "RingBuffer.h"

#include <QDebug>
#include <sys/sysinfo.h>

namespace {
enum {TYPE_FREE,TYPE_DATA,TYPE_TRASH};

quint64 getFreeRam() {
    return 1e12;
    struct sysinfo info;
    int ret = sysinfo(&info);
    if(ret)
        return 0;
    auto freeBytes = info.freeram * info.mem_unit;
    return freeBytes;
}
}

template<class T> RingBuffer<T>::RingBuffer(int _size):
    size(_size),
    writePos(0),
    readPos(0),
    trashPos(0),
    freeSpace(size),
    dataSpace(0),
    trashSpace(0)
{
    Q_ASSERT((int)size+1>0);
    ringData.resize(size+1);
}

template<class T> void RingBuffer<T>::clear()
{
    int val = getContinuosDataSize(0);
    trashData(val);
    val = getContinuosDataSize(0);
    if(val)
        trashData(val);
    Q_ASSERT(getDataSize() == 0);
    val = getContinuosTrashSize();
    freeData(val);
    val = getContinuosTrashSize();
    if(val)
        freeData(val);
    Q_ASSERT(getFreeSize() == getAllSize());
}

template<class T>
bool RingBuffer<T>::resize(qint64 newSize)
{
    Q_ASSERT((int)newSize+1 > 0);
    bool ok = freeSpace.available()==size;
    if(ok)
        ok &= dataSpace.available()==0;
    if(ok)
        ok &= trashSpace.available()==0;
    if(!ok)
        return ok;
    if(size < newSize) {
//        if(writePos+freeSpace.available()<size+1)
//            return false;

        constexpr static const auto RESERVED_RAM_SIZE = 2e9;
        if(newSize > ringData.size()) {
            auto freeRam = getFreeRam();
            auto needFreeRam = newSize-ringData.size()+RESERVED_RAM_SIZE;
            if(freeRam < needFreeRam) {
                return false;
            }
        }

        freeSpace.release(newSize-size);
    } else {
        if(!freeSpace.tryAcquire(size-newSize))
            qFatal("Failed to shrink buffer from:%d to:%d", size, newSize);
    }

    ringData.resize(newSize+1);
    ringData.shrink_to_fit();
    size = newSize;
    writePos = 0;
    readPos = 0;
    trashPos = 0;
    return true;
}

// =========== Writer part ===========
template<class T> int RingBuffer<T>::getFreeSize() const
{
    int avail = freeSpace.available();
    return avail;
}

template<class T> double RingBuffer<T>::getFreePart()
{
    return (double) getFreeSize()  / getAllSize();
}

template<class T> int RingBuffer<T>::getContinuosFreeSize() const
{
    return getContinuosTypeSize(TYPE_FREE);
}

template<class T> void RingBuffer<T>::writeData(int continuousSize)
{
    addTypeData(TYPE_DATA, continuousSize);
}

template<class T> bool RingBuffer<T>::waitFreeSpace(int s, int timeout)
{
    return waitTypeSize(TYPE_FREE, s, timeout);
}

template<class T> T *RingBuffer<T>::getWritePointer()
{
    return getTypePointer(TYPE_FREE, 0);
}

template<class T> int RingBuffer<T>::getWriterPos() const
{
    return writePos;
}



// =========== Reader part ===========
template<class T> int RingBuffer<T>::getDataSize() const
{
    int avail = dataSpace.available();
    return avail;
}

template<class T> double RingBuffer<T>::getDataPart()
{
    return (double) getDataSize()  / getAllSize();
}

template<class T> int RingBuffer<T>::getContinuosDataSize(int shift) const
{
    return getContinuosTypeSize(TYPE_DATA, shift);
}

template<class T> void RingBuffer<T>::trashData(int continuousSize)
{
    addTypeData(TYPE_TRASH, continuousSize);
}

template<class T> void RingBuffer<T>::trashData(int continuousSize, int checkReaderPos)
{
    addTypeData(TYPE_TRASH, continuousSize, checkReaderPos);
}

template<class T> bool RingBuffer<T>::waitDataSpace(int s, int timeout)
{
    return waitTypeSize(TYPE_DATA, s, timeout);
}

template<class T> T *RingBuffer<T>::getReadPointer(int shift)
{
    return getTypePointer(TYPE_DATA, shift);
}

template<class T> int RingBuffer<T>::getReaderPos(int shift) const
{
    Q_ASSERT(shift<=dataSpace.available());
    return (readPos+shift)%(size+1);
}

template<class T> void RingBuffer<T>::readData(char *dst, int shift, int s)
{
    int a = dataSpace.available();
    Q_ASSERT(shift>=0 && s>=0 && s+shift<=a);

    int absPos = (readPos + shift)%(size+1);
    int c1 = getContinuosDataSize(shift);
    if(c1<s){
        memcpy(dst, ringData.data()+absPos, c1);
        s -= c1;
        absPos=0;
        dst += c1;
    }
    memcpy(dst, ringData.data()+absPos, s);
}


// =========== Cleaner part ===========
template<class T> int RingBuffer<T>::getTrashSize() const
{
    int avail = trashSpace.available();
    return avail;
}

template<class T> double RingBuffer<T>::getTrashPart()
{
    return (double) getTrashSize()  / getAllSize();
}

template<class T> int RingBuffer<T>::getContinuosTrashSize() const
{
    return getContinuosTypeSize(TYPE_TRASH, 0);
}

template<class T> void RingBuffer<T>::freeData(int continuousSize, int checkTrashPos)
{
    addTypeData(TYPE_FREE, continuousSize, checkTrashPos);
}

template<class T> void RingBuffer<T>::freeData(int continuousSize)
{
    addTypeData(TYPE_FREE, continuousSize);
}

template<class T> bool RingBuffer<T>::waitTrashSpace(int s, int timeout)
{
    return waitTypeSize(TYPE_TRASH, s, timeout);
}

template<class T> T *RingBuffer<T>::getClearPointer(int shift)
{
    int avail = trashSpace.available();
    Q_ASSERT(shift>=0 && shift<avail);
    qint64 absPos = trashPos+shift;
    if(absPos>size)
        absPos -= size+1;
    return ringData.data()+absPos;
}



// =========== Uniform part ===========

template<class T> int RingBuffer<T>::getContinuosTypeSize(int type, int shift) const
{
    qint64 avail, pos;
    if(type==TYPE_FREE) {
        avail = freeSpace.available();
        pos = writePos;
    } else if(type==TYPE_DATA) {
        avail = dataSpace.available();
        pos = readPos;
    } else if(type==TYPE_TRASH) {
        avail = trashSpace.available();
        pos = trashPos;
    } else {
        Q_ASSERT(false);
    }

    Q_ASSERT(shift>=0 && shift<=avail);

    if(pos+shift > size) {
        return avail-shift;
    } else {
        return qMin(avail-shift, size+1-pos-shift);
    }
}

template<class T> bool RingBuffer<T>::waitTypeSize(int type, int s, int timeout)
{
    QSemaphore *sem = nullptr;
    if(type==TYPE_FREE) {
        sem = &freeSpace;
    } else if(type==TYPE_DATA) {
        sem = &dataSpace;
    } else if(type==TYPE_TRASH) {
        sem = &trashSpace;
    } else {
        Q_ASSERT(false);
    }

    if(sem->available()>=s)
        return true;
    bool ret = timeout ? sem->tryAcquire(s, timeout) : sem->tryAcquire(s);
    if(ret)
        sem->release(s);
    return ret;
}

template<class T> void RingBuffer<T>::addTypeData(int type, int continuousSize, int pos)
{
    if(type==TYPE_FREE) {
        Q_ASSERT(pos == trashPos);
    } else if(type==TYPE_DATA) {
        Q_ASSERT(pos == writePos);
    } else if(type==TYPE_TRASH) {
        Q_ASSERT(pos == readPos);
    } else {
        Q_ASSERT(false);
    }
    addTypeData(type, continuousSize);
}
#include <QThread>
template<class T> void RingBuffer<T>::addTypeData(int type, int continuousSize)
{
    QSemaphore *from = nullptr, *to = nullptr;
    qint64 *pos = nullptr;
    if(type==TYPE_FREE) {
        pos = &trashPos;
        from = &trashSpace;
        to = &freeSpace;
//        qDebug()<<size<<"RingBuffer free data:"<<s;
    } else if(type==TYPE_DATA) {
        pos = &writePos;
        from = &freeSpace;
        to = &dataSpace;
//        qDebug()<<size<<"RingBuffer write data:"<<s;
    } else if(type==TYPE_TRASH) {
        pos = &readPos;
        from = &dataSpace;
        to = &trashSpace;
//        qDebug()<<size<<"RingBuffer trash data:"<<s;
    } else {
        Q_ASSERT(false);
    }

//    int fromS = from->available();
//    int toS = to->available();
    const bool res = from->tryAcquire(continuousSize);
    if(!res) {
        auto avail = from->available();
        qFatal("%s", QString("%1: Error in RingBuffer<T>::addTypeData type:%2 size:%3 totSize:%4 avail:%5")
                         .arg(QThread::currentThread()->objectName())
                         .arg(type).arg(continuousSize).arg(getAllSize())
                         .arg(avail)
                         .toStdString().c_str());
    }
//    Q_ASSERT(res);
    *pos += continuousSize;
    if(*pos == size+1)
        *pos = 0;
    if(*pos > size+1)
        qFatal("%s", QString("%1: Out of range in RingBuffer<T>::addTypeData type:%2 size:%3 totSize:%4 pos:%5")
               .arg(QThread::currentThread()->objectName())
               .arg(type).arg(continuousSize).arg(getAllSize()).arg(*pos)
               .toStdString().c_str());
    Q_ASSERT(*pos <= size);
    to->release(continuousSize);
//    qDebug()<<"new pos="<<*pos;
}

template<class T> T *RingBuffer<T>::getTypePointer(int type, int shift)
{
    qint64 pos;
    qint64 avail;
    Q_ASSERT(shift>=0);
    if(type==TYPE_FREE) {
        pos = writePos;
        avail = freeSpace.available();
//        Q_ASSERT(shift<freeSpace.available());
    } else if(type==TYPE_DATA) {
        pos = readPos;
        avail = dataSpace.available();
//        Q_ASSERT(shift<dataSpace.available());
    } else if(type==TYPE_TRASH) {
        pos = trashPos;
        avail = trashSpace.available();
//        Q_ASSERT(shift<trashSpace.available());
    } else {
        Q_ASSERT(false);
    }
    if(shift > avail) {
        qInfo() << __func__ << pos << shift << avail << getAllSize();
        Q_ASSERT(shift<avail);
    }

    pos += shift;
    if(pos>size)
        pos -= size+1;
    return ringData.data()+pos;
}

template<class T> int RingBuffer<T>::getTypePos(int type, int shift) const
{
    qint64 pos;
    Q_ASSERT(shift>=0);
    if(type==TYPE_FREE) {
        pos = writePos; // direct result
        Q_ASSERT(shift<freeSpace.available());
    } else if(type==TYPE_DATA) {
        pos = readPos;
        Q_ASSERT(shift<dataSpace.available());
    } else if(type==TYPE_TRASH) {
        pos = trashPos;
        Q_ASSERT(shift<trashSpace.available());
    } else {
        Q_ASSERT(false);
    }

    pos += shift;
    if(pos>size)
        pos -= size+1;
    return pos;
}

