#include "EvbClientStatusModel.h"

#include <QColor>

#include "app/AppStyle.h"
#include "base/ClientIndex.h"
#include "base/ClientIndexEnabled.h"
#include "base/DeviceId.h"
#include "pnp-server/PNPMessage.h"

#include "TcpClientStatus.h"
#include "CombinerStats.h"
#include "EvbProgramDescriptionHelper.h"
#include "SizeHelper.h"

#include <QApplication>
#include <QStyle>
#include <QPixmap>

namespace {
QColor getTcpClientStateColor(const TcpClientState &st, bool selected)
{
    switch (st) {
    case TcpClientState::Connected:
        return QColor();
    case TcpClientState::Disabled:
        if(selected)
            return {90, 90, 90};
        else
            return Qt::gray; // a0a0a4
    default:
        return Qt::red;
    }
}
QString getMStreamTypeStr(const TcpClientStatus &status)
{
    auto progDescr = status.lastProgDescription;
    auto usedInterface = status.usedInterface;
    QString typeStr;

    if(typeStr.isEmpty() && usedInterface.type == ProgramInterfaceMStream) {
        typeStr = usedInterface.devTypeStr;
        if(typeStr.isEmpty())
            typeStr = usedInterface.devIndex.getDeviceIdName();

    }

    if(typeStr.isEmpty() &&
        status.index.getProgramType() == DaqModule(DaqModule::Type::MStream)) {
        typeStr = progDescr.options.value("dev_id_str");

        if(typeStr.isEmpty()) {
            auto idStr = progDescr.options.value("dev_id");
            typeStr = QString(getDeviceTypeName(idStr.toInt(nullptr, 16)));
        }
    }

    return typeStr;
}
} // namespace

struct EvbClientStatusModelPrivate
{
    ClientIndexSet clientSet;
    CombinerStats combinerStatus;

    QHash<ClientIndex, TcpClientStatus> tcpState;
    QHash<ClientIndex, QString> clientDevIdStr; // cache for Type (up on pnpCacheUpdated)
    ClientIndexSet clientNonEvbSet; // cache for NonEvb (updates on tcpState)
};

EvbClientStatusModel::EvbClientStatusModel(QObject *parent)
    : ClientIndexTableModel{parent}
    , d_ptr(new EvbClientStatusModelPrivate)
{
    Q_D(EvbClientStatusModel);
    setObjectName(__func__);
}

EvbClientStatusModel::~EvbClientStatusModel()
{
    delete d_ptr;
}

ClientIndex EvbClientStatusModel::clientIndexOf(int row) const
{
    Q_D(const EvbClientStatusModel);
    int r = 0;
    for (auto it = d->clientSet.cbegin(); it != d->clientSet.cend(); ++it)
        if (r++ == row)
            return *it;
    return {};
}

int EvbClientStatusModel::rowOf(const ClientIndex &sourceIndex) const
{
    Q_D(const EvbClientStatusModel);
    return std::distance(d->clientSet.cbegin(),
                         d->clientSet.constFind(sourceIndex));
}

int EvbClientStatusModel::rowCount(const QModelIndex &) const
{
    Q_D(const EvbClientStatusModel);
    return d->clientSet.size();
}

int EvbClientStatusModel::columnCount(const QModelIndex &) const
{
    return COL_TOTAL;
}

QVariant EvbClientStatusModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto clIndex = clientIndexOf(index.row());
    if (clIndex.isEmpty())
        return {};

    switch (role) {
    case Qt::DisplayRole: return data_DisplayRole(clIndex, column);
    case Qt::ToolTipRole: return data_ToolTip(clIndex, column);
    case Qt::ForegroundRole: return data_ForegroundRole(clIndex, column);
    case Qt::DecorationRole: return data_DecorationRole(clIndex, column);
    }
    return {};
}

QVariant EvbClientStatusModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (section) {
        case COL_TYPE: return "Type";
        case COL_INDEX: return "Index";
        case COL_DATA_TYPE: return "Data type";
        case COL_CONNECTION_STATE: return "Online";
        case COL_PROG_HOST: return "host";
        case COL_STATUS: return "state";
        case COL_EV_NUM: return "evNum";
        case COL_TRASHED_EVENTS: return "trashed";
        case COL_ALL_EVENTS: return "all";
        case COL_FLUSHED_SIZE: return "flushedSize";
        case COL_TIME_DISTR: return "Time distr";
        case COL_TCP_BUFF: return "Raw data buff";
        case COL_EV_BUFF: return "Ev. meta-info";
        }
    }
    return ClientIndexTableModel::headerData(section, orientation, role);
}

void EvbClientStatusModel::setClientList(const ClientIndexEnabledMap &map)
{
    Q_D(EvbClientStatusModel);
    auto set = ClientIndexSet(map);
    if(d->clientSet == set)
        return;

    beginResetModel();
    d->clientSet = set;
    d->clientNonEvbSet &= set;
    for(auto i : d->clientDevIdStr.keys())
        if(!set.contains(i))
            d->clientDevIdStr.remove(i);
    endResetModel();
    emit colHiddenUpdated(COL_DATA_TYPE, d->clientNonEvbSet.isEmpty());
}

void EvbClientStatusModel::setTcpClientStatus(const TcpClientStatus &status)
{
    Q_D(EvbClientStatusModel);
    const auto &clIndex = status.index;
    if(!d->clientSet.contains(clIndex))
        return;

    auto &oldSt = d->tcpState[clIndex];
    if(oldSt == status)
        return;
    oldSt = status;
    bool isNonEvb = isNonEventBaseProgram(status.lastProgDescription);
    if(isNonEvb) {
        if(!d->clientNonEvbSet.contains(clIndex)) {
            d->clientNonEvbSet.insert(clIndex);
            emit colHiddenUpdated(COL_DATA_TYPE, d->clientNonEvbSet.isEmpty());
        }
    } else {
        if(d->clientNonEvbSet.remove(clIndex))
            emit colHiddenUpdated(COL_DATA_TYPE, d->clientNonEvbSet.isEmpty());
    }

    if(clIndex.getProgramType() == DaqModule(DaqModule::Type::MStream)
        || status.usedInterface.type == ProgramInterfaceMStream) {
        auto typeStr = getMStreamTypeStr(status);
        if(!typeStr.isEmpty()) {
            if(d->clientDevIdStr.value(clIndex) != typeStr) {
                // do update
                d->clientDevIdStr[clIndex] = typeStr;
                invalidateRow(clIndex);
            }
        } else {
            if(d->clientDevIdStr.remove(clIndex)) {
                // do update
                invalidateRow(clIndex);
            }
        }
    }

    invalidateRow(clIndex, {Qt::DisplayRole, Qt::BackgroundRole});
}

void EvbClientStatusModel::setCombinerStats(const CombinerStats &cs)
{
    Q_D(EvbClientStatusModel);
    d->combinerStatus = cs;
    for(auto i : d->clientSet)
        invalidateRow(i, {Qt::DisplayRole, Qt::BackgroundRole});
}

QVariant EvbClientStatusModel::data_DisplayRole(const ClientIndex &clIndex, int column) const
{
    Q_D(const EvbClientStatusModel);
    switch (column) {
    case COL_TYPE:
    {
        auto str = d->clientDevIdStr.value(clIndex);
        return str.isEmpty() ? clIndex.getProgramType().toString() : str;
    }
    case COL_INDEX:
        return clIndex.getProgramIndex().toString();
    case COL_DATA_TYPE:
        return isNonEventBaseClient(clIndex) ? "nonEvBased" : "EvBased";
    case COL_PROG_HOST:
        return d->tcpState.value(clIndex).lastProgDescription.
            hostName.split('.').first();
    case COL_STATUS:
        return d->tcpState.value(clIndex).stateStr;
    case COL_EV_NUM:
    {
        auto v = d->combinerStatus.clientsStat.value(clIndex).lastEventNum;
        if(v.has_value())
            return QString("%L1").arg(v.value());
        return "-";
    }
    case COL_TRASHED_EVENTS:
    {
        auto v = d->combinerStatus.clientsStat.value(clIndex).trashedEvents;
        return QString("%L1").arg(v);
    }
    case COL_ALL_EVENTS:
    {
        auto v = d->combinerStatus.clientsStat.value(clIndex).eventBlocks;
        return QString("%L1").arg(v);
    }
    case COL_FLUSHED_SIZE:
    {
        auto v = d->combinerStatus.clientsStat.value(clIndex).flushedSize;
        return humanReadableSize(v);
    }
    case COL_TIME_DISTR:
    {
        auto v = d->combinerStatus.clientsStat.value(clIndex).timeDistr;
        if(v)
            return QString::number(v.value()/d->combinerStatus.timeDistrCnt);
        return "-";
    }
    case COL_TCP_BUFF:
    {
        auto &st = d->tcpState.value(clIndex).statistic;
        return QString("%1/%2/%3")
            .arg(st.freePart, 4, 'f', 1)
            .arg(st.dataPart, 4, 'f', 1)
            .arg(st.trashPart, 4, 'f', 1);
    }
    case COL_EV_BUFF:
    {
        auto clSt = d->combinerStatus.clientsStat.value(clIndex);
        return QString("%1/%2/%3").arg(clSt.freeEvPart)
            .arg(clSt.preparedEvPart)
            .arg(clSt.flushedEvPart);
    }
    default:;
    }

    return {};
}

QVariant EvbClientStatusModel::data_ToolTip(const ClientIndex &clIndex, int column) const
{
    Q_D(const EvbClientStatusModel);
    switch (column) {
    case COL_CONNECTION_STATE:
    {
        auto ok = d->tcpState.value(clIndex).state == TcpClientState::Connected;
        return ok ? "Client sends discover packets"
                  : "Device doesn't send discover packets";
    }
    case COL_TIME_DISTR: return "Aver. time diff for earliest/\nAver. time full, ms";
    case COL_TCP_BUFF: return "free/data/trash";
    case COL_EV_BUFF: return "free/prepared/flushed";
    case COL_FLUSHED_SIZE:
    {
        auto v = d->combinerStatus.clientsStat.value(clIndex).flushedSize;
        return QString("%L1").arg(v);
    }
    }
    return {};
}

QVariant EvbClientStatusModel::data_ForegroundRole(const ClientIndex &clIndex, int column) const
{
    Q_D(const EvbClientStatusModel);
    switch (column) {
    case COL_STATUS:
        return getTcpClientStateColor(d->tcpState.value(clIndex).state,
                                      false);
    case COL_TRASHED_EVENTS:
    {
        if(d->combinerStatus.clientsStat.value(clIndex).trashedEvents)
            return QColor(Qt::red);
        break;
    }
    case COL_TCP_BUFF:
        if(d->tcpState.value(clIndex).statistic.dataPart > 80)
            return QColor(Qt::red);
        break;
    default:;
    }

    return {};
}

QVariant EvbClientStatusModel::data_DecorationRole(const ClientIndex &clIndex, int column) const
{
    if(column != COL_CONNECTION_STATE)
        return {};

    Q_D(const EvbClientStatusModel);
    auto st = d->tcpState.value(clIndex).state;
    auto p = (st == TcpClientState::Connected) ? QStyle::SP_DialogApplyButton
                                               : QStyle::SP_DialogCloseButton;
    return QApplication::style()->standardPixmap(p);
}

void EvbClientStatusModel::invalidateRow(const ClientIndex &i, const QVector<int> roles)
{
    auto row = rowOf(i);
    emit dataChanged(index(row, 0), index(row, COL_TOTAL-1), roles);
}

bool EvbClientStatusModel::isNonEventBaseClient(const ClientIndex &index) const
{
    Q_D(const EvbClientStatusModel);

    return d->clientNonEvbSet.contains(index);
}
