#include "MpdTrcCore.h"

#include "MpdTrcSignalRouter.h"
#include "MpdTrcDeviceGlobal.h"

#include "base/DeviceIndexEnabled.h"
#include "base/RunDocument.h"
#include "fsm/MStreamDeviceMgrFsm.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "mldiscover/DeviceDispatcher.h"
#include "remote-control-server/RemoteControlServer.h"

MpdTrcCore::MpdTrcCore(QObject *parent)
    : ReadoutAppCore(parent)
    , deviceMgrFsm(new MStreamDeviceMgrFsm(this))
    , configDb(new ReadoutAppConfigDb(this))
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<RunDocument>();

    auto signalRouter = &MpdTrcSignalRouter::instance();
    signalRouter->setConfigDb(configDb);
    signalRouter->setCore(this);
    signalRouter->setDeviceMgrFsm(deviceMgrFsm);
    signalRouter->connectAll();
}

const BaseFsm *MpdTrcCore::getFsm() const
{
    return deviceMgrFsm;
}

void MpdTrcCore::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(config == newConfig)
        return;
    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;

    const auto &toRemove = oldSet-newSet;
    if(!toRemove.isEmpty()) {
        for (const auto &i : toRemove)
            removeDevice(i);
    }

    const auto &toCreate = newSet-oldSet;
    for (const auto &i : toCreate) {
        if (! devices.contains(i))
            addDevice(i);
    }
    emit allowRemoteControl(config.allowRemoteControl);
}

void MpdTrcCore::performLocalStart()
{
    deviceMgrFsm->setRequestState(FsmState::Configure);
}

void MpdTrcCore::performLocalStop()
{
    deviceMgrFsm->setRequestState(FsmState::Stopping);
}

void MpdTrcCore::remoteStartRequest(const RcRunId &newRunId, const QVariant &startEvNum)
{
    if(deviceMgrFsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(deviceMgrFsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(newRunId, msg);
        return;
    }

    rcRunId = newRunId;
    runStartEvNum = startEvNum;
    deviceMgrFsm->setRequestState(FsmState::Configure);
}

void MpdTrcCore::remoteStopRequest(const RcRunId &newRunId, const QVariant &stopEvNum)
{
    Q_UNUSED(stopEvNum)
    if(rcRunId.runIndex != newRunId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); "
                             "remote control runIndex (%2)")
                     .arg(newRunId.runIndex)
                     .arg(newRunId.runIndex);
    }

    if(deviceMgrFsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(deviceMgrFsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(newRunId, msg);
    }

    deviceMgrFsm->setRequestState(FsmState::Stopping);
}

void MpdTrcCore::remoteStatusUpdated(const RemoteControlServerStatus &status)
{
    ReadoutAppCore::remoteStatusUpdated(status);

    // TODO:: should we move this into ReadoutAppCore with lastRcStatus?
    if(!status.portInUse && lastRcStatus.portInUse) {
        qWarning() << "Remote control disconnected.";

        if(deviceMgrFsm->getState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            deviceMgrFsm->setRequestState(deviceMgrFsm->getState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }
    lastRcStatus = status;
}

void MpdTrcCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppCore::onFsmStateChange(state, oldState);

    switch (state) {
    case FsmState::Error :
        if(oldState != FsmState::Init) {
            runStartEvNum.clear();
            config.setRunState(false);
            emit configChanged(config);
        }
        break;
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Calibrate :
        case FsmState::FindAdcDelay :
        case FsmState::Stopping:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:{}
        }
        break;
    }
    case FsmState::Configure :
    {
        if(++mstreamSeqNum < 100) mstreamSeqNum = 100;
        deviceMgrFsm->setMStreamSeqNum(mstreamSeqNum);
        emit mstreamResetAll(mstreamSeqNum);
        deviceMgrFsm->setRequestState(FsmState::Starting);
        break;
    }
    case FsmState::Starting :
    {
        emit preRunReset(runStartEvNum);
        config.setRunState(true);
        emit configChanged(config);
        deviceMgrFsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Run :
    {
        qInfo() << QString("Start completed successfully");
        sendRunDocument();
        break;
    }
    case FsmState::Stopping :
    {
        runStartEvNum.clear();
        config.setRunState(false);
        emit configChanged(config);
        deviceMgrFsm->setRequestState(FsmState::Ready);
        break;
    }
    default:;
    }
}

void MpdTrcCore::addDevice(const DeviceIndex &index)
{
    try {
        devices.insert(index, QSharedPointer<MstreamMregDevGlobal>(new MpdTrcDeviceGlobal(index)));
    } catch (std::exception &e){
        qCritical() << QString("[%1] Error on connect to device: %2")
                       .arg(index.getDeviceIdName()).arg(e.what());
    }
}

void MpdTrcCore::removeDevice(const DeviceIndex &index)
{
    devices.remove(index);
}

void MpdTrcCore::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(rcRunId, configDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}
