#include "MpdTrcChModel.h"

#include "MpdTrcChColumns.h"

#include <QColor>
#include <QTimer>

#include "base/DeviceId.h"

namespace {
inline Qt::CheckState boolToCheck(bool v)
{ return v ? Qt::Checked : Qt::Unchecked; }

MpdTrcChColumns getColumnByIndex(const QModelIndex &index)
{
    return static_cast<MpdTrcChColumns>(index.column());
}

QVector<quint16> MpdTrcConfig::* getConfigUIntVectorByCol(MpdTrcChColumns col) {
    switch(col) {
    case MpdTrcChColumns::COL_INPUT_DELAY: return &MpdTrcConfig::inputDelays;
    case MpdTrcChColumns::COL_INPUT_SHAPER: return &MpdTrcConfig::inputShapers;
    default: return nullptr;
    }
}

QVector<bool> Tlu16SfpTriggerConfig::* getConfigBoolVectorByCol(MpdTrcChColumns col) {
    switch(col) {
    case MpdTrcChColumns::COL_TOF_EN: return &Tlu16SfpTriggerConfig::tofChEn;
    case MpdTrcChColumns::COL_LUM_EN: return &Tlu16SfpTriggerConfig::lumChEn;
    default: return nullptr;
    }
}

static QMap<DeviceIndex, int> buildNchMap(const BaseDeviceAppConfig &config,
                                          QMap<DeviceIndex, MpdTrcBuildParams> buildParamsDecDes)
{
    QMap<DeviceIndex, int> nch;
    for(auto it = config.knownConfigs.cbegin(); it!=config.knownConfigs.cend(); ++it) {
        if(it.key().getDevId() != DEVICE_ID_UT24VE_MPD_TRC)
            continue;
        if(!it->enabled)
            continue;
        quint16 v = std::max(it->mpdTrc.inputDelays.size(), it->mpdTrc.inputDelays.size());
        auto bpIt = buildParamsDecDes.find(it.key());
        if(bpIt != buildParamsDecDes.end())
            v = std::max(bpIt->inputChNum, v);
        nch.insert(it.key(), v);
    }
    return nch;
}

static bool needReset(const BaseDeviceAppConfig &config,
                      const BaseDeviceAppConfig &newConfig,
                      QMap<DeviceIndex, MpdTrcBuildParams> buildParamsDecDes)
{
    const auto &oldNch = buildNchMap(config, buildParamsDecDes);
    const auto &newNch = buildNchMap(newConfig, buildParamsDecDes);
    if (oldNch != newNch)
        return true;

    return false;
}

static bool hasChanges(const ModularDeviceConfig &c1, const ModularDeviceConfig &c2)
{
    if(c1.tlu16SfpTrigger.tofChEn != c1.tlu16SfpTrigger.tofChEn)
        return true;
    if(c1.tlu16SfpTrigger.lumChEn != c1.tlu16SfpTrigger.lumChEn)
        return true;
    return false;
}

static bool hasChanges(const BaseDeviceAppConfig &c1, const BaseDeviceAppConfig &c2)
{
    for (auto it1=c1.knownConfigs.cbegin(); it1!=c1.knownConfigs.cend(); ++it1) {
        auto it2 = c2.knownConfigs.constFind(it1.key());
        if(it2 == c2.knownConfigs.cend())
            return true;
        if (hasChanges(*it1, *it2))
            return true;
    }
    return false;
}
}

MpdTrcChModel::MpdTrcChModel(QObject *parent)
    : QAbstractItemModel(parent)
    , delayedSendConfigTimer(new QTimer(this))
{
    delayedSendConfigTimer->setInterval(100);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &MpdTrcChModel::postCheckStateChange);
}

QVariant MpdTrcChModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::ToolTipRole:
        return toolTipData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::ForegroundRole:
        return foregroundData(index);
    case Qt::UserRole:
        return std::numeric_limits<quint16>::max();
    }
    return QVariant();
}

bool MpdTrcChModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;

    const auto item = getItemByIndex(index);
    auto devCfgIt = config.knownConfigs.find(item->dev);
    if(devCfgIt == config.knownConfigs.end())
        return false;

    switch (role) {
    case Qt::CheckStateRole:
    {
        Q_ASSERT(value!=Qt::PartiallyChecked);
        if(item->ch == -1) {
            auto rows = rowCount(index);
            bool ok = true;
            for(int r=0; r<rows; ++r){
                ok &= setData(this->index(r,index.column(),index), value, role);
            }
            return ok;
        } else {
            Q_ASSERT(!item->dev.isNull());
            auto col = getColumnByIndex(index);
            if(auto ptr = getConfigBoolVectorByCol(col)) {
                auto &trigCfg = devCfgIt->tlu16SfpTrigger;
                auto &vec = (trigCfg.*ptr);
                if(vec.size() <= item->ch)
                    return false;
                vec[item->ch] = (value==Qt::Checked) ? true : false;
            } else
                return false;

            enqueueConfigSend();
        }
        return true;
    }
    case Qt::EditRole:
    {
        const auto item = getItemByIndex(index);
        auto col = getColumnByIndex(index);

        auto ptr = getConfigUIntVectorByCol(col);
        if(!ptr)
            return false;

        auto &devModCfg = devCfgIt->mpdTrc.*ptr;
        if(item->ch == -1)
            return false;

        // auto &devModCfg = (devCfgIt->mpdTrc.*ptr);
        while(devModCfg.size() < item->ch)
            devModCfg.push_back(0);
        devModCfg[item->ch] = value.toUInt();
        enqueueConfigSend();
        emit dataChanged(index, index, {Qt::DisplayRole, Qt::EditRole});
        return true;
    }
    }

    return false;
}

Qt::ItemFlags MpdTrcChModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = getColumnByIndex(index);
    switch (col) {
    case MpdTrcChColumns::COL_INPUT_SHAPER:
    case MpdTrcChColumns::COL_INPUT_DELAY:
        f |= Qt::ItemIsEditable | Qt::ItemIsEnabled;
    case MpdTrcChColumns::COL_TOF_EN:
    case MpdTrcChColumns::COL_LUM_EN:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    default:;
    }
    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant MpdTrcChModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return {};
    if(role != Qt::DisplayRole)
        return {};
    switch (static_cast<MpdTrcChColumns>(section)) {
    case MpdTrcChColumns::COL_NAME: return "Name";
    case MpdTrcChColumns::COL_INPUT_SHAPER: return "In. shaper";
    case MpdTrcChColumns::COL_INPUT_DELAY: return "In. delay";
    case MpdTrcChColumns::COL_TOF_EN: return "TOF en";
    case MpdTrcChColumns::COL_LUM_EN: return "Lum en";

    case MpdTrcChColumns::COL_STATUS: return "Status";
    case MpdTrcChColumns::COL_BITSLIP: return "BitSlip";
    case MpdTrcChColumns::COL_IDELAY: return "iDelay";
    case MpdTrcChColumns::COL_RX_CODE_ERR_CNT: return "Rx code err cnt";
    default:;
    }
    return {};
}

QModelIndex MpdTrcChModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const auto *parentItem = parent.isValid() ?
        static_cast<MpdTrcChModelItem*>(parent.internalPointer()) :
        &rootItem;
    MpdTrcChModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex MpdTrcChModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const auto *childItem = static_cast<MpdTrcChModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    MpdTrcChModelItem *parentItem = childItem->parent;
    if(parentItem == &rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int MpdTrcChModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<MpdTrcChModelItem*>(parent.internalPointer()) :
        &rootItem;
    return parentItem->childCount();
}

int MpdTrcChModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return static_cast<int>(MpdTrcChColumns::COL_TOTAL);
}

void MpdTrcChModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}

int MpdTrcChModel::getDevRow(const DeviceIndex &di) const
{
    return -1;
}

void MpdTrcChModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if (config == newConfig)
        return;

    bool performReset = needReset(config, newConfig, buildParams);
    bool hasAnyChanges = performReset ? true : hasChanges(config, newConfig);

    config = newConfig;
    if (performReset) {
        doReset();
    } else {
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void MpdTrcChModel::setTlu16SfpTrigBp(const DeviceIndex &index,
                                      const Tlu16SfpTriggerBuildParams &bp)
{
    if(index.getDevId() != DEVICE_ID_TLU16SFP)
        return;
    auto it = buildParamsTrig.find(index);
    if(it != buildParamsTrig.end() || *it == bp)
        return;
    buildParamsTrig[index] = bp;

    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;

    bool modified = false;
    auto &tofChEn = devIt->tlu16SfpTrigger.tofChEn;
    if(tofChEn.size() != bp.chNum) {
        while(tofChEn.size() < bp.chNum)
            tofChEn.push_back(Tlu16SfpTriggerConfig::DEF_CH_EN);
        tofChEn.resize(bp.chNum);
        modified = true;
    }
    auto &lumChEn = devIt->tlu16SfpTrigger.lumChEn;
    if(devIt->tlu16SfpTrigger.lumChEn.size() != bp.chNum) {
        while(lumChEn.size() < bp.chNum)
            lumChEn.push_back(Tlu16SfpTriggerConfig::DEF_CH_EN);
        devIt->tlu16SfpTrigger.lumChEn.resize(bp.chNum);
        modified = true;
    }

    if(modified) {
        doReset();
        postCheckStateChange();
    }
}

void MpdTrcChModel::setMpdTrcBp(const DeviceIndex &index,
                                   const MpdTrcBuildParams &bp)
{
    auto it = buildParams.find(index);
    if(it != buildParams.end() || *it == bp)
        return;
    buildParams[index] = bp;

    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;
    if(bp.inputChNum == devIt->mpdTrc.inputDelays.size() &&
        bp.inputChNum == devIt->mpdTrc.inputShapers.size())
        return;

    devIt->mpdTrc.inputDelays.resize(bp.inputChNum);
    devIt->mpdTrc.inputShapers.resize(bp.inputChNum);

    doReset();
    postCheckStateChange();
}

void MpdTrcChModel::setTluDecDesStatus(const DeviceIndex &index, const TluDecDesStatus &status)
{
    if(index.getDevId() != DEVICE_ID_TLU16SFP)
        return;
    if(!config.knownConfigs.contains(index))
        return;
    devStatus[index] = status;

    sendDataChanged({Qt::DisplayRole});
}

void MpdTrcChModel::resetCounters()
{
    devStatusOffset = devStatus;
    sendDataChanged();
}

void MpdTrcChModel::onDeviceConnected(const DeviceIndex &devIndex)
{
    devStatusOffset.remove(devIndex);
    sendDataChanged();
}

void MpdTrcChModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    emit configChanged(this->config);
}

void MpdTrcChModel::enqueueConfigSend()
{
    delayedSendConfigTimer->start();
}

void MpdTrcChModel::doReset()
{
    beginResetModel();
    rootItem.children.clear();
    const auto &nchMap = buildNchMap(config, buildParams);

    if (!nchMap.isEmpty()) {
        auto allItem = rootItem.createChild();

        for (auto devIt = nchMap.cbegin(); devIt != nchMap.cend(); ++devIt) {
            auto devItem = allItem->createChild();
            devItem->dev = devIt.key();

            for (int ch = 0; ch<*devIt; ch++) {
                auto chItem = devItem->createChild();
                chItem->dev = devIt.key();
                chItem->ch = ch;
            }
        }
    }
    endResetModel();
}

void MpdTrcChModel::sendDataChanged(QVector<int> roles, const QModelIndex parent)
{
    auto rows = rowCount(parent);

    if(!rows)
        return;

    auto top = index(0,0, parent);
    auto last = index(rows-1, columnCount(top)-1, parent);
    emit dataChanged(top, last, roles);

    for(auto r=0; r<rows; ++r) {
        auto chIndex = index(r, 0, parent);
        sendDataChanged(roles, chIndex);
    }
}

QVariant MpdTrcChModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case MpdTrcChColumns::COL_NAME:
        if(item->dev.isNull())
            return "All";
        if(item->ch == -1)
            return item->dev.getSerialStr();
        return QString::number(item->ch);
    case MpdTrcChColumns::COL_STATUS:
        if(auto stPtr = getChStatusPtr(item)) {
            return stPtr->isStatusOk() ? "Ok" :
                QString("%1").arg(stPtr->status, 3, 2, QChar('0'));
        }
        break;
    case MpdTrcChColumns::COL_BITSLIP:
        if(auto stPtr = getChStatusPtr(item))
            return stPtr->bitSlip;
        break;
    case MpdTrcChColumns::COL_IDELAY:
        if(auto stPtr = getChStatusPtr(item))
            return stPtr->iDelay;
        break;
    case MpdTrcChColumns::COL_RX_CODE_ERR_CNT:
    {
        auto p = getCounterPair(item);
        if(!p.first)
            return {};
        return QString("%L1").arg(*p.first - p.second.value_or(0));
    }
    default:;
    }

    return {};
}

QVariant MpdTrcChModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto &col = getColumnByIndex(index);
    auto ptr = getConfigUIntVectorByCol(col);
    if(!ptr)
        return {};

    const auto &deviceIndex = item->dev;
    if (deviceIndex.isNull())
        return {};

    const auto &devModCfg = config.knownConfigs.value(deviceIndex).mpdTrc.*ptr;
    if (item->ch == -1)
        return {};

    return devModCfg.value(item->ch);
}

QVariant MpdTrcChModel::toolTipData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case MpdTrcChColumns::COL_STATUS:
        if(auto stPtr = getChStatusPtr(item))
            return stPtr->getStatusStringList().join('\n');
        break;
    case MpdTrcChColumns::COL_RX_CODE_ERR_CNT:
    {
        auto p = getCounterPair(item);
        if(!p.first)
            return {};
        auto ret = QString("%L1").arg(*p.first);
        if(p.second)
            ret += QString(" - %L2").arg(*p.second);
        return ret;
    }
    default:;
    }

    return {};
}

QVariant MpdTrcChModel::checkStateData(const QModelIndex &index) const
{
    const auto &col = getColumnByIndex(index);
    auto ptr = getConfigBoolVectorByCol(col);
    if(!ptr)
        return {};

    return getRecursiveState(index);
}

QVariant MpdTrcChModel::foregroundData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case MpdTrcChColumns::COL_STATUS:
        if(auto stPtr = getChStatusPtr(item))
            if(!stPtr->isStatusOk())
                return QColor(Qt::red);
        break;
    default:;
    }

    return {};
}

MpdTrcChModelItem *MpdTrcChModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<MpdTrcChModelItem*>(index.internalPointer());
}

QVariant MpdTrcChModel::getRecursiveState(const QModelIndex &mIndex) const
{
    auto rows = rowCount(mIndex);
    if(rows) {
        QVariant allSt;
        for(int r=0; r<rows; ++r) {
            auto st = getRecursiveState(index(r, mIndex.column(), mIndex));
            if(st.isNull())
                continue;
            if(allSt.isNull())
                allSt = st;
            else if(allSt != st)
                return Qt::PartiallyChecked;
        }
        return allSt;
    } else {
        auto col = getColumnByIndex(mIndex);
        auto ptr = getConfigBoolVectorByCol(col);
        auto item = getItemByIndex(mIndex);
        auto devIt = config.knownConfigs.constFind(item->dev);
        if(devIt == config.knownConfigs.cend())
            return {};
        if(ptr) {
            const auto &chVector = devIt->tlu16SfpTrigger.*ptr;
            if(chVector.size() <= item->ch)
                return {};
            return boolToCheck(chVector.value(item->ch));
        } else {
            const auto &chCfg = devIt->tluDecDes.chReset;
            if(chCfg.size() <= item->ch)
                return {};
            return boolToCheck(chCfg.value(item->ch));
        }
    }
}

const TluDecDesChStatus *MpdTrcChModel::getChStatusPtr(MpdTrcChModelItem *mItem, bool offset) const
{
    if(mItem->ch == -1)
        return nullptr;
    const auto &statMap = offset ? devStatusOffset : devStatus;
    auto it = statMap.constFind(mItem->dev);
    if(it == statMap.cend())
        return nullptr;
    if(it->chSt.size() <= mItem->ch)
        return nullptr;

    return &it->chSt[mItem->ch];
}

QPair<std::optional<quint32>, std::optional<quint32> > MpdTrcChModel::getCounterPair(MpdTrcChModelItem *item) const
{
    QPair<std::optional<quint32>, std::optional<quint32> > ret;
    if(auto stPtr = getChStatusPtr(item)) {
        ret.first = stPtr->rxCodeErrCnt;
        if(auto stOffPtr = getChStatusPtr(item, true))
            ret.second = stOffPtr->rxCodeErrCnt;
    }
    return ret;
}
