#include "MpdTrcLogicForm.h"
#include "ui_MpdTrcLogicForm.h"

#include <QBitArray>
#include <QByteArray>
#include <QCheckBox>
#include <QSignalBlocker>

#include "base/gui_utils.h"

namespace {
constexpr const char *k_val = "val";

static const QVector<QString> MpdTrcPhyChNames =
    {
    "FFD East",
    "FFD West",
    "FFD Vertex",
    "TOF",
    "Luminosity",
    "FHCAL",
    "TPC Laser #0",
    "TPC Laser #1"
};

static QString getMpdTrcPhyChName(const int ch)
{
    if(ch < MpdTrcPhyChNames.size())
        return MpdTrcPhyChNames[ch];
    return QString("Unknown ch #%1").arg(ch);
}
}

MpdTrcLogicForm::MpdTrcLogicForm(QWidget *parent)
    : QWidget(parent)
    , ui(new Ui::MpdTrcLogicForm)
{
    ui->setupUi(this);
    initFhcalLogic(ui->comboBoxEast, ui->spinBoxEastThr);
    initFhcalLogic(ui->comboBoxWest, ui->spinBoxWestThr);
    initFhcalOutLogic();
    initPhyTrigs(MpdTrcPhyChNames.size());
    // ui->horizontalLayout->setAlignment(ui->groupBoxFhcal, Qt::AlignBottom);
}

MpdTrcLogicForm::~MpdTrcLogicForm()
{
    delete ui;
}

void MpdTrcLogicForm::setMpdTrcConfig(const MpdTrcConfig &newConfig)
{
    if(config == newConfig)
        return;

    config = newConfig;
    updateGui();
}

void MpdTrcLogicForm::setMpdTrcBp(const DeviceIndex &index, const MpdTrcBuildParams &bp)
{
    auto it = buildParams.find(index);
    if(it != buildParams.end() || *it == bp)
        return;
    buildParams[index] = bp;

    initPhyTrigs(bp.phyChNum);
}

void MpdTrcLogicForm::updateLutConfigByGui()
{
    bool orLogic = ui->radioButtonPhyOr->isChecked();
    QBitArray data(1<<8);
    quint16 enabledBits = 0;
    for(auto i=0; i<ui->verticalLayoutPhyTrigs->count(); ++i) {
        auto item = ui->verticalLayoutPhyTrigs->itemAt(i);
        if(!item)
            continue;
        auto cb = qobject_cast<QCheckBox*>(item->widget());
        if(!cb || !cb->isChecked())
            continue;
        auto chVal = cb->property(k_val);
        Q_ASSERT(chVal.isValid());
        enabledBits |= 1<<chVal.toUInt();
    }
    // qInfo() << __PRETTY_FUNCTION__ << "enabledBits:" << Qt::hex << enabledBits;
    if(orLogic) {
        for(quint64 i=0; i<(1<<8); ++i) {
            if(i&enabledBits)
                data.toggleBit(i);
        }
    } else {
        data.toggleBit(enabledBits);
    }
    // auto rawData = reinterpret_cast<const quint16*>(data.bits());
    int expSize = (1<<8)/8/sizeof(quint16);
    QVector<quint16> lut;
    if(expSize > lut.size()) {
        lut.resize(expSize);
    }
    memcpy(lut.data(), data.bits(), expSize*sizeof(quint16));
    // qInfo() << __PRETTY_FUNCTION__ << data;
    if(config.lut == lut)
        return;

    config.lut = lut;
    emit mpdTrcConfigChanged(config);
}

void MpdTrcLogicForm::on_comboBoxEast_currentIndexChanged(int index)
{
    auto l = static_cast<FhcalInputLogic>(ui->comboBoxEast->itemData(index).toUInt());
    if(config.fhcalEastLogic == l)
        return;
    config.fhcalEastLogic = l;
    emit mpdTrcConfigChanged(config);
}

void MpdTrcLogicForm::on_spinBoxEastThr_valueChanged(int arg1)
{
    if(config.fhcalEastMajThr == arg1)
        return;
    config.fhcalEastMajThr = arg1;
    emit mpdTrcConfigChanged(config);
}

void MpdTrcLogicForm::on_comboBoxWest_currentIndexChanged(int index)
{
    auto l = static_cast<FhcalInputLogic>(ui->comboBoxWest->itemData(index).toUInt());
    if(config.fhcalWestLogic == l)
        return;
    config.fhcalWestLogic = l;
    emit mpdTrcConfigChanged(config);
}

void MpdTrcLogicForm::on_spinBoxWestThr_valueChanged(int arg1)
{
    if(config.fhcalWestMajThr == arg1)
        return;
    config.fhcalWestMajThr = arg1;
    emit mpdTrcConfigChanged(config);
}

void MpdTrcLogicForm::on_comboBoxOut_currentIndexChanged(int index)
{
    auto l = static_cast<FhcalOutputLogic>(ui->comboBoxOut->itemData(index).toUInt());
    if(config.fhcalOutputLogic == l)
        return;
    config.fhcalOutputLogic = l;
    emit mpdTrcConfigChanged(config);
}

void MpdTrcLogicForm::updateGui()
{
    QSignalBlocker(ui->comboBoxEast);
    ui->comboBoxEast->setCurrentIndex(ui->comboBoxEast->findData(static_cast<quint16>(config.fhcalEastLogic)));

    QSignalBlocker(ui->spinBoxEastThr);
    ui->spinBoxEastThr->setValue(config.fhcalEastMajThr);

    QSignalBlocker(ui->comboBoxWest);
    ui->comboBoxWest->setCurrentIndex(ui->comboBoxWest->findData(static_cast<quint16>(config.fhcalWestLogic)));

    QSignalBlocker(ui->spinBoxWestThr);
    ui->spinBoxWestThr->setValue(config.fhcalWestMajThr);

    QSignalBlocker(ui->comboBoxOut);
    ui->comboBoxOut->setCurrentIndex(ui->comboBoxOut->findData(static_cast<quint16>(config.fhcalOutputLogic)));


    auto bitArr = QBitArray::fromBits(reinterpret_cast<const char *>(config.lut.data()),
                                      config.lut.size()*sizeof(quint16)*8);
    bitArr.resize(1<<8);
    quint32 mask = 0;
    quint32 maskOff = 0;
    int bitCount = 0;
    for(auto i = 0; i < bitArr.size(); ++i) {
        if(bitArr.testBit(i)) {
            ++bitCount;
            mask |= i;
        } else {
            maskOff |= i;
        }
    }
    if(bitCount>1) {
        mask &= ~maskOff;
        // qInfo() << __PRETTY_FUNCTION__ << "bitCount:" << bitCount
        //         << Qt::hex << "mask:" << mask;
        mask = 0;
        for(int i=0; i<16; ++i) {
            auto pos = 1<<i;
            if(bitArr.size() <= pos)
                break;
            if(bitArr.testBit(pos))
                mask |= pos;
        }
    }
    // qInfo() << __PRETTY_FUNCTION__ << "bitCount:" << bitCount
    //         << Qt::hex << "mask:" << mask;

    RecursiveWidgetSignalBlocker groupBlocker(ui->groupBoxPhy);
    (bitCount > 1 ? ui->radioButtonPhyOr : ui->radioButtonPhyAnd)
        ->setChecked(true);

    for(auto i = 0; i < ui->verticalLayoutPhyTrigs->count(); ++i) {
        auto item = ui->verticalLayoutPhyTrigs->itemAt(i);
        if(!item)
            continue;
        auto w = qobject_cast<QCheckBox*>(item->widget());
        if(!w)
            continue;
        auto ch = w->property(k_val).toUInt();
        w->setChecked(mask & (1<<ch));
    }
}

void MpdTrcLogicForm::initFhcalLogic(QComboBox *cb, QSpinBox *sb)
{
    QSignalBlocker cbBlock(cb);
    QSignalBlocker sbBlock(sb);
    Q_UNUSED(cbBlock);
    Q_UNUSED(sbBlock);
    cb->addItem("Common OR", 0);
    cb->addItem("Common AND", 1);
    cb->addItem("Majority", 3);

    sb->setMaximum(5);
}

void MpdTrcLogicForm::initFhcalOutLogic()
{
    auto cb = ui->comboBoxOut;
    QSignalBlocker cbBlock(cb);
    Q_UNUSED(cbBlock);
    cb->addItem("OR between East-West", 0);
    cb->addItem("AND between East-West", 1);
}

void MpdTrcLogicForm::initPhyTrigs(quint16 chNum)
{
    if(phyChNum >= chNum)
        return;
    phyChNum = chNum;
    for(auto i=0; i<ui->verticalLayoutPhyTrigs->count(); ++i) {
        auto item = ui->verticalLayoutPhyTrigs->itemAt(i);
        if(!item)
            continue;
        auto w = item->widget();
        if(!w)
            continue;
        if(w->property(k_val).toUInt() >= phyChNum)
            delete w;
    }

    for(auto ch = ui->verticalLayoutPhyTrigs->count(); ch < phyChNum; ++ch) {
        auto cb = new QCheckBox(getMpdTrcPhyChName(ch), ui->groupBoxPhy);
        cb->setProperty(k_val, ch);
        ui->verticalLayoutPhyTrigs->addWidget(cb);
        connect(cb, &QCheckBox::toggled, this, &MpdTrcLogicForm::updateLutConfigByGui);
    }
    updateLutConfigByGui();
}
