#include "MscCore.h"

#include "base/RunDocument.h"
#include "base/ThreadController.h"
#include "fsm/MStreamDeviceMgrFsm.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PnpDiscovery.h"

#include "config/MscConfigDispatcher.h"
#include "data/MscHistData.h"

#include "MscDevice.h"
#include "MscDeviceGlobal.h"
#include "MscSignalRouter.h"
#include "MscStatusPublisher.h"

MscCore::MscCore(QObject *parent)
    : ReadoutAppCore(parent)
    , fsm(new MStreamDeviceMgrFsm(this))
    , configDB(new ReadoutAppConfigDb(this))
    , statusPublisher(new MscStatusPublisher(this))
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<MscStatus>();
    qRegisterMetaType<MscDevHists>();
    qRegisterMetaType<LldpInfoCache>();
    qRegisterMetaType<MscBuildParams>();
    qRegisterMetaType<RunDocument>();

    deviceDispatcher->subscribeDevices(fsm);

    auto &configDispatcher = MscConfigDispatcher::instance();
    configDispatcher.subscribeConfig(this);
    configDispatcher.subscribeConfig(configDB);
    configDispatcher.subscribeCS(this);
    configDispatcher.subscribeCS(configDB);
    configDispatcher.subscribeCS(statusPublisher);
    deviceDispatcher->subscribeDevices(&configDispatcher);
    new ThreadController(statusPublisher, this);

    PnpDiscovery::subscribeMyProgram(this);

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.subscribe(deviceDispatcher);

    auto &router = MscSignalRouter::instance();
    router.setCore(this);
    router.setDeviceMgrFsm(fsm);
    router.setDeviceDispatcher(deviceDispatcher);

    router.connectAll();
}

const BaseFsm *MscCore::getFsm() const
{
    return fsm;
}

void MscCore::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(newConfig == config){
        return;
    }

    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;
    const auto toDel = oldSet-newSet;
    const auto toAdd = newSet-oldSet;

    for (auto i : toDel) {
        devices.remove(i);
    }

    for (auto i : toAdd) {
        if (! devices.contains(i)) {
            addDevice(i);
        }
    }

    if(!devices.isEmpty())
        curDevice = devices.firstKey();
    emit allowRemoteControl(config.allowRemoteControl);
    if(oldSet != newSet)
        updateMStreamInterfaces();
}

void MscCore::statusUpdated(const MscStatus &status)
{
    sendMetrics(status);
    this->status = status;
}

void MscCore::configSelectorChanged(const ConfigIndexName &conf)
{
    ReadoutAppCore::configSelectorChanged(conf);

    cs = ConfigSelector(DaqModule(DaqModule::Type::Msc16Ve), conf);
}

void MscCore::addDevice(const DeviceIndex &index)
{
    try {
        auto &router = MscSignalRouter::instance();
        auto *devGlob = new MscDeviceGlobal(index);
        devices.insert(index, QSharedPointer<MscDeviceGlobal>(devGlob));
        auto *dev = devGlob->getDevice<MscDevice>();
        router.connect_MscDevice(devGlob, dev);
    } catch (std::exception &e){
        qCritical() << QString("[%1] Error on connect to device: %2")
                       .arg(index.getDeviceIdName()).arg(e.what());
    }
}

void MscCore::increaseMStreamSeqNum()
{
    if(++mstreamSeqNum < 100)
        mstreamSeqNum = 100;
    fsm->setMStreamSeqNum(mstreamSeqNum);
}

void MscCore::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(rcRunId, configDB->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}

bool MscCore::sendMetrics(const MscStatus &status)
{
    if (!status.cycleStatus.has_value())
        return {};
    const auto &gates = status.cycleStatus.value().gates;

    MetricFields fields;
    for(int i=0; i<gates.size(); ++i){
        fields["fullGateCnt." + QString::number(i)] = gates[i].fullGateCnt;
        fields["activeGateCnt." + QString::number(i)] = gates[i].activeGateCnt;
    }

    QMap<QString, QString> tags({{"serial", status.getDeviceIndex().getSerialStr()}});
    MetricSender::send("daq.msc.counts", fields, tags);
    return true;
}

void MscCore::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    ReadoutAppCore::remoteStatusUpdated(status);

    if(!status.portInUse && lastRcStatus.portInUse) {
        qWarning() << "Remote control disconnected.";

        if(fsm->getState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            fsm->setRequestState(fsm->getState() == FsmState::Run
                                 ? FsmState::Stopping : FsmState::Error);
        }
    }
    lastRcStatus = status;
}

void MscCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppCore::onFsmStateChange(state, oldState);

    switch (state) {
    case FsmState::Run:
        sendRunDocument();
        break;
    case FsmState::Error :
        // increaseMStreamSeqNum();
        // emit resetMStream(mstreamSeqNum);
        break;
    default:;
    }
}
