/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcCoreApp.h"

#include <QCoreApplication>
#include <QJsonDocument>
#include <QThread>
#include <QTimer>

#include "RcConfigDb.h"
#include "RcConfigDispatcher.h"
#include "RcSignalRouter.h"
#include "SlowControlServer.h"
#include "TrigCore.h"
#include "TrigDevice.h"
#include "TrigDeviceController.h"
#include "VmeTrigsController.h"
#include "VmeTrigsCore.h"
#include "app/AppVersion.h"
#include "base/Globals.h"
#include "base/RunDocument.h"
#include "base/RunDocumentConverter.h"
#include "base/SyslogMessageSender.h"
#include "base/ThreadController.h"
#include "base/config/ConfigSelector.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "run-control-fsm/RcFsm.h"
#include "run-control-client-manager/RcClientManager.h"
#include "run-control-client-manager/RunStates.h"
#include "StatusPublisher.h"

static const int CONFIG_TIMEOUT = 4000; // time to collect configuration from all clients
static const int START_TIMEOUT = 3e3;
static const int STOP_TIMEOUT = 10e3;

static const QMap<FsmState, QString> map_fsm_to_sc_state = {
    {FsmState::Init, SC_STATE_STANDBY},
    {FsmState::Error, SC_STATE_FAULT},
    {FsmState::Ready, SC_STATE_STANDBY},
//    {FsmState::Configure, SC_STATE_INIT},
    {FsmState::Run, SC_STATE_RUNNING},
    {FsmState::Starting, SC_STATE_INIT},
    {FsmState::Stopping, SC_STATE_INIT},
};

RcCoreApp::RcCoreApp(QObject *parent):
    AbstractConfigurable(parent)
    , configDb(new RcConfigDb(this))
    , rcConfigDispatcher(new RcConfigDispatcher(this))
    , trigCore(new TrigCore(this))
    , clientCore(new RcClientManager(this))
    , vmeTrigsCore(new VmeTrigsCore(this))
    , slowControl(new SlowControlServer(this))
    , remoteControl(new RemoteControlServer(this, 0))
    , evNumChecker(new EvNumChecker(this))
    , config_timer(new QTimer(this))
    , fsm(new RcFsm(this))
    , statusPublisher(new StatusPublisher(this))
{
    AbstractConfigConverter<RunDocument>::registerTypes();
    vmeTrigsCore->setConfigDispatcher(rcConfigDispatcher);

    auto trigDeviceController = trigCore->getTrigDeviceController();
    trigDeviceController->setConfigDispatcher(rcConfigDispatcher);

    connect(this, &RcCoreApp::sendSlaveMessage,
            this, &RcCoreApp::slaveMessageReceived);

    progDesc.setClientIndex({DaqModule(DaqModule::Type::RunControl)});

    auto &router = RcSignalRouter::instance();
    router.addRcCoreApp(this);
    router.addRcConfigDb(configDb);
    router.addRcConfigDispatcher(rcConfigDispatcher);
    router.addVmeTrigsController(vmeTrigsController());
    router.addClientManager(clientCore);
    router.addTrigDeviceController(trigCore->getTrigDeviceController());
    router.addRemoteControlServer(remoteControl);
    router.addSlowControlServer(slowControl);
    router.addEvNumChecker(evNumChecker);
    router.addRcFsm(fsm);
    router.addStatusPublisher(statusPublisher);
    router.connectCore();

    //--- configurations transfer
    connect(config_timer, &QTimer::timeout,
            this, &RcCoreApp::config_timeout);
    config_timer->setSingleShot(true);
    config_timer->setInterval(CONFIG_TIMEOUT);

    new ThreadController(remoteControl, this);
    new ThreadController(slowControl, this);
}

RcCoreApp::~RcCoreApp()
{
}

DeviceDispatcher *RcCoreApp::getRcDeviceDispatcher()
{
    return trigCore->getDeviceDispatcher();
}

VmeTrigsController *RcCoreApp::vmeTrigsController()
{
    return vmeTrigsCore->getVmeTrigsController();
}

void RcCoreApp::quit()
{
    emit closeProgram(progDesc);

    RcSignalRouter::instance().removeRemoteControlServer();

    deleteLater();
}

void RcCoreApp::configSelectorChanged(const ConfigIndexName &conf)
{
    localProgramIndex = conf.getIndex();
    progDesc.setProgramIndex(conf.getIndex());
    ClientIndex ci{Globals::getDaqModule(), localProgramIndex};
    if(rcProgramState.index != ci) {
        rcProgramState.index = ci;
        sendMyRcProgramState();
    }
    emit updateProgramDescription(progDesc);
}

void RcCoreApp::setConfig(const RcConfig &newConfig)
{
    if(config == newConfig)
        return;

    const auto oldConfig = config;
    config = newConfig;
    emit allowRemoteControl(config.pc.allowRemoteControl);
    refreshProgramDescription();
    updateRemoteState();
}

void RcCoreApp::actionOnStateRun()
{
    QCoreApplication::processEvents(QEventLoop::AllEvents, 2000); // why?
    DeviceIndexSet devSet;
    if (!megamaster_flg)
        devSet << config.rcTrigConfig.index;
    evNumChecker->set_dev_serials(devSet);
    if (megamaster_flg)
        evNumChecker->startCheck();

    config_timer->start();
    checkCanSendRunDocument();
}

void RcCoreApp::actionPostRun()
{
//    incrementRunNumber();
}

void RcCoreApp::incrementRunNumber()
{
    config.pc.localRunNumber++;
    emit configChanged(config);
}

void RcCoreApp::onRcFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    logRcFsmStateChange(oldState, state);
    if(rcProgramState.fsm != state) {
        rcProgramState.fsm = state;
        sendMyRcProgramState();
    }

    if(state == FsmState::Error && (oldState != FsmState::Ready && oldState != FsmState::Init))
        emit printFsmInput();

    switch (state) {
    case FsmState::CloseGlobTrig: {
        emit closeGlobTrig();
        const bool waitSpillEnd = megamaster_flg && hasVme;
        fsm->setRequestState(waitSpillEnd ? FsmState::WaitSpillEnd : FsmState::Starting);
        break;
    }
    case FsmState::OpenGlobTrig: {
        emit openGlobTrig();
        fsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Starting:
        emit runStart(rcRunInfo.get_run_id(), rcRunInfo.get_start_ev_num());
        fsm->setRequestState(megamaster_flg ? FsmState::OpenGlobTrig : FsmState::Run);
        break;
    case FsmState::Run:
        actionOnStateRun();
        break;
    case FsmState::PrepareStopping:
        Q_ASSERT(megamaster_flg);
        emit prepareStop();
        fsm->setRequestState(FsmState::Stopping);
        break;
    case FsmState::Stopping:
    {
        auto stopEvNum = rcRunInfo.get_stop_ev_num();
        if(megamaster_flg) {
            evNumChecker->stopCheck();
            stopEvNum = getTrigCore()->getTrigDeviceController()->getStopEvNum();
        }
        stopEvNum.clear();
        emit runStop(stopEvNum);
        config.pc.runEnable = false;
        emit configChanged(config);
        break;
    }
    case FsmState::Error:
        if(config.pc.runEnable) {
            if(megamaster_flg)
                evNumChecker->stopCheck();

            setTrigState(false);
            config.pc.runEnable = false;
            emit configChanged(config);

            emit runStop({}, true);
        }
        break;
    default:
    {}
    }

    if (oldState == FsmState::Run || state != FsmState::Run) {
        actionPostRun();
    }
    emit slowControlStateUpdated(map_fsm_to_sc_state[state], "");
}

void RcCoreApp::logRcFsmStateChange(const FsmState &previousState, const FsmState &state)
{
    auto RunIdStr = QString(rcRunInfo.get_run_id());
    auto previousStateStr = fsmStateToString(previousState);

    if (state == FsmState::Starting) {
        qInfo() << QString("!!! Preparing Run %1... !!!").arg(RunIdStr);
    }

    if (state == FsmState::Stopping) {
        qInfo() << QString("!!! Stopping Run %1... !!!").arg(RunIdStr);
    }

    if (state == FsmState::Run) {
        qInfo() << "!!! Run started !!!";
    }
    if (state == FsmState::Ready) {
        if (previousState == FsmState::Stopping) {
            qInfo() << "!!! Run stopped !!!";
        }
    }
    if (state == FsmState::Error) {
        switch (previousState) {
        case FsmState::Starting:;
        case FsmState::Stopping:;
        case FsmState::Run: {
            qCritical() << QString("!!! %1 failed !!!").arg(previousStateStr);
            break;
        }
        default:;
        }
    }
}

void RcCoreApp::trigManagerRunStateChanged(const FsmState &state)
{
    fsm->setTrigDeviceMgrState(state);
}

void RcCoreApp::clientManagerRunStateChanged(const FsmState &state)
{
    fsm->setClientManagerRunState(state);
}

void RcCoreApp::clientRcProgramStateChanged(const RcProgramState &state)
{
    if(rcProgramState.clients == state.clients)
        return;
    rcProgramState.clients = state.clients;

    sendMyRcProgramState();
}

void RcCoreApp::vmeTrigsRunStateChanged(const FsmState &state)
{
    fsm->setVmeTrigsRunState(state);
}

void RcCoreApp::setTrigState(bool state)
{
    config.rcTrigConfig.trig.trigOpen = state;
    emit configChanged(config);
}

void RcCoreApp::config_transfer_from_clients(const ClientIndex &id, const QJsonObject &json)
{
    if(config.clients.contains(id)){
        clients_config[id] = json;
    }

    checkCanSendRunDocument();
}

void RcCoreApp::checkCanSendRunDocument()
{
    if(!config_timer->isActive())
        return;

    bool allClientsConfig = getMissingConfigClients().isEmpty();
    if(allClientsConfig)
        sendRunDocument();
}

void RcCoreApp::sendRunDocument()
{
    const auto &myCfgObj = QVariant::fromValue(config).value<QJsonObject>();
    const auto &myClientIndex = configDb->getClientIndex();
    RunDocument runDoc(rcRunInfo.get_run_id(), myClientIndex, myCfgObj);

    for(auto it=clients_config.cbegin();it!=clients_config.cend();++it){
        auto clRcDoc = QVariant::fromValue(*it).value<RunDocument>();
        if(!runDoc.merge(clRcDoc)) {
            qWarning() << QString("Failed merge RunDocument from client:%1")
                          .arg(it.key().toString());
        }
    }

    auto missingClients = getMissingConfigClients();
    if(!missingClients.isEmpty()) {
        QStringList sl;

        for(auto i : getMissingConfigClients())
            sl << i.toString();

        qWarning() << "Missing config (for RunDocument) from clients: "
                   << sl.join(", ");
    }

    if(megamaster_flg) {
        runDoc.setMeta(myClientIndex);
        configDb->saveRunDocument(runDoc);
        emit sendRunDocumentToClients(runDoc);
    } else {
        emit sendRunDocumentToRc(runDoc);
    }
    clients_config.clear();
    config_timer->stop();
}

void RcCoreApp::config_timeout()
{
    sendRunDocument();
}

bool RcCoreApp::perfomeLocalStart()
{
    incrementRunNumber();
    const RcRunId newRunId(localProgramIndex.toString(),
                           config.pc.localRunNumber,
                           QUuid::createUuid());
    setMegaMasterFlag(true);
    rcRunInfo.setStartEvNum(1);
    rcRunInfo.setId(newRunId);
    emit rcRunIdChanged(rcRunInfo.get_run_id());
    return performStart();
}

void RcCoreApp::perfomeLocalStop()
{
    return perfomeStop(false);
}

void RcCoreApp::onRemoteStartRequest(const RcRunId &runId, const QVariant &startEvNumber)
{
    setMegaMasterFlag(false);
    rcRunInfo.setStartEvNum(startEvNumber);
    rcRunInfo.setId(runId);
    emit rcRunIdChanged(rcRunInfo.get_run_id());
    performStart();

    refreshProgramDescription();
}

void RcCoreApp::onRemoteStopRequest(const RcRunId &runId, const QVariant &stopEvNumber)
{
    const auto &loc_runId = rcRunInfo.get_run_id();
    if(loc_runId.runIndex != runId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                          .arg(loc_runId.runIndex)
                          .arg(runId.runIndex);
        return;
    }
    if(loc_runId.runNumber != runId.runNumber) {
        qWarning()<< QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                          .arg(loc_runId.runNumber)
                          .arg(runId.runNumber);
        return;
    }

    rcRunInfo.setStopEvNum(stopEvNumber);
    perfomeStop(false);
    refreshProgramDescription();
}

void RcCoreApp::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = progDesc.updateProgramInterface(
                ProgramInterfaceRemoteControl, 0,
                status.tcpPort, status.listening, !status.portInUse,
                status.peerHost, status.peerPort);
    if(changed)
        emit updateProgramDescription(progDesc);

    const bool doStop = isInRun() && isRemoteConnected() && !status.portInUse;
    if(remoteConnected != status.portInUse) {
        remoteConnected = status.portInUse;
        sendMyRcProgramState();
    }

    if(doStop) {
        qWarning() << "Stop run due to remoteControl disconnect";
        perfomeStop(true);
    }
    // TODO updateStatStopButton();
}

void RcCoreApp::trigStatusUpdated(const TrigStatus &status)
{
    lastTrigStatus = status;
    if (status.spillStatus) {
        fsm->setSpillState(status.spillStatus->status.reg.bit.phase == 4);
    }
    if(!status.conn.isReady() && isInRun()) {
        auto msg = SeverityMessage(SlaveMessageSeverity::Warning, QString(status.getDeviceIdent() + " not ready"));

        const MessageTreeItem m(RcSlaveId(SlaveType::Trig, status.getDeviceIdent()),
                               rcRunInfo.get_run_id(),
                               msg);
        emit slaveMessageReceived(m);
    }
}

void RcCoreApp::onSlaveErrorMessageReceived(const MessageTreeItem &m)
{
    if (isRemoteConnected()) {
        QString mess = QString("Retransmit runError to the remoteControl in Run %1")
                           .arg(QString(rcRunInfo.get_run_id()));
        if (!m.isEmpty())
            mess.append(": ").append(m.text());
        qWarning() << mess;
        emit sendRunError(m.runId, m.text());
    } else {
        QString mess = QString("Errors in Run %1").arg(QString(rcRunInfo.get_run_id()));
        if (!m.isEmpty())
            mess.append(": ").append(m.text());
        qWarning() << mess;

        perfomeStop(true);
    }
}

void RcCoreApp::slaveMessageReceived(const MessageTreeItem &m)
{
    if (!isInRun() || rcRunInfo.get_run_id() != m.runId) {
        // qWarning() << QString("Got RunError from %1 run").arg(config.pc.runEnable?"previous":"stopped");
        return;
    }
    switch (m.severity()) {
    case SlaveMessageSeverity::Error:
        onSlaveErrorMessageReceived(m);
        break;
    case SlaveMessageSeverity::Warning:
        break;
    }
    // append and send to upstream

}

void RcCoreApp::onRunWarning(const RcRunId &runId, QString warnStr)
{
    const auto slaveType = mapSenderToSlaveType(sender());
    const MessageTreeItem m(RcSlaveId(slaveType),
                           runId,
                           SeverityMessage(SlaveMessageSeverity::Warning, warnStr));
    emit sendSlaveMessage(m);

}

void RcCoreApp::onRunError(const RcRunId &runId, QString warnStr)
{
    const auto slaveType = mapSenderToSlaveType(sender());
    const MessageTreeItem m(RcSlaveId(slaveType),
                           runId,
                           SeverityMessage(SlaveMessageSeverity::Error, warnStr));
    emit sendSlaveMessage(m);

}

bool RcCoreApp::performStart()
{
    config.pc.runEnable = true;
    clients_config.clear();
    emit configChanged(config);

    fsm->setRequestState(megamaster_flg ? FsmState::CloseGlobTrig : FsmState::Starting);
    return true;
}

void RcCoreApp::perfomeStop(bool isError)
{
    auto nextSt = FsmState::Error;
    if(!isError) {
        auto curSt = fsm->getState();
        if(curSt == FsmState::Run && megamaster_flg) {
            nextSt = FsmState::PrepareStopping;
        } else if(curSt == FsmState::Run ||
                  (megamaster_flg && curSt == FsmState::WaitSpillEnd)) {
            nextSt = FsmState::Stopping;
        }
    }

    fsm->setRequestState(nextSt);
}

void RcCoreApp::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if(!isInRun() || rcRunInfo.get_run_id() != runId)
        return;

    if (isRemoteConnected()) {
        QString mess = QString("Retransmit backPressure(%2) to the remoteControl in Run %1")
                           .arg(QString(rcRunInfo.get_run_id()))
                           .arg(backPressure);
        if(backPressure)
            qWarning() << mess;
        else
            qInfo() << mess;
        emit sendBackPressure(runId, backPressure);
    } else {
        if(backPressure) {
            setTrigState(false);
            qWarning()<< "BackPressure enabled";
        } else {
            setTrigState(true);
            qInfo()<< "BackPressure disabled";
        }
    }
}

void RcCoreApp::onVmeFlagUpdated(bool _hasVme)
{
    hasVme = _hasVme;
    emit sendVmeFlag(hasVme);
}

QMap<QString, QString> RcCoreApp::createProgramDescriptionOptions()
{
    QMap<QString, QString> options;

    options["fsm"] = fsmStateToString(fsm->getState());
    const auto &runId = rcRunInfo.get_run_id();
    options["run_index"] = runId.runIndex;
    options["run_number"] = QVariant(runId.runNumber).toString();

    progDesc.ver_date = getAppDateStr();
    progDesc.ver_hash = getAppVersion();

    return options;
}

QString RcCoreApp::createRemoteState(const RcClientConfigMap &clients, const QMap<DeviceIndex, VmeTrigConfig> &vmeTrigs)
{
    QStringList remoteState;

    for (auto it=clients.cbegin(); it!=clients.cend(); ++it) {
        if(it->en)
            remoteState << QString("%1").arg(it.key().toString());
    }

    for (auto it = vmeTrigs.begin(); it != vmeTrigs.end(); ++it) {
        const auto index = it.key();
        const auto &vme = it.value();
        if (vme.enabled && (vme.checkEvNum || vme.checkWr))
            remoteState << QString("tmwr:%1").arg(index.getIdent());
    }
    return remoteState.join("; ");
}

void RcCoreApp::refreshProgramDescription()
{
    const auto newOptions = createProgramDescriptionOptions();
    progDesc.options = newOptions;
    if (newOptions != progDesc.options) {
        progDesc.options = newOptions;
        emit updateProgramDescription(progDesc);
    }
}

void RcCoreApp::sendMyRcProgramState() const
{
    emit myRcProgramStateChanged(rcProgramState);
    if(!isRemoteConnected())
        emit allRcProgramStateChanged(rcProgramState);
}

void RcCoreApp::updateRemoteState()
{
    static QString oldState;
    const auto &state = createRemoteState(config.clients, config.vmeTrigs);
    if (state == oldState)
        return;
    oldState = state;
    emit sendRemoteState(state);
}

SlaveType RcCoreApp::mapSenderToSlaveType(QObject *sender) const
{
    if (sender == clientCore) {
        return SlaveType::TcpClient;
//    } else if(sender == ui->trigMainForm) {
//        return SlaveType::Trig;
//    } else if(sender == ui->vmeTrigsWidget) {
//        return SlaveType::VmeTrig;
    } else {
        Q_ASSERT(false);
        return SlaveType::Unknown;
    }
}

bool RcCoreApp::isInRun() const
{
    auto st = fsm->getState();
    return st!=FsmState::Error && st!=FsmState::Ready;
}

ClientIndexSet RcCoreApp::getMissingConfigClients() const
{
    return config.clients.getClientIndexEnabledMap().getEnabled()
        - makeClientIndexSet(clients_config.keys());
}

void RcCoreApp::setMegaMasterFlag(bool mm)
{
    auto dc = trigCore->getTrigDeviceController();
    dc->setMegaMasterFlag(mm);
    megamaster_flg = mm;
}
