/*
**    Copyright 2014-2015 Ivan Filippov
**    Copyright 2014-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef RCMAINWINDOW_H
#define RCMAINWINDOW_H

#include "CoreTypes.h"
#include "RcConfig.h"
#include "TrigStatus.h"
#include "app/BaseMainWindow.h"
#include "fsm/FsmState.h"

#include <QMainWindow>
#include <QMap>
#include <QPointer>
#include <QTime>

class ConfigIndexName;
class ConfigSelector;
class ConfigurationManager;
class DeviceDispatcher;
class DeviceIndex;
class DeviceIndexEnabledMap;
class DiscoverDialog;
class DlnpCalibrationTriggerDialog;
class LldpDiscovery;
class PNPServer;
class QHostAddress;
class QLabel;
class QTimer;
class RcConfigDb;
class RcCoreApp;
class RcSignalRouter;
class RemoteControlServer;
struct RemoteControlServerStatus;
class SlowControlDialog;
class TrigCore;
class TrigDevice;
class TrigDeviceProxy;
struct TrigStatus;
class VmeTrigsController;
class VmeTrigsCore;
class DeviceStatusDialog;

namespace Ui {
class RcMainWindow;
}

class RcMainWindow : public BaseMainWindow
{
    Q_OBJECT

    friend class RcSignalRouter;

public:
    enum TAB_INDEX { TAB_TRIG_DEV=0, TAB_CLIENTS=1, TAB_VME_TRIGS=2, TAB_LOGS=3};

    explicit RcMainWindow(QWidget *parent = nullptr);
    virtual ~RcMainWindow() override;

public Q_SLOTS:
    void setConfig(const RcConfig &newConfig);
    void configSelectorChanged(const ConfigIndexName &conf);
    void setDeviceList(const DeviceIndexEnabledMap &devices);

    void onRcFsmStateChange(const FsmState &state, const FsmState &previousState);
    void trigManagerRunStateChanged(const FsmState &state);
    void clientManagerRunStateChanged(const FsmState &state);
    void vmeTrigsRunStateChanged(const FsmState &state);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);
    void setRcRunId(const RcRunId &id);
    void dlnpCalibrationTriggerBpUpdated() const;

protected slots:

private Q_SLOTS:
    void init();

    void trigStatusUpdated(const TrigStatus &status); // GUI

    void on_actionFindDevice_triggered();
    void on_actionAllow_Remote_Control_toggled(bool val);
    void on_actionEnableStatusDump_toggled(bool val);

    void on_actionSetStatusDumpDirectory_triggered();

    void on_tabWidget_currentChanged(int index);

    void onSlowCntStatusUpdated(quint16 tcpPort, bool listening, bool portInUse); // GUI

    void on_actionConfiguration_manager_triggered();
    void slaveMessageReceived(const MessageTreeItem &m);

    void on_actionPrint_Fsm_input_triggered();
    void on_actionSet_local_Run_number_triggered();
    void on_actionSlow_control_triggered();
    void on_actionDevice_status_triggered();
    void on_actionCalibration_trigger_triggered();

signals:
    void configChanged(const RcConfig &config);
    void switch_to_config(const ConfigIndexName &conf) const;
    void deviceListChanged(const DeviceIndexEnabledMap &); // unused

    void destroyAll() const;
    void perfomeLocalStart() const;
    void perfomeLocalStop() const;
    void sendRunWarn(const RcRunId &, QString) const;
    void deleteAll() const;
    // void deviceDiscovered(const DeviceDescription &) const;
    void setStatusDumpDirName(const QString &);
    void setStatusDumpEnable(bool);
    void printFsmInput();

protected:
    void closeEvent(QCloseEvent *event) override;

private:
    void updateGuiControlBlock(); // GUI
    void subFsmStateChanged(int tabIndex, const FsmState &state);
    void setTabState(bool isOk, int tabIndex); // GUI
    void logRcFsmStateChange(const FsmState &previousState, const FsmState &state);
//    bool waitVmeTrigsStarted(const QElapsedTimer &startTime);
//    bool waitClientsStarted(const QElapsedTimer &startTime);
    void updateDeviceLabel();
    void updateDeviceStatusLabel(const TrigStatus &status);
    // unused int addClient(QHostAddress host, int port, int trig, int clientIndex=-1);
    void loadSettings();
    void updateStatStopButton();
    void updateGroupBoxWarn();
    void updateRemoteBlocking();
    void clearRunWarning();
    bool isInRun() const;

    Ui::RcMainWindow *ui;

    QLabel * const labelFsmStateRc;
    QLabel * const labelFsmStateTrig;
    QLabel * const labelFsmStateClient;
    QLabel * const labelFsmStateVme;

    QLabel * const labelDevState;
    QLabel * const labelSlowControl;
    QLabel * const labelDevice;
    QLabel * const labelFirmware;
    QLabel * const labelTemp;
    DiscoverDialog * const discoverDialog;
    SlowControlDialog * const slowControlDialog;
    DeviceStatusDialog * const deviceStatusDialog;
    DlnpCalibrationTriggerDialog * const dlnpCalibrationTriggerDialog;

    RcCoreApp * const rcCoreApp;
    RcConfig config;
    RcRunId rcRunId;
    QPointer<ConfigurationManager> configuration_manager;

    __attribute__((deprecated("change to specified BaseFsm state")))
    QMap<int, bool> tabOk;
    FsmState curRcFsmState = FsmState::Init;
    bool underRemoteCtrl = false;
};

#endif // RCMAINWINDOW_H
