/*
**    Copyright 2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcSignalRouter.h"

#include <QObject>
#include "RcMainWindow.h"
#include "RcConfigDb.h"
#include "RcConfigDispatcher.h"
#include "RcCoreApp.h"
#include "SlowControlDialog.h"
#include "SlowControlServer.h"
#include "TrigDevice.h"
#include "TrigDeviceController.h"
#include "TrigMainForm.h"
#include "VmeTrigsController.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-widgets/DlnpCalibrationTrigger/DlnpCalibrationTriggerDialog.h"
#include "pnp-server/PnpDiscovery.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "run-control-fsm/RcFsm.h"
#include "run-control-client-manager/RcClientManager.h"
#include "run-control-client-manager/gui/RcClientManagerWidget.h"
#include "StatusPublisher.h"

namespace {

template <typename Func1, typename Func2>
inline void connectUnique(const typename QtPrivate::FunctionPointer<Func1>::Object *sender, Func1 signal,
                    const typename QtPrivate::FunctionPointer<Func2>::Object *receiver, Func2 slot)
{
    QObject::connect(sender, signal, receiver, slot, Qt::UniqueConnection);
}

template <typename Func1>
static inline bool disconnect(const typename QtPrivate::FunctionPointer<Func1>::Object *sender, Func1 signal)
{
    return QObject::disconnect(sender, signal, nullptr, nullptr);
}
}

RcSignalRouter::RcSignalRouter()
{
}

void RcSignalRouter::connectCore()
{
    Q_ASSERT(rcConfigDispatcher);
    Q_ASSERT(vmeTrigsController);

    connect_Configuration();

    connect_EvNumChecker();
    connect_StatusPublisher();
    connect_PnpServer();
    connect_RcClientManager();
    connect_RemoteControlServer();
    connect_SlowControlServer();
    connect_VmeTrigsController();
    connect_TrigDeviceController();
}
void RcSignalRouter::connectAll()
{
    connectCore(); // again

    connect_RcClientManagerWidget();
    connect_RcMainWindow();
    connect_TrigMainForm();
}

void RcSignalRouter::connect_RcMainWindow()
{
    Q_ASSERT(clientManager);
    Q_ASSERT(mainWindow);
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(remoteControlServer);

    connectUnique(clientManager, &RcClientManager::config_transfer,
            rcCoreApp, &RcCoreApp::config_transfer_from_clients);

    connectUnique(rcCoreApp, &RcCoreApp::sendRunDocumentToRc,
            remoteControlServer, &RemoteControlServer::config_transfer);
    connectUnique(mainWindow, &RcMainWindow::perfomeLocalStart,
                  rcCoreApp, &RcCoreApp::perfomeLocalStart);
    connectUnique(mainWindow, &RcMainWindow::perfomeLocalStop,
                  rcCoreApp, &RcCoreApp::perfomeLocalStop);
    connectUnique(mainWindow, &RcMainWindow::printFsmInput,
                  rcCoreApp, &RcCoreApp::printFsmInput);
    connectUnique(rcCoreApp, &RcCoreApp::rcRunIdChanged,
                  mainWindow, &RcMainWindow::setRcRunId);
}

void RcSignalRouter::connect_StatusPublisher()
{
    Q_ASSERT(statusPublisher);
}

void RcSignalRouter::connect_PnpServer()
{
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(rcFsm);
    Q_ASSERT(vmeTrigsController);

    PnpDiscovery::subscribeMyProgram(rcCoreApp);
    if(clientManager)
        PnpDiscovery::subscribeCacheUpdaed(clientManager);
    PnpDiscovery::subscribeCacheUpdaed(vmeTrigsController);

    connectUnique(rcFsm, &RcFsm::stateChanged,
                  rcCoreApp, &RcCoreApp::onRcFsmStateChange);
    connectUnique(rcCoreApp, &RcCoreApp::printFsmInput,
                  rcFsm, &RcFsm::printFsmInput);
}

void RcSignalRouter::connect_TrigDevice(TrigDevice *trigDevice)
{
    if (!trigDevice)
        return;

//    Q_ASSERT(deviceDispatcher);
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(slowControlServer);
    Q_ASSERT(trigDevice);

    if (trigMainForm) {
        connectUnique(trigDevice, &TrigDevice::firmwareBuildParamsChanged,
                trigMainForm, &TrigMainForm::setFirmwareBuildParams);
    }
    if (mainWindow) {
        connectUnique(trigDevice, &TrigDevice::updateStatus,
                mainWindow, &RcMainWindow::trigStatusUpdated);
    }
    connectUnique(trigDevice, &TrigDevice::updateStatus,
            rcCoreApp, &RcCoreApp::trigStatusUpdated);

    if (trigMainForm) {
        connectUnique(trigDevice, &TrigDevice::updateStatus,
                trigMainForm, &TrigMainForm::trigStatusUpdated);

        connectUnique(trigDevice, &TrigDevice::spillPlotUpdated,
                trigMainForm, &TrigMainForm::spillPlotUpdated);

        connectUnique(trigDevice, &TrigDevice::triggerHistogramUpdated,
                trigMainForm, &TrigMainForm::triggerHistogramUpdated);
    }

    connectUnique(trigDevice, &TrigDevice::updateStatus,
            slowControlServer, &SlowControlServer::trigStatusUpdated);

    connectUnique(trigDevice, &TrigDevice::spillPlotUpdated,
            slowControlServer, &SlowControlServer::spillPlotUpdated);

    connectUnique(trigDevice, &TrigDevice::triggerHistogramUpdated,
            slowControlServer, &SlowControlServer::triggerHistogramUpdated);

    if (mainWindow) {
        connectUnique(mainWindow, &RcMainWindow::setStatusDumpEnable,
                trigDevice, &TrigDevice::setStatusDumpEnable);

        connectUnique(mainWindow, &RcMainWindow::setStatusDumpDirName,
                trigDevice, &TrigDevice::setStatusDumpDirName);
//    connect(trigDevice, &TrigDevice::setupWritten,
//            mainWindow, &RcMainWindow::writeSetup);
    }

    if (trigMainForm) {
//        connect(trigMainForm, &TrigMainForm::trigSetupUpdated,
//                trigDevice, &TrigDevice::writeSetup);

        connectUnique(trigMainForm, &TrigMainForm::clearTrigStats,
                trigDevice, &TrigDevice::clearStats);

        connectUnique(trigMainForm, &TrigMainForm::resetAll,
                trigDevice, &TrigDevice::resetAll);
    }

    connect_EvNumChecker(trigDevice);
//    deviceDispatcher->subscribeHostAddressUpdates(trigDevice->getDeviceIndex(), trigDevice);
    if(statusPublisher) {
        connectUnique(trigDevice, &TrigDevice::updateStatus,
                statusPublisher, &StatusPublisher::onTrigStatusUpdated);

        connectUnique(trigDevice, &TrigDevice::trigSpillStatisticsUpdated,
                statusPublisher, &StatusPublisher::onTrigSpillStatisticsUpdated);
    }
    if(dlnpCalibrationTriggerDialog) {
        connectUnique(trigDevice, &ModularMregDevice::dlnpCalibrationTriggerBpUpdated,
                      dlnpCalibrationTriggerDialog, &DlnpCalibrationTriggerDialog::dlnpCalibrationTriggerBpUpdated);
        if (mainWindow) {
            connectUnique(trigDevice, &ModularMregDevice::dlnpCalibrationTriggerBpUpdated,
                          mainWindow, &RcMainWindow::dlnpCalibrationTriggerBpUpdated);
        }
    }
}

void RcSignalRouter::connect_RemoteControlServer()
{
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(rcFsm);
    Q_ASSERT(remoteControlServer);

    connectUnique(remoteControlServer, &RemoteControlServer::startRequest,
            rcCoreApp, &RcCoreApp::onRemoteStartRequest);

    connectUnique(remoteControlServer, &RemoteControlServer::stopRequest,
            rcCoreApp, &RcCoreApp::onRemoteStopRequest);

    connectUnique(rcCoreApp, &RcCoreApp::sendRemoteState,
            remoteControlServer, &RemoteControlServer::sendState);

    connectUnique(rcCoreApp, &RcCoreApp::sendVmeFlag,
            remoteControlServer, &RemoteControlServer::onVmeFlag);

    connectUnique(rcCoreApp, &RcCoreApp::sendRunError,
            remoteControlServer, &RemoteControlServer::onRunError);

    if (mainWindow) {
        connectUnique(mainWindow, &RcMainWindow::sendRunWarn,
                remoteControlServer, &RemoteControlServer::onRunWarning);
    }

    connectUnique(rcCoreApp, &RcCoreApp::sendBackPressure,
            remoteControlServer, &RemoteControlServer::onBackPressure);

    connectUnique(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
            rcCoreApp, &RcCoreApp::onRemoteStatusUpdated);
    if (mainWindow) {
        connectUnique(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                mainWindow, &RcMainWindow::onRemoteStatusUpdated);
    }
    connectUnique(rcCoreApp, &RcCoreApp::allowRemoteControl,
            remoteControlServer, &RemoteControlServer::startListen);

    if (mainWindow) {
        connectUnique(mainWindow, &RcMainWindow::destroyAll,
                remoteControlServer, &RemoteControlServer::deleteLater);
    }
    connectUnique(rcFsm, &RcFsm::stateChanged,
            remoteControlServer, &RemoteControlServer::onFsmStateChange);

    connectUnique(rcCoreApp, &RcCoreApp::myRcProgramStateChanged,
                  remoteControlServer, &RemoteControlServer::onRcProgramStateChange);
}

void RcSignalRouter::removeRemoteControlServer()
{
    Q_ASSERT(remoteControlServer);
    // TODO: use disconnect(remoteControlServer, nullptr, nullptr, nullptr);
    disconnect(remoteControlServer, &RemoteControlServer::startRequest);
    disconnect(remoteControlServer, &RemoteControlServer::stopRequest);
    disconnect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated);
    remoteControlServer = nullptr;
}

void RcSignalRouter::connect_RcClientManager()
{
    Q_ASSERT(clientManager);
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(rcConfigDispatcher);

    connectUnique(clientManager, &RcClientManager::runWarn,
                  rcCoreApp, &RcCoreApp::onRunWarning);
    connectUnique(clientManager, &RcClientManager::runError,
                  rcCoreApp, &RcCoreApp::onRunError);

    connectUnique(rcCoreApp, &RcCoreApp::runStart,
            clientManager, &RcClientManager::onRestartAll);

    connectUnique(rcCoreApp, &RcCoreApp::runStop,
            clientManager, &RcClientManager::onStopAll);

    connectUnique(clientManager, &RcClientManager::backPressure,
            rcCoreApp, &RcCoreApp::onBackPressure);

    connectUnique(clientManager, &RcClientManager::vmeFlagUpdated,
            rcCoreApp, &RcCoreApp::onVmeFlagUpdated);

    connectUnique(clientManager, &RcClientManager::runStateChanged,
                  rcCoreApp, &RcCoreApp::clientManagerRunStateChanged);

    connectUnique(clientManager, &RcClientManager::runStateChanged,
            rcCoreApp, &RcCoreApp::clientManagerRunStateChanged_);

    connectUnique(clientManager, &RcClientManager::rcProgramStateUpdated,
                  rcCoreApp, &RcCoreApp::clientRcProgramStateChanged);

    connectUnique(rcCoreApp, &RcCoreApp::printFsmInput,
            clientManager, &RcClientManager::printFsmInput);

    connectUnique(rcCoreApp, &RcCoreApp::sendRunDocumentToClients,
            clientManager, &RcClientManager::sendRunDocumentToClients);

    connectUnique(remoteControlServer, &RemoteControlServer::gotRunDocument,
                  clientManager, &RcClientManager::sendRunDocumentToClients);

    connectUnique(remoteControlServer, &RemoteControlServer::gotRcProgramState,
                  clientManager, &RcClientManager::allRcProgramStateChanged);

    connectUnique(rcCoreApp, &RcCoreApp::allRcProgramStateChanged,
                  clientManager, &RcClientManager::allRcProgramStateChanged);

    rcConfigDispatcher->subscribeRcClientConfigMap(clientManager);
}

void RcSignalRouter::connect_RcClientManagerWidget()
{
    if(!clientManagerWidget)
        return;
    Q_ASSERT(rcConfigDispatcher);
    rcConfigDispatcher->subscribeRcClientConfigMap(clientManagerWidget);

    if(clientManager) {
        connectUnique(clientManager, &RcClientManager::clientInfoUpdated,
                      clientManagerWidget, &RcClientManagerWidget::clientInfoUpdated);

        connectUnique(clientManager, &RcClientManager::runStateChanged,
                      clientManagerWidget, &RcClientManagerWidget::onFsmStateChange);
    }
    if(rcCoreApp) {
        connectUnique(rcCoreApp, &RcCoreApp::allRcProgramStateChanged,
                      clientManagerWidget, &RcClientManagerWidget::clientRcProgramStateUpdated);
    }
    if(remoteControlServer) {
        connectUnique(remoteControlServer, &RemoteControlServer::gotRcProgramState,
                      clientManagerWidget, &RcClientManagerWidget::clientRcProgramStateUpdated);
    }
}

void RcSignalRouter::connect_Configuration()
{
    if (mainWindow) {
        rcConfigDispatcher->subscribe(mainWindow);
        rcConfigDispatcher->subscribeCS(mainWindow);
    }
    Q_ASSERT(rcCoreApp);
    rcConfigDispatcher->subscribe(rcCoreApp);
    rcConfigDispatcher->subscribeCS(rcCoreApp);

    if (configurationManager) {
        rcConfigDispatcher->subscribeCS(configurationManager);
    }
    if(clientManagerWidget) {
        connectUnique(rcConfigDispatcher, &RcConfigDispatcher::clientIndexChanged,
                      clientManagerWidget, &RcClientManagerWidget::myClientIndexUpdated);
    }

    Q_ASSERT(configDb);
    rcConfigDispatcher->subscribe(configDb);
    rcConfigDispatcher->subscribeCS(configDb);

    if(dlnpCalibrationTriggerDialog) {
        rcConfigDispatcher->subscribeDlnpCalibrationTriggerConfig(dlnpCalibrationTriggerDialog);
    }
}

void RcSignalRouter::connect_EvNumChecker()
{
    if(!evNumChecker)
        return;

    if(vmeTrigsController) {
        connectUnique(evNumChecker, &EvNumChecker::evNumRequest,
                vmeTrigsController, &VmeTrigsController::evNumRequest);

        connectUnique(vmeTrigsController, &VmeTrigsController::evNumResponse,
                evNumChecker, &EvNumChecker::gotVmeTrigEvNum);

        connectUnique(vmeTrigsController, &VmeTrigsController::vmeMapUpdated,
                evNumChecker, &EvNumChecker::updateVmeMap);
    }

    if(remoteControlServer) {
        connectUnique(remoteControlServer, &RemoteControlServer::evNumCheckRequest,
                evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);

        connectUnique(evNumChecker, &EvNumChecker::completedEvNumCheck,
                remoteControlServer, &RemoteControlServer::evNumCheckResponse);
    }

    if(clientManager) {
        connectUnique(evNumChecker, &EvNumChecker::evNumRequest,
                clientManager, &RcClientManager::evNumRequest);

        connectUnique(clientManager, &RcClientManager::evNumResponse,
                evNumChecker, &EvNumChecker::gotClientEvNum);

        connectUnique(clientManager, &RcClientManager::clientMapUpdated,
                evNumChecker, &EvNumChecker::updateClientMap);
    }
}

void RcSignalRouter::connect_EvNumChecker(const TrigDevice *trigDevice)
{
    if(!evNumChecker || !trigDevice)
        return;

    connectUnique(evNumChecker, &EvNumChecker::getTrigEvNum,
            trigDevice, &TrigDevice::evNumRequest);

    connectUnique(trigDevice, &TrigDevice::topEvNumResponse,
            evNumChecker, &EvNumChecker::gotTopLevelEvNum);

    connectUnique(trigDevice, &TrigDevice::evNumResponse,
            evNumChecker, &EvNumChecker::gotDevEvNum);
}


void RcSignalRouter::connect_SlowControlServer()
{
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(slowControlServer);

    rcConfigDispatcher->subscribeSlowControl(slowControlDialog);

    if(rcConfigDispatcher)
        rcConfigDispatcher->subscribeSlowControl(slowControlServer);

    connectUnique(rcCoreApp, &RcCoreApp::runStart,
            slowControlServer, &SlowControlServer::onRunStart);

    if(slowControlDialog)
        connectUnique(slowControlServer, &SlowControlServer::connectionStatusUpdated,
                slowControlDialog, &SlowControlDialog::onSlowCntStatusUpdated);

    if (mainWindow) {
        connectUnique(mainWindow, &RcMainWindow::destroyAll,
                slowControlServer, &SlowControlServer::deleteLater);

        connectUnique(slowControlServer, &SlowControlServer::connectionStatusUpdated,
                mainWindow, &RcMainWindow::onSlowCntStatusUpdated);
    }

    connectUnique(rcCoreApp, &RcCoreApp::slowControlStateUpdated,
            slowControlServer, &SlowControlServer::slowControlStateUpdated);
}

void RcSignalRouter::connect_TrigMainForm()
{
    Q_ASSERT(mainWindow);
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(trigMainForm);

    if(rcFsm) {
        connectUnique(rcFsm, &RcFsm::stateChanged,
                mainWindow, &RcMainWindow::onRcFsmStateChange);
        connectUnique(rcFsm, &RcFsm::stateChanged,
                trigMainForm, &TrigMainForm::rcFsmStateChanged);
    }
    connectUnique(rcCoreApp, &RcCoreApp::clientManagerRunStateChanged_,
            mainWindow, &RcMainWindow::clientManagerRunStateChanged);
    connectUnique(rcCoreApp, &RcCoreApp::vmeTrigsRunStateChanged_,
            mainWindow, &RcMainWindow::vmeTrigsRunStateChanged);
    connectUnique(rcCoreApp, &RcCoreApp::trigManagerRunStateChanged_,
            mainWindow, &RcMainWindow::trigManagerRunStateChanged);
}

void RcSignalRouter::connect_VmeTrigsController()
{
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(vmeTrigsController);

    connectUnique(vmeTrigsController, &VmeTrigsController::sendSlaveMessage,
            rcCoreApp, &RcCoreApp::sendSlaveMessage);

    connectUnique(rcCoreApp, &RcCoreApp::runStart,
            vmeTrigsController, &VmeTrigsController::onRunStart);

    connectUnique(rcCoreApp, &RcCoreApp::runStop,
            vmeTrigsController, &VmeTrigsController::onRunStop);

    connectUnique(vmeTrigsController, &VmeTrigsController::stateChanged,
            rcCoreApp, &RcCoreApp::vmeTrigsRunStateChanged);

    connectUnique(vmeTrigsController, &VmeTrigsController::stateChanged,
            rcCoreApp, &RcCoreApp::vmeTrigsRunStateChanged_); //proxy

    connectUnique(rcCoreApp, &RcCoreApp::printFsmInput,
            vmeTrigsController, &VmeTrigsController::printFsmInput);
}

void RcSignalRouter::connect_TrigDeviceController()
{
    Q_ASSERT(rcCoreApp);
    Q_ASSERT(trigDeviceController);

    connectUnique(rcCoreApp, &RcCoreApp::runStart,
            trigDeviceController, &TrigDeviceController::onRunStart);
    connectUnique(rcCoreApp, &RcCoreApp::prepareStop,
            trigDeviceController, &TrigDeviceController::onPrepareStop);
    connectUnique(rcCoreApp, &RcCoreApp::runStop,
            trigDeviceController, &TrigDeviceController::onRunStop);
    connectUnique(rcCoreApp, &RcCoreApp::closeGlobTrig,
            trigDeviceController, &TrigDeviceController::onCloseGlobTrig);
    connectUnique(rcCoreApp, &RcCoreApp::openGlobTrig,
            trigDeviceController, &TrigDeviceController::onOpenGlobTrig);
    connectUnique(trigDeviceController, &TrigDeviceController::stateChanged,
            rcCoreApp, &RcCoreApp::trigManagerRunStateChanged);
    connectUnique(trigDeviceController, &TrigDeviceController::stateChanged,
            rcCoreApp, &RcCoreApp::trigManagerRunStateChanged_);
    if (mainWindow) {
        connectUnique(mainWindow, &RcMainWindow::printFsmInput,
                trigDeviceController, &TrigDeviceController::printFsmInput);
    }
}
