#include "TqdcCore.h"

#include <QElapsedTimer>

#include "TqdcSignalRouter.h"

#include "cal/CalResult.h"
#include "cal/CalConfigDb.h"
#include "fsm/AdcDeviceMgrFsm.h"
#include "base/DeviceIndexEnabled.h"
#include "base/RunDocument.h"
#include "base/ThreadController.h"
#include "device-modules/TdcCore/TdcCoreStatus.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "mldiscover/LldpDiscovery.h"
#include "remote-control-server/RemoteControlServerStatus.h"
#include "RunLogic/RunLogicMultiStatus.h"
#include "TqdcTelnetRemoteControl.h"
#include "TqdcStartupOptions.h"

TqdcCore::TqdcCore(const TqdcStartupOptions &so, QObject *parent)
    : ReadoutAppCore(parent)
    , fsm(new AdcDeviceMgrFsm(this))
    , tqdcConfigDb(new ReadoutAppConfigDb(this))
    , calConfigDb(new CalConfigDb(this))
    , telnetRc(so.telnetPort ? new TqdcTelnetRemoteControl(*so.telnetPort) : nullptr)
{
    qRegisterMetaType<CalResult>();
    qRegisterMetaType<CalConfig>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<LldpInfoCache>();
    qRegisterMetaType<RunDocument>();
    qRegisterMetaType<TdcCoreStatus>();

    auto &signalRouter = TqdcSignalRouter::instance();
    signalRouter.setDeviceDispatcher(deviceDispatcher);
    signalRouter.setCalConfigDb(calConfigDb);
    signalRouter.setConfigDb(tqdcConfigDb);
    signalRouter.setCore(this);
    signalRouter.setDeviceMgrFsm(fsm);
    if(telnetRc) {
        signalRouter.setTelnetRemoteControl(telnetRc);
        new ThreadController(telnetRc, this);
    }
    signalRouter.connectAll();

    initTqdcProgDescr();
}

bool TqdcCore::isRun() const
{
    return config.isRun();
}

const BaseFsm *TqdcCore::getFsm() const
{
    return fsm;
}

void TqdcCore::setConfig(const ConfigKey &configKey, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(configKey);
    if (config == newConfig)
        return;
    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;

    const auto toDel = oldSet - newSet;
    const auto toAdd = newSet - oldSet;

    if (!toDel.isEmpty()) {
        for (const auto &i : toDel)
            removeDevice(i);
    }
    for (const auto &i : toAdd) {
        if (!devices.contains(i))
            addDevice(i);
    }
//    cfgDispatcher->sendCache();
    updateTqdcProgDescrOptions();
    emit allowRemoteControl(config.allowRemoteControl);
    if(oldSet != newSet)
        updateMStreamInterfaces();
}

void TqdcCore::performLocalStart()
{
    runStartEvNum.clear();
    fsm->setRequestState(FsmState::Starting);
}

void TqdcCore::performLocalStop()
{
    fsm->setRequestState(FsmState::Stopping);
}

void TqdcCore::remoteStartRequest(const RcRunId &runId_, const QVariant &startEvNum)
{
    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
        return;
    }

    rcRunId = runId_;
    runStartEvNum = startEvNum;
    fsm->setRequestState(FsmState::Starting);
}

void TqdcCore::remoteStopRequest(const RcRunId &runId_, const QVariant &stopEvNum)
{
    Q_UNUSED(stopEvNum)
    if(rcRunId.runIndex != runId_.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); "
                             "remote control runIndex (%2)")
                     .arg(runId_.runIndex)
                     .arg(runId_.runIndex);
    }

    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
    }

    if(fsm->getState() == FsmState::Starting)
        fsm->setRequestState(FsmState::Error);
    if(fsm->getState() == FsmState::Run)
        fsm->setRequestState(FsmState::Stopping);
}

void TqdcCore::remoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool prevUnderRemoteCtrl = underRemoteCtrl;
    ReadoutAppCore::remoteStatusUpdated(status);
    if(underRemoteCtrl != prevUnderRemoteCtrl) {
        if(isRun()) {
            qCritical() << "Stopping current run";
            fsm->setRequestState(fsm->getState() == FsmState::Run
                                     ? FsmState::Stopping : FsmState::Error);
        }
    }
}

void TqdcCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppCore::onFsmStateChange(state, oldState);

    if(telnetRun && *telnetRun) {
        if(state == FsmState::Error || state == FsmState::Ready)
            telnetRun = false;
    }
    switch (state) {
    case FsmState::Error :
        runStartEvNum.clear();
        if(config.isRun()) {
            config.setRunState(false);
            emit configChanged(config);
        }
        break;
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Calibrate :
        case FsmState::FindAdcDelay :
        case FsmState::Stopping:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:{}
        }
        break;
    }
    case FsmState::Starting :
    {
        emit preRunReset(runStartEvNum, rcRunId);
        emit preRunCheck();
        lastDevStatus.clear();
        config.setRunState(true);
        emit configChanged(config);
        fsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Calibrate :
    case FsmState::FindAdcDelay :
    {
        ++cal_key;
        fsm->setMainCalKey(cal_key);
        if(state == FsmState::Calibrate)
            emit startTqdcCalibration(cal_key);
        else
            emit startAdcSrdDesCalibration(cal_key, adcSerDesCalHardMode, adcSerDesCalResDir);
        fsm->setRequestState(FsmState::Ready);
        break;
    }
    case FsmState::Run :
    {
        qInfo() << "Start completed successfully:" << rcRunId.toString();
        sendRunDocument();
        break;
    }
    case FsmState::Stopping :
    {
        runStartEvNum.clear();
        config.setRunState(false);
        emit configChanged(config);
        fsm->setRequestState(FsmState::Ready);
        break;
    }
    default:
    {}
    }
}

void TqdcCore::addDevice(const DeviceIndex &index)
{
    auto *devGlob = new TqdcDeviceGlobal(index, this);
    devices.insert(index, QSharedPointer<TqdcDeviceGlobal>(devGlob));
}

void TqdcCore::removeDevice(const DeviceIndex &index)
{
    devices.remove(index);
}

void TqdcCore::setDeviceStatus(const ModularDeviceStatus &status)
{
    if(status.runStatistic) {
        if(telnetRun && !(*telnetRun)) {
            telnetRunBlockedEv[status.getDeviceIndex()] = status.runStatistic->getCounter(RunStatistic::Blocked);
            auto gotSet = makeDeviceIndexSet(telnetRunBlockedEv.keys());
            if(config.getDeviceIndexEnabledMap().getEnabled() == gotSet) {
                emit timedRunFinished(telnetRunBlockedEv);
                telnetRun.reset();
                telnetRunBlockedEv.clear();
            }
        }
    }

    ReadoutAppCore::setDeviceStatus(status);
}

void TqdcCore::startCalibration_()
{
    fsm->setRequestState(FsmState::Calibrate);
}

void TqdcCore::requestAdcSerDesCalibration(bool hardMode, const QString &dir)
{
    if(fsm->getState() != FsmState::Ready){
        qWarning() << "AdcSerDesCalibration can be performed just in Ready state";
        return;
    }
    adcSerDesCalHardMode = hardMode;
    adcSerDesCalResDir = dir;
    fsm->setRequestState(FsmState::FindAdcDelay);
}

void TqdcCore::startTelnetRun()
{
    if(fsm->getState() == FsmState::Ready) {
        telnetRun = true;
        performLocalStart();
    }
}

void TqdcCore::setRunLogicStatus(const RunLogicMultiStatus &st)
{
    if(st.allFinished() && fsm->getState() == FsmState::Run)
        performLocalStop();
}

void TqdcCore::initTqdcProgDescr()
{
    progDescr.setClientIndex(tqdcConfigDb->getClientIndex());

    //todo: transfer it from libs.pro
#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    emit updateProgramDescription(progDescr);
}

void TqdcCore::updateTqdcProgDescrOptions()
{
    QStringList devList;
    for(const auto &dg : devices)
        devList << dg->getDeviceIndex().toString();
    progDescr.options["DevList"] = devList.join(";");
    progDescr.options["run"] = QVariant(config.isRun()).toString();

    emit updateProgramDescription(progDescr);
}

void TqdcCore::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(rcRunId, tqdcConfigDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}
