#include "TqdcSignalRouter.h"

#include <QObject>

#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"
#include "CalibrationDialog.h"
#include "CalibrationModel.h"
#include "DeviceChTree/DeviceChModel.h"
#include "DeviceChTree/DeviceChTreeView.h"
#include "FirFilter/FirFilterWidget.h"
#include "NetworkPortWidget/NetworkPortConfigDialog.h"
#include "RunLogic/RunLogicController.h"
#include "RunLogic/RunLogicWidget.h"
#include "TqdcTelnetRemoteControl.h"
#include "TqdcConfigDispatcher.h"
#include "TqdcCore.h"
#include "TqdcDevice.h"
#include "TqdcDeviceGlobal.h"
#include "TqdcMainWindow.h"
#include "TrigCtrlStat/TrigCtrlStatWidget.h"
#include "TtlIoWidget/TtlIoWidget.h"
#include "WaveBlc/WaveBlcWidget.h"
#include "base/DeviceId.h"
#include "cal/CalConfigDb.h"
#include "configuration-manager/ConfigurationManager.h"
#include "configuration-manager/ConfigurationManager.h"
#include "fsm/AdcDeviceMgrFsm.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "mldiscover/DeviceDispatcher.h"
#include "remote-control-server/RemoteControlServer.h"

static constexpr const auto device_filter_list = {
    DEVICE_ID_TDC64VHLE,
    DEVICE_ID_TDC64VLE,
    DEVICE_ID_TDC72VXS,
    DEVICE_ID_TQDC16VS_E
};

TqdcSignalRouter::TqdcSignalRouter()
{
    setDeviceIdFilter(device_filter_list);
    setConfigDispatcher(&TqdcConfigDispatcher::instance());
}

void TqdcSignalRouter::setCore(TqdcCore *p)
{
    core = p;
    setReadoutAppCore(p);
}

void TqdcSignalRouter::setMainWindow(TqdcMainWindow *p)
{
    // set mainWindow must be before core
    mainWindow = p;
    setReadoutAppMainWindow(p);
}

void TqdcSignalRouter::connect_ConfigurationDispatcher() const
{
    ReadoutAppSignalRouter::connect_ConfigurationDispatcher();

    auto tqdcConfigDispatcher = &TqdcConfigDispatcher::instance();

    if(core) {
        configDispatcher->subscribeConfig(core);
    }
    if(mainWindow) {
        configDispatcher->subscribeConfig(mainWindow->getDeviceTreeModel());
        tqdcConfigDispatcher->subscribeCalConfig(mainWindow->calibrationDialog->getCalibrationModel());
        auto widgetTtl = mainWindow->getWidgetTtl();
        if(widgetTtl) {
            configDispatcher->subscribeTtlCfg(widgetTtl);
        }
        auto trigCtrlStatWidget = mainWindow->getTrigCtrlStatWidget();
        if (trigCtrlStatWidget) {
            configDispatcher->subscribeTrigCtrlStatConfig(trigCtrlStatWidget);
        }
        if (auto widget = mainWindow->getFirFilterWidget()) {
            configDispatcher->subscribeFirFilterConfig(widget);
        }
        if (auto widget = mainWindow->getWaveBlcWidget()) {
            configDispatcher->subscribeWaveBlcConfig(widget);
        }
    }
    if(tqdcConfigDb) {
        configDispatcher->subscribeConfig(tqdcConfigDb);
        configDispatcher->subscribeCS(tqdcConfigDb);
    }
    if(calConfigDb){
        tqdcConfigDispatcher->subscribeCalConfig(calConfigDb);
    }
}

void TqdcSignalRouter::connect_Core() const
{
    ReadoutAppSignalRouter::connect_Core();

    if(core)
    connect(core->runLogicController, &RunLogicController::runLogicStatusChanged,
            core, &TqdcCore::setRunLogicStatus);
}

void TqdcSignalRouter::connect_deviceDispatcher() const
{
    ReadoutAppSignalRouter::connect_deviceDispatcher();

    if(!deviceDispatcher)
        return;

    if(mainWindow){
        deviceDispatcher->subscribeDevices(mainWindow->calibrationDialog->getCalibrationModel());
        if (auto w = mainWindow->getFirFilterWidget()) {
            connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                    w, &FirFilterWidget::setDeviceList);
        }
        if (auto w = mainWindow->getTrigCtrlStatWidget()) {
            connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                    w, &TrigCtrlStatWidget::setDeviceList);
        }
        if(auto chView = mainWindow->getDeviceChTreeView()) {
            connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                    chView, &DeviceChTreeView::setDeviceList);
        }
    }
    if (configDispatcher) {
        deviceDispatcher->subscribeDevices(configDispatcher);
    }
}

void TqdcSignalRouter::connect_TelnetRemoteControl() const
{
    if(!telnetRc)
        return;

    if(deviceMgrFsm)
        connect(deviceMgrFsm, &BaseFsm::stateChanged,
            telnetRc, &TelnetRemoteControl::onFsmStateChange);

    if(telnetRc) {
        connect(telnetRc, &TelnetRemoteControl::startRun, core, &TqdcCore::startTelnetRun);
        connect(telnetRc, &TelnetRemoteControl::stopRun, core, &TqdcCore::performLocalStop);
        connect(core, &TqdcCore::timedRunFinished, telnetRc, &TelnetRemoteControl::timedRunFinished);
        configDispatcher->subscribeConfig(telnetRc);
    }
}

void TqdcSignalRouter::connect_MainWindow() const
{
    ReadoutAppSignalRouter::connect_MainWindow();
    if(!core)
        return;

    if(adcSerDesCalDialog)
        connect(adcSerDesCalDialog, &AdcSerDesCalDialog::requestCalibration,
                core, &TqdcCore::requestAdcSerDesCalibration); // TODO: move to BaseSignalRouter

    if(!mainWindow)
        return;

    connect(mainWindow, &TqdcMainWindow::closeAll,
            core, &TqdcCore::deleteLater);

    connect(mainWindow, &TqdcMainWindow::performLocalStart,
            core, &TqdcCore::performLocalStart);

    connect(mainWindow, &TqdcMainWindow::performLocalStop,
            core, &TqdcCore::performLocalStop);

    if (auto runLogicWidget = mainWindow->getRunLogicWidget()) {
        runLogicWidget->setController(core->runLogicController);
    }
}

void TqdcSignalRouter::connect_RemoteControlServer() const
{
    ReadoutAppSignalRouter::connect_RemoteControlServer();

    if(!remoteControlServer)
        return;

    if (core) {

        connect(remoteControlServer, &RemoteControlServer::startRequest,
                core, &TqdcCore::remoteStartRequest);
        connect(remoteControlServer, &RemoteControlServer::stopRequest,
                core, &TqdcCore::remoteStopRequest);

        connect(core, &TqdcCore::sendRunError,
                remoteControlServer,
                QOverload<const RcRunId &, QString>::of(&RemoteControlServer::onRunError));

        connect(core, &TqdcCore::sendRunWarning,
                remoteControlServer, &RemoteControlServer::onRunWarning);
        connect(core, &TqdcCore::allowRemoteControl,
                remoteControlServer, &RemoteControlServer::startListen);
    }
    if(mainWindow){
        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                mainWindow->calibrationDialog, &CalibrationDialog::onRemoteStatusUpdated);
    }
}

void TqdcSignalRouter::connect_FSM() const
{
    ReadoutAppSignalRouter::connect_FSM();

    if (!deviceMgrFsm)
        return;

    if(mainWindow){
        connect(deviceMgrFsm, &BaseFsm::stateChanged,
                mainWindow, &TqdcMainWindow::onFsmStateChange);

        connect(deviceMgrFsm, &BaseFsm::stateChanged,
                mainWindow->calibrationDialog, &CalibrationDialog::onFsmStateChange);

        connect(mainWindow, &TqdcMainWindow::printFsmInputs,
                deviceMgrFsm, &BaseFsm::printFsmInput);

        auto widgetTtl = mainWindow->getWidgetTtl();
        if(widgetTtl) {
            connect(deviceMgrFsm, &BaseFsm::stateChanged,
                    widgetTtl, &TtlIoWidget::onFsmStateChange);
        }
        if(auto w = mainWindow->getTrigCtrlStatWidget())
            connect(deviceMgrFsm, &BaseFsm::stateChanged,
                    w, &TrigCtrlStatWidget::onFsmStateChange);
    }
}

void TqdcSignalRouter::connect_TqdcDeviceGlobal(const TqdcDeviceGlobal *devGlob, const TqdcDevice *dev) const
{
    ReadoutAppSignalRouter::connect_device(devGlob, dev);

    if (!devGlob || !dev)
        return;

    connect(dev, &TqdcDevice::statusUpdated,
            core, &TqdcCore::setDeviceStatus);

    if(mainWindow){
        connect(dev, &TqdcDevice::deviceInfoUpdated,
                mainWindow, &TqdcMainWindow::deviceInfoUpdated);

        connect(dev, &TqdcDevice::statusUpdated,
                mainWindow, &TqdcMainWindow::deviceStatusUpdated);

        connect(mainWindow, &TqdcMainWindow::softReset,
                dev, &TqdcDevice::softReset);

        connect(mainWindow, &TqdcMainWindow::mstream_reset_all,
                devGlob, &TqdcDeviceGlobal::resetMStream);

        connect(mainWindow->calibrationDialog, &CalibrationDialog::startCalibration,
                core, &TqdcCore::startCalibration_);

        if(auto w = mainWindow->getWidgetTtl()) {
            connect(dev, &TqdcDevice::ttlIoStartupCfg,
                    w, &TtlIoWidget::setStartUpCfg);
        }

        if (auto w = mainWindow->getTrigCtrlStatWidget()) {
            connect(dev, &ModularMregDevice::trigCtrlBpUpdated,
                    w, &TrigCtrlStatWidget::setTrigCtrlBp);
        }
        if (auto w = mainWindow->getFirFilterWidget()) {
            connect(dev, &ModularMregDevice::firFilterBpUpdated,
                    w, &FirFilterWidget::setFirFilterBp);
        }
    }

    connect(core, &TqdcCore::startTqdcCalibration,
            dev, &TqdcDevice::startTqdcCalibration);
    connect(core, &TqdcCore::startAdcSrdDesCalibration,
            dev, &TqdcDevice::startAdcSrdDesCalibration);
    connect(core, &TqdcCore::preRunReset,
            dev, &TqdcDevice::preRunReset);
    connect(core, &TqdcCore::preRunCheck,
            devGlob, &TqdcDeviceGlobal::preRunCheck);

    auto adcDeviceMgrFsm = qobject_cast<AdcDeviceMgrFsm *>(deviceMgrFsm);
    if (adcDeviceMgrFsm) {
        connect(dev, &TqdcDevice::calibrateFinished,
                adcDeviceMgrFsm, &AdcDeviceMgrFsm::setDevCalKey);
    }

    deviceDispatcher->subscribeHostAddressUpdates(devGlob->getDeviceIndex(),
                                                  dev);
    if (configDispatcher)
        configDispatcher->forwardDeviceConfig(dev);
    auto tqdcConfigDispatcher = &TqdcConfigDispatcher::instance();
    tqdcConfigDispatcher->subscribeCalResult(dev);
}

void TqdcSignalRouter::connectAll() const
{
    ReadoutAppSignalRouter::connectAll();
    connect_TelnetRemoteControl();
}
