#include "TtvxsDevice.h"

#include "TrigCtrlStat/TrigCtrlStatModule.h"
#include "TtvxsRegs.h"

#include "LegacyDeviceModule.h"
#include "StatReadout/StatReadoutModule.h"
#include "TrigInfo/TrigInfoModule.h"
#include "regio/regop.h"

namespace {
static const double TTVXS_CORE_CLOCK_HZ = 125.0e6;
typedef union reg_busy_src {
    quint16 all=0;
    struct {
        unsigned int pulse: 1;
        unsigned int feLink: 1;
        unsigned int unused2: 14;
    } bit;
} reg_busy_src;
}

TtvxsDevice::TtvxsDevice(const DeviceIndex &devIndex, QObject *parent) :
    ModularMregDevice(devIndex, status, parent)
    , status(devIndex)
    , legacyDevice(createModule<LegacyDeviceModule>())
{}

reg_base_ctrl TtvxsDevice::currentBaseCtrl() const
{
    reg_base_ctrl ctrl = {};
    ctrl.bit.run_csr = configCache.ttvxsCore.trig_open;
    return ctrl;
}

reg_rc_ctrl TtvxsDevice::currentRegRcCtrl() const
{
    reg_rc_ctrl rc_ctrl;
    rc_ctrl.bit.en_aux_cnt = configCache.ttvxsCore.aux_cnt_for_trig;
    return rc_ctrl;
}

RegOpVector TtvxsDevice::writeCoreConfigCommand(const TtvxsCoreConfig &cfg) const
{
    RegOpVector r;

    reg_base_ctrl base_ctrl = currentBaseCtrl();
    r.RegWrite16(REG_DEVICE_CTRL, base_ctrl.all);

    r.RegWrite16(REG_RC_CTRL, currentRegRcCtrl().all);

    r.RegWrite16(REG_RC_TRIG_SRC, cfg.trigSrc.all);

    r.RegWrite32(REG32_RC_BUSY_MASK, cfg.vxsChMask);

    reg_busy_src busySource;
    busySource.bit.feLink = cfg.busySource.vxsFeLink ? 1 : 0;
    busySource.bit.pulse = cfg.busySource.vxsPulse ? 1 : 0;
    r.RegWrite16(REG_BUSY_SOURCE, busySource.all);

    return r;
}

bool TtvxsDevice::writeTrigCtrlStatConfig(const TrigCtrlStatConfig &cfg)
{
    RegOpVector r;

    quint32 period = static_cast<quint32>(TTVXS_CORE_CLOCK_HZ/cfg.trigTimerFreq);
    r.RegWrite32(REG32_FIXED_FREQ_PULSER_PERIOD, period);

    double prob = 1.0 * cfg.randomPulserFreq * (quint32)(-1) / TTVXS_CORE_CLOCK_HZ;
    std::clamp(prob, 1.0, (double)TTVXS_CORE_CLOCK_HZ);
    quint32 regValue = static_cast<quint32>(prob);
    r.RegWrite32(REG32_RANDOM_PULSER_PROBABILITY, regValue);

    quint32 deadTime = static_cast<quint32>(cfg.randomPulserMinPeriod * TTVXS_CORE_CLOCK_HZ);
    r.RegWrite32(REG32_RANDOM_PULSER_DEAD_TIME, deadTime);

    return legacyDevice->regOpExecRebased(r);
}

bool TtvxsDevice::writeTtvxsCoreConfig(const TtvxsCoreConfig &cfg)
{
    RegOpVector r;
    r << writeCoreConfigCommand(cfg);
    return legacyDevice->regOpExecRebased(r);
}

bool TtvxsDevice::writeConfig()
{
    ModularMregDevice::writeConfig();
    bool ok = true;
    if (!isConfigCacheValid() || (configCache.trigCtrlStat != config.trigCtrlStat)) {
        ok &= writeTrigCtrlStatConfig(config.trigCtrlStat);
    }
    if (!isConfigCacheValid() || (configCache.ttvxsCore != config.ttvxsCore)) {
        ok &= writeTtvxsCoreConfig(config.ttvxsCore);
    }
    return ok;
}

bool TtvxsDevice::evNumReset(const DeviceIndex &index, const QVariant &startEvNum)
{
    if(!index.isNull() && index!=getDeviceIndex())
        return true;
    return preRunActions(startEvNum);
}

void TtvxsDevice::onRestartAll(const RcRunId &runId, const QVariant &startEvNum)
{
    Q_UNUSED(runId)
    softClear({});
    evNumReset(getDeviceIndex(), startEvNum);
    resetTimestamp();
    clearStat({});
}

bool TtvxsDevice::readStatus()
{
    return
        ModularMregDevice::readStatus() &&
        readTtvxsStatus();
}

bool TtvxsDevice::readTtvxsBuildParams()
{
    RegOpVector r;
    r.RegRead32(REG128_BUILD_PARAMS, &status.buildParams[0]);
    r.RegRead32(REG128_BUILD_PARAMS+2, &status.buildParams[1]);
    r.RegRead32(REG128_BUILD_PARAMS+4, &status.buildParams[2]);
    r.RegRead32(REG128_BUILD_PARAMS+6, &status.buildParams[3]);
    return legacyDevice->regOpExecRebased(r);
}

std::optional<TtvxsDeviceCoreStatus> TtvxsDevice::readTtvxsDeviceCoreStatus()
{
    TtvxsDeviceCoreStatus st;
    RegOpVector r;

    reg_base_ctrl base_ctrl = currentBaseCtrl();
    base_ctrl.bit.run_csr = configCache.ttvxsCore.trig_open;
    base_ctrl.bit.cntLock = 1; // Lock counters
    r.RegWrite16(REG_DEVICE_CTRL, base_ctrl.all);
    r.RegRead16(REG_RC_TRIG_STATE, &st.trigState);
    r.RegRead32(REG32_BUSY_INPUT_STATE, &st.inputState);
    r.RegRead32(REG32_BUSY_TIMEOUT_STATE, &st.timeoutState);
    for (int i=0; i<TtvxsAuxCounters::CNT_NUM; ++i) {
        r.RegRead64(REG_AUX_CNT_BASE+i*4, &st.aux_cnt.data[i]);
    }
    r.RegRead16(REG_TRIG_QUEUE, &st.trigQueue);

    base_ctrl.bit.cntLock = 0; // Unlock counters
    r.RegWrite16(REG_DEVICE_CTRL, base_ctrl.all);

    if (!legacyDevice->regOpExecRebased(r))
        return {};
    return st;
}

bool TtvxsDevice::readTtvxsStatus()
{
    auto coreStatusOpt = readTtvxsDeviceCoreStatus();
    if (!coreStatusOpt.has_value())
        return false;
    status.coreStatus = coreStatusOpt.value();
    emit statusUpdated(status);
    return true;
}

QString TtvxsDevice::networkPortName(int portIndex)
{
    switch(portIndex) {
    case NetworkPortType::SFP0: return "sfp0";
    case NetworkPortType::SFP1: return "sfp1";
    case NetworkPortType::SFP2: return "sfp2";
    case NetworkPortType::SFP3: return "sfp3";
    }
    return {};
}

void TtvxsDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();

    resetTimestamp();
    evNumReset();
    setDeviceConfig(config_key, configCache);
}

void TtvxsDevice::on_deviceDisconnected()
{
    ModularMregDevice::on_deviceDisconnected();
    emit statusUpdated(status);
}

bool TtvxsDevice::resetTimestamp(const DeviceIndex &i)
{
    if (!i.isNull() && i != getDeviceIndex())
        return true;

    reg_rc_ctrl regVal = currentRegRcCtrl();
    RegOpVector r;
    regVal.bit.sendTimeReset = 0;
    r.RegWrite16(REG_RC_CTRL, regVal.all);
    regVal.bit.sendTimeReset = 1;
    r.RegWrite16(REG_RC_CTRL, regVal.all);
    return legacyDevice->regOpExecRebased(r);
}

bool TtvxsDevice::softClear(const DeviceIndex &i)
{
    if (!i.isNull() && i != getDeviceIndex())
        return true;

    dropConfigCache();

    RegOpVector r;
    reg_base_ctrl base_ctrl = currentBaseCtrl();
    base_ctrl.bit.run_csr = config.ttvxsCore.trig_open;
    base_ctrl.bit.softClear = 1;
    r.RegWrite16(REG_DEVICE_CTRL, base_ctrl.all);
    base_ctrl.bit.softClear = 0;
    r.RegWrite16(REG_DEVICE_CTRL, base_ctrl.all);

    return legacyDevice->regOpExecRebased(r);
}

bool TtvxsDevice::clearStat(DeviceIndex i)
{
    // TODO remember up-time;
    if(!i.isNull() && i != getDeviceIndex()) return true;

    auto rc_ctrl = currentRegRcCtrl();

    RegOpVector r;

    rc_ctrl.bit.clearStat = 1;
    r.RegWrite16(REG_RC_CTRL, rc_ctrl.all);

    rc_ctrl.bit.clearStat = 0;
    r.RegWrite16(REG_RC_CTRL, rc_ctrl.all);

    return legacyDevice->regOpExecRebased(r);
}

bool TtvxsDevice::getTrigEvNum(int check_index, bool top_level)
{
    auto trigInfo = getFirstModule<TrigInfoModule>();
    if (!trigInfo)
        return false;
    auto st = trigInfo->readStatus();
    if(!st)
        return false;

    auto evNum = st->eventNumber;
    if (top_level)
        emit topEvNumResponse(check_index, evNum);
    else
        emit evNumResponse(check_index, getDeviceIndex(), evNum);
    return true;
}
