#ifndef TTVXSMAINWINDOW_H
#define TTVXSMAINWINDOW_H

#include <QProcess>
#include <QSharedPointer>
#include <optional>

#include "app/BaseMainWindow.h"
#include "base/ConfigKey.h"
#include "base/DeviceIndexEnabled.h"
#include "base/ProgramIndex.h"
#include "fsm/FsmState.h"
#include "lib-common/RunControlAppConfig.h"
#include "pnp-server/PNPMessage.h"

namespace Ui {
class TtvxsMainWindow;
}

class RcClientManager;
class ConfigIndexName;
class ConfigSelector;
class ConfigurationManager;
class DeviceDispatcher;
class TtvxsDeviceGlobal;
class DeviceIndex;
class DiscoverDialog;
class EvNumChecker;
class DeviceStatusDialog;
class FeLinkStatusDialog;
class LldpInfo;
class LldpInfoCache;
class QLabel;
class QTableWidget;
class QTableWidgetItem;
struct RcRunId;
class RemoteControlServer;
class RemoteControlServerStatus;
struct RunDocument;
class TtvxsConfig;
class TtvxsConfigDb;
class TtvxsStatus;
class RcFsm;
class TrigDeviceMgrFsm;

class TtvxsMainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit TtvxsMainWindow(QWidget *parent = nullptr);
    ~TtvxsMainWindow() override;

signals:
    void switch_to_config(const ConfigIndexName &conf) const;
    void configChanged(const RunControlAppConfig &config);

    //=== signals to dev
    void closeAll();
    void resetTimestamp(DeviceIndex=DeviceIndex());
    void clearStat(DeviceIndex=DeviceIndex());
    void evNumReset(DeviceIndex=DeviceIndex(), const QVariant &startEvNum = {});
    void softClear(DeviceIndex=DeviceIndex());
    void restartAll(const RcRunId &runId, const QVariant &startEvNum = {}) const;

    void updateProgramDescription(const ProgramDescription &) const;
    void closeProgram(const ProgramDescription &) const;
    void allowRemoteControl(bool) const;
    void sendRemoteState(QString) const;
    void sendRunError(const RcRunId &, QString) const;
    void sendRunWarn(const RcRunId &, QString) const;
    void sendBackPressure(const RcRunId &, bool) const;
    void write_history_config(const RcRunId &,
                              const QJsonObject &,
                              bool megamaster_flg) const;

    void sendRunDocumentToRc(const RunDocument &data);
    void stopAll(quint64 stopEvNum=0) const;
    void remoteStartCompleted(bool, const RcRunId &) const;
    void remoteStopCompleted(bool, const RcRunId &) const;
    void set_run_state(bool run);
    void printFsmInput() const;

public slots:
    void setConfig(ConfigKey config_key, const RunControlAppConfig &config_);
    void configSelectorChanged(const ConfigIndexName &conf) override;
    void lldpCacheUpdated(const LldpInfoCache &lldpCache);
    void onRcFsmStateChange(const FsmState &state, const FsmState &oldState);
    void onTrigFsmStateChange(const FsmState &state);
    void clientManagerRunStateChanged(const FsmState &state);

private:
    void init_GUI_style();
    void init_GUI_default_settings();
    void initTableDev();
    void initStatPresets();
    void writeSetup(DeviceIndex index=DeviceIndex());
    void addDevice(const DeviceIndex &index);
    void disconnectFromDev(const DeviceIndex &index);
    void updateGui();
    int getChCfgCol(const DeviceIndex &i) const;
    DeviceIndex getDevByTableItem(const QTableWidgetItem *item) const;
    DeviceIndex getDevByCol(int col) const;
    bool isSlotBlocked(int row);
    bool performStart(const RcRunId &runId);
    void openGlobTrig();
    void closeGlobTrig();
    void setTrigState(bool open);
    void actionOnStateRun();
    void perfomeStop(bool isError);
    bool isReadForRun() const;
    void refreshProgramDescription();
    void updateControlEnable();
    void updateStatStopButton();
    void updateGroupBoxWarn();
    void clearWarnings();
    void logRcFsmStateChange(const FsmState &state, const FsmState &oldState);
    void block_widgets_signal(bool block);
    bool isReadyErrorSt() const;
    void checkCanSendRunDocument();
    ClientIndexSet getMissingConfigClients() const;
    bool isMegamaster() const;
    void setMegamaster(bool newMegamaster_flg);

private slots:
    void sendRunDocument();
    void statusUpdated(const TtvxsStatus &st, const TtvxsStatus &prev_st);
    void updateDeviceLldpInfo(const DeviceIndex &index, const LldpInfo &lldp);
    int get_col_by_serial(QString serial);
    static void update_table_size(QTableWidget * const table, bool widthOnly=false);
    void onRunError(const RcRunId &runId, QString errStr);
    void onRunWarning(const RcRunId &runId, QString warnStr);
    void onBackPressure(const RcRunId &runId, bool backPressure);
    void onEvNumCheckFail(QString mess);
    void setTabState(bool isOk, int tabIndex);
    void config_transfer_from_clients(const ClientIndex &id, const QJsonObject &json);
    void onRemoteStartRequest(const RcRunId &runId, const QVariant &startEvNumber);
    void onRemoteStopRequest(const RcRunId &runId, const QVariant &stopEvNumber);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);

    //GUI slots
    void on_actionSelect_devices_triggered();
    void on_actionReset_timestamp_triggered();
    void on_actionSoftclear_triggered();
    void on_actionClear_stats_triggered();
    void on_actionSNMP_dialog_triggered();
    void on_actionFeLinkStatusDialog_triggered();
    void on_actionEvent_number_reset_triggered();
    void on_actionAllow_Remote_Control_toggled(bool val);
    void on_pushButton_start_clicked();
    void on_pushButton_stop_clicked();
    void on_radioButton_open_toggled(bool checked);
    void on_checkBox_external_ttl_toggled(bool checked);
    void on_checkBox_feLinkSfp_toggled(bool checked);
    void on_checkBox_timer_toggled(bool checked);
    void on_checkBox_random_toggled(bool checked);
    void on_spinBox_timer_valueChanged(int val);
    void on_spinBox_random_valueChanged(int val);
    void on_doubleSpinBox_dead_time_valueChanged(double val);
    void on_checkBox_aux_cnt_for_trig_toggled(bool checked);
    void on_groupBoxStat_toggled(bool checked);
    void on_spinBoxStatPeriod_valueChanged(int period_ms);
    void on_comboBoxPreset_currentIndexChanged(int index);
    void on_tableWidgetDev_cellChanged(int row, int column);
    void on_actionConfiguration_manager_triggered();
    void on_actionDevice_Status_triggered();
    void on_tableWidgetDev_currentItemChanged(QTableWidgetItem *current, QTableWidgetItem *previous);
    void on_tableWidgetDev_itemDoubleClicked(QTableWidgetItem *item);

private:
    Ui::TtvxsMainWindow *ui;
    ConfigurationManager * const configuration_manager;
    QString program_index;
    RunControlAppConfig config;
    RcRunId rcRunId;
    DeviceIndexEnabledMap devMap;
    ProgramIndex localProgramIndex;
    TtvxsConfigDb * const configDb;
    QMap<ClientIndex, QJsonObject> clients_config;
    DiscoverDialog * const discoverDialog;
    DeviceStatusDialog * const deviceStatusDialog;
    FeLinkStatusDialog * const feLinkStatusDialog;
    RcClientManager * const clientCore;
    DeviceDispatcher * const deviceDispatcher;
    ProgramDescription progDesc;
    RemoteControlServer * const remoteControlServer;
    bool remoteConnected;
    EvNumChecker * const evNumChecker;
    QMap<DeviceIndex, QSharedPointer<TtvxsDeviceGlobal> > devices;
    QTimer * const writeRunHistoryTimer;
    QMap<int, bool> tabOk;
    bool megamaster = false;

    bool lookingForStopEvNum = false;
    QVariant startEvNum;
    QVariant stopEvNum;

    RcFsm * const rc_fsm;
    TrigDeviceMgrFsm * const trig_fsm;

    QLabel *labelWarnTtvxs;
    QLabel *labelWarnClients;
    QLabel *labelWarnOther;

    QLabel * const labelFsmStateRc;
    QLabel * const labelFsmStateTrig;
    QLabel * const labelFsmStateClient;

    QList<QObject*> widgets_to_block;
};


#endif // TTVXSMAINWINDOW_H
