/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DaqBaseApp.h"

#include <QCommandLineParser>
#include <QCoreApplication>
#include <QSettings>

#include "base/Globals.h"
#include "base/SyslogMessageSender.h"
#include "AppVersion.h"

static void initLibraries()
{
//    qInfo() << __PRETTY_FUNCTION__;
}

Q_COREAPP_STARTUP_FUNCTION(initLibraries);

DaqBaseApp::DaqBaseApp(int &_argc, char *_argv[])
    :
    argc(_argc),
    argv(_argv),
    nosyslogOption("no-syslog", QCoreApplication::translate("main", "Disable logging to syslog.")),
    debugOption({"d", "debug"}, "Run in debug mode"),
    commonCfgFileOption("common-config", "Common config file path", "commonCfgFile"),
    configNameOption("config",
                     QCoreApplication::translate("main", "Startup with certain configuration"),
                     "[config]"),
    configListOption("config-list", "List available configurations"),
    // will be overwritten in DaqBaseApp::parseOptions()
    versionOption({"v", "version"}, "Current program version"),
    helpOption("h", "help")
{
    QCoreApplication::setApplicationVersion(getAppVersion());
}

void DaqBaseApp::init()
{
    setupOptions();
    parseOptions();

    app.reset(createApplication());

    QCoreApplication::setOrganizationName(Globals::organizationName);
    QCoreApplication::setOrganizationDomain(Globals::organizationDomain);
    QCoreApplication::setApplicationName(Globals::getDaqModule().prettyName());


    if (! option_no_syslog)
        SyslogMessageSender::instance().install();

    QSettings::setDefaultFormat(QSettings::IniFormat);
}

void DaqBaseApp::setupOptions()
{
    options << nosyslogOption;
    options << debugOption;
    options << commonCfgFileOption;
    options << configNameOption;
    options << configListOption;
}

void DaqBaseApp::processParseResults(QCommandLineParser &parser)
{
    if (parser.isSet(helpOption) || parser.isSet("help-all")) {
        new QCoreApplication(argc, argv);
        parser.showHelp();
        Q_UNREACHABLE();
    }

    if (parser.isSet(versionOption)) {
//        parser.showVersion();
        showVersion();
        Q_UNREACHABLE();
    }

    Globals::setDebugMode(parser.isSet(debugOption));
    option_no_syslog = parser.isSet(nosyslogOption);
    option_config_list = parser.isSet(configListOption);

    if(parser.isSet(commonCfgFileOption)) {
        const auto &cfgFile = parser.value(commonCfgFileOption);
        Globals::setCommonConfPath(cfgFile);
    }

    if(hasConfigSelectorOption()) {
        const QStringList args = parser.positionalArguments();
        if (!args.empty()) {
            const auto &programIndex = ProgramIndex::fromString(args[0]);
            Globals::setInitialProgramIndex(programIndex);
        }
        if(parser.isSet(configNameOption)) {
            Globals::setInitialConfigurationName(parser.value(configNameOption));
        }
    }
}

QCoreApplication *DaqBaseApp::createApplication() const
{
    return new QCoreApplication(argc, argv);
}

int DaqBaseApp::exec()
{
    Q_ASSERT(app);
    return app->exec();
}

bool DaqBaseApp::hasConfigSelectorOption() const
{
    return true;
}

void DaqBaseApp::showVersion() const
{
    const auto &program_version = QCoreApplication::applicationVersion();
    if (program_version.isEmpty()) {
        fprintf(stderr, "\033[91m" "Unknown program version" "\033[0m\n");
    } else {
        fprintf(stdout, "\033[32m" "program version: %s" "\033[0m\n",
                program_version.toStdString().c_str());
    }
    exit(0);
    Q_UNREACHABLE();
}

void DaqBaseApp::parseOptions()
{
    QCommandLineParser parser;
    parser.setApplicationDescription(QCoreApplication::translate(
        "DaqBaseApp", "An example demonstrating the class DaqBaseApp."
        ));
    if(hasConfigSelectorOption())
        parser.addPositionalArgument("programIndex", "set application's programIndex", "[programIndex]");
    for (const auto &option : options)
        parser.addOption(option);

    versionOption = parser.addVersionOption();
    helpOption = parser.addHelpOption();

    QStringList args;
    for(int i=0; i< argc; ++i)
        args << QString(argv[i]);
    parser.parse(args);
    processParseResults(parser);
}
