/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef IPV4PREFIX_H
#define IPV4PREFIX_H

#include <QString>
#include <QHostAddress>
#include <QList>

class QDebug;

struct Ipv4Prefix {
    Ipv4Prefix(quint32 ip = 0, int maskbits = 32)
        : _maskbits(maskbits), _mask(makemask(maskbits)), _ip(ip & _mask) { }
    Ipv4Prefix(const QString &s)
        : Ipv4Prefix(parse(s)) { }

    bool operator==(const Ipv4Prefix &r) const {
        return r._ip == _ip && r._maskbits == _maskbits;
    }
    bool operator!=(const Ipv4Prefix &r) const {
        return !(*this == r);
    }
    bool operator<(const Ipv4Prefix &r) const {
        if (_ip == r._ip)
            return _mask < r. _mask;
        return _ip < r._ip;
    }
    bool operator>(const Ipv4Prefix &r) const {
        return (*this == r) ? false : !(*this < r);
    }
    operator QString() const;
    static Ipv4Prefix parse(const QString &s);
    static quint32 makemask(int bits) {
        if (bits <= 0)
            return 0;
        if (bits > 32)
            bits = 32;
        return ~((1 << (32 - bits)) - 1);
    }
    quint32 ip() const { return _ip; }
    quint32 mask() const { return _mask; }
    int maskbits() const { return _maskbits; }

private:
    int _maskbits = 32;
    quint32 _mask = 0xFFFFFFFFu;
    quint32 _ip = 0;
};

class Ipv4PrefixList
{
public:
    bool isEmpty() const { return list.isEmpty(); }
    bool contains(const Ipv4Prefix &r) const { return list.contains(r); }
    void insert(const Ipv4Prefix &r) { list.append(r); }
    bool match(quint32 ip) const;
    Ipv4Prefix findLongestMatch(quint32 ip, bool *ok = nullptr) const;
    bool operator==(const Ipv4PrefixList &r) const {
        return r.list == list;
    }
    bool operator!=(const Ipv4PrefixList &r) const {
        return !(*this == r);
    }
    friend QDebug operator<<(QDebug dbg, const Ipv4PrefixList &r);

private:
    QList<Ipv4Prefix> list;
};


#endif // IPV4PREFIX_H
