/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef DEVICEDISPATCHER_H
#define DEVICEDISPATCHER_H

#include "base/AbstractDispatcher.h"
#include "base/DeviceDispatcherFilter.h"
#include "base/DeviceIndexEnabled.h"

class QHostAddress;
class LldpInfo;
class LldpInfoCache;

class DeviceDispatcher : public AbstractDispatcher
{
    Q_OBJECT

public:
    DeviceDispatcher(QObject *parent = nullptr);
    ~DeviceDispatcher() override;
    DeviceIndexSet getDevices() const { return DeviceIndexSet(devices); }
    QString deviceIdent(const DeviceIndex &index) const;

    template<class T>
    void subscribeDevices(T *obj) {
        connect(obj, &T::deviceListChanged,
                this, &DeviceDispatcher::setDeviceList,
                Qt::UniqueConnection);
        connect(this, &DeviceDispatcher::deviceListChanged,
                obj, &T::setDeviceList,
                Qt::UniqueConnection);
        logConnection(obj, "DeviceList");
    }

    template<class T>
    void subscribeHostAddressUpdates(const DeviceIndex &index, T *obj)
    {
        auto filter = new DeviceDispatcherFilter(index, this);
        connect(this, &DeviceDispatcher::setHostAddress,
                filter, &DeviceDispatcherFilter::setHostAddress,
                Qt::UniqueConnection);
        connect(filter, &DeviceDispatcherFilter::setHostAddressFiltered,
                obj, &T::setHostAddress,
                Qt::UniqueConnection);
        connect(obj, &T::destroyed,
                filter, &QObject::deleteLater);
        connect(this, &DeviceDispatcher::unsubscribeHostAddressUpdates,
                filter, &DeviceDispatcherFilter::destroy,
                Qt::UniqueConnection);
        logConnection(obj, "HostAddress");
        tryConnect(index);
    }

signals:
    void setHostAddress(const DeviceIndex &, const QHostAddress &, int ttl);
    void deviceListChanged(const DeviceIndexEnabledMap &);
    void unsubscribeHostAddressUpdates(const DeviceIndex &i, QObject *_parent = nullptr) const;

public slots:
    void lldpCacheUpdated(const LldpInfoCache &new_cache);
    void setDeviceList(const DeviceIndexEnabledMap &new_devices);

private:
    void logDevicesChange(const DeviceIndexSet &set, QString action);
    void logDeviceStateChange(const DeviceIndex &index, bool discovered);
    void logIpAddressChange(const DeviceIndex &index, const QHostAddress &old_ip, const QHostAddress &new_ip);
    void tryConnect(const DeviceIndex &index);
    void tryConnect(const DeviceIndex &index, const LldpInfo &info);
    static bool isMyAddress(const QHostAddress &address);
    bool isDeviceEnabled(const DeviceIndex &index) const;

    DeviceIndexEnabledMap devices;
    LldpInfoCache * const cache;
};

#endif // DEVICEDISPATCHER_H
