/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "LldpDiscovery_p.h"

#include <QDebug>
#include <QSet>
#include <QTimer>

#include "LldpDiscovery.h"
#include "MldpListener.h"

const int validateCacheTimeMs = 1000;
const qint64 throttleTimeMs = 200;

LldpDiscoveryPrivate::LldpDiscoveryPrivate(QObject *parent)
    : QObject(parent),
    mldpListener(new MldpListener(this)),
    cacheValidateTimer(new QTimer(this)),
    throttleTimer(new QTimer(this))
{
    connect(mldpListener, &MldpListener::receivedLldpInfo,
            this, &LldpDiscoveryPrivate::receivedLldpInfo);

    connect(cacheValidateTimer, &QTimer::timeout,
            this, &LldpDiscoveryPrivate::validateCache);

    cacheValidateTimer->start(validateCacheTimeMs);

    connect(throttleTimer, &QTimer::timeout,
            this, &LldpDiscoveryPrivate::throttleTimeout);

    connect(this, &LldpDiscoveryPrivate::setDeviceIdFilter,
            mldpListener, &MldpListener::setDeviceIdFilter);

    throttleTimer->setSingleShot(true);

    sendUpdates(); // send initial empty
}

void LldpDiscoveryPrivate::receivedLldpInfo(const LldpInfo &lldpInfo)
{
    if (lldpInfo.isNull())
        return;
    const DeviceIndex &index = lldpInfo.getDeviceIndex();
    if (index.isNull())
        return;
    map.insert(index, lldpInfo);
    requestUpdate();
}

void LldpDiscoveryPrivate::validateCache()
{
    QSet<DeviceIndex> expired;
    for (auto it = map.begin(); it != map.end(); ++it) {
        if (it.value().hasExpired())
            expired.insert(it.key());
    }
    bool modified = false;
    for (auto it = expired.constBegin(); it != expired.constEnd(); ++it) {
        map.remove(*it);
        modified = true;
    }
    if (modified)
        requestUpdate();
}

void LldpDiscoveryPrivate::requestUpdate()
{
    mapModified = true;
    qint64 t = updateTime.elapsed() - throttleTimeMs;
    if (t < 0) {
        throttleTimer->start(-t);
    } else {
        throttleTimeout();
    }
}

void LldpDiscoveryPrivate::throttleTimeout()
{
    sendUpdates();
}

void LldpDiscoveryPrivate::sendUpdates()
{
    LldpInfoCache cache;
    for (auto it = map.begin(); it != map.end(); ++it) {
        cache.insert(it.key(), it.value());
    }
    emit cacheUpdated(cache);

    static QSet<DeviceIndex> prevDevSet;
    const auto list = cache.keys();
    const auto curDevSet = makeDeviceIndexSet(list);
    if (curDevSet != prevDevSet) {
        const auto &discoverSet = curDevSet - prevDevSet;
        const auto &expireSet = prevDevSet - curDevSet;
        for (auto it = discoverSet.begin(); it != discoverSet.end(); ++it) {
            emit deviceStateChanged(DeviceStatePair(*it, DeviceDiscovered));
        }
        for (auto it = expireSet.begin(); it != expireSet.end(); ++it) {
            emit deviceStateChanged(DeviceStatePair(*it, DeviceExpired));
        }
        prevDevSet = curDevSet;
    }

    updateTime.restart();
}
