#include "MldpListener.h"

#include <arpa/inet.h>

#include <QHostAddress>

#include "base/DeviceId.h"
#include "multicast-listener/MultiCastListener.h"
#include "mldp.h"

namespace {
const quint16 MLDP_UDP_PORT = 33303;
const char *MLDP_MULTICAST_IP_ADDR = "239.192.1.1";

static const quint32 OUI_TIA_0012BB = 0x0012bb;
static const quint32 OUI_AFI = 0x02a6b8;

#if 0x41424344UL != ('ABCD')
#error "big endian architecture support not implemented"
#endif

typedef uint64_t uint48_t;
uint48_t ntoh_48(uint48_t x) {
    return ((((x) & 0x0000000000ff) << 40) |    \
            (((x) & 0x00000000ff00) << 24) |    \
            (((x) & 0x000000ff0000) << 8)  |    \
            (((x) & 0x0000ff000000) >> 8)  |    \
            (((x) & 0x00ff00000000) >> 24) |    \
            (((x) & 0xff0000000000) >> 40));
}

QString formatSerialNumber(const QString &input)
{
    QString s = input.trimmed().replace("0000-", "");
    return s.toUpper();
}

QString getString(const QByteArray& data, int startPos){
    return QString::fromLatin1(data.data()+startPos,data.size()-startPos);
}

QString hostLogPrefix(const QHostAddress &host) {
    return QString("[src_ip=\"%1\"] ").arg(host.toString());
}
}

MldpListener::MldpListener(QObject *parent) :
    MldpListener(QHostAddress(MLDP_MULTICAST_IP_ADDR), MLDP_UDP_PORT, parent)
{}

MldpListener::MldpListener(const QHostAddress &_host, quint16 _port, QObject *parent) :
    QObject(parent),
    rawListener(new MultiCastListener(_host, _port, this))
{
    qRegisterMetaType<LldpInfo>();
    connect(rawListener, &MultiCastListener::received,
            this, &MldpListener::analyzeData);
}

// see IEEE 802.1AB for more details
void MldpListener::analyzeData(const QByteArray &data, const QHostAddress &host) const
{
    bool ok;
    auto lldp = decodeData(data, host, &ok);
    if(ok) {
        send_info(lldp);
    }
}

QString errStr(const QHostAddress &host, const QString &msg)
{
    return QString("%1 %2").arg(hostLogPrefix(host), msg);
}

LldpInfo MldpListener::decodeData(const QByteArray &data, const QHostAddress &host, bool *ok)
{
    int size = data.size();

    LldpInfo mldp;
    bool &_ok = mldp.valid = true;

    for (int pos=0;pos+1<size;){
        quint8 lldpType = (quint8)data.at(pos) >> 1; // first 7 bits
        // following 9 bits
        quint16 lldpLen = (((quint16)data.at(pos) & 1) << 8) + (quint8)data.at(pos+1);
        pos += 2;

        if (lldpLen==0) continue;
        if (pos+lldpLen > data.size()){
            qWarning() << errStr(host, "Invalid length in TLV format or it is a truncated block");
            _ok = false;
            break;
        }

        QByteArray block = QByteArray(data.data()+pos, lldpLen);
        pos += lldpLen;

        switch (lldpType) {
        case 1:
        {
            if (block.size() < 1) {
                _ok = false;
                qWarning() << errStr(host, "Invalid TLV 1 format");
                break;
            }
            quint8 tlvSubType = (quint8) block.at(0);
            if (tlvSubType == 4 && block.size() == 1+6) {
                mldp.common.haddr = 0;
                for (int j=1;j<7;++j) {
                    quint64 v = (quint8)block.at(j);
                    mldp.common.haddr += v << (6-j)*8;
                }
            } else {
                qWarning() << errStr(host, "Invalid TLV 1 format (chassisID is not a MAC)");
            }
            break;
        }
        case 2: // Port ID
        {
            if (block.size() < 1) {
                _ok = false;
                qWarning() << errStr(host, "Invalid TLV 2 format");
                break;
            }
            quint8 tlvSubType = (quint8) block.at(0);
            QString portID = QString(block.mid(1, -1));
            // in our case tlvSubType=7 (Locally assigned)
            if(tlvSubType==7)
                mldp.common.portID = portID;
            break;
        }
        case 3: // TTL
        {
            if (block.size() != 2) {
                _ok = false;
                qWarning() << errStr(host, "Invalid TLV 3 format");
                break;
            }
            quint32 mldpTtl = ((quint16)block.at(0)) << 8;
            mldpTtl += block.at(1);
            mldp.common.ttl = mldpTtl;
            break;
        }
        case 5: // System Name
        {
            // string that indicates the system's administratively assigned name
//            QString sys_name(QString(block).trimmed()); // "u40ve-rc056959F2"
            break;
        }
        case 6: // System Description
        {
            // textual description of the network entity
//            QString sys_desc(QString(block).trimmed()); // "UT24VE"
            break;
        }
        case 7: // System Capabilities
        {
            if (block.size() == 4) {
//                quint16 sysCap = ((quint16)block.at(1) << 8) + (quint16)block.at(0);
//                quint16 enCap = ((quint16)block.at(3) << 8) + (quint16)block.at(2);
//                qDebug()<<QString("[System Capabilities]: %1").arg(sysCap, 2, 16);
//                qDebug()<<QString("[Enabled Capabilities]: %1").arg(enCap, 2, 16);
            } else {
                qWarning() << errStr(host, "Invalid TLV 7 format");
            }
            break;
        }
        case 8: // Management Address
        {
            if (block.size() <= 9) {
                _ok = false;
                qWarning() << errStr(host, "Invalid TLV 8 format");
                break;
            }
            quint8 mngAddrLen = block.at(0);
            quint8 mngAddrSubtype = block.at(1);
            QByteArray mngAddrA = block.mid(2, mngAddrLen-1);
//            quint8 intNumSubtype = block.mid(1+mngAddrLen, 1).at(0);
//            QByteArray intNumA = block.mid(1+mngAddrLen+1, 4);
            quint8 oidLen = block.mid(1+mngAddrLen+1+4, 1).at(0);
//            QString oid = block.mid(1+mngAddrLen+1+4+1);

            if (block.size() != mngAddrLen + oidLen + 7) {
                _ok = false;
                qWarning() << errStr(host, "Invalid TLV 8 format");
                break;
            }
            if (mngAddrSubtype == 1 && mngAddrLen == 5) {
                mldp.common.ip_addr = 0;
                for (int j=0;j<4;++j)
                    mldp.common.ip_addr += (quint8)mngAddrA.at(j) << ((3-j)*8);
            } else {
                _ok = false;
                qWarning() << errStr(host, "Invalid TLV 8 format");
                break;
            }
            // should be same as portId field
//            quint32 intNum = 0;
//            for (int j=0; j<4; j++)
//                intNum += (quint32)intNumA.at(j) << (3-j)*8;
            break;
        }
        case 127: // Organizationally Specific
        {
            if (block.size() < 4) {
                _ok = false;
                qWarning() << errStr(host, "Invalid TLV 127 format");
                break;
            }
            // Organizationally Unique Identifier
            quint32 oui = ((quint8)block.at(0)<<16)+((quint8)block.at(1)<<8)+((quint8)block.at(2));
            quint8 subtype = (quint8)block.at(3);
            const QByteArray data(block.data() + 4, block.size() - 4);
            switch (oui) {
            case OUI_TIA_0012BB: {
                // LLDP-Media Endpoint Discovery (ANSI/TIA-1057)
                switch(subtype){
                case 1:
                    // ignoring
                    break;
                case 5:
                {
                    // 5 Hardware Revision: "01.00"
                    mldp.med.hw_str = LldpInfo::formatVersion(getString(data, 0));
                    break;
                }
                case 6:
                {
                    // 6 Firmware Revision: "02.00.19676"
                    mldp.med.fw_str = LldpInfo::formatVersion(getString(data, 0));
                    break;
                }
                case 8:
                {
                    // 8 Serial Number: "0000-0569-59F2"
                    mldp.med.sn_str = formatSerialNumber(getString(data, 0));
                    break;
                }
                case 9:
                {
                    // 9 Manufacturer Name: "JINR AFI"
                    mldp.med.mn_str = getString(data, 0).trimmed();
                    break;
                }
                case 10:
                {
                    mldp.med.modelname_str = getString(data, 0).trimmed();
                    break;
                }
                } // switch(subtype)

                break;
            }
            case OUI_AFI: {
                // LOCAL
                switch (subtype) {
                case 1: {
                    if (block.size() < 10) {
                        _ok = false;
                        qWarning() << errStr(host, "Invalid TLV 127.LOCAL.1 format");
                        break;
                    }
                    // device info
                    auto p = reinterpret_cast<const afi_tlv_1_t *> (data.constData());
                    mldp.ext.device_id = ntohs(p->device_id);
                    mldp.ext.serial_id = ntoh_48(p->serial_id);
                    if (block.size() >= 12) {
                        auto p = reinterpret_cast<const afi_tlv_1_ext_t *> (data.constData());
                        mldp.ext.chassis_slot = ntohs(p->chassis_slot);
                    }
                    break;
                }
                case 2: {
                    if (block.size() < 29) {
                        _ok = false;
                        qWarning() << errStr(host, "Invalid TLV 127.LOCAL.2 format");
                        break;
                    }
                    auto p = reinterpret_cast<const afi_tlv_2_t *> (data.constData());
                    mldp.ext.master.ip = ntohl(p->master_ip);
                    mldp.ext.master.port = ntohs(p->master_udp_port);
                    mldp.ext.mstream.ip = ntohl(p->mstream_ip);
                    mldp.ext.mstream.port = ntohs(p->mstream_udp_port);
                    mldp.ext.master_mac = ntoh_48(p->master_mac);
                    mldp.ext.mstream_mac = ntoh_48(p->mstream_mac);
                    mldp.ext.master_locked = p->flags.master_locked;
                    mldp.ext.mstream_locked = p->flags.mstream_locked;
                    break;
                }
                } // switch (subtype)
                break;
            }
            } // switch (oui)
            break;
        }
        default:
        {
            QString s;
            for (int j=0; j<lldpLen; ++j)
                s.append(QString("%1 ").arg((quint8)block.at(j),2,16,QChar('0')));
            qWarning() << errStr(host, QString("unknown lldpType=%1 dump:%2")
                                    .arg(lldpType).arg(s));
            break;
        }
        } // switch (lldpType)
    }
    mldp.startTimer();
    if(ok)
        *ok = _ok;
    return mldp;
}

void MldpListener::send_info(const LldpInfo &lldp) const
{
    if(deviceIdFilter.isEmpty()
            || deviceIdFilter.contains(lldp.getDeviceIndex().getDevId())
            || deviceIdFilter.contains(DEVICE_ID_AnyDevice)) {
        emit receivedLldpInfo(lldp);
    }
}

