#include "FlashProgModule.h"

#include "DeviceCommandType.h"
#include "DeviceModuleCommand.h"
#include "base/DeviceIndex.h"
#include "base/fw_version.h"
#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"

#include "FlashProgStatus.h"
#include "sdb/AfiSdbId.h"
#include "FlashDevImpl.h"

FlashProgModule::FlashProgModule(const MregBusInterface &bus, QObject *parent)
    : QObject(parent)
    , BaseDeviceModule(bus)
    , flashDev(new FlashDevImpl(this, this))
    , flashIdData(new FlashIdData)
{
    connect(flashDev, &FlashDev::taskFinished, this, &FlashProgModule::taskFinished);
}

FlashProgModule::~FlashProgModule()
{
    delete flashIdData;
}

QList<AFI_SDB_ID> FlashProgModule::getSupportedSdbIds()
{
    return { AFI_SDB_Flash };
}

void FlashProgModule::post_init()
{
    if (!isAvailable()) {
        // no-SDB compatibility
            bus.setAddrRange({0x7000, 0x00FF});
    }
    static const fw_version_t ver_2_0(2,0,0);
    configureFlash(getVersion() >= ver_2_0);
}

std::optional<FlashProgStatus> FlashProgModule::readStatus()
{
    FlashProgStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}

bool FlashProgModule::readStatus(FlashProgStatus *result)
{
    FlashProgStatus status;
    auto flashIdData = flashDev->getFlashIdData();
    status.chip_valid = flashIdData.isValid();
    status.chip_capacity_Mib = flashIdData.getFlashCapacity_Mib();
    status.addr4bMode = flashDev->isAddr4bMode();

    auto task = flashDev->getTasks();
    if (!task.has_value())
        task = flashDev->getFinishedTasks();
    if (task.has_value()) {
        status.taskName = task->getOpStr();
        status.taskAborted = task->isAborted();
        status.taskRunning = task->isRunning();
        status.progress = (double)task->pos.offset() / task->pos.size();
    } else {
        status.taskName = QString();
        status.taskAborted = false;
        status.taskRunning = false;
        status.progress = 0.;
    }

    if (result)
        *result = status;
    return true;
}

bool FlashProgModule::abortTask()
{
    return flashDev->abortTask();
}

void FlashProgModule::configureFlash(bool supportAddr4bMode)
{
    flashDev->configureFlash(flashIdData, supportAddr4bMode);
}

bool FlashProgModule::isConfigured() const
{
    return flashDev->isConfigured();
}

std::optional<FlashDevTask> FlashProgModule::getTasks() const
{
    return flashDev->getTasks();
}

std::optional<FlashDevTask> FlashProgModule::getFinishedTasks() const
{
    return flashDev->getFinishedTasks();
}

bool FlashProgModule::runTasks()
{
    return flashDev->runTasks();
}

bool FlashProgModule::isReady() const
{
    return flashDev->isReady();
}

bool FlashProgModule::submitTask(const FlashImage &image, FlashOp op)
{
    FlashDevTask task(image, *flashIdData);
    task.op = op;
    return flashDev->submitTask(task);
}

void FlashProgModule::setSingleCheckMode(bool v) const
{
    flashDev->setSingleCheckMode(v);
}

bool FlashProgModule::softwareReboot()
{
    bool ok = (bus.dev != 0 && bus.dev->isConnected());
    if (!ok)
        return {};
    qInfo().noquote() << QString("[%1] software reboot").arg(getDeviceIndex().getIdent());
    RegOpVector r;
    r.RegWrite16(0x000F, 0xDEAD);
    regOpExecRebased(r); // no return expected
    return true;
}

DeviceModuleCommandResult FlashProgModule::runCommand(const DeviceModuleCommand &command)
{
    DeviceModuleCommandResult result;
    result.seq = command.seq;
    bool ok = true;
    switch(command.type) {
    case DeviceCommandType::Reboot: {
        ok &= softwareReboot();
        break;
    }
    case DeviceCommandType::Upload: {
        if (isReady()) {
            if (!command.data.canConvert<FlashImage>()) {
                result.error = EINVAL;
                return result;
            }
            const auto &image = command.data.value<FlashImage>();
            ok &= submitTask(image, FlashOp::Write);
        } else {
            result.error = EBUSY;
            return result;
        }
        break;
    }
    default:
        result.error = EINVAL;
        return result;
    }

    result.error = ok ? 0 : EIO;
    return result;
}
