/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ModularMregDevice.h"

#include "Ad5328/Ad5328Module.h"
#include "Ad5622/Ad5622Module.h"
#include "AdcDataTest/AdcDataTestModule.h"
#include "AdcErc/AdcErcModule.h"
#include "AdcLemoMux/AdcLemoMuxModule.h"
#include "AdcMpdTrig/AdcMpdTrigModule.h"
#include "AdcSpiLtm9011/AdcSpiLtm9011Module.h"
#include "AdcSpiAds52J90/AdcSpiAds52J90Module.h"
#include "AdcSpiAd9249/AdcSpiAd9249Module.h"
#include "Bmc/BmcModule.h"
#include "BoardGpio/BoardGpioModule.h"
#include "ClockControl/ClockControlModule.h"
#include "ClockControl/ClockStatus.h"
#include "ClockDividerControl/ClockDividerControlModule.h"
#include "CrosspointSwitch/CrosspointSwitchModule.h"
#include "DeserializersControl/AdcSerDesDelay.h"
#include "DeserializersControl/DeserializersControlModule.h"
#include "DeviceCommand.h"
#include "DeviceModuleMapper.h"
#include "DlnpCalibrationTrigger/DlnpCalibrationTriggerModule.h"
#include "FirFilter/FirFilterModule.h"
#include "FeLinkStatus/FeLinkStatus.h"
#include "FeLinkStatus/FeLinkStatusModule.h"
#include "FlashProg/FlashProgModule.h"
#include "HitCore/HitCoreModule.h"
#include "MlinkCsr/MlinkCsrModule.h"
#include "MpdTrc/MpdTrcModule.h"
#include "NetworkPort/NetworkPortModule.h"
#include "NetworkPort/NetworkPortStatus.h"
#include "RunLogic/RunLogicModule.h"
#include "RunStatistic/RunStatisticModule.h"
#include "SpillControl/SpillControlModule.h"
#include "StatReadout/StatReadoutModule.h"
#include "TdcCore/TdcCoreModule.h"

#include "Tlu16SfpTrigger/Tlu16SfpTriggerModule.h"
#include "Tlu40LvdsLuminosity/Tlu40LvdsLuminosityModule.h"
#include "Tlu40LvdsTofTrigger/Tlu40LvdsTofTriggerModule.h"
#include "TluDecDes/TluDecDesModule.h"
#include "TluTtlIoControl/TluTtlIoControlModule.h"
#include "Ut24TtlIo/Ut24TtlIoModule.h"

#include "TrigCtrlStat/TrigCtrlStatModule.h"
#include "TrigInfo/TrigInfoModule.h"
#include "TtlIo/TtlIoModule.h"
#include "WaveBlc/WaveBlcModule.h"
#include "WaveRec/WaveRecModule.h"
#include "WaveTrig/WaveTrigModule.h"
#include "WrStatus/WrStatusModule.h"
#include "WrTimeEmulator/WrTimeEmulatorModule.h"
#include "ModularDeviceStatus.h"
#include "device-utils/AdcSerDesCalibration.h"
#include "device-utils/ClockDividerReset.h"

class ModularMregDevicePrivate
{
public:
    ModularMregDevicePrivate(ModularMregDevice *q, const DeviceIndex &deviceIndex)
        : q_ptr(q)
        , deviceModuleMapper(deviceIndex, q, q)
    {}

private:
    Q_DECLARE_PUBLIC(ModularMregDevice)
    ModularMregDevice * const q_ptr;

public:
    DeviceModuleMapper deviceModuleMapper;
    QMap<int, QSharedPointer<Ad5328Module>> ad5328Modules;
    QMap<int, QSharedPointer<Ad5622Module>> ad5622Module;
    QMap<int, QSharedPointer<AdcDataTestModule>> adcDataTestModule;
    QMap<int, QSharedPointer<AdcErcModule>> adcErcModule;
    QMap<int, QSharedPointer<AdcLemoMuxModule>> adcLemoMuxModule;
    QMap<int, QSharedPointer<AdcMpdTrigModule>> adcMpdTrigModule;
    QMap<int, QSharedPointer<AdcSpiLtm9011Module>> adcSpiLtm9011Module;
    QMap<int, QSharedPointer<AdcSpiAds52J90Module>> adcSpiAds52J90Module;
    QMap<int, QSharedPointer<AdcSpiAd9249Module>> adcSpiAd9249Module;
    QMap<int, QSharedPointer<BmcModule>> bmcModules;
    QMap<int, QSharedPointer<BoardGpioModule>> boardGpioModule;
    QMap<int, QSharedPointer<ClockControlModule>> clockControlModule;
    QMap<int, QSharedPointer<ClockDividerControlModule>> clockDividerControlModule;
    QMap<int, QSharedPointer<CrosspointSwitchModule>> crosspointSwitchModule;
    QMap<int, QSharedPointer<DeserializersControlModule>> deserializersControlModules;
    QMap<int, QSharedPointer<DlnpCalibrationTriggerModule>> dlnpCalibrationTriggerModules;
    QMap<int, QSharedPointer<FeLinkStatusModule>> feLinkStatusModule;
    QMap<int, QSharedPointer<FirFilterModule>> firFilterModule;
    QMap<int, QSharedPointer<FlashProgModule>> flashProgModule;
    QMap<int, QSharedPointer<HitCoreModule>> hitCoreModule;
    QMap<int, QSharedPointer<MlinkCsrModule>> mlinkCsrModule; // Note: QtMregDevice contains its own MlinkCsrModule
    QMap<int, QSharedPointer<MpdTrcModule>> mpdTrcModules;
    QMap<int, QSharedPointer<NetworkPortModule>> networkPortModules;
    QMap<int, QSharedPointer<RunLogicModule>> runLogicModule;
    QMap<int, QSharedPointer<RunStatisticModule>> runStatisticModule;
    QMap<int, QSharedPointer<SpillControlModule>> spillControlModule;
    QMap<int, QSharedPointer<StatReadoutModule>> statReadoutModule;
    QMap<int, QSharedPointer<TdcCoreModule>> tdcCoreModule;
    QMap<int, QSharedPointer<Tlu16SfpTriggerModule>> tlu16SfpTriggerModules;
    QMap<int, QSharedPointer<Tlu40LvdsLuminosityModule>> tlu40LvdsLuminosityModules;
    QMap<int, QSharedPointer<Tlu40LvdsTofTriggerModule>> tlu40LvdsTofTriggerModules;
    QMap<int, QSharedPointer<TluDecDesModule>> tluDecDesModules;
    QMap<int, QSharedPointer<TluTtlIoControlModule>> tluTtlIoModules;
    QMap<int, QSharedPointer<TrigInfoModule>> trigInfoModule;
    QMap<int, QSharedPointer<TrigCtrlStatModule>> trigCtrlStatModule;
    QMap<int, QSharedPointer<TtlIoModule>> ttlIoModules;
    QMap<int, QSharedPointer<Ut24TtlIoModule>> ut24TtlIoModules;
    QMap<int, QSharedPointer<WaveBlcModule>> waveBlcModule;
    QMap<int, QSharedPointer<WaveRecModule>> waveRecModule;
    QMap<int, QSharedPointer<WaveTrigModule>> waveTrigModule;
    QMap<int, QSharedPointer<WrStatusModule>> wrStatusModules;
    QMap<int, QSharedPointer<WrTimeEmulatorModule>> wrTimeEmulatorModules;
};

ModularMregDevice::ModularMregDevice(const DeviceIndex &deviceIndex, ModularDeviceStatus &status, QObject *parent)
    : QtMregDevice(deviceIndex, parent)
    , d_ptr(new ModularMregDevicePrivate(this, deviceIndex))
    , deviceIndex(deviceIndex)
    , status(status)
    , adcSerDesCal (new AdcSerDesCalibration(this))
    , timerAdcDesCheck(new QTimer(this))
{
    // register types this class emits
    qRegisterMetaType<ConfigKey>();
    qRegisterMetaType<DeserializersControlBuildParams>();
    qRegisterMetaType<DlnpCalibrationTriggerBuildParams>();
    qRegisterMetaType<ModularDeviceStatus>();
    qRegisterMetaType<FeLinkStatus>();
    qRegisterMetaType<FeLinkBuildParams>();
    qRegisterMetaType<FirFilterBuildParams>();
    qRegisterMetaType<QMap<int, FeLinkStatus> >();
    qRegisterMetaType<TdcCoreStatus>();
    qRegisterMetaType<TrigCtrlBuildParams>();
    qRegisterMetaType<ClockBuildParams>();
    qRegisterMetaType<ClockStatus>();
    qRegisterMetaType<ClockDividerStatus>();
    qRegisterMetaType<Tlu16SfpTriggerBuildParams>();
    qRegisterMetaType<Tlu40LvdsLuminosityBuildParams>();
    qRegisterMetaType<Tlu40LvdsLuminosityStatus>();
    qRegisterMetaType<Tlu40LvdsTofTriggerBuildParams>();
    qRegisterMetaType<Tlu40LvdsTofTriggerStatus>();
    qRegisterMetaType<TluDecDesBuildParams>();
    qRegisterMetaType<TluDecDesStatus>();
    qRegisterMetaType<TluTtlIoControlBuildParams>();
    qRegisterMetaType<CrosspointSwitchBuildParams>();
    qRegisterMetaType<AdcBuildParams>();
    qRegisterMetaType<AdcSerDesDelayMap>();
    qRegisterMetaType<AdcLemoMuxBuildParams>();
    qRegisterMetaType<AdcMpdTrigBuildParams>();
    qRegisterMetaType<RunLogicStatus>();

    connect(adcSerDesCal, &AdcSerDesCalibration::progressUpdated,
            this, &ModularMregDevice::on_adcSerDesProgressUpdated);
    connect(adcSerDesCal, &AdcSerDesCalibration::adcDeserConfigChecked,
            this, &ModularMregDevice::adcDeserConfigChecked);

    connect(timerAdcDesCheck, &QTimer::timeout, this, &ModularMregDevice::checkAdcDeserConfig);
    timerAdcDesCheck->start(10000);
}

ModularMregDevice::~ModularMregDevice()
{
    delete d_ptr;
}

QList<SdbModuleIndex> ModularMregDevice::getModuleIndexes() const
{
    Q_D(const ModularMregDevice);
    return d->deviceModuleMapper.getModuleIndexes();
}

QSharedPointer<AbstractDeviceModule> ModularMregDevice::findModule(const SdbModuleIndex &moduleIndex)
{
    Q_D(ModularMregDevice);
    return d->deviceModuleMapper.findModule(moduleIndex);
}

QSharedPointer<const AbstractDeviceModule> ModularMregDevice::findModule(const SdbModuleIndex &moduleIndex) const
{
    Q_D(const ModularMregDevice);
    return d->deviceModuleMapper.findModule(moduleIndex);
}

QList<QSharedPointer<AbstractDeviceModule>> ModularMregDevice::findModules(AFI_SDB_ID id)
{
    Q_D(ModularMregDevice);
    return d->deviceModuleMapper.findModules(id);
}

QList<QSharedPointer<const AbstractDeviceModule>> ModularMregDevice::findModules(AFI_SDB_ID id) const
{
    Q_D(const ModularMregDevice);
    return d->deviceModuleMapper.findModules(id);
}

QSharedPointer<AbstractDeviceModule> ModularMregDevice::getFirstModule(AFI_SDB_ID id)
{
    const auto &list = findModules(id);
    if (list.empty())
        return {};
    return list.first();
}

QSharedPointer<const AbstractDeviceModule> ModularMregDevice::getFirstModule(AFI_SDB_ID id) const
{
    const auto &list = findModules(id);
    if (list.empty())
        return {};
    return list.first();
}

QSharedPointer<AdcSpiBaseModule> ModularMregDevice::getAdcSpiModule() const
{
    QSharedPointer<AdcSpiBaseModule> mod;
    if(mod = getFirstModule<AdcSpiLtm9011Module>())
        return mod;
    if(mod = getFirstModule<AdcSpiAds52J90Module>())
        return mod;
    if(mod = getFirstModule<AdcSpiAd9249Module>())
        return mod;
    return {};
}

QSharedPointer<AbstractDeviceModule> ModularMregDevice::insertModuleImpl(QSharedPointer<AbstractDeviceModule> module)
{
    Q_D(ModularMregDevice);
    return d->deviceModuleMapper.insert(module);
}

bool ModularMregDevice::modWriteConfig(QSharedPointer<AbstractDeviceModule> mod, bool modWriteOk)
{
    if(!modWriteOk) {
        QPair<AFI_SDB_ID, quint32> id{mod->getSdbId(), mod->getBaseAddr()};
        if(!failedModules.contains(id)) {
            qWarning() << QString("[%1] Module writeConfig failed for:'%2' on 0x%3 base addr")
                              .arg(getIdent()).arg(mod->name())
                              .arg(mod->getBaseAddr(), 0, 16);
            failedModules << id;
        }
    }
    return modWriteOk;
}

template<>
QMap<int, QSharedPointer<Ad5328Module>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->ad5328Modules;
}

template<>
QMap<int, QSharedPointer<Ad5622Module>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->ad5622Module;
}

template<>
QMap<int, QSharedPointer<AdcDataTestModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->adcDataTestModule;
}

template<>
QMap<int, QSharedPointer<AdcErcModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->adcErcModule;
}

template<>
QMap<int, QSharedPointer<AdcLemoMuxModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->adcLemoMuxModule;
}

template<>
QMap<int, QSharedPointer<AdcMpdTrigModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->adcMpdTrigModule;
}

template<>
QMap<int, QSharedPointer<AdcSpiLtm9011Module>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->adcSpiLtm9011Module;
}

template<>
QMap<int, QSharedPointer<AdcSpiAds52J90Module>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->adcSpiAds52J90Module;
}

template<>
QMap<int, QSharedPointer<AdcSpiAd9249Module>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->adcSpiAd9249Module;
}

template<>
QMap<int, QSharedPointer<BmcModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->bmcModules;
}

template<>
QMap<int, QSharedPointer<BoardGpioModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->boardGpioModule;
}

template<>
QMap<int, QSharedPointer<ClockControlModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->clockControlModule;
}

template<>
QMap<int, QSharedPointer<ClockDividerControlModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->clockDividerControlModule;
}

template<>
QMap<int, QSharedPointer<CrosspointSwitchModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->crosspointSwitchModule;
}

template<>
QMap<int, QSharedPointer<DeserializersControlModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->deserializersControlModules;
}

template<>
QMap<int, QSharedPointer<DlnpCalibrationTriggerModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->dlnpCalibrationTriggerModules;
}

template<>
QMap<int, QSharedPointer<FeLinkStatusModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->feLinkStatusModule;
}

template<>
QMap<int, QSharedPointer<FirFilterModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->firFilterModule;
}

template<>
QMap<int, QSharedPointer<FlashProgModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->flashProgModule;
}

template<>
QMap<int, QSharedPointer<HitCoreModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->hitCoreModule;
}

template<>
QMap<int, QSharedPointer<MlinkCsrModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->mlinkCsrModule;
}

template<>
QMap<int, QSharedPointer<MpdTrcModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->mpdTrcModules;
}

template<>
QMap<int, QSharedPointer<NetworkPortModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->networkPortModules;
}

template<>
QMap<int, QSharedPointer<RunLogicModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->runLogicModule;
}

template<>
QMap<int, QSharedPointer<RunStatisticModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->runStatisticModule;
}

template<>
QMap<int, QSharedPointer<SpillControlModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->spillControlModule;
}

template<>
QMap<int, QSharedPointer<StatReadoutModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->statReadoutModule;
}

template<>
QMap<int, QSharedPointer<TdcCoreModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->tdcCoreModule;
}

template<>
QMap<int, QSharedPointer<Tlu16SfpTriggerModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->tlu16SfpTriggerModules;
}

template<>
QMap<int, QSharedPointer<Tlu40LvdsLuminosityModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->tlu40LvdsLuminosityModules;
}

template<>
QMap<int, QSharedPointer<Tlu40LvdsTofTriggerModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->tlu40LvdsTofTriggerModules;
}

template<>
QMap<int, QSharedPointer<TluDecDesModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->tluDecDesModules;
}

template<>
QMap<int, QSharedPointer<TluTtlIoControlModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->tluTtlIoModules;
}

template<>
QMap<int, QSharedPointer<TrigCtrlStatModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->trigCtrlStatModule;
}

template<>
QMap<int, QSharedPointer<TrigInfoModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->trigInfoModule;
}

template<>
QMap<int, QSharedPointer<TtlIoModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->ttlIoModules;
}

template<>
QMap<int, QSharedPointer<Ut24TtlIoModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->ut24TtlIoModules;
}

template<>
QMap<int, QSharedPointer<WrStatusModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->wrStatusModules;
}

template<>
QMap<int, QSharedPointer<WrTimeEmulatorModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->wrTimeEmulatorModules;
}

template<>
QMap<int, QSharedPointer<WaveBlcModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->waveBlcModule;
}

template<>
QMap<int, QSharedPointer<WaveRecModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->waveRecModule;
}

template<>
QMap<int, QSharedPointer<WaveTrigModule>> ModularMregDevice::getModules() const
{
    Q_D(const ModularMregDevice);
    return d->waveTrigModule;
}

void ModularMregDevice::createModules()
{
    Q_D(ModularMregDevice);
    d->ad5328Modules = d->deviceModuleMapper.createModules<Ad5328Module>();
    d->ad5622Module = d->deviceModuleMapper.createModules<Ad5622Module>();
    d->adcDataTestModule = d->deviceModuleMapper.createModules<AdcDataTestModule>();
    d->adcErcModule = d->deviceModuleMapper.createModules<AdcErcModule>();
    d->adcLemoMuxModule = d->deviceModuleMapper.createModules<AdcLemoMuxModule>();
    d->adcMpdTrigModule = d->deviceModuleMapper.createModules<AdcMpdTrigModule>();
    d->adcSpiLtm9011Module = d->deviceModuleMapper.createModules<AdcSpiLtm9011Module>();
    d->adcSpiAds52J90Module = d->deviceModuleMapper.createModules<AdcSpiAds52J90Module>();
    d->adcSpiAd9249Module = d->deviceModuleMapper.createModules<AdcSpiAd9249Module>();
    d->bmcModules = d->deviceModuleMapper.createModules<BmcModule>();
    d->boardGpioModule = d->deviceModuleMapper.createModules<BoardGpioModule>();
    d->clockControlModule = d->deviceModuleMapper.createModules<ClockControlModule>();
    d->clockDividerControlModule = d->deviceModuleMapper.createModules<ClockDividerControlModule>();
    d->crosspointSwitchModule = d->deviceModuleMapper.createModules<CrosspointSwitchModule>();
    d->deserializersControlModules = d->deviceModuleMapper.createModules<DeserializersControlModule>();
    d->dlnpCalibrationTriggerModules = d->deviceModuleMapper.createModules<DlnpCalibrationTriggerModule>();
    d->feLinkStatusModule = d->deviceModuleMapper.createModules<FeLinkStatusModule>();
    d->firFilterModule = d->deviceModuleMapper.createModules<FirFilterModule>();
    d->flashProgModule = d->deviceModuleMapper.createModules<FlashProgModule>(this);
    d->hitCoreModule = d->deviceModuleMapper.createModules<HitCoreModule>();
    d->mlinkCsrModule = d->deviceModuleMapper.createModules<MlinkCsrModule>();
    d->mpdTrcModules = d->deviceModuleMapper.createModules<MpdTrcModule>();
    d->networkPortModules = d->deviceModuleMapper.createModules<NetworkPortModule>();
    d->runLogicModule = d->deviceModuleMapper.createModules<RunLogicModule>();
    d->runStatisticModule = d->deviceModuleMapper.createModules<RunStatisticModule>();
    d->spillControlModule = d->deviceModuleMapper.createModules<SpillControlModule>();
    d->statReadoutModule = d->deviceModuleMapper.createModules<StatReadoutModule>();
    d->tdcCoreModule = d->deviceModuleMapper.createModules<TdcCoreModule>();
    d->tlu16SfpTriggerModules = d->deviceModuleMapper.createModules<Tlu16SfpTriggerModule>();
    d->tlu40LvdsLuminosityModules = d->deviceModuleMapper.createModules<Tlu40LvdsLuminosityModule>();
    d->tlu40LvdsTofTriggerModules = d->deviceModuleMapper.createModules<Tlu40LvdsTofTriggerModule>();
    d->tluDecDesModules = d->deviceModuleMapper.createModules<TluDecDesModule>();
    d->tluTtlIoModules = d->deviceModuleMapper.createModules<TluTtlIoControlModule>();
    d->trigCtrlStatModule = d->deviceModuleMapper.createModules<TrigCtrlStatModule>();
    d->trigInfoModule = d->deviceModuleMapper.createModules<TrigInfoModule>();
    d->ttlIoModules = d->deviceModuleMapper.createModules<TtlIoModule>();
    d->ut24TtlIoModules = d->deviceModuleMapper.createModules<Ut24TtlIoModule>();
    d->waveBlcModule = d->deviceModuleMapper.createModules<WaveBlcModule>();
    d->waveRecModule = d->deviceModuleMapper.createModules<WaveRecModule>();
    d->waveTrigModule = d->deviceModuleMapper.createModules<WaveTrigModule>();
    d->wrStatusModules = d->deviceModuleMapper.createModules<WrStatusModule>();
    d->wrTimeEmulatorModules = d->deviceModuleMapper.createModules<WrTimeEmulatorModule>();
}

void ModularMregDevice::initModules()
{
    Q_D(ModularMregDevice);
    d->deviceModuleMapper.initModules();

    if(auto mod = getFirstModule<DeserializersControlModule>()) {
        auto bp = mod->getBuildParams();
        if (bp) {
            emit deserializersControlBpUpdated(getDeviceIndex(), *bp);
        }
    }
    if (auto mod = getFirstModule<DlnpCalibrationTriggerModule>()) {
        auto bp = mod->getBuildParam();
        if (bp)
            emit dlnpCalibrationTriggerBpUpdated(getDeviceIndex(), *bp);
    }

    if (auto mod = getFirstModule<FeLinkStatusModule>()) {
        auto bp = mod->getBuildParams();
        if (bp)
            emit feLinkBpUpdated(getDeviceIndex(), *bp);
    }

    if (auto mod = getFirstModule<FirFilterModule>()) {
        auto bp = mod->getBuildParam();
        if (bp)
            emit firFilterBpUpdated(getDeviceIndex(), *bp);
    }

    if (auto mod = getFirstModule<WaveRecModule>()) {
        auto bp = mod->getBuildParams();
        if (bp)
            emit adcBuildParamsUpdated(getDeviceIndex(), *bp);
    }

    if (auto mod = getFirstModule<AdcLemoMuxModule>()) {
        auto bp = mod->getBuildParam();
        if (bp)
            emit adcLemoMuxBpUpdated(getDeviceIndex(), *bp);
    }

    if (auto mod = getFirstModule<AdcMpdTrigModule>()) {
        auto bp = mod->getBuildParam();
        if (bp)
            emit adcMpdTrigBpUpdated(getDeviceIndex(), *bp);
    }

    if (auto m = getFirstModule<ClockControlModule>()) {
        auto bp = m->getBuildParam();
        if (bp)
            emit clockControlBpUpdated(getDeviceIndex(), *bp);
    }

    if (auto m = getFirstModule<CrosspointSwitchModule>()) {
        auto bp = m->getBuildParam();
        if (bp)
            emit crosspointSwitchBpUpdated(getDeviceIndex(), *bp);
    }

    if (auto m = getFirstModule<Tlu16SfpTriggerModule>())
        if (auto bp = m->getBuildParam())
            emit tlu16SfpTrigBpUpdated(getDeviceIndex(), *bp);
    if (auto m = getFirstModule<Tlu40LvdsLuminosityModule>())
        if (auto bp = m->getBuildParam())
            emit tlu40LvdsLuminosityBpUpdated(getDeviceIndex(), *bp);
    if (auto m = getFirstModule<Tlu40LvdsTofTriggerModule>())
        if (auto bp = m->getBuildParam())
            emit tlu40LvdsTofTriggerBpUpdated(getDeviceIndex(), *bp);
    if (auto m = getFirstModule<TluDecDesModule>())
        if (auto bp = m->getBuildParam())
            emit tluDecDesBpUpdated(getDeviceIndex(), *bp);
    if (auto m = getFirstModule<TluTtlIoControlModule>())
        if (auto bp = m->getBuildParam())
            emit tluTtlIoControlBpUpdated(getDeviceIndex(), *bp);

    if (auto mod = getFirstModule<TrigCtrlStatModule>()) {
        auto bp = mod->getBuildParam();
        if (bp)
            emit trigCtrlBpUpdated(getDeviceIndex(), *bp);
    }
}

void ModularMregDevice::on_deviceConnected()
{
    QtMregDevice::on_deviceConnected();
    status.connected = true;

    createModules();
    failedModules.clear();
    initModules();

    QVariantList sdbList;
    for(auto v : getModuleIndexes())
        sdbList << v.id;
    emit deviceSdbListUpdated(getDeviceIndex(), sdbList);

    dropConfigCache();
    if (config_key.isValid())
        syncConfigDeferred();

    adcSerDesCal->init();
}

void ModularMregDevice::on_deviceDisconnected()
{
    Q_D(ModularMregDevice);
    d->deviceModuleMapper.deinitModules();
    QtMregDevice::on_deviceDisconnected();
    adcSerDesCal->deinit();
    status = ModularDeviceStatus(status.getDeviceIndex());
    status.connected = false;
    dropConfigCache();
    emit deviceConfigWritten(getDeviceIndex(), {});
    emit statusUpdated(status);
}

bool ModularMregDevice::readStatus()
{
    status.connected = isConnected();
    status.sdb = getSdb();
    status.moduleIndexes = getModuleIndexes();
    if (!readModulesStatus()) return false;
    emit statusUpdated(status);
    return true;
}

bool ModularMregDevice::readModulesStatus()
{
    Q_D(ModularMregDevice);
    bool ok = true;

    status.adcDataTestStatus.reset();
    for (auto it = d->adcDataTestModule.begin(); it != d->adcDataTestModule.end(); ++it) {
        const auto &module = it.value();
        status.adcDataTestStatus = module->readStatus();
        ok &= status.adcDataTestStatus.has_value();
    }

    status.bmcStatus.reset();
    for (auto it = d->bmcModules.begin(); it != d->bmcModules.end(); ++it) {
        const auto &module = it.value();
        status.bmcStatus = module->readStatus();
        ok &= status.bmcStatus.has_value();
    }

    status.boardGpio.reset();
    for (auto it = d->boardGpioModule.begin(); it != d->boardGpioModule.end(); ++it) {
        const auto &module = it.value();
        status.boardGpio = module->readStatus();
        ok &= status.boardGpio.has_value();
    }

    status.clockStatus.reset();
    for (auto it = d->clockControlModule.begin(); it != d->clockControlModule.end(); ++it) {
        const auto &module = it.value();

        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.clockStatus = result;
        emit clockControlStatusUpdated(getDeviceIndex(), *result);
    }

    status.clockDividerStatus.reset();
    for (auto it = d->clockDividerControlModule.begin(); it != d->clockDividerControlModule.end(); ++it) {
        const auto &module = it.value();

        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.clockDividerStatus = result;
        emit clockDividerStatusUpdated(getDeviceIndex(), *result);
    }

    status.desStatus.reset();
    for (auto it = d->deserializersControlModules.begin(); it != d->deserializersControlModules.end(); ++it) {
        const auto &module = it.value();
        status.desStatus = module->readStatus();
        ok &= status.desStatus.has_value();
    }

    status.feLinkStatus.clear();
    for (auto it = d->feLinkStatusModule.begin(); it != d->feLinkStatusModule.end(); ++it) {
        auto portIndex = it.key() * 100; // module_index * 100
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        const auto feLinkStatusMap = result.value();
        for (auto jt = feLinkStatusMap.begin(); jt != feLinkStatusMap.end(); ++jt) {
            status.feLinkStatus.insert(portIndex + jt.key(), jt.value()); // ... + ch
        }
    }
    emit feLinkStatusUpdated(getDeviceIndex(), status.feLinkStatus);

    status.flashProgStatus.reset();
    for (auto it = d->flashProgModule.begin(); it != d->flashProgModule.end(); ++it) {
        const auto &module = it.value();
        status.flashProgStatus = module->readStatus();
        ok &= status.flashProgStatus.has_value();
    }

    status.firFilterStatus.reset();
    for (auto it = d->firFilterModule.begin(); it != d->firFilterModule.end(); ++it) {
        const auto &module = it.value();
        status.firFilterStatus = module->readStatus();
        ok &= status.firFilterStatus.has_value();
    }

    status.hitCoreStatus.reset();
    for (auto it = d->hitCoreModule.begin(); it != d->hitCoreModule.end(); ++it) {
        const auto &module = it.value();
        status.hitCoreStatus = module->readStatus();
        ok &= status.hitCoreStatus.has_value();
    }

    status.mlinkCsrStatus.reset();
    for (auto it = d->mlinkCsrModule.begin(); it != d->mlinkCsrModule.end(); ++it) {
        const auto &module = it.value();
        status.mlinkCsrStatus = module->readStatus();
        ok &= status.mlinkCsrStatus.has_value();
    }

    status.networkPortStatus.clear();
    for (auto it = d->networkPortModules.begin(); it != d->networkPortModules.end(); ++it) {
        auto portIndex = it.key();
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.networkPortStatus.insert(portIndex, result.value());
    }

    status.runLogicStatus.reset();
    for (auto it = d->runLogicModule.begin(); it != d->runLogicModule.end(); ++it) {
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.runLogicStatus = result;
        emit runLogicStatusUpdated(getDeviceIndex(), *result);
    }

    status.runStatistic.reset();
    for (auto it = d->runStatisticModule.begin(); it != d->runStatisticModule.end(); ++it) {
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.runStatistic = result;
    }

    status.spillStatus.reset();
    for (auto it = d->spillControlModule.begin(); it != d->spillControlModule.end(); ++it) {
        const auto &module = it.value();
        status.spillStatus = module->readStatus();
        ok &= status.spillStatus.has_value();
    }

    status.tdcCoreStatus.reset();
    for (auto it = d->tdcCoreModule.begin(); it != d->tdcCoreModule.end(); ++it) {
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.tdcCoreStatus = result;
        emit hptdcStatusUpdated(getDeviceIndex(), result.value());
    }




    status.tlu40LvdsLuminosityStatus.reset();
    for (auto it = d->tlu40LvdsLuminosityModules.begin(); it != d->tlu40LvdsLuminosityModules.end(); ++it) {
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.tlu40LvdsLuminosityStatus = result;
        emit tlu40LvdsLuminosityStatusUpdated(getDeviceIndex(), result.value());
    }
    status.tlu40LvdsTofTriggerStatus.reset();
    for (auto it = d->tlu40LvdsTofTriggerModules.begin(); it != d->tlu40LvdsTofTriggerModules.end(); ++it) {
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.tlu40LvdsTofTriggerStatus = result;
        emit tlu40LvdsTofTriggerStatusUpdated(getDeviceIndex(), result.value());
    }
    status.tluDecDesStatus.reset();
    for (auto it = d->tluDecDesModules.begin(); it != d->tluDecDesModules.end(); ++it) {
        const auto &module = it.value();
        const auto &result = module->readStatus();
        if (!result.has_value()) {
            ok = false;
            continue;
        }
        status.tluDecDesStatus = result;
        emit tluDecDesStatusUpdated(getDeviceIndex(), result.value());
    }




    status.trigInfo.reset();
    for (auto it = d->trigInfoModule.begin(); it != d->trigInfoModule.end(); ++it) {
        const auto &module = it.value();
        status.trigInfo = module->readStatus();
        ok &= status.trigInfo.has_value();
    }

    status.trigStatus.reset();
    for (auto it = d->trigCtrlStatModule.begin(); it != d->trigCtrlStatModule.end(); ++it) {
        const auto &module = it.value();
        status.trigStatus = module->readStatus();
        ok &= status.trigStatus.has_value();
    }

    status.wrStatus.reset();
    for (auto it = d->wrStatusModules.begin(); it != d->wrStatusModules.end(); ++it) {
        const auto &module = it.value();
        status.wrStatus = module->readStatus();
        ok &= status.wrStatus.has_value();
    }
    status.waveBlc.reset();
    for (auto it = d->waveBlcModule.begin(); it != d->waveBlcModule.end(); ++it) {
        const auto &module = it.value();
        status.waveBlc = module->readStatus();
        ok &= status.waveBlc.has_value();
    }
    status.waveRec.reset();
    for (auto it = d->waveRecModule.begin(); it != d->waveRecModule.end(); ++it) {
        const auto &module = it.value();
        status.waveRec = module->readStatus();
        ok &= status.waveRec.has_value();
    }
    return ok;
}

void ModularMregDevice::on_adcSerDesProgressUpdated(int p) const
{
    emit adcSerDesProgressUpdated(getDeviceIndex(), p);
}

void ModularMregDevice::checkAdcDeserConfig() const
{
    if(!config.isRun() && config.periodicalAdcDelayCheck) {
        adcSerDesCal->checkAdcDeserConfig(config.serDesDelay);
    }
    timerAdcDesCheck->start();
}

QMap<QString, double> ModularMregDevice::readTemp()
{
    Q_D(ModularMregDevice);
    for (auto it = d->bmcModules.begin(); it != d->bmcModules.end(); ++it) {
        auto module = it.value();
        status.bmcStatus = module->readStatus();
    }
    return status.getTemps();
}

void ModularMregDevice::setConfigData(const ConfigKey &key, const ModularDeviceConfig &newConfig)
{
    config_key = key;
    config = newConfig;
}

void ModularMregDevice::setDeviceConfig(const ConfigKey &newConfigKey, const ModularDeviceConfig &newConfig)
{
    if (!newConfig.index.isNull() && newConfig.index != getDeviceIndex()) {
        return;
    }

    setConfigData(newConfigKey, newConfig);

    syncConfigDeferred();
}

void ModularMregDevice::startAdcSrdDesCalibration(const ConfigKey &calibrationKey,
                                                  bool hardMode, const QString &resDir)
{
    if (!adcSerDesCal->calibrate(hardMode, resDir)) {
        if(auto mod = getAdcSpiModule())
            mod->setCustomPatternMode(false);
        qWarning() << "ADC SerDes calibration failed";
    }
    const auto &dd = adcSerDesCal->getStatus().getDevDelay();
    emit devSerDesDelayChanged(getDeviceIndex(), dd);
    emit calibrateFinished(getDeviceIndex(), calibrationKey);

    dropConfigCache();
    syncConfigDeferred();
}

void ModularMregDevice::evNumRequest(int checkIndex)
{
    if(auto mod = getFirstModule<TrigInfoModule>()) {
        auto st = mod->readStatus();
        if(st && st->isValid()) {
            emit evNumResponse(checkIndex, getDeviceIndex(), st->eventNumber);
            return;
        }
    }

    auto mod = getFirstModule<RunStatisticModule>();
    if(!mod)
        return;
    auto st = mod->readStatus();
    if(!st)
        return;

    emit evNumResponse(checkIndex, getDeviceIndex(), st->getCounter(RunStatistic::Accepted));
}

void ModularMregDevice::feLinkPhyReset(const DeviceIndex &dev, int ch)
{
    if(!dev.isNull() && dev != getDeviceIndex())
        return;
    auto mod = getFirstModule<FeLinkStatusModule>();
    if(!mod)
        return;
    mod->phyReset(ch);
}

void ModularMregDevice::adcClkDivSync(/*const ConfigKey &calibrationKey*/)
{
    auto resetUtil = ClockDividerReset(this);
    resetUtil.init();
    resetUtil.syncClock();
    // emit calibrateFinished(getDeviceIndex(), calibrationKey);
}

bool ModularMregDevice::syncConfigDeferred()
{
    if (!isConnected())
        return false; // try again later

    if (!config_key.isValid()) {
        qWarning() << "Invalid config key, skipped";
        return true; // nothing to do
    }

    bool ok = writeConfig();

    if (ok) {
        qDebug().noquote() << "Device configuration synchronized, key" << config_key;
        setConfigCache();
        emit deviceConfigWritten(getDeviceIndex(), config_key);
    } else {
        qWarning().noquote() << "Failed to synchronize device configuration, key" << config_key;
        dropConfigCache();
        emit deviceConfigWritten(getDeviceIndex(), {});
    }
    return ok;
}

void ModularMregDevice::enforceConfigConstraints()
{
    enforceConfigConstraints(&config);
}

bool ModularMregDevice::writeConfig()
{
    Q_D(ModularMregDevice);
    bool ok = true;

    enforceConfigConstraints();

    if (!setupCacheValid || configCache.ad5328Configs != config.ad5328Configs) {
        for (auto it = d->ad5328Modules.begin(); it != d->ad5328Modules.end(); ++it) {
            auto index = it.key();
            auto mod = it.value();
            ok &= modWriteConfig(mod, mod->writeConfig(config.ad5328Configs[index]));
        }
    }
    if (!setupCacheValid || configCache.ad5622 != config.ad5622) {
        auto mod = getFirstModule<Ad5622Module>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.ad5622));
        }
    }
    if (!setupCacheValid || configCache.adcErc != config.adcErc) {
        auto mod = getFirstModule<AdcErcModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.adcErc));
        }
    }
    if (!setupCacheValid || configCache.adcLemoMux != config.adcLemoMux) {
        auto mod = getFirstModule<AdcLemoMuxModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.adcLemoMux));
        }
    }
    if (!setupCacheValid || configCache.adcMpdTrig != config.adcMpdTrig) {
        auto mod = getFirstModule<AdcMpdTrigModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.adcMpdTrig));
        }
    }
    if (!setupCacheValid || configCache.boardGpio != config.boardGpio) {
        auto mod = getFirstModule<BoardGpioModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.boardGpio));
        }
    }
    if (!setupCacheValid || configCache.clockControl != config.clockControl) {
        auto mod = getFirstModule<ClockControlModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.clockControl));
        }
    }
    if (!setupCacheValid || configCache.crosspointSwitch != config.crosspointSwitch) {
        auto mod = getFirstModule<CrosspointSwitchModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.crosspointSwitch));
        }
    }
    if (!setupCacheValid || configCache.serDesDelay != config.serDesDelay) {
        auto mod = getFirstModule<DeserializersControlModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->setSerDesDelay(config.serDesDelay));
    }
    if (!setupCacheValid || configCache.dlnpCalibrationTrigger != config.dlnpCalibrationTrigger) {
        auto mod = getFirstModule<DlnpCalibrationTriggerModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.dlnpCalibrationTrigger));
        }
    }
    if (!setupCacheValid || configCache.feLink != config.feLink) {
        auto mod = getFirstModule<FeLinkStatusModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.feLink));
        }
    }
    if (!setupCacheValid || configCache.firFilter != config.firFilter) {
        auto mod = getFirstModule<FirFilterModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.firFilter));
        }
    }
    if (!setupCacheValid || configCache.hitCore != config.hitCore) {
        auto mod = getFirstModule<HitCoreModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.hitCore));
    }
    if (!setupCacheValid || configCache.mpdTrc != config.mpdTrc) {
        for (auto mod : d->mpdTrcModules) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.mpdTrc));
        }
    }
    if (!setupCacheValid || configCache.networkPort != config.networkPort) {
        for (auto mod : d->networkPortModules) {
            ok &= modWriteConfig(mod, mod->setConfig(config.networkPort));
        }
    }
    if (!setupCacheValid || configCache.runLogic != config.runLogic) {
        auto mod = getFirstModule<RunLogicModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.runLogic));
    }
    if (!setupCacheValid || configCache.spillControl != config.spillControl) {
        auto mod = getFirstModule<SpillControlModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.spillControl));
    }
    if (!setupCacheValid || configCache.statReadout != config.statReadout) {
        auto mod = getFirstModule<StatReadoutModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.statReadout));
    }
    if (!setupCacheValid || configCache.tdcCore != config.tdcCore) {
        auto mod = getFirstModule<TdcCoreModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.tdcCore));
        }
    }
    if (!setupCacheValid || configCache.tlu16SfpTrigger != config.tlu16SfpTrigger) {
        auto mod = getFirstModule<Tlu16SfpTriggerModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.tlu16SfpTrigger));
    }
    if (!setupCacheValid || configCache.tlu40lvdsLuminosity != config.tlu40lvdsLuminosity) {
        auto mod = getFirstModule<Tlu40LvdsLuminosityModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.tlu40lvdsLuminosity));
    }
    if (!setupCacheValid || configCache.tlu40lvdsTofTrigger != config.tlu40lvdsTofTrigger) {
        auto mod = getFirstModule<Tlu40LvdsTofTriggerModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.tlu40lvdsTofTrigger));
    }
    if (!setupCacheValid || configCache.tluDecDes != config.tluDecDes) {
        auto mod = getFirstModule<TluDecDesModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.tluDecDes));
    }
    if (!setupCacheValid || configCache.tluTtlIo != config.tluTtlIo) {
        auto mod = getFirstModule<TluTtlIoControlModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.tluTtlIo));
    }
    if (!setupCacheValid || configCache.ut24TtlIo != config.ut24TtlIo) {
        auto mod = getFirstModule<Ut24TtlIoModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.ut24TtlIo));
    }
    if (!setupCacheValid || configCache.trigCtrlStat != config.trigCtrlStat) {
        auto mod = getFirstModule<TrigCtrlStatModule>();
        if (mod) {
            ok &= modWriteConfig(mod, mod->writeConfig(config.trigCtrlStat));
        }
    }
    if (!setupCacheValid || configCache.ttlIo != config.ttlIo) {
        auto mod = getFirstModule<TtlIoModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.ttlIo));
    }
    if (!setupCacheValid || configCache.waveBlc != config.waveBlc) {
        auto mod = getFirstModule<WaveBlcModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.waveBlc));
    }
    if (!setupCacheValid || configCache.waveRec != config.waveRec) {
        auto mod = getFirstModule<WaveRecModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.waveRec));
    }
    if (!setupCacheValid || configCache.waveTrig != config.waveTrig) {
        auto mod = getFirstModule<WaveTrigModule>();
        if (mod)
            ok &= modWriteConfig(mod, mod->writeConfig(config.waveTrig));
    }
    return ok;
}

bool ModularMregDevice::preRunActions(const QVariant &evNumVar)
{
    bool ok = true;

    // resetHptdcErrors
    if (auto mod = getFirstModule<TdcCoreModule>()) {
        ok &= mod->writeConfig(config.tdcCore);
    }

    if(auto mod = getFirstModule<RunStatisticModule>()) {
        ok &= mod->resetCounters();
    }

    if (auto mod = getFirstModule<TrigInfoModule>()) {
        quint64 evNum = 1;
        if(evNumVar.isValid()) {
            bool evOk;
            auto v = evNumVar.toULongLong(&evOk);
            if(evOk)
                evNum = v;
        }

        ok &= mod->writeGlobEvNum(evNum);
    }

    if (auto mod = getFirstModule<RunLogicModule>()) {
        ok &= mod->writeReset();
    }

    return ok;
}

bool ModularMregDevice::writeSoftRun(bool runState)
{
    bool ok = true;
    auto runLogic = getFirstModule<RunLogicModule>();
    if (runLogic)
        ok &= runLogic->writeSoftRun(runState);
    return ok;
}

//bool ModularMregDevice::writeNetworkPortConfig(const NetworkPortConfig &config)
//{
//    Q_D(ModularMregDevice);
//    bool ok = true;
//    for (auto it = d->networkPortModules.begin(); it != d->networkPortModules.end(); ++it) {
//        auto module = it.value();
//        ok &= module->setConfig(config);
//    }
//    return ok;
//}

void ModularMregDevice::rebootDevice(const DeviceIndex &index)
{
    if (!index.isNull() && index != getDeviceIndex())
        return;
    auto flashProg = getFirstModule<FlashProgModule>();
    if (!flashProg)
        return;
    flashProg->softwareReboot();
}

void ModularMregDevice::submitDeviceCommand(const DeviceIndex &index, const DeviceCommand &command)
{
    Q_D(ModularMregDevice);

    if (!index.isNull() && index != getDeviceIndex())
        return;
    auto result(command);
    result.deviceIndex = getDeviceIndex();
    result.result.seq = result.command.seq;
    // const auto &moduleIndexes = d->deviceModuleMapper.getModuleIndexes();
    // if (!command.moduleIndex.isNull() && !moduleIndexes.contains(command.moduleIndex))
    auto module = d->deviceModuleMapper.findModule(command.moduleIndex);
    if (!module) {
        qWarning().noquote() << QString("%1: module not found (%2 @%3)")
                                    .arg(__func__)
                                    .arg(command.moduleIndex.id, 8, 16, QChar('0'))
                                    .arg(command.moduleIndex.baseAddr, 4, 16, QChar('0'));
        result.result.error = EINVAL;
    } else {
        const auto &moduleResult = module->runCommand(command.command);
        result.result = moduleResult;
    }
    // result.status = DeviceCommand::Complete;
    emit deviceCommandComplete(getDeviceIndex(), result);
}
