#include "MscStreamModule.h"

#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

#include "MscStreamConfig.h"

namespace {
enum {
    //Stream readout CSR
    REG_CTRL = 0x0,
    REG_V1_CE_MONITOR_MASK = 0x1, //External count enable monitor mask
    REG_V1_BUILD_TIMEOUT = 0x3,
    REG32_V1_SLICETIME = 0x6,
    REG32_V1_CLK_FREQ = 0x20, //Clock frequency in KHz, 32-bits, RO static
    REG_V1_CONTERS_WIDTH = 0x22, //Counters width (CNT_BITS), RO static
};

enum {
    REG_CTRL_BIT_ENABLE = 1 << 15,
};

enum {
    REG32_ID        = 0x0,
    REG_CONTROL     = 0x2,

    REG32_V2_CLK_FREQ = 0x4, //Clock frequency in KHz, 32-bits, RO static
    REG_V2_CH_NUM = 0x6, //Number of channels, RO static
    REG_V2_CE_IN_NUM = 0x07, //Number of external CE inputs, RO static
    REG_V2_CONTERS_WIDTH = 0x8, //Counters width (CNT_BITS), RO static

    REG_V2_CE_MONITOR_MASK = 0x10, //External count enable monitor mask
    REG_V2_BUILD_TIMEOUT = 0x11, //Stream packet build timeout in milliseconds, R/W
    REG32_V2_SLICETIME = 0x12, //Slicetimer in nanoseconds, 32-bits, R/W
    REG64_V2_HIST_CE_MASK = 0x14, //Apply histogram count enable mask for stream encoder, 1 bit per channel, 64 bits R/W
};
}

QList<AFI_SDB_ID> MscStreamModule::getSupportedSdbIds()
{
    return {AFI_SDB_MSC_STREAM_READOUT};
}

RegOpVector MscStreamModule::setConfigRegOp(const MscStreamConfig &cfg)
{
    RegOpVector r;
    switch (majorVersion()) {
    case 1:
        r.RegWrite16(REG_CTRL, getControlReg(cfg));
        r.RegWrite16(REG_V1_CE_MONITOR_MASK, getCeMonitorMask(cfg));
        r.RegWrite16(REG_V1_BUILD_TIMEOUT, cfg.timeout);
        r.RegWrite32(REG32_V1_SLICETIME, (cfg.slice/8)*8);
        break;
    case 2:
        r.RegWrite16(REG_CONTROL, getControlReg(cfg));
        r.RegWrite16(REG_V2_CE_MONITOR_MASK, getCeMonitorMask(cfg));
        r.RegWrite16(REG_V2_BUILD_TIMEOUT, cfg.timeout);
        r.RegWrite32(REG32_V2_SLICETIME, (cfg.slice/8)*8);
        r.RegWrite64(REG64_V2_HIST_CE_MASK, getHistCeMask(cfg));
        break;
    default:
        return {};
    }
    return r;
}

bool MscStreamModule::writeConfig(const MscStreamConfig &cfg)
{
    auto r = setConfigRegOp(cfg);
    return regOpExecRebased(r);
}

void MscStreamModule::post_init()
{
    BaseDeviceModule::post_init();
    buildParams.reset();
    quint32 id = AFI_SDB_MSC_STREAM_READOUT;
    MscStreamBuildParams bp;

    RegOpVector r;
    switch (bp.majVer = majorVersion()) {
    case 1:
        r.RegRead32(REG32_ID, &id);
        r.RegRead32(REG32_V1_CLK_FREQ, &bp.clkFreq);
        r.RegRead16(REG_V1_CONTERS_WIDTH, &bp.cntWidth);
        break;
    case 2:
        r.RegRead32(REG32_ID, &id);
        r.RegRead32(REG32_V2_CLK_FREQ, &bp.clkFreq);
        r.RegRead16(REG_V2_CH_NUM, &bp.chNum);
        r.RegRead16(REG_V2_CE_IN_NUM, &bp.ceInNum);
        r.RegRead16(REG_V2_CONTERS_WIDTH, &bp.cntWidth);
        break;
    default:
        return;
    }

    if(regOpExecRebased(r) && getSupportedSdbIds().contains(static_cast<AFI_SDB_ID>(id))) {
        buildParams = bp;
    }
}

quint16 MscStreamModule::getControlReg(const MscStreamConfig &cfg) const
{
    quint16 v=0;
    if(cfg.en)
        v |= REG_CTRL_BIT_ENABLE;
    return v;
}

quint16 MscStreamModule::getCeMonitorMask(const MscStreamConfig &cfg) const
{
    quint16 ret=0;
    for(int i=0; i<cfg.ceMonitor.size(); ++i)
        if(cfg.ceMonitor[i])
            ret |= 1<<i;
    return ret;
}

quint64 MscStreamModule::getHistCeMask(const MscStreamConfig &cfg) const
{
    if(!buildParams)
        return 0;
    quint64 ret=0;
    for(int i=0; i<buildParams->chNum; ++i)
        if(cfg.histCe.value(i, MscStreamConfig::HIST_CE_DEF))
            ret |= 1<<i;
    return ret;
}
