/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RunLogicMultiStatus.h"

#include <QMap>
#include <algorithm>

static QStringList makeDeviceIndexIdentList(const DeviceIndexSet &list)


{
    QStringList result;
    std::transform(list.cbegin(),
                   list.cend(),
                   std::back_inserter(result),
                   [] (const DeviceIndex &i) {return i.getSerialStr();} );
    return result;
}

void RunLogicMultiStatus::setDeviceList(const DeviceIndexEnabledMap &new_devices)
{
    deviceSet = new_devices.getEnabled();
    update();
}

void RunLogicMultiStatus::setRunLogicConfig(const RunLogicConfig &newConfig)
{
    config = newConfig;
    update();
}

void RunLogicMultiStatus::setRunLogicStatus(const DeviceIndex &deviceIndex, const RunLogicStatus &newStatus)
{
    if (!newStatus.isValid())
        return;
    auto found = status.find(deviceIndex);
    if (found != status.end() && *found == newStatus)
        return;
    status[deviceIndex] = newStatus;
    update();
}

void RunLogicMultiStatus::update()
{
    const auto &enabledDevices = deviceSet;
    isEmpty = enabledDevices.empty();
    missing.clear();
    invalid.clear();
    idle.clear();
    running.clear();
    finished.clear();
    for (const auto &k : enabledDevices) {
        auto found = status.constFind(k);
        if (found == status.cend()) {
            missing << k;
            continue;
        }
        const auto &st = *found;
        if (!st.isValid())
            invalid << k;
        if (st.isIdle())
            idle << k;
        if (st.isRunning())
            running << k;
        if (st.hasFinished())
            finished << k;
    }

    // combine multiple devices
    multiProgress.clear();
    multiRunTime.clear();
    multiEventCount.clear();
    anyNotIdle = false;
    for (auto it = status.cbegin(); it != status.cend(); ++it) {
        const auto &deviceIndex = it.key();
        if (!deviceSet.contains(deviceIndex))
            continue;
        const auto &st = it.value();
        double eventProgress = config.isEventLimit() ? (double)st.currentEventCount / config.eventLimit : 0;
        double timeProgress = config.isTimeLimit() ? (double)st.currentRunTime / config.timeLimit: 0;
        multiProgress.push_back(std::max(eventProgress, timeProgress));
        multiRunTime.push_back(st.currentRunTime);
        multiEventCount.push_back(st.currentEventCount);
        anyNotIdle |= !st.isIdle();
    }
}

QStringList RunLogicMultiStatus::statusText() const
{
    QStringList sl;
    if (allFinished())
        sl << "Finished";
    else if (allRunning())
        sl << "Running";
    else if (allIdle())
        sl << "Idle";
    else
        sl << "Mixed state";
    return sl;
}

QString RunLogicMultiStatus::getFullText() const
{
    const QMap<QString, DeviceIndexSet> map =
        {{"Invalid", invalid},
            {"Missing", missing},
            {"Running", running},
            {"Idle", idle},
            {"Finished", finished}
        };
    QStringList sl;
    for (auto it = map.cbegin(); it != map.cend(); ++it) {
        auto indexList = makeDeviceIndexIdentList(it.value());
        if (!indexList.empty())
            sl << it.key() + ":\n" + indexList.join(", ");
    }
    return sl.join("\n");
}

double RunLogicMultiStatus::progress() const
{
    auto minProgress = std::min_element(multiProgress.cbegin(), multiProgress.cend());
    return std::clamp(*minProgress, 0., 1.);
}

double RunLogicMultiStatus::minRunTime() const
{
    auto found = std::min_element(multiRunTime.cbegin(), multiRunTime.cend());
    if (found == multiRunTime.cend())
        return 0;
    return *found;
}

qint64 RunLogicMultiStatus::minEventCount() const
{
    auto found = std::min_element(multiEventCount.cbegin(), multiEventCount.cend());
    if (found == multiEventCount.cend())
        return 0;
    return *found;
}

bool RunLogicMultiStatus::allIdle() const
{
    return idle == deviceSet;
}

bool RunLogicMultiStatus::allRunning() const
{
    return running == deviceSet;
}

bool RunLogicMultiStatus::allFinished() const
{
    return finished == deviceSet;
}
