#include "SpillControlModule.h"

#include "SpillConfig.h"
#include "SpillStatus.h"

#include "base/DeviceId.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

enum {
    REG32_SPILL_DUR_PHASE_1_MS      = 0x00,
    REG32_SPILL_DUR_PHASE_2_MS      = 0x02,
    REG32_SPILL_DUR_PHASE_3_MS      = 0x04,
    REG32_SPILL_DUR_PHASE_4_MS      = 0x06,
    REG32_SPILL_DUR_PHASE_5_MS      = 0x08,
    REG32_SPILL_GATE_EXTEND_MS      = 0x0A,
    REG32_SPILL_COUNT               = 0x0C, // read-only
    REG32_SPILL_CURRENT_TIME_MS     = 0x0E, // read-only
    REG32_SPILL_CTRL                = 0x10,
    REG16_SPILL_STATUS              = 0x12, // read-only
    REG32_SPILL_SOFTSPILL_PERIOD_MS = 0x14,
    REG32_SPILL_DETECT_MIN_MS       = 0x16,
    REG32_SPILL_DETECT_MAX_MS       = 0x18,
    REG32_SPILL_TTL_LAST_PERIOD_MS  = 0x1A, // read-only
    REG32_SPILL_WR_LAST_PERIOD_MS   = 0x1C, // read-only
    REG64_SPILL_LAST_SPILL_TAI      = 0x20, // read-only
};

SpillControlModule::SpillControlModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        switch (getDeviceId()) {
        case DEVICE_ID_U40VE_RC:
        case DEVICE_ID_UT24VE:
            setBaseAddr(0x00C0, 0x3F);
            break;
        case DEVICE_ID_TRC:
            setBaseAddr(0x1300, 0x3F);
            break;
        default:;
        }
    }
}

QList<AFI_SDB_ID> SpillControlModule::getSupportedSdbIds()
{
    return {AFI_SDB_SPILL_CONTROL};
}

mlink::RegOpVector SpillControlModule::setConfigRegOp(const SpillConfig &spill)
{
    RegOpVector r;

    r.RegWrite32(REG32_SPILL_CTRL, 0); // disable spill

    r.RegWrite32(REG32_SPILL_DUR_PHASE_1_MS, spill.durPhase1);
    r.RegWrite32(REG32_SPILL_DUR_PHASE_2_MS, spill.durPhase2);
    r.RegWrite32(REG32_SPILL_DUR_PHASE_3_MS, spill.durPhase3);
    r.RegWrite32(REG32_SPILL_DUR_PHASE_4_MS, spill.durPhase4);
    r.RegWrite32(REG32_SPILL_DUR_PHASE_5_MS, spill.durPhase5);

    r.RegWrite32(REG32_SPILL_CTRL, (spill.source & 0xF) << 1); // enable spill
    r.RegWrite32(REG32_SPILL_SOFTSPILL_PERIOD_MS, spill.timerPeriod);

    return r;
}

bool SpillControlModule::writeConfig(const SpillConfig &config)
{
    auto r = setConfigRegOp(config);
    return regOpExecRebased(r);
}

mlink::RegOpVector SpillControlModule::getStatusRegOp(SpillStatus *st)
{
    RegOpVector r;
    r.RegRead32(REG32_SPILL_COUNT, &st->count);
    r.RegRead32(REG32_SPILL_CURRENT_TIME_MS, &st->curSpillTimeMs);
    r.RegRead16(REG16_SPILL_STATUS, &st->status.reg.all);
    r.RegRead32(REG32_SPILL_TTL_LAST_PERIOD_MS, &st->ttlLastPeriodMs);
    r.RegRead32(REG32_SPILL_WR_LAST_PERIOD_MS, &st->wrLastPeriodMs);
    r.RegRead64(REG64_SPILL_LAST_SPILL_TAI, &st->lastSpillTaiRaw);
    return r;
}

bool SpillControlModule::readStatus(SpillStatus *status)
{
    auto r = getStatusRegOp(status);
    return regOpExecRebased(r);
}

std::optional<SpillStatus> SpillControlModule::readStatus()
{
    SpillStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}
