#include "TdcCoreStatus.h"

#include <array>
#include <QStringList>

static const quint16 HPTDC_ERR_MASK = 0x7FF;
static const std::array<const char *, 16> HPTDC_ST_STR {
    "Vernier Error",
    "Coarse Error",
    "Channel Select Error",
    "L1 Buffer Parity Error",
    "Trigger FIFO Parity Error",
    "Trigger Matching State Error",
    "Readout FIFO Parity Error",
    "Readout State Error",
    "Setup Parity Error",
    "Control Parity Error",
    "JTAG Instruction Error",
    "TDC Have Readout Token",
    "Readout FIFO Full",
    "Readout FIFO Empty",
    "Trigger FIFO Full",
    "Trigger FIFO Empty"
};


bool HptdcStatus::isOk() const
{
    return (regValues[0] & HPTDC_ERR_MASK) && isDllLock();
}

bool HptdcStatus::isDllLock() const
{
    return regValues[1] & (1<<12);
}

bool HptdcStatus::hasError() const
{

    return !isDllLock() || (regValues[0] & 0x7F);
}

QString HptdcStatus::getSummary() const
{
    QStringList sList;
    for(int i=0; i<16; ++i)
        if(regValues[0] & (1<<i) & HPTDC_ERR_MASK)
            sList << HPTDC_ST_STR[i];

    //    quint8 readOutFifoOccup = regValues[1] & 0xFF;
    //    quint8 trigFifoOccup = (regValues[1]>>12) & 0xFF;

    //    quint8 occupL1Buffer[4];
    //    occupL1Buffer[0] = regValues[2] & 0xFF;
    //    occupL1Buffer[1] = (regValues[2]>>12) & 0xFF;
    //    occupL1Buffer[2] = regValues[3] & 0xFF;
    //    occupL1Buffer[3] = (regValues[4]>>12) & 0xFF;
    if(!isDllLock())
        sList << QString("Dll %1lock").arg(isDllLock()?"":"un");

    return sList.join(' ');
}

Severity TdcCoreStatus::getHptdcSeverity() const
{
    if (hptdcStatus.size() <= 0)
        return Severity::Unknown;
    if (getHptdcError())
        return Severity::Error;
    for (int hptdcId=0; hptdcId<hptdcStatus.size(); ++hptdcId) {
        const HptdcStatus &hptdc = hptdcStatus[hptdcId];
        if (hptdc.hasError())
            return Severity::Error;
        // TODO: check for warnings
    }
    return Severity::Normal;
}

bool TdcCoreStatus::hasHptdcErr() const
{
    if (getHptdcError())
        return true;
    for (int hptdcId=0; hptdcId<hptdcStatus.size(); ++hptdcId) {
        const HptdcStatus &hptdc = hptdcStatus[hptdcId];
        if (hptdc.hasError())
            return true;
    }
    return false;
}

QString TdcCoreStatus::getSummary() const
{
    QString str;
    auto hptdcErrorFlags = getHptdcError();
    if(hptdcErrorFlags)
        str += QString("HPTDC Err flags: %1").arg(hptdcErrorFlags, 0, 2);
    for (int hptdcId=0; hptdcId<hptdcStatus.size(); ++hptdcId) {
        const HptdcStatus &hptdc = hptdcStatus[hptdcId];
        if(!hptdc.getSummary().isEmpty()) {
            if(!str.isEmpty()) str += "\n";
            str += QString("HPTDC #%1:\n%2").arg(hptdcId).arg(hptdc.getSummary());
        }
    }
    if(str.isEmpty()) {
        str = "Ok";
    }
    return str;
}
