/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "WaveBlcModule.h"

#include "CalResult.h"
#include "WaveBlcConfig.h"
#include "WaveBlcStatus.h"
#include "base/fw_version.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {

union ctrl_reg_t {
    unsigned all;
    struct {
        unsigned maf_width: 5;
        unsigned unused_7_5: 3;
        unsigned maf_roundoff: 3;
        unsigned unused_11: 1;
        unsigned mode: 2;
        unsigned polarity: 1;
        unsigned enable: 1;
    } bit;
};

enum {
    REG32_ID = 0x0,
    REG_CONTROL = 0x2,
    REG_LOCK = 0x3,
    REG_CHANNEL_COUNT = 0x4,
    REG_COMMON_OFFSET = 0x5,
};

constexpr const std::initializer_list<std::pair<WaveBlcStatus::ChannelRegArray, quint16>> REG_ARRAY {
    {WaveBlcStatus::RawData, 0x08},
    {WaveBlcStatus::StaticOffset, 0x09},
    {WaveBlcStatus::Settings, 0x0A},
    {WaveBlcStatus::LoThr, 0x0B},
    {WaveBlcStatus::HiThr, 0x0C},
};

const quint16 REG_CONTROL_BIT_RESET = 1;
const quint16 REG_LOCK_BIT_LOCK = 1;
} // namespace

WaveBlcModule::WaveBlcModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
}

QList<AFI_SDB_ID> WaveBlcModule::getSupportedSdbIds()
{
    return {AFI_SDB_WAVE_BLC};
}

std::optional<WaveBlcStatus> WaveBlcModule::readStatus()
{
    if (!readStatus(&status))
        return {};
    return status;
}

bool WaveBlcModule::readStatus(WaveBlcStatus *result)
{
    WaveBlcStatus status;
    RegOpVector r;
    r.RegRead32(REG32_ID, &status.id);
    r.RegWrite16(REG_LOCK, 0);
    r.RegWrite16(REG_LOCK, REG_LOCK_BIT_LOCK);
    r.RegRead16(REG_CHANNEL_COUNT, &status.channelCount);
    for (auto it = REG_ARRAY.begin(); it != REG_ARRAY.end(); ++it) {
        r.RegRead16(it->second, &status.arrayAddress[it->first]);
    }
    r.RegWrite16(REG_LOCK, 0);
    if (!regOpExecRebased(r))
        return {};
    if (result)
        *result = status;
    return true;
}

bool WaveBlcModule::writeConfig(const WaveBlcConfig &config)
{
    if (!status.isValid()) {
        if (!readStatus())
            return false;
        if (!status.isValid())
            return false;
    }
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, REG_CONTROL_BIT_RESET);
    for (int i=0; i<status.channelCount; i++) {
        const auto &chConf = config.getChannel(i);
        ctrl_reg_t ctrl = {};
        ctrl.bit.maf_width = std::clamp(chConf.mafLength - 1, 0, 31);
        ctrl.bit.maf_roundoff = std::clamp((int)chConf.mafRoundoff, 0, 7);
        ctrl.bit.mode = chConf.mode;
        ctrl.bit.polarity = (chConf.polarity == WaveBlcChannelConfig::Positive) ? 0 : 1;
        ctrl.bit.enable = (chConf.mode == WaveBlcChannelConfig::Bypass) ? 0 : 1;
        r.RegWrite16(status.arrayAddress[WaveBlcStatus::Settings] + i, ctrl.all);
        r.RegWrite16(status.arrayAddress[WaveBlcStatus::StaticOffset] + i, chConf.staticOffset);
        int thr = chConf.thr;
        if (thr < 0) thr = -thr;
        r.RegWrite16(status.arrayAddress[WaveBlcStatus::LoThr] + i, -thr);
        r.RegWrite16(status.arrayAddress[WaveBlcStatus::HiThr] + i, thr);
    }
    if(isV1_1())
        r.RegWrite16(REG_COMMON_OFFSET, config.commonOffset);
    r.RegWrite16(REG_CONTROL, 0);
    return regOpExecRebased(r);
}

QMap<int, QVector<int> > WaveBlcModule::getAllChannelData(int repeat)
{
    if (!status.isValid())
        return {};
    RegOpVector r;
    QMap<int, QVector<quint16>> rawdata;
    for (int ch = 0; ch < status.channelCount; ch++) {
        auto &samples = rawdata[ch];
        samples.resize(repeat);
        for (int k = 0; k < repeat; k++)
            r.RegRead16(status.arrayAddress[WaveBlcStatus::RawData] + ch, &samples[k]);
    }
    if (!regOpExecRebased(r))
        return {};
    QMap<int, QVector<int> > result;
    for (auto it = rawdata.cbegin(); it != rawdata.cend(); ++it) {
        auto ch = it.key();
        const auto &samples = it.value();
        for (auto sample : samples)
            result[ch].push_back(static_cast<qint16>(sample));
    }
    return result;
}

bool WaveBlcModule::readBaseline(MultichannelBaselineStats *result, int repeat)
{
    const auto &allChannelData = getAllChannelData(repeat);
    if (allChannelData.empty())
        return {};
    auto stats = MultichannelBaselineStats(allChannelData);
    if (result)
        *result = stats;
    return true;
}

bool WaveBlcModule::isV1_1() const
{
    static const fw_version_t ver_1_1(1,1,0);
    return getVersion() >= ver_1_1;
}
