#include "AdcMpdTrigModel.h"

#include "AdcMpdTrigColumns.h"

#include <QTimer>

namespace {
bool AdcMpdTrigChConfig::* getConfigBoolFieldPtr(AdcMpdTrigColumns col) {
    switch (col) {
    case AdcMpdTrigColumns::COL_NAME: return &AdcMpdTrigChConfig::en;
    case AdcMpdTrigColumns::COL_POL: return &AdcMpdTrigChConfig::pol;
    default: return nullptr;
    }
    Q_UNREACHABLE();
}
inline Qt::CheckState boolToCheck(bool v)
{ return v ? Qt::Checked : Qt::Unchecked; }

AdcMpdTrigColumns getColumnByIndex(const QModelIndex &index)
{
    return static_cast<AdcMpdTrigColumns>(index.column());
}

static QMap<DeviceIndex, int> buildNchMap(const BaseDeviceAppConfig &config)
{
    QMap<DeviceIndex, int> nch;
    const auto &keys = config.knownConfigs.keys();
    for (auto index : keys) {
        nch.insert(index, TqdcCoreConfig::getNumberOfChannels(index.getDevId()));
    }
    return nch;
}

static bool needReset(const BaseDeviceAppConfig &config, const BaseDeviceAppConfig &newConfig)
{
    if (config.knownConfigs.keys() != newConfig.knownConfigs.keys())
        return true;
    const auto &oldNch = buildNchMap(config);
    const auto &newNch = buildNchMap(newConfig);
    if (oldNch != newNch)
        return true;
    for (const auto &ds : config.knownConfigs) {
        if (ds.enabled != newConfig.knownConfigs[ds.index].enabled)
            return true;
    }
    return false;
}

static bool hasChanges(const ModularDeviceConfig &c1, const ModularDeviceConfig &c2)
{
    if(c1.adcMpdTrig.groups != c2.adcMpdTrig.groups)
        return true;
    return false;
}

static bool hasChanges(const BaseDeviceAppConfig &c1, const BaseDeviceAppConfig &c2)
{
    if (hasChanges(c1.defaultSetup, c2.defaultSetup))
        return true;

    for (auto it1=c1.knownConfigs.cbegin(); it1!=c1.knownConfigs.cend(); ++it1) {
        auto it2 = c2.knownConfigs.constFind(it1.key());
        if (hasChanges(*it1, *it2))
            return true;
    }
    return false;
}
}

AdcMpdTrigModel::AdcMpdTrigModel(QObject *parent)
    : QAbstractItemModel(parent)
    , delayedSendConfigTimer(new QTimer(this))
{
    delayedSendConfigTimer->setInterval(10);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &AdcMpdTrigModel::postCheckStateChange);
}

QVariant AdcMpdTrigModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return {};
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::CheckStateRole:
        if(getColumnByIndex(index) == AdcMpdTrigColumns::COL_NAME)
            return checkStateData(index);
        break;
    case Qt::UserRole:
        return limitsData(index);
    }
    return {};
}

bool AdcMpdTrigModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;

    switch (role) {
    case Qt::EditRole:
        return setEditRole(index, value);
    case Qt::CheckStateRole:
        return setCheckStateData(index, value);
    }

    return false;
}

Qt::ItemFlags AdcMpdTrigModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = getColumnByIndex(index);
    switch (col) {
    case AdcMpdTrigColumns::COL_NAME:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    default:;
    }
    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant AdcMpdTrigModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return {};
    if(role != Qt::DisplayRole)
        return {};
    switch (static_cast<AdcMpdTrigColumns>(section)) {
    case AdcMpdTrigColumns::COL_NAME: return "Name";
    case AdcMpdTrigColumns::COL_GROUP_THR: return "Thr";
    case AdcMpdTrigColumns::COL_ADC_CH: return "ADC ch";
    case AdcMpdTrigColumns::COL_POL: return "Polarity";
    // case AdcMpdTrigColumns::COL_PEAK_STATUS: return "Peak";
    default:;
    }
    return {};
}

QModelIndex AdcMpdTrigModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const auto *parentItem = parent.isValid() ?
        static_cast<AdcMpdTrigModelItem*>(parent.internalPointer()) :
        &rootItem;
    AdcMpdTrigModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex AdcMpdTrigModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const auto *childItem = static_cast<AdcMpdTrigModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    AdcMpdTrigModelItem *parentItem = childItem->parent;
    if(parentItem == &rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int AdcMpdTrigModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<AdcMpdTrigModelItem*>(parent.internalPointer()) :
        &rootItem;
    return parentItem->childCount();
}

int AdcMpdTrigModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return static_cast<int>(AdcMpdTrigColumns::COL_TOTAL);
}

void AdcMpdTrigModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    sendDataChanged({Qt::EditRole, Qt::UserRole});
}

int AdcMpdTrigModel::getDevRow(const DeviceIndex &di) const
{
    return -1;
}

void AdcMpdTrigModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if (config == newConfig)
        return;


    bool performReset = needReset(config, newConfig);
    bool hasAnyChanges = performReset ? true : hasChanges(config, newConfig);

    config = newConfig;
    if (performReset) {
        doReset();
    } else {
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void AdcMpdTrigModel::setAdcMpdTrigBp(const DeviceIndex &index, const AdcMpdTrigBuildParams &bp)
{
    auto it = devBp.find(index);
    if(it != devBp.end() && *it == bp)
        return;

    devBp[index] = bp;
    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;

    bool hasChanges = false;
    if(devIt->adcMpdTrig.groups.size() != bp.trigGroupNum) {
        devIt->adcMpdTrig.groups.resize(bp.trigGroupNum);
        hasChanges = true;
    }
    for(auto it=devIt->adcMpdTrig.groups.begin(); it!=devIt->adcMpdTrig.groups.end(); ++it) {
        if(it->channels.size() == bp.chNumPerGroup)
            continue;
        it->channels.resize(bp.chNumPerGroup);
        hasChanges = true;
    }

    if(hasChanges) {
        doReset();
        postCheckStateChange();
    }
    sendDataChanged(QVector<int>{Qt::EditRole, Qt::UserRole});
}

void AdcMpdTrigModel::setDeviceSet(const DeviceIndexSet &set)
{
    if(deviceSet == set)
        return;
    deviceSet = set;

    // for(auto i : dev_status.keys()) {
    //     if(!deviceSet.contains(i))
    //         dev_status.remove(i);
    // }

    doReset();
}

void AdcMpdTrigModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    emit configChanged(this->config);
}

void AdcMpdTrigModel::enqueueConfigSend()
{
    delayedSendConfigTimer->start();
}

void AdcMpdTrigModel::doReset()
{
    beginResetModel();
    rootItem.children.clear();
    const auto &setups = config.knownConfigs;
    if (config.hasEnabledDevices()) {
        auto allItem = rootItem.createChild();

        for (auto devIt = setups.cbegin(); devIt != setups.cend(); ++devIt) {
            if (!devIt->enabled)
                continue;
            auto devItem = allItem->createChild();
            devItem->dev = devIt.key();
            auto devCfg = devIt->adcMpdTrig;
            for(auto group=0; group<devCfg.groups.size(); ++group) {
                auto groupItem = devItem->createChild();
                groupItem->group = group;
                auto groupCfg = devIt->adcMpdTrig.groups[group];
                for (int ch = 0; ch<groupCfg.channels.size(); ch++) {
                    auto chItem = groupItem->createChild();
                    chItem->ch = ch;
                }
            }
        }
    }
    endResetModel();
}

void AdcMpdTrigModel::sendDataChanged(QVector<int> roles, const QModelIndex parent)
{
    auto rows = rowCount(parent);

    if(!rows)
        return;

    auto top = index(0,0, parent);
    auto last = index(rows-1, columnCount(top)-1, parent);
    emit dataChanged(top, last, roles);

    for(auto r=0; r<rows; ++r) {
        auto chIndex = index(r, 0, parent);
        sendDataChanged(roles, chIndex);
    }
}

QVariant AdcMpdTrigModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case AdcMpdTrigColumns::COL_NAME:
        if(item->dev.isNull())
            return "All";
        if(item->group == -1)
            return item->dev.getSerialStr();
        if(item->ch == -1)
            return QString("group #%1").arg(item->group);
        return QString::number(item->ch);
    }

    return {};
}

QVariant AdcMpdTrigModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case AdcMpdTrigColumns::COL_ADC_CH:
        if(item->ch == -1)
            return {};
        return config.knownConfigs.value(item->dev).adcMpdTrig.
            groups.value(item->group).channels.value(item->ch).adcCh;
    case AdcMpdTrigColumns::COL_GROUP_THR:
        if(item->group == -1 || item->ch != -1)
            return {};
        return config.knownConfigs.value(item->dev).adcMpdTrig.
            groups.value(item->group).trigThr;
    case AdcMpdTrigColumns::COL_POL:
        return checkStateData(index);
    }

    return {};
}

QVariant AdcMpdTrigModel::checkStateData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    if (!item)
        return {};

    const auto &col = getColumnByIndex(index);
    auto ptr = getConfigBoolFieldPtr(col);
    if(!ptr)
        return {};
    return getRecursiveState(index, ptr);
}

QVariant AdcMpdTrigModel::limitsData(const QModelIndex &index) const
{
    switch (getColumnByIndex(index)) {
    case AdcMpdTrigColumns::COL_ADC_CH:
    {
        const auto item = getItemByIndex(index);
        if(!item)
            return {};
        auto devIt = devBp.constFind(item->dev);
        if(devIt == devBp.cend())
            return {};
        return devIt->adcChNum;
    }
    case AdcMpdTrigColumns::COL_GROUP_THR:
        return std::numeric_limits<int>::max();
    }

    return {};
}

bool AdcMpdTrigModel::setEditRole(const QModelIndex &index, const QVariant &value)
{
    const auto item = getItemByIndex(index);
    auto col = getColumnByIndex(index);
    switch (col) {
    case AdcMpdTrigColumns::COL_GROUP_THR:
    {
        Q_ASSERT(item->ch == -1);
        Q_ASSERT(config.knownConfigs.contains(item->dev));
        if(!config.knownConfigs.contains(item->dev))
            return false;

        auto &devCfg = config.knownConfigs[item->dev].adcMpdTrig;
        if(devCfg.groups.size() <= item->group)
            return false;

        auto &groupCfg = devCfg.groups[item->group];
        groupCfg.trigThr = value.toUInt();
        enqueueConfigSend();
        return true;
    }
    case AdcMpdTrigColumns::COL_ADC_CH:
    {
        if(item->ch == -1)
            return {};

        Q_ASSERT(config.knownConfigs.contains(item->dev));
        if(!config.knownConfigs.contains(item->dev))
            return false;

        auto &devCfg = config.knownConfigs[item->dev].adcMpdTrig;
        if(devCfg.groups.size() <= item->group)
            return false;

        auto &groupCfg = devCfg.groups[item->group];
        if(groupCfg.channels.size() <= item->ch)
            return false;

        auto &chCfg = groupCfg.channels[item->ch];
        chCfg.adcCh = value.toUInt();
        enqueueConfigSend();
        return true;
    }
    default:;
    }
    return false;
}

bool AdcMpdTrigModel::setCheckStateData(const QModelIndex &index, const QVariant &value)
{
    const auto item = getItemByIndex(index);
    if(item->ch == -1) {
        auto rows = rowCount(index);
        bool ok = true;
        for(int r=0; r<rows; ++r) {
            ok &= setCheckStateData(this->index(r,index.column(),index), value);
        }
        return ok;
    } else {
        Q_ASSERT(!item->dev.isNull());
        auto col = getColumnByIndex(index);
        auto ptr = getConfigBoolFieldPtr(col);
        if(!ptr)
            return false;
        Q_ASSERT(config.knownConfigs.contains(item->dev));
        if(!config.knownConfigs.contains(item->dev))
            return false;

        auto &devCfg = config.knownConfigs[item->dev].adcMpdTrig;
        if(devCfg.groups.size() <= item->group)
            return false;

        auto &groupCfg = devCfg.groups[item->group];
        if(groupCfg.channels.size() <= item->ch)
            return false;

        auto &chCfg = groupCfg.channels[item->ch];

        Q_ASSERT(value!=Qt::PartiallyChecked);

        chCfg.*ptr = (value==Qt::Checked) ? true : false;

        enqueueConfigSend();
    }
    return true;
}

AdcMpdTrigModelItem *AdcMpdTrigModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<AdcMpdTrigModelItem*>(index.internalPointer());
}

QVariant AdcMpdTrigModel::getRecursiveState(const QModelIndex &mIndex, bool AdcMpdTrigChConfig::* ptr) const
{
    auto rows = rowCount(mIndex);
    if(rows) {
        QVariant allSt;
        for(int r=0; r<rows; ++r) {
            auto st = getRecursiveState(index(r,0,mIndex), ptr);
            if(allSt.isNull())
                allSt = st;
            else if(allSt != st)
                return Qt::PartiallyChecked;
        }
        return allSt.isNull() ? boolToCheck(AdcMpdTrigChConfig::DEF_CH_EN) : allSt;
    } else {
        auto item = getItemByIndex(mIndex);
        const auto &chCfg = config.knownConfigs.value(item->dev).adcMpdTrig.groups.value(item->group).channels.value(item->ch);
        return boolToCheck(chCfg.*ptr);
    }
}
