#include "AdcSerDesCalDialog.h"
#include "ui_AdcSerDesCalDialog.h"

#include "base/DeviceIndexEnabled.h"
#include "base/DeviceIndexConverter.h"
#include "base/Globals.h"
#include "DeserializersControl/DeserializersControlBuildParams.h"

#include "SingleSerDesDelayWidget.h"

#include <QFileDialog>
#include <QSettings>

namespace {
static const QString GROUP_NAME = "";
}

AdcSerDesCalDialog::AdcSerDesCalDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::AdcSerDesCalDialog)
{
    ui->setupUi(this);
    connect(ui->pushButtonClose, &QAbstractButton::clicked,
            this, &QDialog::close);
    ui->progressBar->setValue(0);

    AbstractConfigConverter<DeviceIndex>::registerTypes();
//    ui->verticalLayoutSingleAdc->addWidget(new SingleSerDesDelayWidget(0, this));
    resoreLocalSettings();
    if(!Globals::getDebugMode())
        ui->groupBoxSaveCalRes->hide();
}

AdcSerDesCalDialog::~AdcSerDesCalDialog()
{
    delete ui;
}

void AdcSerDesCalDialog::onFsmStateChange(const FsmState &state)
{
    if(state == FsmState::FindAdcDelay) {
        for(auto it=devProgress.begin(); it!=devProgress.end(); ++it)
            it.value() = 0;
        ui->progressBar->setValue(0);
    }

    const bool en = state==FsmState::Ready;
    ui->progressBar->setEnabled(state==FsmState::FindAdcDelay);
    ui->pushButtonCalibrate->setEnabled(en);

    for(int i = 0; i<ui->verticalLayoutSingleAdc->count(); ++i) {
        auto item = ui->verticalLayoutSingleAdc->itemAt(i);
        auto ssdd = qobject_cast<SingleSerDesDelayWidget*>(item->widget());
        if(!ssdd)
            continue;
        ssdd->setControlEnabled(en);
    }
}

void AdcSerDesCalDialog::setDeserializersControlBp(const DeviceIndex &index,
                                                   const DeserializersControlBuildParams &bp)
{
    if(!devSet.contains(index))
        return;
    if(devDesCnt.value(index,0) == bp.desCnt)
        return;

    devDesCnt[index] = bp.desCnt;
    if(index == curDev)
        restoreGui();
}

void AdcSerDesCalDialog::setDevSerDesDelay(const DeviceIndex & index, const AdcSerDesDelayMap &delay)
{
    if(!devSet.contains(index))
        return;
    if(allDev.map.value(index).map == delay.map)
        return;

    allDev.map[index] = delay;
    if(curDev == index)
        restoreGui();

    emit devSerDesDelayChanged(index, allDev.map[index]);
}

void AdcSerDesCalDialog::setDeviceList(const DeviceIndexEnabledMap &new_devices)
{
    const auto &newSet = new_devices.getEnabled();
    auto devCb = ui->comboBoxCurDev;
    // to Del
    for(auto index : devSet - newSet) {
        devProgress.remove(index);
        devDesCnt.remove(index);
        allDev.map.remove(index);
        int cbItem = devCb->findData(QVariant::fromValue(index));
        if(cbItem != -1)
            devCb->removeItem(cbItem);
    }

    // to Add
    for(auto index : newSet - devSet) {
        devCb->insertItem(0, index.getIdent(), QVariant::fromValue(index));
    }
    devSet = newSet;
}

void AdcSerDesCalDialog::progressUpdated(const DeviceIndex &index, int progress)
{
    int minProgress = 100;
    for(auto i: devSet) {
        if(i == index)
            devProgress[i] = progress;
        minProgress = std::min(minProgress, devProgress.value(i));
    }
    ui->progressBar->setValue(minProgress);
}

void AdcSerDesCalDialog::onDeviceChange(const DeviceIndex &di)
{
    curDev = di;
    if(di.isNull() && ui->comboBoxCurDev->count())
        ui->comboBoxCurDev->setCurrentIndex(0);
    restoreGui();
}

void AdcSerDesCalDialog::restoreGui()
{
    int cbItem = ui->comboBoxCurDev->findData(QVariant::fromValue(curDev));
    ui->comboBoxCurDev->setCurrentIndex(cbItem);

    const auto &devDelay = allDev.map.value(curDev);
    auto l = ui->verticalLayoutSingleAdc;

    for(int itemId = 0; itemId<l->count();) {
        auto item = l->itemAt(itemId);
        auto ssdd = qobject_cast<SingleSerDesDelayWidget*>(item->widget());
        if(ssdd) {
            auto adcId = ssdd->getId();
            if(adcId>devDesCnt.value(curDev,0)-1) {
                l->removeWidget(ssdd);
                delete ssdd;
                continue;
            }

            ssdd->setDelay(devDelay.map[adcId]);
        }
        ++itemId;
    }
    for(int adcId = l->count(); adcId<devDesCnt.value(curDev,0); ++adcId) {
        auto ssdd = new SingleSerDesDelayWidget(adcId, this);
        ssdd->setDelay(devDelay.map[adcId]);
        connect(ssdd, &SingleSerDesDelayWidget::clockUpdated,
                this, &AdcSerDesCalDialog::onClockDelayChanged);
        connect(ssdd, &SingleSerDesDelayWidget::dataUpdated,
                this, &AdcSerDesCalDialog::onDataDelayChanged);
        l->addWidget(ssdd);
    }
}

void AdcSerDesCalDialog::resoreLocalSettings()
{
    QSettings s;
    if(Globals::getDebugMode()) {
        ui->groupBoxSaveCalRes->setChecked(s.value(GROUP_NAME+"/en").toBool());
        ui->lineEditPath->setText(s.value(GROUP_NAME+"/path").toString());
    }
    ui->checkBoxHardMode->setChecked(s.value(GROUP_NAME+"/hardMode").toBool());
}

void AdcSerDesCalDialog::saveLocalSettings()
{
    QSettings s;
    s.beginGroup(GROUP_NAME);
    if(Globals::getDebugMode()) {
        s.setValue("en", ui->groupBoxSaveCalRes->isChecked());
        s.setValue("path", ui->lineEditPath->text());
    }
    s.setValue("hardMode", ui->checkBoxHardMode->isChecked());
    s.endGroup();
}

void AdcSerDesCalDialog::validateSaveFolder()
{
    if(!ui->groupBoxSaveCalRes->isChecked()) {
        savePathVaild = true;
        return;
    } else {
        QFileInfo fi(ui->lineEditPath->text());
        savePathVaild = fi.isDir() && fi.isWritable();
    }

    auto p = ui->lineEditPath->palette();
    auto b = p.brush(ui->lineEditPath->backgroundRole());
    b.setColor(savePathVaild ? Qt::transparent : QColor(0xFFc0c0));
    p.setBrush(ui->lineEditPath->backgroundRole(), b);
    ui->lineEditPath->setPalette(p);
}

void AdcSerDesCalDialog::on_comboBoxCurDev_currentIndexChanged(int)
{
    onDeviceChange(ui->comboBoxCurDev->currentData().value<DeviceIndex>());
}

void AdcSerDesCalDialog::on_groupBoxSaveCalRes_toggled(bool en)
{
    ui->groupBoxSaveCalRes->setEnabled(en);
    validateSaveFolder();
}

void AdcSerDesCalDialog::on_lineEditPath_textChanged(const QString &)
{
    validateSaveFolder();
}

void AdcSerDesCalDialog::on_pushButtonBrowse_clicked()
{
    const auto &dirPath = QFileDialog::getExistingDirectory(this, "FindAdcClockDelay",
                                                            ui->lineEditPath->text());
    if(dirPath.isEmpty())
        return;

    ui->lineEditPath->setText(dirPath);
    saveLocalSettings();
    validateSaveFolder();
}

void AdcSerDesCalDialog::on_checkBoxHardMode_toggled(bool)
{
    saveLocalSettings();
}

void AdcSerDesCalDialog::on_pushButtonCalibrate_clicked()
{
    emit requestCalibration(ui->checkBoxHardMode->isChecked(), ui->lineEditPath->text());
}

void AdcSerDesCalDialog::onClockDelayChanged(int id, int val)
{
    if(curDev.isNull())
        return;
    Q_ASSERT(id < devDesCnt.value(curDev,0));
    allDev.map[curDev].map[id].clock = val;
    emit devSerDesDelayChanged(curDev, allDev.map[curDev]);
}

void AdcSerDesCalDialog::onDataDelayChanged(int id, int val)
{
    if(curDev.isNull())
        return;
    Q_ASSERT(id < devDesCnt.value(curDev,0));
    allDev.map[curDev].map[id].data = val;
    emit devSerDesDelayChanged(curDev, allDev.map[curDev]);
}
