/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceCommand.h"
#include "RunLogic/RunLogicConfig.h"
#include "RunLogic/RunLogicController.h"
#include "RunLogic/RunLogicFsm.h"
#include "RunLogicWidget.h"
#include "base/DeviceIndex.h"
#include "base/gui_utils.h"
#include "ui_RunLogicWidget.h"

#include <QStyle>
#include <QTimer>
#include <QBitmap>

namespace {
inline const double timeLimitStep = 1e-3;
inline const double progressScale = 1e6;
inline const int throttleTimeMs = 50;
} // namespace


class RunLogicWidgetPrivate
{
public:
    bool controllerWasSet = false;
    RunLogicConfig config;
    RunLogicMultiStatus status;
    RunLogicFsm::State state = RunLogicFsm::Init;
    QTimer timer;
};

RunLogicWidget::RunLogicWidget(QWidget *parent)
    : QWidget(parent)
    , d_ptr(new RunLogicWidgetPrivate())
    , ui(new Ui::RunLogicWidget)
{
    ui->setupUi(this);
    setEnabled(false); // until controller is set
    {
        RecursiveWidgetSignalBlocker blocker(this);
        ui->spinBoxEventLimit->setRange(1, 1'000'000'000);
        ui->doubleSpinBoxTimeLimit->setRange(1, 1e6);
        ui->doubleSpinBoxTimeLimit->setSingleStep(1);
        ui->doubleSpinBoxTimeLimit->setDecimals(3);
        ui->doubleSpinBoxTimeLimit->setSuffix("s");
    }
    ui->progressBar->setRange(0, progressScale);
    ui->labelRunTime->clear();
    ui->labelEventCount->clear();
    ui->labelStatus->clear();

    QFont font{"Monospace"};
    ui->labelEventCount->setFont(font);
    ui->labelRunTime->setFont(font);
    ui->labelStatus->setFont(font);
    // set labels to 13 chars
    ui->labelEventCount->setText("             ");
    ui->labelRunTime->setText("             ");

    QString styleLargeFont = "font-size: 18pt";
    ui->labelState->setFont(styleLargeFont);
    ui->pushButtonStart->setStyleSheet(styleLargeFont);
    ui->pushButtonReset->setStyleSheet(styleLargeFont);
    ui->pushButtonStart->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButtonReset->setIcon(style()->standardIcon(QStyle::SP_MediaStop));

//    setController(new RunLogicController(this));

    updateEnableState();
    updateStatus();

    Q_D(RunLogicWidget);

    connect(&d->timer, &QTimer::timeout,
            this, &RunLogicWidget::notifyProgressUpdate);
    d->timer.setSingleShot(true);
    d->timer.setInterval(1000);
    restoreGuiFromConfig();
}

RunLogicWidget::~RunLogicWidget()
{
    delete ui;
    delete d_ptr;
}

void RunLogicWidget::setController(RunLogicController *controller)
{
    Q_D(RunLogicWidget);
    d->controllerWasSet = controller;
    if(controller) {
        connect(controller, &RunLogicController::runLogicConfigChanged,
                this, &RunLogicWidget::setRunLogicConfig);
        connect(controller, &RunLogicController::runLogicStatusChanged,
                this, &RunLogicWidget::setRunLogicStatus);
        connect(controller, &RunLogicController::stateChanged,
                this, &RunLogicWidget::setState);

        connect(this, &RunLogicWidget::performAction,
                controller, &RunLogicController::performAction);
        connect(this, &RunLogicWidget::runLogicConfigChanged,
                controller, &RunLogicController::setRunLogicConfig);
    }
    updateEnableState();
}

void RunLogicWidget::notifyProgressUpdate()
{
    Q_D(RunLogicWidget);
    if (d->timer.isActive() && (d->timer.interval() - d->timer.remainingTime()) < throttleTimeMs)
        return;
    updateEnableState();
    updateStatus();
    updateProgress();
    d->timer.start();
}

void RunLogicWidget::setRunLogicStatus(const RunLogicMultiStatus &s)
{
    Q_D(RunLogicWidget);
    d->status = s;
    notifyProgressUpdate();
}

void RunLogicWidget::setState(int st)
{
    Q_D(RunLogicWidget);
    d->state = RunLogicFsm::State(st);
    notifyProgressUpdate();
}

void RunLogicWidget::setRunLogicConfig(const RunLogicConfig &newConfig)
{
    Q_D(RunLogicWidget);
    if (newConfig == d->config)
        return;
    d->config = newConfig;
    restoreGuiFromConfig();
}

void RunLogicWidget::updateStatus()
{
    Q_D(RunLogicWidget);
    if (!d->controllerWasSet)
        return;
    auto multiStatus = d->status;

    QPixmap pixmap;
    if (!multiStatus.invalid.empty()) {
        pixmap = style()->standardPixmap(QStyle::SP_MessageBoxCritical);
    } else if (!multiStatus.missing.empty()) {
        pixmap = style()->standardPixmap(QStyle::SP_MessageBoxWarning);
    } else if (!multiStatus.running.empty()) {
        pixmap = style()->standardPixmap(QStyle::SP_MediaPlay);
    } else if (multiStatus.allFinished()) {
        pixmap = style()->standardPixmap(QStyle::SP_MediaPause);
    } else {
        pixmap = style()->standardPixmap(QStyle::SP_MediaStop);
    }
    ui->labelState->setPixmap(pixmap);
    ui->labelState->setToolTip(multiStatus.getFullText());
}

void RunLogicWidget::updateProgress()
{
    Q_D(const RunLogicWidget);
    if (!d->controllerWasSet)
        return;
    const auto &multiStatus = d->status;

    auto runTimeStr = QString("%L1").arg(multiStatus.minRunTime() * timeLimitStep, 13);
    auto eventCountStr = QString("%L1").arg(multiStatus.minEventCount(), 13);
    ui->labelRunTime->setText(runTimeStr);
    ui->labelEventCount->setText(eventCountStr);

    QStringList statusText = multiStatus.statusText();
    ui->labelStatus->setText(statusText.join("\n"));

    ui->progressBar->setValue(multiStatus.progress() * progressScale);
}

void RunLogicWidget::restoreGuiFromConfig()
{
    Q_D(const RunLogicWidget);
    {
        RecursiveWidgetSignalBlocker blocker(this);
        ui->checkBoxEventLimit->setChecked(d->config.eventLimitEnable);
        ui->checkBoxTimeLimit->setChecked(d->config.timeLimitEnable);
        ui->spinBoxEventLimit->setValue(d->config.eventLimit);
        ui->doubleSpinBoxTimeLimit->setValue(d->config.timeLimit * timeLimitStep);
    }
    updateEnableState();
}

void RunLogicWidget::updateEnableState()
{
    Q_D(const RunLogicWidget);
    setEnabled(d->controllerWasSet);
    if (!d->controllerWasSet)
        return;
    const auto &multiStatus = d->status;
    const auto &state = d->state;

    ui->spinBoxEventLimit->setEnabled(d->config.eventLimitEnable);
    ui->doubleSpinBoxTimeLimit->setEnabled(d->config.timeLimitEnable);
    ui->progressBar->setEnabled(d->config.eventLimitEnable || d->config.timeLimitEnable);

    ui->pushButtonReset->setEnabled(RunLogicFsm::allowedReset(state));
    ui->pushButtonStart->setEnabled(RunLogicFsm::allowedRunToggle(state) && !multiStatus.allFinished());

    if (RunLogicFsm::transitionAllowed(state, RunLogicFsm::SentRun)) {
        ui->pushButtonStart->setText("Start");
        ui->pushButtonStart->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    }
    if (RunLogicFsm::transitionAllowed(state, RunLogicFsm::SentPause)) {
        ui->pushButtonStart->setText("Pause");
        ui->pushButtonStart->setIcon(style()->standardIcon(QStyle::SP_MediaPause));
    }
}

void RunLogicWidget::on_checkBoxTimeLimit_toggled(bool checked)
{
    Q_D(RunLogicWidget);
    d->config.timeLimitEnable = checked;
    updateEnableState();
    emit runLogicConfigChanged(d->config);
}

void RunLogicWidget::on_checkBoxEventLimit_toggled(bool checked)
{
    Q_D(RunLogicWidget);
    d->config.eventLimitEnable = checked;
    updateEnableState();
    emit runLogicConfigChanged(d->config);
}

void RunLogicWidget::on_doubleSpinBoxTimeLimit_valueChanged(double arg1)
{
    Q_D(RunLogicWidget);
    d->config.timeLimit = arg1 / timeLimitStep;
    emit runLogicConfigChanged(d->config);
}

void RunLogicWidget::on_spinBoxEventLimit_valueChanged(int arg1)
{
    Q_D(RunLogicWidget);
    d->config.eventLimit = arg1;
    emit runLogicConfigChanged(d->config);
}

void RunLogicWidget::on_pushButtonReset_clicked()
{
    Q_D(RunLogicWidget);
    if (!d->controllerWasSet)
        Q_ASSERT(nullptr);
    emit performAction(RunLogicController::Reset);
}

void RunLogicWidget::on_pushButtonStart_clicked()
{
    Q_D(RunLogicWidget);
    if (!d->controllerWasSet)
        Q_ASSERT(nullptr);
    emit performAction(RunLogicController::Toggle);
}
