/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef FLASHDEVTASK_H
#define FLASHDEVTASK_H

#include <vector>

#include <QByteArray>
#include <QDebug>
#include <QMetaType>

#include "FlashIdData.h"

enum class FlashImageType { Golden, Regular };
Q_DECLARE_METATYPE(FlashImageType)
enum class FlashOp { None, Read, Write, Verify, Erase };

inline auto flashOpStr(FlashOp op)
{
    switch (op) {
    case FlashOp::None: return "none";
    case FlashOp::Read: return "read";
    case FlashOp::Write: return "write";
    case FlashOp::Verify: return "verify";
    case FlashOp::Erase: return "erase";
    }
    return "unknown";
}

inline auto flashImageTypeStr(FlashImageType imageType)
{
    switch (imageType) {
    case FlashImageType::Golden: return "golden";
    case FlashImageType::Regular: return "regular";
    }
    return "unknown";
}

struct FlashImage
{
    QString fileName;
    QByteArray data;
    FlashImageType type = FlashImageType::Golden;
};
Q_DECLARE_METATYPE(FlashImage)

struct FlashDevTask
{
    enum class State {Init, Running, WritingFirstSector, Complete, Aborted };
    static inline auto taskStateStr(State state)
    {
        switch (state) {
        case State::Init: return "init";
        case State::Running: return "running";
        case State::WritingFirstSector: return "writingFirstSector";
        case State::Complete: return "complete";
        case State::Aborted: return "aborted";
        }
        return "unknown";
    }
    struct Pos {
        void start(uint32_t first, uint32_t last) {
            Q_ASSERT(first <= last);
            begin = first;
            end = last;
            current = begin;
            state = State::Running;
        }
        int advance(int value, bool writeCase) {
            Q_ASSERT(current >= begin);
            Q_ASSERT(current <= end);
            Q_ASSERT(begin <= end);
            current += value;
            current = std::clamp(current, begin, end);

            if(writeCase && state == State::WritingFirstSector) {
                state = State::Complete;
                return current;
            }

            if (current >= end) {
                if(writeCase) {
                    state = State::WritingFirstSector;
                    current = 0;
                } else {
                    state = State::Complete;
                }
            }
            return current;
        }
        void abort() {
            state = State::Aborted;
        }
        bool isAborted() const { return state == State::Aborted; }
        bool isRunning() const { return state == State::Running || state == State::WritingFirstSector; }
        bool isWritingFirstSector() const { return state == State::WritingFirstSector; }
        bool isComplete() const { return state == State::Complete; }
        State getState() const { return state; }
        int size() const { return std::max(0, end - begin); }
        int offset() const { return std::clamp(current - begin, 0, size()); }
        int remains() const  { return std::clamp(end - current, 0, size()); }
        int current = 0;
        int begin = 0;
        int end = 0;
    private:
        State state = State::Init;
    };
    FlashDevTask() = default;
    FlashDevTask(const FlashImage &image, const FlashIdData &flashIdData)
        : image(image)
        , flashIdData(flashIdData)
        , flashImageData(makeMultibootImage())
    {
        if (image.data.size() > getMaxImageSize()) {
            qWarning().noquote() << "Image size exceeds memory capacity";
        }
    }

    void start(uint32_t first, uint32_t last) {
        pos.start(first, last);
        qDebug().noquote() << QString("Started %1").arg(flashOpStr(op));
    }
    int advance(int value) { return pos.advance(value, op==FlashOp::Write); }
    void abort() { pos.abort(); }
    bool isAborted() const { return pos.isAborted(); }
    bool isComplete() const { return pos.isComplete(); }
    bool isRunning() const { return pos.isRunning(); }
    QString stateStr() const { return taskStateStr(pos.getState()); }

    uint32_t startAddr() const;
    QString getImageFileName() const { return image.fileName; }
    FlashImageType getImageType() const { return image.type; }
    uint32_t imageSize() const { return flashImageData.size(); }
    int getMaxImageSize() const { return flashIdData.getFlashSizeBytes() / 2; }
    std::vector<uint16_t> getBlockData() const;
    std::vector<uint16_t> getSectorData() const;
    std::vector<uint16_t> getImageData(int chunkSize) const;
    QString getProgressStr() const;
    QString getOpStr() const { return flashOpStr(op); }
    QString taskStatusStr() const;
    void setUseAddr4bMode(bool v) {
        useAddr4bMode = v;
        flashImageData = makeMultibootImage();
    }

    FlashOp op = FlashOp::None;
    Pos pos;

private:
    static bool patchGoldenImage(QByteArray *bitstream, uint32_t bootAddr, bool useAddr4bMode);
    QByteArray makeMultibootImage() const;

    bool useAddr4bMode = false;
    FlashImage image;
    FlashIdData flashIdData;
    QByteArray flashImageData;
};

Q_DECLARE_METATYPE(FlashDevTask)

#endif // FLASHDEVTASK_H
