#include "MpdDataDecoder.h"

#include <QDateTime>
#include <QDebug>

#include <QJsonObject>
#include <QJsonDocument>

#include "base/DeviceId.h"
#include "base/RcRunId.h"
#include "base/bmc/fpga_mcu_regs_addr.h"
#include "mregdev/MregDeviceAdcm.h"
#include "device-modules/OneWire/OneWire.h"

#include "DeviceHelper.h"
#include "HptdcError.h"
#include "HistParams.h"

namespace {
// 14-bit signed value, 1/16 degrees Celsius
double decodeTemp_v1(qint32 val)
{
    val &= 0x1FFF;
    val = (val << 19) >> 19;
    return val / 32.0;
}

QString get_raw_data_prefix(const quint32 *word,const void *firstWord)
{
    const long ev_hdr_offset = word - reinterpret_cast<const quint32*>(firstWord);
    return QString("[%1] %2").arg(ev_hdr_offset,3,10,QChar('0')).arg(*word, 8, 16, QChar('0'));
}

enum {
    REG_TEMPERATURE = mlink::MregDeviceAdcm::ADCM16_TEMPERATURE_REG, //0x4A, // 74
    REG_FW_REV = mlink::MregDeviceAdcm::ADCM16_FW_REV_REG, //0x4A, // 74
};

}

MpdDataDecoder::MpdDataDecoder(QObject *p)
    : QObject(p)
{
}

void MpdDataDecoder::reset()
{
    analyzerStats.clear();
    decodedData.clear();
    lastErrStr.clear();
    rawDataTxt.clear();
}

void MpdDataDecoder::requestRawData(int events)
{
    rawDataTxt.clear();
    rawDataEvents = events;
}

void MpdDataDecoder::setRawDataEveryEv(bool everyEv)
{
    rawDataEveryEv = everyEv;
}

void MpdDataDecoder::invertUpdated(QMap<DeviceIndex, QVector<bool> > _invertMap)
{

}

void MpdDataDecoder::set_can_skip(bool skip)
{
    can_skip = skip;
}

void MpdDataDecoder::prepareForStop(quint32 evNum)
{
    if(analyzerStats.lastEvNum == evNum){
        emit gotStopEvent(evNum);
    } else {
        lookForStop = true;
        stopEvNum = evNum;
    }
}

void MpdDataDecoder::startWrite(bool rawData)
{
    writeDecodedEnabled = !rawData;
}

void MpdDataDecoder::stopWrite()
{
    writeDecodedEnabled = false;
}

void MpdDataDecoder::dataAccuired(const QByteArray &newData)
{
    if(data.size())
        data.append(newData);
    else
        data = newData;

    read();
}

void MpdDataDecoder::read()
{
    curPos = 0;
    curEventStartPos = 0;

    QElapsedTimer decodeTime;
    decodeTime.start();
    quint32 eventDecoded=0;
    while(true) {
        try{
            if(!readRecord()) {
//                qInfo()<<"End of data";
                break;
            }
            ++eventDecoded;
            if(decodedData.event_cnt > DECODED_EVENTS_LIMIT)
                flushData();
        } catch(const char* message) {
            mpdDeviceHdr = nullptr;
            mstreamHdr = nullptr;
            tsTAI = nullptr;
            dataBlockHdr = nullptr;
            ++analyzerStats.formatErrorCnt;

            QString mess(message);
            if(findNextRecord()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
            } else{
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            if(printBadEvent) {
                printBadEvent = false;
                printCurrentDevice();
            }
        } catch(const QString message) {
            mpdDeviceHdr = nullptr;
            mstreamHdr = nullptr;
            tsTAI = nullptr;
            dataBlockHdr = nullptr;
            ++analyzerStats.formatErrorCnt;
            QString mess(message);
            if(findNextRecord()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
            } else {
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            if(printBadEvent) {
                printBadEvent = false;
                printCurrentDevice();
            }
        }
    }

    analyzerStats.decodeTimeNs += decodeTime.nsecsElapsed();
    analyzerStats.eventDecoded += eventDecoded;
    analyzerStats.sizeDecoded += curEventStartPos;
    ++analyzerStats.decodeLoop;

    if(curEventStartPos) {
        if(curEventStartPos==static_cast<uint>(data.size()))
            data.clear();
        else
            data = data.right(data.size()-static_cast<int>(curEventStartPos));
    }

    flushData();
}

bool MpdDataDecoder::checkSkip()
{
//    fn1 = &skip_work(analyzerSem);
    if(skipFunc && (rawDataEveryEv || rawDataEvents==0) && skipFunc()) {
        return true;
    }
    return false;
}

void MpdDataDecoder::sendRunStartStop(bool start, const RcRunId &id)
{
    flushData();

    if(start)
        emit runStarted(id);
    else
        emit runStoped(id);
}

bool MpdDataDecoder::findNextRecord()
{
    uint shift = 0;
    const uint dataSize = static_cast<uint>(data.size());
    const char * dataStart = data.data();
    uint pos;
    while(true) {
        shift += 2;
        pos = curEventStartPos+shift;
        if(pos+sizeof(MpdHdr) > dataSize) {
            break;
        }
        mpdHdr = reinterpret_cast<const MpdHdr *>(dataStart+pos);
        if(mpdHdr->isValid()) {
            curPos = curEventStartPos + shift;

            QString mess = QString("Syncword was found at 0x%1; Shift:0x%2")
                    .arg(curEventStartPos,0,16).arg(shift,0,16).toStdString().c_str();
            qInfo()<<mess;

            return true;
        }
    }
    curEventStartPos = dataSize;
    return false;
}

quint32 getPureMpdPayloadLen(const MpdHdr *h) {
    const quint32 &headerSize = h->sync == MPD_SYNC_NORM
            || h->sync == MPD_SYNC_EOB ? sizeof (MpdEventHdr)
                                            : sizeof (MpdHdr);
    return h->length+headerSize-sizeof (MpdEventHdr);
}

bool MpdDataDecoder::readRecord()
{
    curEventStartPos = curPos;
    if(static_cast<uint>(data.size())<curPos+sizeof(MpdHdr))
        return false;

    mpdHdr = reinterpret_cast<MpdHdr *>(data.data()+curPos);

    if(!mpdHdr->isValid()) {
        QString mess = QString("Format error: wrong MPD_SYNC_WORD: %1; at file pos:%2")
                .arg(mpdHdr->sync, 0, 16).arg(curPos);
        throw(mess.toStdString().c_str());
    }

    const quint32 &recordLen = getPureMpdPayloadLen(mpdHdr) + sizeof (MpdEventHdr);
    if(static_cast<uint>(data.size()) < curPos+recordLen)
        return false;

    if(rawDataEveryEv)
        rawDataEvents = 1;

    switch(mpdHdr->sync) {
    case MPD_SYNC_NORM:
    case MPD_SYNC_NORM_TLV:
    case MPD_SYNC_EOB:
    case MPD_SYNC_EOB_TLV:
        if(checkSkip()) {
            ++analyzerStats.skip;
            curPos += recordLen;
            return true;
        }
    }
    ++analyzerStats.pass;

    printRawCurMpdHdr();

    switch(mpdHdr->sync) {
    case MPD_SYNC_NORM:
    case MPD_SYNC_NORM_TLV:
        readEventRec();
        break;
    case MPD_SYNC_EOB:
    case MPD_SYNC_EOB_TLV:
        readStatRec();
        break;
    case MPD_SYNC_RUN_START:
    case MPD_SYNC_RUN_STOP:
    case MPD_SYNC_FILE_BEGIN:
    case MPD_SYNC_FILE_END:
        readFileRunRec();
        break;
    case MPD_SYNC_JSON_TLV:
        readJson();
        break;
    default:
        printRawTxtBlock(reinterpret_cast<const char*>(mpdHdr),
                         mpdHdr->length, "Skipped record payload",
                         sizeof(MpdHdr));
    }

    curPos += recordLen;
    flushRawDataEvent();
    return true;
}

void MpdDataDecoder::readEventRec()
{
    auto dataPtr = reinterpret_cast<const char *>(mpdHdr) + sizeof (MpdHdr);

    evNumPtr = reinterpret_cast<const quint32*>(dataPtr);
    dataPtr += sizeof (quint32);
    printRawCurEvNum();

    auto pureMpdPayloadLen = getPureMpdPayloadLen(mpdHdr);
    quint32 payloadLeft = pureMpdPayloadLen;

    curCompleteEv.clear();
    curCompleteEv.evNum = *evNumPtr;
    while(payloadLeft) {
        if(payloadLeft < sizeof(MpdDeviceHdr))
            throw("Format error: There is no data for MpdDeviceHdr");

        mpdDeviceHdr = reinterpret_cast<const MpdDeviceHdr *>(dataPtr);
        printRawCurMpdDevHdr();

        auto fullDevLen = sizeof(MpdDeviceHdr) + mpdDeviceHdr->length;
        if(payloadLeft < fullDevLen)
            throw("Format error: There is no data for MpdDevice payload");

        readDeviceEvent();
        mpdDeviceHdr = nullptr;
        dataPtr += fullDevLen;
        payloadLeft -= fullDevLen;
    }
    decodedData.events[curCompleteEv.evNum] = curCompleteEv;
    ++decodedData.event_cnt;
    curCompleteEv.clear();

    prevEventNum = *evNumPtr;
    lastEventNumAtPrevFrame = prevEventNum;
    analyzerStats.lastEvNum = prevEventNum;

    if(lookForStop &&  stopEvNum == prevEventNum) {
        emit gotStopEvent(stopEvNum);
        lookForStop = false;
        stopEvNum = 0;
    }
}

void MpdDataDecoder::readDeviceEvent()
{
    quint32 payloadLeft = mpdHdr->length;
    auto dataPtrGlobal = reinterpret_cast<const char *>(mpdDeviceHdr) + sizeof (MpdDeviceHdr);
//    qInfo()<<" Device:"<<hex<<mpdDeviceHdr->deviceSerial<<dec;

    if(!(isPureAdcDev() || isTadcDev() || isTtvxsDev() || isTrcDev())) {
        printRawTxtBlock(dataPtrGlobal, mpdDeviceHdr->length, "Skipped device block");
        return;
    }
//    qInfo()<<"mpdDevice: length:"<<mpdDeviceHdr->length<<" serial:"<<hex<<mpdDeviceHdr->deviceSerial;

    quint32 readLen = 0;
    curTqdcEv.clear();
    curTrcEv.clear();
    curTtvxsEv.clear();
    curTdcStat.clear();
    curTrcEv.evNum = curTtvxsEv.evNum = curTqdcEv.evNum = *evNumPtr;
    curTrcEv.index = curTtvxsEv.index = curTqdcEv.index = getCurDeviceIndex();
    while(mpdDeviceHdr->length > readLen) {
        if(readLen+sizeof(MpdMStreamHdr)>mpdDeviceHdr->length) {
            throw(QString("Format error: rest size of devPayloadLen is not enough for MStream header(%1):"
                          "devPayloadLen-parsed=%2-%3")
                .arg(sizeof(MpdMStreamHdr))
                .arg(mpdDeviceHdr->length)
                .arg(readLen));
        }
        mstreamHdr = reinterpret_cast<const MpdMStreamHdr *>(dataPtrGlobal+readLen);

//        qInfo()<<"mstreamHdr: length:"<<mstreamHdr->words32b*sizeof(quint32);
        printRawCurMStreamHdr(isPureAdcDev());

        readLen += sizeof(MpdMStreamHdr);
        const quint32 mStreamWholeSize = mstreamHdr->words32b*sizeof(quint32);
        if(readLen+mStreamWholeSize>mpdDeviceHdr->length) {
            throw(QString("Format error: rest size of devPayloadLen is not enough for MStream payload(%1 32-bit words):"
                          "devPayloadLen-parsed=%2-%3")
                .arg(mstreamHdr->words32b)
                .arg(mpdDeviceHdr->length)
                .arg(readLen));
        }

        auto dataPtr = reinterpret_cast<const char *>(mstreamHdr)+sizeof(MpdMStreamHdr);
        switch (mstreamHdr->subtype) {
        case 0:
            decodeMStreamType0();
            break;
        case 1:
            if(isPureAdcDev()) {
                auto payloadPtr = reinterpret_cast<const quint32*>(dataPtr);
                decodeMStreamType1Adc(payloadPtr);
                break;
            }
            [[fallthrough]];
        default:
            printRawTxtBlock(dataPtr, mstreamHdr->words32b*sizeof(quint32),
                             "MStreamPayload");
//            throw( QString("Format error: Invalid subtype number in MStream header: %1.")
//                .arg(mstreamHdr->subtype).toStdString().c_str());
        }

//        curPos += mstreamHdr->words32b*sizeof(quint32);
//        qInfo()<<"MS block:ev="<<*evNumPtr<<"; subtype="<<mstreamHdr->subtype<<"; curpos="<<curpos;
        readLen += mStreamWholeSize;
    }
    mstreamHdr = nullptr;
    if(mpdDeviceHdr->length != readLen)
        throw(QString("Format error: Sum of sizes of MStream blocks(%1) isn't equal to one declared in MpdDevice header(%2)")
            .arg(readLen).arg(mpdDeviceHdr->length).toStdString().c_str());

    if(isPureAdcDev() || isTadcDev()) {
        curCompleteEv.tqdcEv[curTqdcEv.index] = curTqdcEv;
        curTqdcEv.clear();
    } else if(isTtvxsDev()) {
        curCompleteEv.ttvxsEv[curTtvxsEv.index] = curTtvxsEv;
        curTtvxsEv.clear();
    } else if(isTrcDev()) {
        curCompleteEv.trcEv[curTrcEv.index] = curTrcEv;
        curTrcEv.clear();
    } else {
        Q_ASSERT(false);
    }

    if(!curTdcStat.index.isNull()) {
        curCompleteEv.devStats[curTdcStat.index] = curTdcStat;
        curTdcStat.clear();
    }

    if(rawDataEvents) {
        rawDataTxt << "";
    }
}

quint32 MpdDataDecoder::readFileRunRec()
{
    Q_ASSERT(mpdHdr->sync == MPD_SYNC_RUN_START || mpdHdr->sync == MPD_SYNC_RUN_STOP ||
             mpdHdr->sync == MPD_SYNC_FILE_BEGIN || mpdHdr->sync == MPD_SYNC_FILE_END);

    quint32 payloadLeft = mpdHdr->length;
    auto dataPtr = reinterpret_cast<const char *>(mpdHdr) + sizeof (MpdHdr);
    quint32 readLen = 0;

    QString runIndex;
    quint32 runNumber=0;
    QString prefixStr;
    std::optional<quint32> evbIndex, fileSplitId;
    while(readLen < payloadLeft) {
        auto h = reinterpret_cast<const MpdHdr *>(dataPtr+readLen);

        readLen += sizeof (MpdHdr);
        auto p = dataPtr+readLen;
        switch (h->sync) {
        case MPD_SYNC_RUN_INDEX:
            runIndex = QString::fromLatin1(p, static_cast<int>(h->length));
            // remove 0-terminating
            runIndex = runIndex.mid(0, runIndex.indexOf(char(0)));
            prefixStr = QString("RunIndex '%1'").arg(runIndex);
            break;
        case MPD_SYNC_RUN_NUMBER:
            if(h->length != sizeof (quint32))
                throw("Wrong paylod length for MPD_SYNC_RUN_NUMBER block");
            runNumber = *reinterpret_cast<const quint32 *>(p);
            prefixStr = QString("RunNumber #%1").arg(runNumber);
            break;
        case MPD_SYNC_FILE_ID:
            if(h->length != sizeof (quint32))
                throw("Wrong paylod length for MPD_SYNC_FILE_ID block");
            fileSplitId = *reinterpret_cast<const quint32 *>(p);
            prefixStr = QString("FileSplit id #%1").arg(*fileSplitId);
            break;
        case MPD_SYNC_EV_ORDER:
            if(h->length != sizeof (quint32))
                throw("Wrong paylod length for MPD_SYNC_EV_ORDER block");
            evbIndex = *reinterpret_cast<const quint32 *>(p);
            prefixStr = QString("Evb ev index #%1").arg(*evbIndex);
            break;
        default:;
        }
        printRawTxtBlock(reinterpret_cast<const char *>(h),
                         sizeof(MpdHdr)+h->length,
                         prefixStr);
        readLen += h->length;
    }

    auto log = QString("%1: '%2' #%3")
            .arg(MpdHdr::getSyncName(mpdHdr->sync))
            .arg(runIndex).arg(runNumber);
    if(evbIndex.has_value())
        log += QString(" evIndex=%1")
                .arg(*evbIndex);
    if(fileSplitId.has_value())
        log += QString(" fileSplitId=%1")
                .arg(*fileSplitId);
    qInfo() << log;

    switch (mpdHdr->sync) {
    case MPD_SYNC_RUN_START:
    case MPD_SYNC_RUN_STOP:
        sendRunStartStop(mpdHdr->sync == MPD_SYNC_RUN_START,
                         {runIndex, static_cast<int>(runNumber)});
    case MPD_SYNC_FILE_BEGIN:
    case MPD_SYNC_FILE_END:
//    sendFileBeginEnd(mpdHdr->sync == MPD_SYNC_FILE_BEGIN,
//                     {runIndex, static_cast<int>(runNumber)}, ... );
    default:;
    }

    Q_ASSERT(readLen == payloadLeft);
    if(readLen != payloadLeft)
        throw(QString("%1 readLen(%2) != payloadLeft(%3")
            .arg(__func__).arg(readLen).arg(payloadLeft));
    return readLen;
}

quint32 MpdDataDecoder::readStatRec()
{
    Q_ASSERT(mpdHdr->sync == MPD_SYNC_EOB_TLV || mpdHdr->sync == MPD_SYNC_EOB);

    quint32 payloadLeft = getPureMpdPayloadLen(mpdHdr);
    auto dataPtr = reinterpret_cast<const char *>(mpdHdr) + sizeof (MpdEventHdr); // although skip reserved word
    quint32 readLen = 0;

    while(readLen < payloadLeft) {
        mpdDeviceHdr = reinterpret_cast<const MpdDeviceHdr *>(dataPtr+readLen);
        printRawCurMpdDevHdr();
        readLen += sizeof (MpdDeviceHdr);

        switch (mpdDeviceHdr->deviceId) {
        case DEVICE_ID_MSC16VE_E:
        {
            const quint32* payloadPtr =
                reinterpret_cast<const quint32*>(dataPtr+readLen);
            decodeMscBlock(payloadPtr);
            break;
        }
        default:
            printRawTxtBlock(dataPtr+readLen, mpdDeviceHdr->length, "Skipped payload");
        }
        readLen += mpdDeviceHdr->length;
    }
    Q_ASSERT(readLen == payloadLeft);

    mpdDeviceHdr = nullptr;
    return payloadLeft;
}

DeviceIndex MpdDataDecoder::getCurDeviceIndex() const
{
    Q_ASSERT(mpdDeviceHdr);
    return mpdDeviceHdr
            ? DeviceIndex(static_cast<DeviceId>(mpdDeviceHdr->deviceId),
                                      mpdDeviceHdr->deviceSerial)
            : DeviceIndex();
}

void MpdDataDecoder::readJson()
{
    const char* data = reinterpret_cast<const char *>(mpdHdr);
    data += sizeof (MpdHdr) + sizeof (quint32);
    mpdDeviceHdr = reinterpret_cast<const MpdDeviceHdr *>(data);

    data += sizeof (MpdDeviceHdr);
    auto jsonData =QByteArray::fromRawData(data, mpdDeviceHdr->length);
    int pos = jsonData.indexOf(char(0));
    if(pos!=-1)
        jsonData.truncate(pos);

    auto jsonDoc = QJsonDocument::fromJson(jsonData);
    if(rawDataEvents) {
        auto jsonList = QString::fromStdString(jsonDoc.toJson(QJsonDocument::Indented).toStdString()).split('\n');

        printRawCurMpdDevHdr();
        printRawTxtHeader("JSON payload");
        int shift = 0;
        QString suffix;
        while(shift<mpdDeviceHdr->length) {
            if(!jsonList.isEmpty())
                suffix = "\t\t"+jsonList.takeFirst();
            else
                suffix.clear();

            auto wordPtr = data+shift;
            printRawWord(wordPtr, suffix);
            shift += 4;
        }
        while(rawDataEvents && !jsonList.isEmpty()) {
            rawDataTxt << "               \t\t"
                          + jsonList.takeFirst();
        }
    }
    mpdDeviceHdr = nullptr;
}

void MpdDataDecoder::printCurrentDevice() const
{
    if(!mpdDeviceHdr)
        return;
    qWarning() << "Event header:";
    const quint16 WORD_LIMIT = 1000;
    quint16 wordCnt = 0;
    const quint32 *evPtr = reinterpret_cast<const quint32 *>(mpdHdr);
    while(wordCnt*sizeof(quint32) < sizeof(MpdEventHdr)){
        const quint32 *wordPtr = evPtr+wordCnt;
        qWarning()  << get_raw_data_prefix(wordPtr, evPtr);
        ++wordCnt;
    }

    qWarning() << "Device event payload:";
    wordCnt = 0;
    const quint32 *devPtr = reinterpret_cast<const quint32 *>(mpdDeviceHdr);
    while(wordCnt*sizeof(quint32) < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length){
        const quint32 *wordPtr = devPtr+wordCnt;
        qWarning()  << get_raw_data_prefix(wordPtr, devPtr);
        ++wordCnt;
        if(wordCnt > WORD_LIMIT) {
            qWarning("Device payload length excedeed %d 32-bit words. Suppressed.", WORD_LIMIT);
            break;
        }
    }
}

void MpdDataDecoder::printRawCurMpdHdr()
{
    printRawMpdHdr(mpdHdr);
}

void MpdDataDecoder::printRawMpdHdr(const MpdHdr * const h)
{
    if(!rawDataEvents) return;

    rawDataTxt << QString("\t%1:").arg("MpdHdr");
    printRawWord(&h->sync,":\t\tSyncWord: "+h->getSyncName());
    printRawWord(&h->length, QString("\t\tlen:%1")
                 .arg(h->length));
}

void MpdDataDecoder::printRawCurEvNum()
{
    if(!rawDataEvents) return;

    if(mpdHdr->sync == MPD_SYNC_NORM)
        rawDataTxt << QString("\t%1:").arg("evNum");
    if(mpdHdr->sync == MPD_SYNC_NORM || mpdHdr->sync == MPD_SYNC_NORM_TLV)
        printRawWord(evNumPtr, QString("\t\tevNum:%1")
                     .arg(*evNumPtr));
}

void MpdDataDecoder::printRawCurMpdDevHdr()
{
    if(!mpdDeviceHdr)
        return;
    if(!rawDataEvents) return;

    const auto &serStr = DeviceIndex::getSerialStr(mpdDeviceHdr->deviceSerial);
    printRawTxtHeader(QString("MpdDeviceHdr (%1)").arg(serStr));
    const quint32 *word = reinterpret_cast<const quint32 *>(mpdDeviceHdr);
    printRawWord(word, QString(":\t\tSerial: %1").arg(serStr));
    printRawWord(word+1, QString(":\t\tDevId:0x%1(%2) Len:%3")
                 .arg(mpdDeviceHdr->deviceId, 2, 16, QChar('0'))
                 .arg(getDeviceTypeName(mpdDeviceHdr->deviceId))
                 .arg(mpdDeviceHdr->length));
}

void MpdDataDecoder::printRawCurTai()
{
    if(!rawDataEvents) return;
    if(!tsTAI) return;

    printRawTxtHeader("TAI");
    QString suffix = QString(":\t\t%2")
            .arg(QDateTime::fromTime_t(tsTAI->sec).toString());
    printRawWord(tsTAI, suffix);
    suffix = QString(":\t\tns:%L1 %2")
            .arg(tsTAI->nSec,9,10,QChar('0'))
            .arg(tsTAI->flags==2 ? "ok" : "invalid");
    printRawWord(reinterpret_cast<const char *>(tsTAI)+4, suffix);
}

void MpdDataDecoder::printRawCurMStreamHdr(bool isPureAdcDev)
{
    if(!mstreamHdr)
        return;
    if(!rawDataEvents) return;

    printRawTxtHeader("MStreamHdr");
    QString suffix = QString(":\t\tsubtype:%1; words:%2")
            .arg(mstreamHdr->subtype).arg(mstreamHdr->words32b);
    if(mstreamHdr->subtype && isPureAdcDev)
        suffix += QString(" ch:%1").arg(mstreamHdr->usrDefBits);
    printRawWord(mstreamHdr, suffix);
}

void MpdDataDecoder::printRawTxtHeader(const QString &header)
{
    if(!rawDataEvents) return;
    if(!header.isEmpty())
        rawDataTxt << QString("\t%1:").arg(header);
}

void MpdDataDecoder::printRawTxtBlock(const char *ptr, uint size, const QString &header, int byteShift)
{
    if(!rawDataEvents) return;
    printRawTxtHeader(header);
    auto word = reinterpret_cast<const quint32*>(ptr+byteShift);
    for(uint w=0; w<(size/sizeof(quint32)); ++w){
        printRawWord(word+w);
    }
}

void MpdDataDecoder::printRawWord(const void *word, QString suffix)
{
    if(!rawDataEvents)
        return;
    rawDataTxt << get_raw_data_prefix(reinterpret_cast<const quint32*>(word),
                                      mpdHdr).append(suffix);
}

void MpdDataDecoder::flushRawDataEvent()
{
    if(!rawDataEvents)
        return;

    if(--rawDataEvents == 0) {
        emit rawDataUpdated(rawDataTxt);
        rawDataTxt.clear();
    } else
        rawDataTxt << "==============\n";
}

bool MpdDataDecoder::isPureAdcDev() const
{
    return
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64VE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64 ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64S2 ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64VE2 ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64VE_XGE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64ECAL ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64WR ||
            false;
}

bool MpdDataDecoder::isTadcDev() const
{
    return
            mpdDeviceHdr->deviceId == DEVICE_ID_TDC72VXS ||
            mpdDeviceHdr->deviceId == DEVICE_ID_TDC64VHLE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_TDC64VLE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_TQDC16VS_E ||
            mpdDeviceHdr->deviceId == DEVICE_ID_PHOS_FEC ||
            false;
}

bool MpdDataDecoder::isTtvxsDev() const
{
    return
            mpdDeviceHdr->deviceId == DEVICE_ID_TTVXS ||
            false;
}

bool MpdDataDecoder::isTrcDev() const
{
    return
            mpdDeviceHdr->deviceId == DEVICE_ID_TRC ||
            false;
}

void MpdDataDecoder::decodeMStreamType0()
{
    auto dataPtr = reinterpret_cast<const char *>(mstreamHdr)+sizeof(MpdMStreamHdr);
    //            if(mstreamHdr->words32b*sizeof(quint32) != sizeof(AdcDataType1))
    //                throw( QString("Format error: Invalid size(%1 Bytes) of MStream subtype=0 block; should be %2 Bytes.")
    //                    .arg(mstreamHdr->words32b*sizeof(quint32)).arg(sizeof(AdcDataType0)).toStdString().c_str());
    decodeTsTai(dataPtr);

    const quint32 mStreamWholeSize = mstreamHdr->words32b*sizeof(quint32);
    quint32 readLen = sizeof(TaiRawData);

    if(isPureAdcDev()) {
        curTqdcEv.tai = tsTAI->toTaiTime();
        auto chMaskPtr = reinterpret_cast<const quint64*>(dataPtr+readLen);
        curTqdcEv.chMask = *chMaskPtr;

        auto wPtr = reinterpret_cast<const quint32*>(chMaskPtr);
        printRawWord(wPtr, ":\t\tchMask (Low)");
        printRawWord(wPtr+1, ":\t\tchMask (Hi)");
        return;
    }
    if(isTrcDev()) {
        decodeTrcBlock();
        return;
    }

    while(mStreamWholeSize > readLen) {
        dataBlockHdr = reinterpret_cast<const DataBlockHeader *>(dataPtr+readLen);
        quint16 wholeBlockLen = sizeof(DataBlockHeader)+dataBlockHdr->len;
        if(readLen+wholeBlockLen > mStreamWholeSize)
            throw(QString("Format error: Data Block length (%1) is large then rest of MStream payload:"
                          "MStreamlen-readLen-DataBlockHeader=%2-%3-%4")
                .arg(dataBlockHdr->len)
                .arg(mStreamWholeSize)
                .arg(readLen)
                .arg(sizeof(DataBlockHeader)));

        auto blockPayloadPtr = reinterpret_cast<const quint32*>(dataBlockHdr)
                + sizeof (DataBlockHeader)/sizeof (quint32);
        switch(dataBlockHdr->type()) {
        case 0x0:
            decodeModuleDataBlockTdc(blockPayloadPtr);
            break;
        case 0x1:
            decodeModuleDataBlockAdc(blockPayloadPtr);
            break;
        case 0xA:
            decodeModuleDataBlockTtvxs(blockPayloadPtr);
            break;
        case 0xF:
            decodeModuleDataBlockStat(blockPayloadPtr);
            break;
        default:
            if(rawDataEvents) {
                printRawTxtBlock(reinterpret_cast<const char*>(dataBlockHdr),
                                 wholeBlockLen,
                                 QString("Unknown data block:%1")
                                 .arg(dataBlockHdr->type()));
            }
        }
        readLen += wholeBlockLen;
    }
    dataBlockHdr = nullptr;
}

void MpdDataDecoder::decodeTsTai(const void* ptr)
{
    tsTAI = reinterpret_cast<const TaiRawData*>(ptr);
    printRawCurTai();
}

void MpdDataDecoder::decodeModuleDataBlockTdc(const quint32 *payloadPtr)
{
    curTqdcEv.tai = tsTAI->toTaiTime();

    quint16 tdcWord = 0;
    int tdcWordCnt=0;
    quint16 tdcTS = 0;
    QString wordStr;
    bool hasHptdcError = false;
    QMap<quint8, QVector<TdcEdge> > chEdgeMap;
    const DeviceIndex &index = curTqdcEv.index;
    if(rawDataEvents) {
        printRawTxtHeader("DataBlockHeader (HPTDC)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%1; len:%2")
                     .arg(dataBlockHdr->type()).arg(dataBlockHdr->len));
        if(dataBlockHdr->specific.tdc.fifio_overflow)
            rawDataTxt.last().append(" (event FIFO overflow)");
    }
    while(dataBlockHdr->len > tdcWord*sizeof(quint32)){
        const quint32 &word = *(payloadPtr+tdcWord);
        ++tdcWordCnt;
        const int dt = word >> 28;
//        qInfo() << tdcWord<< ":  " << hex << word << dec;
        wordStr = QString(":\t\t");
        switch (dt) {
        case 2: // tdc ev hdr
        {
            tdcTS = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF; // evNum in TdcHeader
            if(rawDataEvents)
                wordStr.append(QString("ev hdr evNum:%1").arg(tdcEv));
//            qInfo() << "tdc ev hdr: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (*evNumPtr & 0xFFF));
        }
            break;
        case 3: // tdc ev trl
        {
            quint16 wordCnt = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF;
            if(rawDataEvents)
                wordStr.append(QString("ev trl evNum:%1; words:%2").arg(tdcEv).arg(wordCnt));
//            qInfo() << "tdc ev trl: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (*evNumPtr & 0xFFF));
            if(tdcWordCnt != wordCnt) {
                int wc = 0;
                qWarning("tdcWordCnt != wordCnt devIndex: 0x%02x 0x%08x",
                         mpdDeviceHdr->deviceId, mpdDeviceHdr->deviceSerial);
                const quint32 *devPtr = reinterpret_cast<const quint32 *>(mpdDeviceHdr);
                while(wc*sizeof(quint32) < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length){
                    const quint32 *wordPtr = devPtr+wc;
                    qWarning()  << get_raw_data_prefix(wordPtr, devPtr);
                    ++wc;
                }
            }
            Q_ASSERT(tdcWordCnt == wordCnt);
            tdcWordCnt=0;
        }
            break;
        case 4: // tdc leading edge
//            qInfo() << tdcWord<< ":  " << hex << word << dec;
        case 5: // tdc trl edge
        {
            const quint8 ch = (word>>21)&0x7F;
            const quint32 time = word&0x1FFFFF;
            const bool risening = (dt==5)==signalInverted;
            if(rawDataEvents)
                wordStr.append(QString("%1 ch:%2; t(ps):%L3")
                               .arg(risening?"leading":"trailing").arg(ch,2).arg(24000.*time/1024));
            chEdgeMap[ch].push_back(TdcEdge(tdcTS,time, risening));
            break;
        }
        case 6:
        {
            hasHptdcError = true;
            curTqdcEv.errors.push_back(word);
            const int tdcId = (word>>24)&0xF;
            QStringList errList;
            for(int errBit=0; errBit< 15; ++errBit){
                if((word & 0x7FFF) & 1<<errBit) {
                    errList << QString("%1\t\t%2").arg(1<<errBit, 14, 16).arg(HPTDC_ERR_STR[errBit]);
                }
            }
            QString errStr = QString("Got HPTDC error (0x%1 tdcId=%2):\n%3")
                    .arg(mpdDeviceHdr->deviceSerial, 8, 16, QChar('0'))
                    .arg(tdcId).arg(errList.join("\n"));
            if(rawDataEvents)
                wordStr.append(errStr);
            if(!lastErrStr.contains(errStr)) {
                qWarning().noquote() << errStr;
                lastErrStr << errStr;
                while(lastErrStr.size() > 10)
                    lastErrStr.removeFirst();
            }

            break;
        }
        default:
            break;
        }
        if(rawDataEvents)
            printRawWord(&word, wordStr);
        ++tdcWord;
    }
    if(hasHptdcError) {
        if(printBadEvent) {
            printBadEvent = false;
            qWarning() << "Got HPTDC error";
            printCurrentDevice();
        }
    }


    const auto &evNum = *evNumPtr;
    auto &errWordMap = analyzerStats.dev_stat[index].err_cnt;
    for(quint32 word : curTqdcEv.errors) {
        auto &errEvRangesList = errWordMap[word];
        if(errEvRangesList.isEmpty() || errEvRangesList.last().second+1 != evNum) {
            errEvRangesList<< QPair<quint32, quint32>(evNum,evNum);
        } else {
            ++errEvRangesList.last().second;
        }
    }


    // Combine TdcEdge into TdcChHit and fill into
    bool hasTdcChHit = false;
    auto &devInvMap = invertMap[index];
    auto &devHits = analyzerStats.dev_stat[index].dev_ch_hits;
    for (quint8 ch: chEdgeMap.keys()) {
        for (const TdcEdge &edge: chEdgeMap[ch]) {
            if(edge.rising != devInvMap.value(ch, false)) {
                if(hasTdcChHit) {
                    curTqdcEv.hits[ch].tdcHits.push_back(curTdcHit);
                    ++devHits[ch].tdc_hits;
                    curTdcHit.clear();
                } else {
                    hasTdcChHit = true;
                }
                curTdcHit.ch = ch;
                curTdcHit.leadingTime = edge.time;
                curTdcHit.tdcTs = edge.tdcTs;
            } else {
                curTdcHit.ch = ch;
                curTdcHit.trailingTime = edge.time;
                if(!hasTdcChHit)
                    curTdcHit.tdcTs = edge.tdcTs;
                curTqdcEv.hits[ch].tdcHits.push_back(curTdcHit);
                ++devHits[ch].tdc_hits;
                curTdcHit.clear();
                hasTdcChHit = false;
            }
        }
        if(hasTdcChHit) {
            curTqdcEv.hits[curTdcHit.ch].tdcHits.push_back(curTdcHit);
            ++devHits[curTdcHit.ch].tdc_hits;
            curTdcHit.clear();
            hasTdcChHit = false;
        }
    }
}

void MpdDataDecoder::decodeModuleDataBlockAdc(const quint32 *payloadPtr)
{
    curTqdcEv.tai = tsTAI->toTaiTime();

    int blockWord = 0;
    int totHitLen = -1;
    int curHitLen =0;
    const int &dataBlockCh = mpdDeviceHdr->deviceId == DEVICE_ID_PHOS_FEC ?
                dataBlockHdr->specific.phose_adc.ch : dataBlockHdr->specific.adc.ch;
    qint16 sample1 = 0, sample2 = 0;
    std::vector<double> &waveForm = curAdcHit.wf;
    if(rawDataEvents) {
        printRawTxtHeader("DataBlockHeader (ADC)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%1; len:%2 ch:%3")
                     .arg(dataBlockHdr->type()).arg(dataBlockHdr->len)
                     .arg(dataBlockCh));
    }

    while(dataBlockHdr->len > blockWord*sizeof(quint32)) {
        const quint32 &word = *(payloadPtr+blockWord);
        ++blockWord;
        if(totHitLen==-1) {
            curAdcHit.ch = dataBlockCh;
            curAdcHit.adcTs = word & 0xFFFF;
            totHitLen = word >> 16;
            if(totHitLen%sizeof(quint16))
                throw(QString("Format error: bad size of adc data len: %1").arg(totHitLen));
            if(dataBlockHdr->len < blockWord*sizeof(quint32)+totHitLen) {
                throw(QString("Format error:there is no enough data "
                              "in adc data block for hit with length %1")
                        .arg(totHitLen));
            }
            totHitLen /= sizeof(qint16);

            waveForm.reserve(totHitLen);
            if(rawDataEvents) {
                rawDataTxt << QString("%1:\t\tlen:%2; ts:%3")
                              .arg(*(payloadPtr+blockWord-1), 8, 16, QChar('0'))
                              .arg(totHitLen).arg(curAdcHit.adcTs);
            }
            if(totHitLen==0) {
                totHitLen = -1;
            }
        } else {
            sample1 = static_cast<qint16>( word      & 0xFFFF);
            sample2 = static_cast<qint16>((word>>16) & 0xFFFF);
            waveForm.push_back(sample1);
            ++curHitLen;
            // skip sample2 in case totHitLent is odd
            if(curHitLen != totHitLen) {
                waveForm.push_back(sample2);
                ++curHitLen;
            }
            if(curHitLen == totHitLen) {
                curTqdcEv.hits[curAdcHit.ch].adcHits.push_back(curAdcHit);
                ++analyzerStats.dev_stat[curTqdcEv.index].dev_ch_hits[curAdcHit.ch].adc_hits;
                curAdcHit.clear();
                totHitLen = -1;
                curHitLen=0;
            }
            if(rawDataEvents) {
                rawDataTxt << QString("%1:\t\t%2 %3")
                              .arg(*(payloadPtr+blockWord-1), 8, 16, QChar('0'))
                              .arg(sample2, 10).arg(sample1, 10);
            }
        }
    }
}

void MpdDataDecoder::decodeMStreamType1Adc(const quint32 *payloadPtr)
{
    curAdcHit.ch = mstreamHdr->usrDefBits;
    decodeTsTai(payloadPtr);
    curAdcHit.taiTime = tsTAI->toTaiTime();

    const quint16 adcOffset = adcRawDataSigned ? 0 : 0x8000;

    int wordIndex = sizeof (TaiRawData)/sizeof (quint32);
    for(; wordIndex<mstreamHdr->words32b;++wordIndex) {
        auto wfPtr = payloadPtr+wordIndex;
        const qint16 &veven = qint16((*wfPtr >> 16) & 0xFFFF) - adcOffset;
        const qint16 &vodd  = qint16((*wfPtr >> 0)  & 0xFFFF) - adcOffset;
        curAdcHit.wf.push_back(veven);
        curAdcHit.wf.push_back(vodd);
        if(rawDataEvents) {
            printRawWord(wfPtr, QString(":\t\t%1 %2").arg(veven).arg(vodd));
        }
    }
    curTqdcEv.hits[curAdcHit.ch].adcHits.push_back(curAdcHit);
    ++analyzerStats.dev_stat[curTqdcEv.index].dev_ch_hits[curAdcHit.ch].adc_hits;
    curAdcHit.clear();
}

void MpdDataDecoder::decodeModuleDataBlockTtvxs(const quint32 *payloadPtr)
{
    curTtvxsEv.tai = tsTAI->toTaiTime();
    if(dataBlockHdr->len < sizeof (curTtvxsEv.trig_src))
        throw(QString("Format error: TTVXS Trigger Data Block length (%1) should contains at lest 'Trigger word'")
            .arg(dataBlockHdr->len));

    curTtvxsEv.trig_src.all = *payloadPtr;
    if(!analyzerStats.ttvxsList.contains(curTtvxsEv.index))
        analyzerStats.ttvxsList << curTtvxsEv.index;
    if(rawDataEvents) {
        QStringList trigList;
        if(curTtvxsEv.trig_src.bit.pulser) trigList << "pulser";
        if(curTtvxsEv.trig_src.bit.random_pulser) trigList << "random_pulser";
        if(curTtvxsEv.trig_src.bit.external_ttl) trigList << "external_ttl";
        if(curTtvxsEv.trig_src.bit.digital_sfp) trigList << "digital_sfp";
        printRawWord(payloadPtr, QString("\t\ttrigSrc: %1").arg(trigList.join(',')));
        for(uint word=1; word<dataBlockHdr->len/sizeof (quint32); ++word) {
            QString suffix;
            switch (word) {
            case 1:
                suffix = QString("\t\ttrigCandid: %L1").arg(*(payloadPtr+word));
                break;
            case 2:
                suffix = QString("\t\ttrigAccept: %L1").arg(*(payloadPtr+word));
                break;
            }
            printRawWord(payloadPtr+word, suffix);
        }
    }
}

void MpdDataDecoder::decodeModuleDataBlockStat(const quint32 *payloadPtr)
{
    curTdcStat.index = getCurDeviceIndex();
    curTdcStat.regIoTimeout = dataBlockHdr->specific.stat.regio_timeout;
    curTdcStat.regIoError = dataBlockHdr->specific.stat.regio_error;
    if(rawDataEvents) {
        printRawTxtHeader("DataBlockHeader (Stats)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%2; len:%3")
                     .arg(dataBlockHdr->type()).arg(dataBlockHdr->len));

        if(curTdcStat.regIoTimeout)
            rawDataTxt.last().append(" (has RegIO timeout)");
        if(curTdcStat.regIoError)
            rawDataTxt.last().append(" (has RegIO error)");
    }

    int parsed_len = 0;
    while(parsed_len < dataBlockHdr->len) {
        const quint32 &word32 = *payloadPtr;
        quint16 addr = word32>>16;
        quint16 val = word32 & 0xFFFF;
        curTdcStat.stats[addr] = val;
        if(rawDataEvents) {
            printRawWord(payloadPtr, QString(":\t\tAddr:%2\tval:%3")
                          .arg(addr, 4, 16, QChar('0'))
                          .arg(val, 4, 16, QChar('0')));
        }
        switch (addr) {
        case REG_TEMPERATURE:
        case REG_BMC_TEMP_1:
        case REG_BMC_TEMP_2:
        case REG_BMC_TEMP_3:
        case REG_BMC_TEMP_4: {
            double temp = addr==REG_TEMPERATURE ? OneWire::convertOneWireTemp(val)
                                                : decodeTemp_v1(val);
            int temp_id = addr==REG_TEMPERATURE ? 0
                                                : addr-REG_BMC_TEMP_1+1;
            if(val)
                curTqdcEv.temp[temp_id] = temp;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" temp[%1]:%2").arg(temp_id).arg(temp));
            }
            break;
        case REG_FW_REV:
            if(val)
                curTqdcEv.fw_rev = val;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" fw_rev:%1").arg(val));
            break;
        case REG_BMC_REVISION:
            if(val)
                curTqdcEv.bmc_rev = val;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" bmc_rev:%1").arg(val));
            break;
        }
        ++payloadPtr;
        parsed_len += sizeof(quint32);
    }
}

void MpdDataDecoder::decodeTrcBlock()
{
    curTrcEv.tai = tsTAI->toTaiTime();
    auto dataPtr = reinterpret_cast<const char *>(mstreamHdr)
            +sizeof(MpdMStreamHdr)+sizeof(TaiRawData);
    auto wordPtr = reinterpret_cast<const quint32*>(dataPtr);

    if(rawDataEvents) {
        printRawWord(wordPtr,   ": ChState before reduct.");
        printRawWord(wordPtr+1, ": ChState after  reduct.");
    }

    curTrcEv.chStateBefore = *wordPtr; ++wordPtr;
    curTrcEv.chStateAfter  = *wordPtr; ++wordPtr;

    int wfLen  = mstreamHdr->words32b - sizeof (TaiRawData)/sizeof (quint32) - 2;
    static QVector<quint32> wf;
    if(wf.capacity() < wfLen) {
        wf.reserve(wfLen);
    }
    if(wfLen && rawDataEvents)
        printRawTxtHeader("Channel state waveform");
    while(wfLen>0) {
        --wfLen;
        if(rawDataEvents) {
            printRawWord(wordPtr);
        }
        wf.push_back(*wordPtr);
        ++wordPtr;
    }
    curTrcEv.wf = wf;
    wf.clear();
}

void MpdDataDecoder::decodeMscBlock(const quint32 *payloadPtr)
{
    mstreamHdr = reinterpret_cast<const MpdMStreamHdr *>(payloadPtr);
    printRawCurMStreamHdr();
    payloadPtr += sizeof (MpdMStreamHdr)/sizeof (quint32);

    decodeTsTai(payloadPtr);
    payloadPtr += sizeof (TaiRawData)/sizeof (quint32);

    auto ver = (*payloadPtr)>>28;
    switch (ver) {
    case 0:
        decodeMscBlock_v0(payloadPtr);
        break;
    case 1:
        decodeMscBlock_v1(payloadPtr);
        break;
    default:
    {
        static bool warMsg = true;
        if(warMsg) {
            warMsg = false;
            qWarning() << QString("[%1] Unsuppurted MSC version: %2")
                              .arg(getCurDeviceIndex().getIdent())
                              .arg(ver);
        }
    }
    }
}

void MpdDataDecoder::decodeMscBlock_v0(const quint32 *payloadPtr)
{
    auto &curMscSlices = decodedData.mscData0[getCurDeviceIndex()].slices;
    curMscSlice.ch.clear();
    curMscSlice.ver = 0;
    curMscSlice.bits = (*payloadPtr)&0xF;
    if(rawDataEvents) {
        printRawWord(payloadPtr,   QString("\t\tver:%1; bits:%2").arg(curMscSlice.ver).arg(curMscSlice.bits));
    }
    ++payloadPtr;

    curMscSlice.sliceNs = *(payloadPtr);
    if(rawDataEvents) {
        printRawWord(payloadPtr,   QString("\t\tslice:%1ns").arg(curMscSlice.sliceNs));
    }
    ++payloadPtr;
    auto &devHits = analyzerStats.dev_stat[getCurDeviceIndex()].dev_ch_hits;

    static constexpr const int sliceHdrSize = sizeof (TaiRawData)/sizeof (quint32) + 2;
    int sliceLen  = mstreamHdr->words32b - sliceHdrSize;
    if(sliceLen && rawDataEvents)
        printRawTxtHeader("Slices");
    quint8 slicePerWord = 28/curMscSlice.bits;
    const quint32 valueMask = (1<<curMscSlice.bits)-1;
    quint8 dt, ch;
    while(sliceLen>0) {
        dt = (*payloadPtr)>>28;
        --sliceLen;
        switch (dt) {
        case 0xF:
            if(rawDataEvents)
                printRawWord(payloadPtr, "\t\tpadding");
            break;
        case 0xE:
            curMscSlice.ceState = ((*payloadPtr)>>24) & 0xF;
            curMscSlice.number = (*payloadPtr) & 0xFFFFFF;
            curMscSlice.tai = tsTAI->toTaiTime();
            curMscSlice.tai.addNs(qint64(curMscSlice.number)*curMscSlice.sliceNs);


            if(rawDataEvents) {
                printRawWord(payloadPtr, QString("\t\tce:%1; slice:%2")
                                 .arg(curMscSlice.ceState,0,16)
                                 .arg(curMscSlice.number));
            }
            curMscSlices.push_back(curMscSlice);
            curMscSlice.ch.clear();
            break;
        default:
            for(int i=0; i<slicePerWord; ++i) {
                auto v = ((*payloadPtr)>>(i*curMscSlice.bits)) & valueMask;
                if(v) {
                    ch = static_cast<quint8>(dt*slicePerWord+i);
                    curMscSlice.ch.push_back({ch, v});
                    devHits[ch].msc_hits += v;
                }
            }
            if(rawDataEvents) {
                QStringList suffix;
                for(int i=0; i<slicePerWord; ++i) {
                    auto v = ((*payloadPtr)>>(i*curMscSlice.bits)) & valueMask;
                    if(v) {
                        ch = static_cast<quint8>(dt*slicePerWord+i);
                        suffix << QString("Ch%1:%2")
                                      .arg(ch, 0, 10, QChar(' '))
                                      .arg(v, 0, 10, QChar(' '));
                    }
                }
                printRawWord(payloadPtr, "\t\t\t"+suffix.join("; "));
            }
        }
        ++payloadPtr;
    }
    Q_ASSERT(curMscSlice.ch.empty());
    mstreamHdr = nullptr;
}

void MpdDataDecoder::decodeMscBlock_v1(const quint32 *payloadPtr)
{
    MscChData chData;
    chData.tai = tsTAI->toTaiTime();
    chData.ver = 1;
    quint8 cntBits = (*payloadPtr)&0xF;
    chData.ch = (*payloadPtr >> 4)&0xFF;
    quint8 ceBits = (*payloadPtr >> 12)&0xF;
    if(rawDataEvents) {
        printRawWord(payloadPtr,   QString("\t\tver:%1; ceBits:%2; ch:%3; cntBits:%4")
                         .arg(chData.ver).arg(ceBits)
                         .arg(chData.ch).arg(cntBits));
    }
    ++payloadPtr;

    chData.sliceNs = *(payloadPtr);
    if(rawDataEvents) {
        printRawWord(payloadPtr,   QString("\t\tslice:%1ns").arg(chData.sliceNs));
    }
    ++payloadPtr;

    chData.missedCnt = *payloadPtr;
    if(rawDataEvents) {
        printRawWord(payloadPtr,   QString("\t\missedCnt:%1ns").arg(chData.missedCnt));
    }
    ++payloadPtr;
    auto &devChHits = analyzerStats.dev_stat[getCurDeviceIndex()]
                          .dev_ch_hits[chData.ch].msc_hits;

    static constexpr const int sliceHdrSize_v2 = sizeof (TaiRawData)/sizeof (quint32) + 3;
    int sliceLen  = mstreamHdr->words32b - sliceHdrSize_v2;
    if(sliceLen && rawDataEvents)
        printRawTxtHeader("Slices");
    const quint32 cntMask = (1<<cntBits)-1;
    const quint32 ceMask = (1<<ceBits)-1;
    chData.data.resize(sliceLen);
    for(auto i=0; i<sliceLen; ++i) {
        MscTimeSliceData &sl = chData.data[i];
        sl.cnt  = (*payloadPtr) & cntMask;
        sl.ceState = (*payloadPtr>>cntBits) & ceMask;
        sl.sliceNumber = *payloadPtr >> (cntBits+ceBits);
        devChHits += sl.cnt;
        if(rawDataEvents) {
            printRawWord(payloadPtr, QString("\t\tslice:%L1; ce:%2; cnt:%L3")
                             .arg(sl.sliceNumber)
                             .arg(sl.ceState,0,16)
                             .arg(sl.cnt));
        }
        ++payloadPtr;
    }
    mstreamHdr = nullptr;
    decodedData.mscData1[getCurDeviceIndex()].chMap[chData.ch].push_back(chData);
}

void MpdDataDecoder::flushData()
{
    if(flushFunc)
        flushFunc();

    decodedData.clear();
}
