/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MpdRawTypes.h"

#include "base/DeviceIndex.h"

bool MpdHdr::knownSync(quint32 d)
{
    switch (d) {
    case MPD_SYNC_NORM:
    case MPD_SYNC_EOB:
    case MPD_SYNC_RUN_START:
    case MPD_SYNC_RUN_STOP:
    case MPD_SYNC_FILE_BEGIN:
    case MPD_SYNC_FILE_END:
    case MPD_SYNC_NORM_TLV:
    case MPD_SYNC_EOB_TLV:
    case MPD_SYNC_JSON_TLV:
        return true;
    }
    return false;
}

QString MpdHdr::getSyncName(const quint32 &s)
{
    switch (s) {
    case MPD_SYNC_NORM: return "Event(Old)";
    case MPD_SYNC_EOB: return "Stat(EOB)(Old)";
    case MPD_SYNC_RUN_START: return "RunStart";
    case MPD_SYNC_RUN_STOP: return "RunStop";
    case MPD_SYNC_FILE_BEGIN: return "FileBegin";
    case MPD_SYNC_FILE_END: return "FileEnd";
    case MPD_SYNC_RUN_NUMBER: return "RunNumber";
    case MPD_SYNC_RUN_INDEX: return "RunIndex";
    case MPD_SYNC_NORM_TLV: return "Event";
    case MPD_SYNC_EOB_TLV: return "Stat(EOB)";
    case MPD_SYNC_JSON_TLV: return "JSON";
    case MPD_SYNC_FILE_ID: return "FileId";
    case MPD_SYNC_EV_ORDER: return "EvbEvIndex";
    default: return "UNKNOWN";
    }
}

bool MpdHdr::isValid() const
{
    return MpdHdr::knownSync(sync);
}

bool MpdEventHdr::isValid() const
{
    return MpdHdr::knownSync(sync);
}

bool MpdRunNumberHdr::isValid() const
{
    return MpdHdr::knownSync(sync);
}

bool MpdRawDataDescriptor::isValid() const
{
    return eventHdr.isValid() &&
           (eventHdr.length == sizeof(deviceHdr) + deviceHdr.length) &&
           (deviceHdr.length == packet.size());
}

void MpdRawDataDescriptor::setDeviceIndex(const DeviceIndex &i)
{
    deviceHdr.deviceId = i.getDevId();
    deviceHdr.deviceSerial = i.getSerial();
}

void MpdRawDataDescriptor::setData(const QByteArray &d)
{
    setData(d.constData(), d.size());
}

void MpdRawDataDescriptor::setData(const char *d, int size)
{
    deviceHdr.length = size;
    eventHdr.length = sizeof(deviceHdr) + deviceHdr.length;
    packet.clear();
    packet.append(reinterpret_cast<const char *>(&eventHdr), sizeof(eventHdr));
    packet.append(reinterpret_cast<const char *>(&deviceHdr), sizeof(deviceHdr));
    packet.append(d, size);
}

std::optional<quint32> MpdRawDataDevRecord::getSync() const
{
    if(mpdHdr.has_value())
        return mpdHdr->sync;
    if(eventHdr.has_value())
        return eventHdr->sync;
    return {};
}

void MpdRawDataDevRecord::clear()
{
    mpdHdr.reset();
    eventHdr.reset();
    deviceHdr.reset();

    data.resize(0);
}

//template <>
//void MpdRawDataDevRecord::addData(const char *d, int size)
//{
//    data.append(static_cast<const char*>(std::as_const(d)), size);
//}

//template <>
//void MpdRawDataDevRecord::addData(char *d, int size) {
//    addData(static_cast<const char *>(d), size);
//}

//template <>
//void MpdRawDataDevRecord::setHeader(const MpdHdr &h)
//{
//    mpdHdr = h;
//}

//template <>
//void MpdRawDataDevRecord::setHeader(const MpdEventHdr &h)
//{
//    eventHdr = h;
//}

//template <>
//void MpdRawDataDevRecord::setHeader(const MpdDeviceHdr &h)
//{
//    deviceHdr = h;
//}
