/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ReadoutAppCore.h"

#include <QDebug>
#include <QLoggingCategory>

#include "AsyncDeviceCommander.h"
#include "fsm/BaseFsm.h"
#include "DeviceStatusZmqPublisher/DeviceStatusZmqPublisher.h"
#include "lib-common/BaseDeviceAppConfig.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mstream-lib/MStreamStat.h"
#include "remote-control-server/EvNumChecker.h"
#include "RunLogic/RunLogicController.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_appcore)
Q_LOGGING_CATEGORY(lc_appcore, "afi.appcore")
}

ReadoutAppCore::ReadoutAppCore(QObject *parent)
    : BaseAppCore{parent}
    , deviceCommander(new AsyncDeviceCommander(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , deviceStatusZmqPublisher(new DeviceStatusZmqPublisher(this))
    , evNumChecker(new EvNumChecker(this))
    , runLogicController(new RunLogicController(this))
{
}

ReadoutAppCore::~ReadoutAppCore()
{}

void ReadoutAppCore::deviceDisconnected(const DeviceIndex &index)
{
    if (isRun()) {
        emit sendRunError(rcRunId, index.getIdent() + " disconnected");
    }
}

void ReadoutAppCore::onEvNumCheckFail(const QString &mess) const
{
    if(!rcRunId.isValid())
        return;
    emit sendRunWarning(rcRunId, mess);
}

void ReadoutAppCore::requestAdcClkDivSync() const
{
    if (!isRun())
        emit adcClkDivSync();
}

void ReadoutAppCore::setDeviceStatus(const ModularDeviceStatus &st)
{
    feLinkRunCheck(st);
    adcSerDesRunCheck(st);
    trigOnXOffRunCheck(st);

    lastDevStatus[st.getDeviceIndex()] = st;
}

void ReadoutAppCore::setMStreamTcpStatus(const MStreamStat &st)
{
    if(st.deleted) {
        if(mstreamTcpStatus.remove(st.index))
            updateMStreamInterfaces();
        return;
    }
    QPair<QString, MsTcpServerConnectionStatus> p{st.devIdStr, st.tcpStatus.connectionSt};
    auto it = mstreamTcpStatus.constFind(st.index);
    if(it == mstreamTcpStatus.cend() || *it != p) {
        mstreamTcpStatus[st.index] = p;
        updateMStreamInterfaces();
    }
}

void ReadoutAppCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    BaseAppCore::onFsmStateChange(state, oldState);

    switch (state) {
    case FsmState::Run:
        feLinkStatusRunReper.clear();
        break;
    default:;
    }
}

bool ReadoutAppCore::isRun() const
{
    auto st = getFsm()->getState();
    return st==FsmState::Run
        || st==FsmState::Starting
        || st==FsmState::Configure;
}

void ReadoutAppCore::feLinkRunCheck(const ModularDeviceStatus &status)
{
    if(!isRun())
        return;
    auto index = status.getDeviceIndex();
    auto stIt = feLinkStatusRunReper.find(index);
    if(stIt == feLinkStatusRunReper.end()) {
        feLinkStatusRunReper.insert(index, status);
        return;
    }
    bool isErr = getBaseAppConfig().errOnFeLinkErr;
    bool isWarn = getBaseAppConfig().warnOnFeLinkErr;
    if(!isErr && !isWarn) {
        // In this case just remember 1st device status
        return;
    }

    auto oldIt=stIt->feLinkStatus.cbegin();
    QStringList errStrList;
    for(auto curIt=status.feLinkStatus.cbegin(); curIt!=status.feLinkStatus.cend(); ++curIt){
        if(oldIt == stIt->feLinkStatus.cend() || curIt.key() != oldIt.key()) {
            qCritical()
            << QString("[%1]").arg(index.getSerialStr())
            << QString("Ch:%1").arg(curIt.key())
            << "Changes in feLinkStatus keys oldKeys:"
            << stIt->feLinkStatus.keys()
            << " curKeys:"
            << status.feLinkStatus.keys();
            *stIt = status;
            return;
        }
        if(curIt->baseStatus.allOk() != oldIt->baseStatus.allOk()) {
            auto msg = QString("[%1][%2] feLinkStatus change ch:%3")
            .arg(programIndex.toString())
                .arg(index.toString())
                .arg(curIt.key());
            qWarning() << msg;
            errStrList << msg;
            ++oldIt;
            continue;
        }
        auto checkList = {&FeLinkCounter::RX_code_error,
                          &FeLinkCounter::RX_CRC_error,
                          &FeLinkCounter::unknown_received,
                          &FeLinkCounter::phy_reset};
        for(quint32 FeLinkCounter::*p : checkList) {
            if(curIt->counter.*p != oldIt->counter.*p) {
                auto msg = QString("[%1][%2] feLinkStatus err counter change ch:%3 (%L4->%L5)")
                .arg(programIndex.toString())
                    .arg(index.getIdent())
                    .arg(curIt.key())
                    .arg(oldIt->counter.*p)
                    .arg(curIt->counter.*p);
                qWarning() << msg;
                errStrList << msg;
                break;
            }
        }
        ++oldIt;
    }

    if(!errStrList.isEmpty()) {
        auto msg = QString("[%1] FeLink error: %2")
        .arg(programIndex.toString())
            .arg(errStrList.join('\n'));
        if(isErr) {
            emit sendRunError(rcRunId, msg);
        } else if(isWarn) {
            emit sendRunWarning(rcRunId, msg);
        }
        qWarning() << QString("[%1][%2] old:\n")
                          .arg(programIndex.toString())
                          .arg(index.getIdent())
                   << stIt->feLinkStatus
                   <<"\nCurrent:\n"
                   << status.feLinkStatus;
        // update device reper status
        *stIt = status;
    }
}

void ReadoutAppCore::adcSerDesRunCheck(const ModularDeviceStatus &st)
{
    if(!isRun() || !getBaseAppConfig().errOnSerDesUnlock || !st.desStatus)
        return;
    if(!st.desStatus->isOk()) {
        auto msg = QString("SerDes unlock(%1) detected in %2")
                       .arg(st.desStatus->toString())
                       .arg(st.getDeviceIndex().getIdent());
        emit sendRunError(rcRunId, QString("[%1] %2")
                                       .arg(programIndex.toString())
                                       .arg(msg));
    }
}

void ReadoutAppCore::trigOnXOffRunCheck(const ModularDeviceStatus &status)
{
    if(!isRun() || !status.runStatistic)
        return;

    auto index = status.getDeviceIndex();
    auto it = lastDevStatus.constFind(index);
    if(it == lastDevStatus.cend() || !it->runStatistic)
        return;
    auto blockCnt = (status-*it).runStatistic->getCounter(RunStatistic::Blocked);
    if(getBaseAppConfig().errOnTrigOnXOff && blockCnt && underRemoteCtrl) {
        auto msg = QString("Blocked trig(%1) in %2").arg(blockCnt)
                       .arg(index.getIdent());
        emit sendRunError(rcRunId, QString("[%1] %2")
                                       .arg(programIndex.toString())
                                       .arg(msg));
    }
}

void ReadoutAppCore::updateMStreamInterfaces()
{
    // Remove all ProgramInterfaceMStream
    for(auto j=0; j<progDescr.interfaces.size(); ) {
        if(progDescr.interfaces[j].type == ProgramInterfaceMStream)
            progDescr.interfaces.removeAt(j);
        else
            ++j;
    }

    int intId = 0;
    for(auto die : getBaseAppConfig().getDeviceIndexEnabledMap()) {
        auto it = mstreamTcpStatus.constFind(die.index);
        if(it == mstreamTcpStatus.cend())
            continue;

        const auto &connSt = it->second;
        ProgramInterface interface;
        interface.setType(ProgramInterfaceMStream);
        interface.devIndex = die.index;
        interface.devTypeStr = it->first;
        interface.id = intId;
        interface.port = connSt.tcpPort;
        interface.enabled = true;
        interface.isFree = connSt.isFree;

        if(!connSt.isFree)
            interface.setPeer(connSt.conHost, connSt.conPort);
        progDescr.interfaces.push_back(interface);

        ++intId;
    }

    emit updateProgramDescription(progDescr);
}

