/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MStreamStatisticTableModel.h"

#include <optional>

#include "app/AppStyle.h"

struct MStreamStatisticItem
{
    std::optional<bool> connection;
    std::optional<MStreamStat> mstream;
    std::optional<MStreamStat> prev_mstream;
    std::optional<ReceiverStat> receiver;
    std::optional<ReceiverStat> prev_receiver;
};

class MStreamStatisticTable : public QMap<DeviceIndex, MStreamStatisticItem>
{
public:
    DeviceIndex indexOf(int row) const;
    int rowOf(const DeviceIndex &sourceIndex) const;
};

int MStreamStatisticTable::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(begin(), find(deviceIndex));
}

DeviceIndex MStreamStatisticTable::indexOf(int row) const
{
    int r = 0;
    for (auto it = begin(); it != end(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

MStreamStatisticTableModel::MStreamStatisticTableModel(QObject *parent)
    : DeviceIndexTableModel{parent}
    , devices(new MStreamStatisticTable)
{}

MStreamStatisticTableModel::~MStreamStatisticTableModel()
{
    delete devices;
}

DeviceIndex MStreamStatisticTableModel::deviceIndexOf(int row) const
{
    return devices->indexOf(row);
}

int MStreamStatisticTableModel::rowOf(const DeviceIndex &sourceIndex) const
{
    return devices->rowOf(sourceIndex);
}

static const double ResetModelRatioThreshold = 0; // always use resetModel

void MStreamStatisticTableModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    const auto &newSet = map.getEnabled();
    const auto &oldSet = makeDeviceIndexSet(devices->keys());
    const auto toRemove = oldSet-newSet;
    const auto toAdd = newSet-oldSet;
    bool reset = false;
    reset |= (toRemove.size() >= devices->size() * ResetModelRatioThreshold);
    reset |= (toAdd.size() >= devices->size() * ResetModelRatioThreshold);
    if (reset) {
        beginResetModel();
        for(auto i : toRemove)
            devices->remove(i);
        for(auto i : toAdd)
            devices->insert(i, {});
        endResetModel();
    } else {
        for(auto i : toRemove) {
            removeRow(rowOf(i));
            devices->remove(i);
        }
        for(auto i : toAdd) {
            devices->insert(i, {});
            insertRow(rowOf(i));
        }
    }
}

void MStreamStatisticTableModel::stat_updated(const MStreamStat &ms_stat)
{
    const auto &deviceIndex = ms_stat.index;
    auto it = devices->find(deviceIndex);
    if (it == devices->end())
        return;
    it->prev_mstream = it->mstream;
    it->mstream = ms_stat;
    invalidateRow(deviceIndex);
}

void MStreamStatisticTableModel::stat_from_receiver_updated(const ReceiverStat &stat)
{
    const auto &deviceIndex = stat.index;
    auto it = devices->find(deviceIndex);
    if (it == devices->end())
        return;
    it->prev_receiver = it->receiver;
    it->receiver = stat;
    invalidateRow(deviceIndex);
}

void MStreamStatisticTableModel::mstreamConnectionStateChanged(const DeviceIndex &deviceIndex, bool connectionState)
{
    auto it = devices->find(deviceIndex);
    if (it == devices->end())
        return;
    it->connection = connectionState;

    int row = rowOf(deviceIndex);
    if (row == -1)
        return;
    const auto &indexFrom = createIndex(row, COL_CONNECTION);
    emit dataChanged(indexFrom, indexFrom);
}

int MStreamStatisticTableModel::rowCount(const QModelIndex &parent) const
{
    return  devices->size();
}

int MStreamStatisticTableModel::columnCount(const QModelIndex &parent) const
{
    return COL_TOTAL;
}

QVariant MStreamStatisticTableModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto row = index.row();
    auto deviceIndex = devices->indexOf(row);
    if (deviceIndex.isNull())
        return {};

    switch (role) {
    case Qt::BackgroundRole: return data_BackgroundRole(deviceIndex, column);
    case Qt::DisplayRole: return data_DisplayRole(deviceIndex, column);
    case Qt::ToolTipRole: return data_ToolTipRole(deviceIndex, column);
    }
    return {};
}

QVariant MStreamStatisticTableModel::data_BackgroundRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    const auto &mstream = device.mstream;
    const auto &receiver = device.receiver;
    if (device.connection.has_value()) {
        const auto &connection = device.connection.value();
        switch (column) {
        case COL_CONNECTION: if (!connection) return AppStyle::tableBg(Severity::Error);
        }
    }

    return {};
}

QVariant MStreamStatisticTableModel::data_DisplayRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    if (device.connection.has_value()) {
        const auto &connection = device.connection.value();
        switch (column) {
        case COL_CONNECTION: return connection ? "Open" : "Closed";
        }
    }
    if (device.mstream.has_value()) {
        const auto &mstream = device.mstream.value();
        if (device.prev_mstream.has_value()) {
            const auto delta = mstream - device.prev_mstream.value();
            const double time = 1e-3 * delta.acquiredAtMs;
            if (time > 0) {
                switch (column) {
                case COL_PKG_RATE: return QString("%L1").arg(delta.packageCount / time, 0, 'f', 0);
                case COL_RECV_DATA_RATE: return QString("%L1").arg(delta.recvDataSize / 1e6 / time, 0, 'f', 1);
                case COL_EVENT_RATE: return QString("%L1").arg(delta.evCount / time, 0, 'f', 0);
                }
            }
        }
        switch (column) {
        case COL_PKG_COUNT: return QString("%L1").arg(mstream.packageCount);
        case COL_RECV_DATA_SIZE: return QString("%L1").arg(mstream.recvDataSize);
        case COL_RETRANS_FRAG_PARTS: return QString("%L1").arg(mstream.retransmitedFragParts);
        case COL_FRAG_TOTAL: return QString("%L1").arg(mstream.fragTotal);
        case COL_FRAG_CLOSED: return QString("%L1").arg(mstream.fragClosed);
        case COL_GOOD_FORMAT_PKG: return QString("%L1").arg(mstream.goodFormatPackage);
        case COL_MISSED_EVENTS: return QString("%L1").arg(mstream.missedEvents);
        case COL_LAST_EVENT_NUM: return QString("%L1").arg(mstream.lastEvNum);
        case COL_EVENT_COUNT: return QString("%L1").arg(mstream.evCount);
            // case COL_SLOW_CONN_PERCENT: return slow_connection_percent;
        case COL_FRAG_DROPED: return QString("%L1").arg(mstream.fragDroped);
        case COL_FRAG_MISSED: return QString("%L1").arg(mstream.fragMissed);
        case COL_FRAG_INCOMPLETED: return QString("%L1").arg(mstream.fragIncompled);
        case COL_BAD_FORMAT_PKG: return QString("%L1").arg(mstream.badFormatPackage);
        case COL_MISSED_EVENTS_RANGES: return QString("%L1").arg(mstream.missedEventsRanges);
        case COL_INCOMPLETE_EVENTS: return QString("%L1").arg(mstream.incompleteEv);
        }
    }
    if (device.receiver.has_value()) {
        const auto &receiver = device.receiver.value();
        switch (column) {
        case COL_NO_FREE: return QString("%L1").arg(receiver.noFree);
        case COL_DIDNT_HELP: return QString("%L1").arg(receiver.didntHelp);
        case COL_FREE: return QString("%L1").arg(receiver.free);
        case COL_CHECK_FAIL: return QString("%L1").arg(receiver.checkFail);
        case COL_NO_DATA: return QString("%L1").arg(receiver.noData);
        }
    }
    return {};
}

QVariant MStreamStatisticTableModel::data_ToolTipRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    const auto &mstream = device.mstream;
    const auto &receiver = device.receiver;

    switch (column) {
    default: return {};
    }
}

QVariant MStreamStatisticTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal) {
        if (role == Qt::ToolTipRole) {
            switch (static_cast<Columns>(section)) {
            case COL_TYPE: return "Type";
            case COL_SERIAL: return "Serial";
            case COL_CONNECTION: return "MStream connection state";
            case COL_PKG_COUNT: return "Package count";
            case COL_PKG_RATE: return "Package rate";
            case COL_RECV_DATA_SIZE: return "Receieve data size";
            case COL_RECV_DATA_RATE: return "Receieve data rate (MB/s)";
            case COL_RETRANS_FRAG_PARTS: return "Retransmited frag parts";
            case COL_FRAG_TOTAL: return "Fragment total";
            case COL_FRAG_CLOSED: return "Fragment closed";
            case COL_FRAG_DROPED: return "Fragment dropped";
            case COL_FRAG_MISSED: return "Fragment missed";
            case COL_FRAG_INCOMPLETED: return "Fragment incompleted";
            case COL_GOOD_FORMAT_PKG: return "Good format package";
            case COL_BAD_FORMAT_PKG: return "Bad format package";
            case COL_MISSED_EVENTS: return "Missed events";
            case COL_MISSED_EVENTS_RANGES: return "Missed events ranges";
            case COL_INCOMPLETE_EVENTS: return "Incomplete events";
            case COL_LAST_EVENT_NUM: return "Last event number";
            case COL_EVENT_COUNT: return "Event count";
            case COL_EVENT_RATE: return "Event rate";
            case COL_SLOW_CONN_PERCENT: return "Slow connection percent";
            case COL_NO_FREE: return "No free";
            case COL_DIDNT_HELP: return "Didnt help";
            case COL_FREE: return "Free";
            case COL_CHECK_FAIL: return "Check fail";
            case COL_NO_DATA: return "No data";
            case COL_TOTAL: return {};
            }
        }
        if (role == Qt::DisplayRole) {
            switch (static_cast<Columns>(section)) {
            case COL_TYPE: return "Type";
            case COL_SERIAL: return "Serial";
            case COL_CONNECTION: return "MStream";
            case COL_PKG_COUNT: return "Pkt";
            case COL_PKG_RATE: return "Pkt/s";
            case COL_RECV_DATA_SIZE: return "Bytes";
            case COL_RECV_DATA_RATE: return "MB/s";
            case COL_RETRANS_FRAG_PARTS: return "ReTx";
            case COL_FRAG_TOTAL: return "F:Total";
            case COL_FRAG_CLOSED: return "F:Closed";
            case COL_FRAG_DROPED: return "F:Dropped";
            case COL_FRAG_MISSED: return "F:Missed";
            case COL_FRAG_INCOMPLETED: return "F:Incomplete";
            case COL_GOOD_FORMAT_PKG: return "Pkt Ok";
            case COL_BAD_FORMAT_PKG: return "Pkt Bad";
            case COL_MISSED_EVENTS: return "Ev. miss";
            case COL_MISSED_EVENTS_RANGES: return "Ev.R. miss";
            case COL_INCOMPLETE_EVENTS: return "Ev. incomplete";
            case COL_LAST_EVENT_NUM: return "Last Ev. #";
            case COL_EVENT_COUNT: return "Ev. count";
            case COL_EVENT_RATE: return "Ev. rate";
            case COL_SLOW_CONN_PERCENT: return "Slow con %";
            case COL_NO_FREE: return "No free";
            case COL_DIDNT_HELP: return "Didnt help";
            case COL_FREE: return "Free";
            case COL_CHECK_FAIL: return "Check fail";
            case COL_NO_DATA: return "No data";
            case COL_TOTAL: return {};
            }
        }
    }
    return DeviceIndexTableModel::headerData(section, orientation, role);
}

QMap<int, bool> MStreamStatisticTableModel::defaultColumnState() const
{
    return {
        {COL_TYPE, false},
        {COL_SERIAL, false},
        {COL_PKG_COUNT, false},
        {COL_RECV_DATA_SIZE, false},
        {COL_RETRANS_FRAG_PARTS, false},

        {COL_SLOW_CONN_PERCENT, false},
        {COL_FRAG_TOTAL, false},
        {COL_FRAG_CLOSED, false},
        {COL_GOOD_FORMAT_PKG, false},
        {COL_MISSED_EVENTS, false},

        {COL_FRAG_DROPED, false},
        {COL_FRAG_MISSED, false},
        {COL_FRAG_INCOMPLETED, false},
        {COL_BAD_FORMAT_PKG, false},
        {COL_MISSED_EVENTS_RANGES, false},
        {COL_INCOMPLETE_EVENTS, false},

        {COL_NO_FREE, false},
        {COL_DIDNT_HELP, false},
        {COL_FREE, false},
        {COL_CHECK_FAIL, false},
        {COL_NO_DATA, false}
    };
}
