#include "RcClientStatusModel.h"

#include <QColor>

#include "app/AppStyle.h"
#include "base/ClientIndex.h"
#include "base/ClientIndexEnabled.h"
#include "RcClientData.h"

class RcClientStatusTable : public QMap<ClientIndex, ClientInfo>
{};

RcClientStatusModel::RcClientStatusModel(QObject *parent)
    : ClientIndexTableModel{parent}
    , clients(new RcClientStatusTable)
{}

RcClientStatusModel::~RcClientStatusModel()
{
    delete clients;
}

ClientIndex RcClientStatusModel::clientIndexOf(int row) const
{
    int r = 0;
    for (auto it = config.begin(); it != config.end(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

int RcClientStatusModel::rowOf(const ClientIndex &sourceIndex) const
{
    return std::distance(config.begin(), config.find(sourceIndex));
}

int RcClientStatusModel::rowCount(const QModelIndex &) const
{
    return config.size();
}

int RcClientStatusModel::columnCount(const QModelIndex &) const
{
    return COL_TOTAL;
}

QVariant RcClientStatusModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto clIndex = clientIndexOf(index.row());
    if (clIndex.isEmpty())
        return {};

    switch (role) {
    case Qt::DisplayRole: return data_DisplayRole(clIndex, column);
    case Qt::CheckStateRole: return data_CheckStateRole(clIndex, column);
    case Qt::ToolTipRole: return data_ToolTip(clIndex, column);
    case Qt::BackgroundRole: return data_BackgroundRole(clIndex, column);
    }
    return {};
}

QVariant RcClientStatusModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (section) {
        case COL_TYPE: return "Type";
        case COL_INDEX: return "Index";
        case COL_PROTOCOL: return "Proto";
        case COL_IS_VME: return "VME";
        case COL_EV_NUM_CHECK: return "EvNumCheck";
        case COL_HOST: return "Host";
        case COL_STATUS: return "Status";
        case COL_FSM: return "FSM";
        }
    }
    return ClientIndexTableModel::headerData(section, orientation, role);
}

Qt::ItemFlags RcClientStatusModel::flags(const QModelIndex &index) const
{
    auto fl = ClientIndexTableModel::flags(index);
    switch (index.column()) {
    case COL_EV_NUM_CHECK:
        fl |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    default:;
    }

    return fl;
}

bool RcClientStatusModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    auto column = index.column();
    auto clientIndex = clientIndexOf(index.row());
    if (!clientIndex.isEmpty()) {
        switch (column) {
        case COL_EV_NUM_CHECK:
            if(role == Qt::CheckStateRole) {
                auto &cc = config[clientIndex];
                bool en = value == Qt::Checked;
                if(cc.evNumCheck != en) {
                    cc.evNumCheck = en;
                    emit rcClientsConfigChanged(config);
                }
                return true;
            }
            break;
        default:;
        }
    }
    return ClientIndexTableModel::setData(index, value, role);
}

void RcClientStatusModel::setExpectedFsmState(const FsmState &st)
{
    expectedFsmState = st;
    auto indexFrom = createIndex(0, COL_FSM);
    auto indexTo = createIndex(rowCount()-1, COL_FSM);
    emit dataChanged(indexFrom, indexTo);
}

void RcClientStatusModel::setRcClientsConfig(const RcClientConfigMap &map)
{
    if(config == map)
        return;

    const auto &newSet = makeClientIndexSet(map.keys());
    const auto &oldSet = makeClientIndexSet(config.keys());
    const bool doReset = newSet!=oldSet;

    if(doReset) {
        beginResetModel();
        config = map;
    } else {
        for(auto it=config.begin(); it!=config.end(); ++it) {
            const auto &newCfg = map[it.key()];
            if(*it == newCfg)
                continue;
            *it = newCfg;
            if(!doReset)
                invalidateRow(it.key());
        }
    }
    config = map;

    if(doReset)
        endResetModel();
}

void RcClientStatusModel::updateClientStatus(const ClientInfo &status)
{
    const auto &clIndex = status.getIndex();
    if(!config.contains(clIndex))
        return;
    auto it = clients->find(clIndex);
    if(it != clients->end() && *it == status)
        return;
    clients->insert(clIndex, status);
    invalidateRow(clIndex, {Qt::DisplayRole, Qt::BackgroundRole});
}

QVariant RcClientStatusModel::data_DisplayRole(const ClientIndex &clIndex, int column) const
{
    switch (column) {
    case COL_TYPE:
        return clIndex.getProgramType().toString();
    case COL_INDEX:
        return clIndex.getProgramIndex().toString();
    default:;
    }

    auto infoIt = clients->constFind(clIndex);
    if(infoIt==clients->cend())
        return {};
    const auto ci = *infoIt;
    switch (column) {
    case COL_PROTOCOL: return getProtocolIdStr(ci.protoId);
    case COL_IS_VME: return ci.hasVme() ? QChar(0x2713) : QString();
    case COL_HOST: return ci.clientStatus.hostName.isEmpty()
                ? ci.clientStatus.host.toString() : ci.clientStatus.hostName;
    case COL_STATUS: return ci.warn.active ? ci.warn.msg : ci.status;
    case COL_FSM: return fsmStateToString(ci.fsmState);
    }
    return {};
}

QVariant RcClientStatusModel::data_CheckStateRole(const ClientIndex &clIndex, int column) const
{
    auto it=config.constFind(clIndex);
    if(it==config.cend())
        return {};
    switch (column) {
    case COL_EV_NUM_CHECK:
        return it->evNumCheck ? Qt::Checked : Qt::Unchecked;
    default:;
    }
    return {};
}

QVariant RcClientStatusModel::data_ToolTip(const ClientIndex &clIndex, int column) const
{
    auto infoIt = clients->constFind(clIndex);
    if(infoIt==clients->cend())
        return {};

    switch (column) {
    case COL_PROTOCOL: return getProtocolIdStr(infoIt->protoId, false);
    case COL_HOST:
        if(!infoIt->clientStatus.hostName.isEmpty())
            return infoIt->clientStatus.host.toString();
        break;
    }
    return {};
}

QVariant RcClientStatusModel::data_BackgroundRole(const ClientIndex &clIndex, int column) const
{
    auto cfgIt=config.constFind(clIndex);
    if(cfgIt==config.cend() || !cfgIt->en)
        return {};

    auto infoIt = clients->constFind(clIndex);
    if(infoIt==clients->cend())
        return {};
    auto ci = *infoIt;
    bool suportedProto = ci.protoId == RC_CLIENT_PROTO_V4 ||
        ci.protoId == RC_CLIENT_PROTO_V5;

    switch (column) {
    case COL_PROTOCOL:
        if(!suportedProto)
            return AppStyle::tableBg(Severity::Error);
        break;
    case COL_STATUS:
        if(suportedProto && ci.warn.active)
            return QColor("#FF9500");
        break;
    case COL_FSM:
    {
        if(!suportedProto)
            break;
        bool isOk = (!cfgIt->en) || ci.fsmState == expectedFsmState;
        return isOk ? QColor(60, 179, 113) : QColor(255, 99, 71);
    }
    }

    return {};
}

void RcClientStatusModel::invalidateRow(const ClientIndex &i, const QVector<int> roles)
{
    auto row = rowOf(i);
    emit dataChanged(index(row, 0), index(row, COL_TOTAL-1), roles);
}
