/*
**    Copyright 2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RedisPublisher.h"

#include <QDebug>
#include <QEvent>
#include <QFuture>
#include <QLoggingCategory>
#include <QSettings>
#include <QTimer>

#include "base/Globals.h"
#include "base/ServiceConfigUrl.h"
#include "qredisclient/connectionconfig.h"
#include "qredisclient/redisclient.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_redis)
Q_LOGGING_CATEGORY(lc_redis, "afi.redis")
static const char *DEFAULT_URL = "redis://localhost";
static const int DEFAULT_CONNECT_TIMEOUT = 5000; // > 1000
static const int DEFAULT_EXEC_TIMEOUT = 2000; // > 1000
static const bool DEFAULT_ENABLE = true;
static const int DEFERRED_INIT_TIMEOUT = 0;
}

RedisPublisher::RedisPublisher(const QString &serviceName, QObject *parent)
    : AbstractPublisher(parent), serviceName(serviceName)
{
    setObjectName("RedisPublisher");
    ServiceConfigUrl u(serviceName, QUrl::fromUserInput(DEFAULT_URL), DEFAULT_ENABLE);

    enabled = u.isEnabled();
    url = u.getUrl();
    if (url.isEmpty())
        enabled = false;
    if (!enabled)
        return;

    initRedisClient(); // qRegisterMetaType for redis client
    config = getConfig();

    // thread may change after ctor, use QueuedConnection
    reconnectTimer = new QTimer(this);
    reconnectTimer->setSingleShot(true);
    reconnectTimer->start(DEFERRED_INIT_TIMEOUT);
    connect(reconnectTimer, &QTimer::timeout, this, &RedisPublisher::deferredInit, Qt::QueuedConnection);
}

RedisPublisher::~RedisPublisher()
{
    if (connection)
        connection->disconnect();
    deleteConnection();
}

void RedisPublisher::createConnection()
{
    deleteConnection();
    bool autoconnect = false;
    connection = new RedisClient::Connection(config, autoconnect);

    connect(connection, &RedisClient::Connection::connected, this, &RedisPublisher::connected, Qt::QueuedConnection);
    connect(connection, &RedisClient::Connection::disconnected, this, &RedisPublisher::disconnected, Qt::QueuedConnection);
    //connect(connection, &RedisClient::Connection::shutdownStart, this, &RedisPublisher::disconnected, Qt::QueuedConnection);
    connect(connection, &RedisClient::Connection::error, this, &RedisPublisher::error, Qt::QueuedConnection);
    // connect(connection, &RedisClient::Connection::log, this, &RedisPublisher::log, Qt::QueuedConnection);
}

void RedisPublisher::deleteConnection()
{
    delete connection;
    connection = nullptr;
}

bool RedisPublisher::event(QEvent* ev)
{
    if (ev->type() == QEvent::ThreadChange) {
        if (connection) {
            deleteConnection();
            QTimer::singleShot(0, this, &RedisPublisher::deferredInit);
        }
        return true;
    }
    return QObject::event(ev);
}

void RedisPublisher::deferredInit()
{
    qInfo(lc_redis).noquote().nospace() << "connecting to " + url.toString();
    try {
        createConnection();
        connection->connect(true);
    }  catch (std::exception &e) {
        qCritical(lc_redis).noquote().nospace() << QString(e.what());
    }
    //    if (connection->isConnected())
    //        return;
    //startReconnectTimer();
}

void RedisPublisher::startReconnectTimer()
{
    // exponential backoff timer
    const int max_timeout = 300; // 30 seconds
    int timeout = 0;
    if (tryCount > 8)
        timeout = max_timeout;
    else if (tryCount >= 1)
        timeout = 1 << (tryCount-1);
    reconnectTimer->start(100 * timeout);
    tryCount++;
}

RedisClient::ConnectionConfig RedisPublisher::getConfig()
{
    const uint16_t port = (url.port() >= 0) ? url.port() : RedisClient::ConnectionConfig::DEFAULT_REDIS_PORT;
    RedisClient::ConnectionConfig config{url.host(), QString(), port,
                                         QCoreApplication::applicationName()};
    const auto &settings = Globals::getSettings();
    int connect_timeout = DEFAULT_CONNECT_TIMEOUT;
    {
        QString key = Globals::getEnvironment() + "/" + serviceName + "_connect_timeout";
        if (settings->contains(key)) {
            bool ok;
            int value = settings->value(key).toUInt(&ok);
            if (ok)
                connect_timeout = value;
        }
    }
    config.setConnectionTimeout(connect_timeout);
    int exec_timeout = DEFAULT_EXEC_TIMEOUT;
    {
        QString key = Globals::getEnvironment() + "/" + serviceName + "_exec_timeout";
        if (settings->contains(key)) {
            bool ok;
            int value = settings->value(key).toUInt(&ok);
            if (ok)
                exec_timeout = value;
        }
    }
    config.setExecutionTimeout(exec_timeout);
    Q_ASSERT(config.isValid());
    return config;
}

void RedisPublisher::pub(const QString &key, const QByteArray &data)
{
    Q_ASSERT(data.startsWith('{'));
    Q_ASSERT(data.endsWith('}'));

    if (!enabled)
        return;

    if (!connection) {
        qWarning(lc_redis).noquote().nospace() << __func__ << ": connection not initialized, message dropped";
        return;
    }
    try {
        if (connection->isConnected()) {
            // connection->command({"SET", key.toUtf8(), data}).result();
            auto cmd = QList<QByteArray>{"JSON.SET", key.toUtf8(), ".", data};
            auto ret = connection->command(cmd);
            // ret.waitForFinished();
        } else {
            qWarning(lc_redis).nospace() << __func__ << ": not connected, message dropped";
        }
    }  catch (std::exception &e) {
        qCritical(lc_redis).noquote().nospace() << QString(e.what());
    } catch (...) {
        qCritical(lc_redis).noquote().nospace() << "unhandled exception";
    }
}

void RedisPublisher::connected()
{
    qInfo(lc_redis).noquote().nospace() << "connected to " + url.toString();
    tryCount = 0;
}

void RedisPublisher::disconnected()
{
    qInfo(lc_redis).noquote().nospace() << "disconnected from " + url.toString();
    // deleteConnection();
    startReconnectTimer();
}

void RedisPublisher::error(const QString &s)
{
    qCritical(lc_redis).noquote().nospace() << s;
    // deleteConnection();
    startReconnectTimer();
}

void RedisPublisher::log(const QString &s)
{
    qDebug(lc_redis).noquote().nospace() << s;
}
