/*
**    Copyright 2011-2022 Ivan Filippov
**    Copyright 2011-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MStreamDump.h"
#include "base/DeviceId.h"

#include <cstring> // memcpy
#include <iostream> // cerr
#ifdef Q_OS_UNIX
#include <unistd.h>
#endif

#include <QCoreApplication>
#include <QDebug>
#include <QNetworkInterface>
#include <QStringList>
#include <QThread>
#include <QTime>
#include <QTimer>

#include "BaseFragment.h"
#include "MStreamFileWriter.h"
#include "MStreamStat.h"
#include "MlinkStreamReceiver.h"
#include "PacketRingBuffer.h"
#include "RawCheck.h"
#include "app/AppVersion.h"
#include "base/ConfigIndexName.h"
#include "base/DaqModule.h"
#include "base/Globals.h"
#include "base/HexDump.h"
#include "base/ThreadController.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PnpDiscovery.h"
#include "tcpserver/TcpServer.h"

#define PRINT_FRAMES false

namespace {
const quint32 MLINK_DATA_PADDING_MAGIC = 0x12206249;
const int BUF_HISTORY_SIZE = 800;

QByteArray allineData(const QByteArray &data)
{
    auto allinedData(data);
    if(int rest = (allinedData.size()%4))
        allinedData.append(4-rest, 0);
    return allinedData;
}
}

MStreamDump::MStreamDump(const DeviceIndex &index, bool _doPrintStat)
    :MStreamDump(MStreamDumpParameters(index), _doPrintStat)
{
    qRegisterMetaType<MpdRawDataDevRecord>();
    PnpDiscovery::instance().subscribeMyProgram(this);
}

MStreamDump::MStreamDump(MStreamDumpParameters msParams, bool _dumpAsApp) :
    QObject(),
    MStreamDumpParameters(msParams),
    sharedBuffer(new PacketRingBuffer(16 * 100)), // FIXME: hwBufSize*100
    mr(nullptr),
    fw(nullptr),
    tcpServer(new TcpServer(tcpPort, nullptr, QString("%1").arg(msParams.deviceIndex.getIdent()))),
    statTime(new QTime()),
    statTimer(new QTimer(this)),
    waitDataTimer(new QTimer(this)),
    metricTimer(new QTimer(this)),
    isDevBusy(false),
    myReceiverPort(0),
    dumpAsApp(_dumpAsApp)
{
    qRegisterMetaType<MLinkFrameInfo>();
    qRegisterMetaType<TcpServerStatus>();
    qRegisterMetaType<QVector<PacketBuffer> >();
    qRegisterMetaType<MStreamStat>();
    qRegisterMetaType<ReceiverStat>();
    stat.index = msParams.deviceIndex;

    msdump_reset(deviceIndex, 0, fileName, true);

    tcpPort = tcpServer->getServerPort();
    tcpServer->setObjectName(QString("TcpS_%1").arg(getSerialIdStr(true)));
    new ThreadController(tcpServer, this);

    connect(this, &MStreamDump::closeProgram, tcpServer, &TcpServer::deleteLater);
    connect(tcpServer, &TcpServer::statusUpdated, this, &MStreamDump::tcpServerStatusUpdated);
    connect(this, &MStreamDump::reset, tcpServer, &TcpServer::reset);
    connect(this, &MStreamDump::preRunCheck, this, &MStreamDump::onPreRunCheck);
    connect(tcpServer, &TcpServer::reseted,
            [=](int s) { emit reseted(deviceIndex, s); });
    PnpDiscovery::instance().subscribeMyProgram(this);

    updateProgDescription();

    connect(statTimer, &QTimer::timeout, this, &MStreamDump::printStat);
    statTimer->setSingleShot(false);
    statTimer->start(1000);

#if SEM_INTERACT
    connect(waitDataTimer, &QTimer::timeout, this, &MStreamDump::waitForData);
    waitDataTimer->setInterval(0);
    waitDataTimer->setSingleShot(true);
#endif

    metricTimer->start(5000);
    connect(metricTimer, &QTimer::timeout, this, &MStreamDump::sendMetrics);
    dataIndex = 0;

    createReciever();
}

MStreamDump::~MStreamDump()
{
    setDeleteFlag();

    statTimer->stop();
    emit closeProgram(progDescr);

    if(fw) delete fw;

    destroyReciever();

    delete statTime;
}

void MStreamDump::setDeleteFlag()
{
    stat.deleted = true;
    emit stat_updated(stat);
    disconnect(this, &MStreamDump::stat_updated, nullptr, nullptr);
    deleteFlag.store(true, std::memory_order_relaxed);
    if (mr)
        mr->setDeleteFlag();
}

bool MStreamDump::hasDeleteFlag() const
{
    return deleteFlag.load(std::memory_order_relaxed);
}

void MStreamDump::start()
{
    statTimer->start(1000);
}

void MStreamDump::gotData(const QVector<PacketBuffer> &tasks)
{
    if(!mr)
        return;
    if(mr_destroy) {
        Q_ASSERT(mr);
        const int size = tasks.size();
        emit releaseData(size);
        return;
    }
#if SEM_INTERACT
    Q_UNUSED(tasks)
#else
    for (const PacketBuffer &task : tasks) {
        processData(task);
    }
    int size = tasks.size();
    emit releaseData(size);
#endif
}

bool MStreamDump::verboseDataCheck(const char *buf, int recvSize)
{
    Q_ASSERT(recvSize>0);

    if(PRINT_FRAMES) {
        qInfo().noquote()
            << Qt::endl << Qt::endl << Qt::endl
            << logPrefix() << " recvSize="<<recvSize;
    }

    if (recvSize < sizeof(MStreamFragmentHeader)) {
        return false;
    }

    int curPos = 0;

    bool mstreamError = false;
    while(recvSize > curPos) {
        if (recvSize-curPos < (int)sizeof(MStreamHeader)) {
            qWarning() << logPrefix() + QString("Can't read MStreamHeader: truncated frame, curPos=%1, recvSize=%2")
                                            .arg(curPos).arg(recvSize);
            mstreamError = true;
            break;
        }
        const auto fragmentHdr = reinterpret_cast<const MStreamFragmentHeader *>(&buf[curPos]);
        curPos += sizeof(MStreamFragmentHeader);
        if (recvSize-curPos < (int)(fragmentHdr->fragmentLength)){
            qWarning() << logPrefix() + QString("Can't read MStream fragment: truncated frame, curPos=%1").arg(curPos);
            qWarning() << logPrefix() + "fragmentHdr->fragmentLength=" << fragmentHdr->fragmentLength;
            qWarning() << logPrefix() + "rest size:" << recvSize-curPos;
            mstreamError = true;
            break;
        }
        if(PRINT_FRAMES) {
            qInfo()<< logPrefix() + "fid="<<fragmentHdr->getFragId(bigFragId)<<" len="<<fragmentHdr->fragmentLength;
        }

        if (fragmentHdr->fragmentLength == 0){
            qWarning() << logPrefix() + QString("Invalid MStream header: fragmentLength=0, curPos=%1").arg(curPos);
            curPos += fragmentHdr->fragmentLength; // payload
            mstreamError = true;
            continue;
        }
        curPos += fragmentHdr->fragmentLength; // payload
    }
    return !mstreamError;
}

void MStreamDump::processData(const PacketBuffer &pkbuf)
{
    // verboseDataCheck(buf, recvSize);

    stat.recvDataSize += pkbuf.payloadSize;
    ++stat.packageCount;

    const auto firstFragment = pkbuf.first();
    if (!firstFragment)
        return;
    const auto firstFragmentId = firstFragment->getFragId(bigFragId);

    bool mstreamError = false;
    for (auto f = firstFragment; f; f = pkbuf.next(f)) {
        fragmentHdr = f;
        const auto fragmentId = f->getFragId(bigFragId);
        addHistory(fragmentId);

        quint16 idDiff = lastClosedFragmentId - fragmentId;
        if(!bigFragId) idDiff &= 0xFF;
        if(!initStage){
        if (idDiff < getAcceptedRange() || idDiff == 0) {
            ++stat.retransmitedFragParts;
            // Smart check for fragmentId resetted
//                bool ret = isFragIdResetted();
//                quint16 curFrId = fragmentId;
            // qDebug()<<"Got retransmited fragment: lastClosedFragmentId="<<lastClosedFragmentId<<" curFrId="<<curFrId;
            break;
/*
//                //Frame re-transmited
//                QSet<quint16> previousFrId;
//                previousFrId.insert(curFrId);
//                for(int shift=1; shift<BUF_HISTORY_SIZE; ++shift){
//                    int pos = histPos-shift;
//                    if(pos<0)
//                        pos += BUF_HISTORY_SIZE;
//                    previousFrId.insert(histFrags[pos]);
//                    if(previousFrId.size() == getAcceptedRange()+1)
//                        break;
//                }
//                QString prevIdStr("previous fragId:");
//                foreach (quint16 id, previousFrId) {
//                    prevIdStr.append(QString("%1 ").arg(id));
//                }
//                qDebug()<<prevIdStr;

//                bool ok=true;
//                if(previousFrId.size() != getAcceptedRange()+1){
//                    qWarning() << logPrefix() + "Insufficient fragId history";
//                    ok=false;
//                } else {
//                    qDebug()<<"lastClosedFragmentId="<<lastClosedFragmentId<<" curFrId="<<curFrId;

//                    foreach (quint16 id, previousFrId) {
//                        quint16 diff = curFrId-id;
//                        if(!bigFragId) diff &= 0xFF;
//                        if(diff > 2*getAcceptedRange()-1){
//                            qDebug()<<"diff(curFrId-if)="<<diff<<" curFrId="<<curFrId<<" id="<<id;
//                            ok=false;
//                            break;
//                        }
//                    }
//                }

//                if(ok){
//                    qDebug()<<"Check passed; Start closing all pending fragments";
//                    quint16 bakFrId = lastClosedFragmentId;
//                    quint16 frId=lastClosedFragmentId+1;
//                    if(!bigFragId) frId &= 0xFF;
//                    for(; frId!=bakFrId; ++frId) {
//                        if(frId==fragmentHdr->getFragId(bigFragId))
//                            continue;
//                        BaseFragment &fragment = fragments[frId];
//                        if(fragment.free) {
//                            lastClosedFragmentId = frId;
//                            continue;
//                        }
//                        qDebug()<<"Force closing: fragmentId="<<frId;
//                        closeFragment(frId, true);
//                    }

//                    lastClosedFragmentId = fragmentHdr->getFragId(bigFragId)-1;
//                    if(!bigFragId) lastClosedFragmentId &= 0xFF;
//                    qDebug() << "Set lastClosedFragmentId=" << lastClosedFragmentId
//                             << " previous lastClosedFragmentId=" << bakFrId;
//                } else {
//                    qDebug()<<"Check failed";
//                    ++stat.retransmitedFragParts;
//                    continue;
//                }
*/
        } else {
            idDiff = fragmentId - lastClosedFragmentId;
            if(!bigFragId) idDiff &= 0xFF;

            if(idDiff>getAcceptedRange()) {
                // Check on fragment loss
//                qWarning() << logPrefix() + "Force closing previous; reinit."<<"lastClosedFragmentId:"<<lastClosedFragmentId
//                           <<"curFrId:"<< fragmentId << "idDiff:" << (qint16)(idDiff);
                if(!adcTrigInfo.free) {
                    closeAdc();
                }
                initStage=true;
            }
        }
        } // if(!initStage)


//            addHistory(fragmentId);
        BaseFragment &fragment = fragments[fragmentId];
        if (fragment.free) {
            ++stat.fragTotal;
            fragment.init(f);
        }

        const int start = f->getFragOffset(bigFragId);
        const int end = start + f->fragmentLength;

        if (fragment.parts.contains(start) && fragment.parts[start]==end){
            //Got re-transmited fragment's part of incomplited package
            ++stat.retransmitedFragParts;
            /*
            qDebug() << "got retransmited fragment part:lastClosedFragmentId=" << lastClosedFragmentId
                     << "cur fragId=" << f->fragmentId
                     << "start=" << start
                     << "end=" << end ;
                     */
            break;
        }

        if (initStage)
            initPackageSet.insert(firstFragmentId);

        if(fragment.completed){
//            qWarning() << logPrefix() + QString("Recived fragment is already completed; fId=%1").arg(fragment.fragmentId);
#ifdef QT_DEBUG
//            printHistory();
#endif
            continue;
        }
        Q_ASSERT(!fragment.completed);

        fragment.insert(f);

        const bool lastFragment = (f->flgs >> 5) & 0x1;          //last fragment
        if (lastFragment){
            fragment.expectedLen = end;
        }
//            qDebug() << "add new part of fragment"<<fragmentId << " from:"<<start<<"to:"<<end<<(lastFragment?" last fragment":"");
        if (initStage) {
            if (initPackageSet.size() > hwBufSize)
                initialize();
        }
        if (fragment.expectedLen){
            while(fragment.parts.contains(fragment.curLen)) {
                fragment.curLen = fragment.parts[fragment.curLen];
            }
            if (fragment.curLen == fragment.expectedLen){
                fragment.completed = true;
                if (!initStage) {
                    if (fragmentId == ((lastClosedFragmentId+1)%(bigFragId?65536:256)))
                        closeFragment(fragmentId);
                    else {
                        quint16 diff = fragmentId-lastClosedFragmentId;
                        if(!bigFragId) diff &= 0xFF;
                        if (diff > getAcceptedRange()){
                            // Got completed fragment after initialization stage which frId is too big
//                            qWarning() << logPrefix() <<"some fragments stocked;"
//                                       <<"lastClosedFragmentId="<<lastClosedFragmentId
//                                      <<"bufSize="<<hwBufSize
//                                     <<"curFrId="<<fragmentId; //TODO: sad
//                            printHistory();
                            closeFragment((lastClosedFragmentId+1)%(bigFragId?65536:256), true);
//                            qWarning() << logPrefix() + QString("framesTotal=%1").arg(stat.fragTotal);
//                            qWarning() << logPrefix() + QString("framesSkipped=%1").arg(stat.fragDroped);
                        }
                    }
                    // else got fragment for uncompleted event (adc64 case)
                }
            }
        }
    }

    if (mstreamError)
        ++stat.badFormatPackage;
    else
        ++stat.goodFormatPackage;

    if(debugPrint){
        debugPrintBufInHex(pkbuf.payload, pkbuf.payloadSize);
        debugPrint=false;
    }
}

bool MStreamDump::isMyAddress(const QHostAddress &addr) const
{
    for (const auto &i : QNetworkInterface::allAddresses()) {
        if (i.protocol() != QAbstractSocket::IPv4Protocol)
            continue;
        if (i == QHostAddress(QHostAddress::LocalHost))
            continue;
        if (i == addr)
            return true;
    }
    return false;
}

void MStreamDump::updateLastRecord()
{
    lastRec = curRec;
    lastRecSent = false;
    curRec.clear();
    flushData(lastRec.data);
}

void MStreamDump::sendJsonBlockInternal(const DeviceIndex &index, const QByteArray &data)
{
    static const quint32 reservedField = 0;

    // allign to 32-bit size
    auto allinedData = allineData(data);

    mpdDeviceHdr.deviceId = index.getDevId();
    mpdDeviceHdr.deviceSerial = index.getSerial();
    mpdDeviceHdr.length = static_cast<quint32>(allinedData.size());

    mpdHdr.sync = MPD_SYNC_JSON_TLV;
    mpdHdr.length = mpdDeviceHdr.length + sizeof(MpdDeviceHdr) + sizeof(reservedField);

    curRec.addData(&mpdHdr, sizeof(mpdHdr));
    curRec.addData(&reservedField, sizeof(reservedField));
    curRec.addData(&mpdDeviceHdr, sizeof(mpdDeviceHdr));
    curRec.addData(allinedData.data(), mpdDeviceHdr.length);
    ++stat.jsonBlocks;
    updateLastRecord();
}

void MStreamDump::lldpCacheUpdated(const LldpInfoCache &dd)
{
    auto it = dd.find(deviceIndex);
    if (it == dd.cend())
        return;

    const auto &lldp = it.value();
//    bool isMySocket = mr ? (lldp.ext.mstream_locked && (mr->getPeerAddress() == lldp.ext.mstream)) : false;
//    const auto lldpHostAddress = QHostAddress(lldp.common.ip_addr);
//    bool isMyConnection = isMySocket && (lldpHostAddress == deviceAddress);
//    if (isMyConnection)
//        return;

    bool isLockedByOther = lldp.ext.mstream_locked && !isMyAddress(lldp.ext.mstream.hostAddress());
    if (isLockedByOther) {
        if (!isDevBusy) {
            qWarning() << logPrefix() +  QString("Device is locked by %1").arg(lldp.ext.mstream.toString());
            isDevBusy = true;
        }
        destroyReciever();
        return;
    }
    isDevBusy = false;
    auto modelname_str = lldp.med.modelname_str.trimmed();
    bool doUpdateProgDescription = false;
    if(stat.devIdStr != modelname_str) {
        stat.devIdStr = modelname_str;
        doUpdateProgDescription = true;
    }
    QHostAddress newHost(lldp.common.ip_addr);
    if(deviceAddress != newHost) {
        deviceAddress = newHost;
        doUpdateProgDescription = true;
        emit hostDiscovered(deviceAddress);
        //            runLoop();
    } else {
        if(!lldp.ext.mstream_locked || myReceiverPort != lldp.ext.mstream.port) {
            emit hostDiscovered(deviceAddress);
            if(!reconnectTimer.isValid() || reconnectTimer.elapsed() > 5000) {
                reconnectTimer.start();
                emit connectToHardware(true);
            }
        }
    }
    if(doUpdateProgDescription)
        updateProgDescription();
}

void MStreamDump::tcpServerStatusUpdated(const TcpServerStatus &status)
{
    tcpStatus = status;
    stat.tcpStatus = tcpStatus;
    updateProgDescription();
}


//On force closing we should send this fragment (if we can, if trig frame completed)
//and others assoshiated with this event
void MStreamDump::closeFragment(quint16 fragmentId, bool force)
{
    BaseFragment *fragment = &fragments[fragmentId];

//    printHistory();
    Q_ASSERT((lastClosedFragmentId+1)%(bigFragId?65536:256) == fragmentId);
    if (fragment->free || !fragment->completed) {
        Q_ASSERT(force);
        if(fragment->free) {
            // Device skiped all fragments with id=fragmentId
//            qWarning() << logPrefix() + QString("Closing empty fragment #%1").arg(fragmentId);
        } else {
            // Device skiped some/any fragment(s) with id=fragmentId
//            qWarning() << logPrefix() + QString("Closing uncompleted fragment #%1").arg(fragmentId);
            fragment->clear();
            ++stat.fragIncompled;
        }

        ++lastClosedFragmentId; // same as lastClosedFragmentId = fragmentId;
        ++stat.fragDroped;
        ++fragmentId;
        if(!bigFragId) fragmentId &= 0xFF;
        fragment = &fragments[fragmentId];
        if(fragment->free || !fragment->completed)
            return;
        force = false;
        MStreamHeader *mStreamHdr = reinterpret_cast<MStreamHeader *>(fragment->data.data());
        const quint32 evNum = mStreamHdr->evNum;
        fragment->evNum = evNum;
        qDebug()<< logPrefix() + "After force closing; Good fragment fragId="<<fragmentId<<"; evNum="<<fragment->evNum;
    }

    do {
        // Normal completed fragment
        MStreamHeader *mStreamHdr = reinterpret_cast<MStreamHeader *>(fragment->data.data());
        const quint32 evNum = mStreamHdr->evNum;
        fragment->evNum = evNum;
        fragment->deviceSerial = mStreamHdr->deviceSerial;

        checkDeviceSerial(mStreamHdr->deviceSerial);

        if (fragment->closed && !force) {
            qWarning() << logPrefix() << "double close fragment frId="<<fragmentId
                     <<" ev="<<fragment->evNum<<" subType="<<fragment->subType;
            printHistory();
            Q_ASSERT(!fragment->closed);
        }
        ++lastClosedFragmentId; // same as lastClosedFragmentId = fragmentId;
        if(!bigFragId) lastClosedFragmentId &= 0xFF;
        Q_ASSERT(lastClosedFragmentId == fragmentId);
        ++stat.fragClosed;
        fragment->closed=true;
        //        qDebug() << "Fragment"<<fragmentId<<"closed evNum"<<fragment->evNum<<" subType="<<fragment->subType;
        switch (fragment->deviceId) {
        case DEVICE_ID_ADC64WR:
        case DEVICE_ID_ADC64S2:
        case DEVICE_ID_ADC64VE:
        case DEVICE_ID_ADC64VE_XGE:
        case DEVICE_ID_ADC64ECAL:
        case DEVICE_ID_ADC64VE2:
            if(!adcTrigInfo.free && adcTrigInfo.evNum != evNum){
                qWarning() << logPrefix() << "force closing previous adc event evNum="<<adcTrigInfo.evNum << "curEvNum="<<evNum << " curFrId=" <<fragmentId;
//                printHistory();
//                qDebug()<<"last Event fragments: "<<lastEventFragments.first<<" - "<<lastEventFragments.second;
                closeAdc();
//                qDebug()<<"last Event fragments: "<<lastEventFragments.first<<" - "<<lastEventFragments.second;
                qWarning()<< logPrefix() + QString("curFrId=%1 type=%2 evNum=%3 ch=%4")
                          .arg(fragment->fragmentId)
                          .arg(fragment->subType)
                          .arg(fragment->evNum)
                          .arg(mStreamHdr->userDefBits,2,16,QChar('0'));
                qDebug()<< logPrefix() + "Start Bad fragment";
                debugPrintBufInHex(fragment->data.data(), fragment->data.size());
                qDebug()<< logPrefix() + "End Bad fragment";
                debugPrint=true;
            }
            if(adcTrigInfo.free){
                adcTrigInfo.free = false;
                adcTrigInfo.evNum = evNum;
            }
            adcTrigInfo.totalDataSize += fragment->curLen - (sizeof(MStreamHeader)-sizeof(MStreamFragmentHeader));
            if (fragment->subType==MS_SUB_TYPE_TRIG){
                if(fragment->curLen != (sizeof(MStreamHeader)-sizeof(MStreamFragmentHeader))+sizeof(AdcMStreamTrigPayload))
                    qWarning() << logPrefix() + "fragment->curLen:" << fragment->curLen;
                Q_ASSERT(fragment->curLen == (sizeof(MStreamHeader)-sizeof(MStreamFragmentHeader))+sizeof(AdcMStreamTrigPayload));

                if (adcTrigInfo.gotInfo){
                    qWarning() << logPrefix() + QString("Got duplicate AdcTrigInfo for evNum=%1").arg(evNum);
                }
                adcTrigInfo.gotInfo = true;
                adcTrigInfo.trigFrag = fragment;

                AdcMStreamTrigPayload *adcTrig = (AdcMStreamTrigPayload*)(fragment->data.data()+sizeof(MStreamHeader));
                adcTrigInfo.workedChannels = ((uint64_t)adcTrig->hiCh << 32) | adcTrig->lowCh;

                if (adcTrigInfo.workedChannels == adcTrigInfo.currentChannels){
//                    qDebug() << "Event "<<evNum<<" completed";
                    closeAdc();
                }
            } else {
                Q_ASSERT(fragment->subType==MS_SUB_TYPE_DATA);

                quint8 &chNum = mStreamHdr->userDefBits;
//                qDebug()<<fragmentId<<": data ch="<<chNum<<" evNum="<<evNum;
                adcTrigInfo.currentChannels |= (quint64)1<<chNum;
                adcTrigInfo.fragments[chNum] = fragment;
                if (adcTrigInfo.gotInfo && adcTrigInfo.workedChannels == adcTrigInfo.currentChannels){
//                    qDebug() << "Event "<<evNum<<" completed";
                    closeAdc();
                }
            }
            break;
        case DEVICE_ID_HRB6ASD:
        case DEVICE_ID_TTVXS:
        case DEVICE_ID_TDC72VXS:
        case DEVICE_ID_TDC64VHLE:
        case DEVICE_ID_TDC64VLE:
        case DEVICE_ID_TQDC16VS_E:
        case DEVICE_ID_TRC:
        case DEVICE_ID_PHOS_FEC:
        {
            Q_ASSERT(fragment->subType==MS_SUB_TYPE_TRIG);
            closeHrb(*fragment, mStreamHdr->userDefBits, fragment->deviceId != DEVICE_ID_HRB6ASD);
//            qDebug() << "Fragment id="<<fragment->fragmentId<<" cleared.";
            break;
        }
        case DEVICE_ID_MSC16VE_E:
        {
            Q_ASSERT(fragment->subType==MS_SUB_TYPE_TIMESLICE);
            closeTimeSlice(*fragment, mStreamHdr->userDefBits);
            break;
        }
        default:
            static QSet<quint8> warnUnknownId;
            if(!warnUnknownId.contains(fragment->deviceId)) {
                warnUnknownId.insert(fragment->deviceId);
                qWarning() << logPrefix() + QString("Unknown device type %1").arg(fragment->deviceId, 2, 16, QChar('0'));
            }
            closeHrb(*fragment, mStreamHdr->userDefBits);
            break;
        }
        //check is there a completed fargment
        ++fragmentId;
        if(!bigFragId) fragmentId &= 0xFF;
        fragment = &fragments[fragmentId];
    } while(fragment->completed);
}


void MStreamDump::closeAdc()
{
    QMap<quint8, BaseFragment*>::const_iterator it;
    if(!adcTrigInfo.gotInfo) {
        // Can't send without trig fragment
        qWarning() << logPrefix() + "closeAdc without adcTrigInfo";
        for (it=adcTrigInfo.fragments.constBegin(); it!=adcTrigInfo.fragments.constEnd(); ++it){
            qInfo() << logPrefix() +  "drop frId:"<<it.value()->fragmentId;
            it.value()->clear();
            ++stat.fragDroped;
        }
        adcTrigInfo.clear();
        return;
    }
    // fragment.curLen - length in bytes =8+Subtype 0 Payload
    int fragmentsNumber = adcTrigInfo.fragments.size();
    if(adcTrigInfo.gotInfo) {
        if(mpdDeviceHdr.deviceSerial == adcTrigInfo.trigFrag->deviceSerial){
            // Skip first adc-event
            const quint32 expectedEvNum = (mpdEventHdr.evNum+1)&0xFFFFFFu;
            if(expectedEvNum != adcTrigInfo.trigFrag->evNum){
                const quint32 diff = (adcTrigInfo.trigFrag->evNum-expectedEvNum)&0xFFFFFFu;
                ++stat.missedEventsRanges;
                stat.missedEvents += diff;
            }
        }
        mpdDeviceHdr.deviceId = adcTrigInfo.trigFrag->deviceId;
        mpdDeviceHdr.deviceSerial = adcTrigInfo.trigFrag->deviceSerial;
        mpdEventHdr.evNum = adcTrigInfo.trigFrag->evNum;
        ++fragmentsNumber;
    } else {
        mpdDeviceHdr.deviceId = DEVICE_ID_ADC64; // adc64 DeviceId
        mpdDeviceHdr.deviceSerial = 0;
        mpdEventHdr.evNum = 0;
    }
    mpdDeviceHdr.length = adcTrigInfo.totalDataSize + fragmentsNumber*sizeof(MpdMStreamHdr);

    mpdEventHdr.length = mpdDeviceHdr.length + sizeof(MpdDeviceHdr);

    sendHeaders();

    if(adcTrigInfo.workedChannels != adcTrigInfo.currentChannels){
        qWarning() << logPrefix() << "Incomplete Event workedChannels:"<<hex<<adcTrigInfo.workedChannels<<" currentChannels:"<<adcTrigInfo.currentChannels;
        debugPrint=true;
        ++stat.incompleteEv;
    }
    // Send trig MStream block
    quint16 minFr = 0, maxDiff, diff;
    BaseFragment *frag;
    if(adcTrigInfo.gotInfo) {
        frag = adcTrigInfo.trigFrag;
        minFr = frag->fragmentId;
        maxDiff = 0;
        mpdMStreamHdr.subtype = frag->subType;
        mpdMStreamHdr.words32b = frag->curLen - (sizeof(MStreamHeader)-sizeof(MStreamFragmentHeader));
        mpdMStreamHdr.words32b /= 4;
        mpdMStreamHdr.usrDefBits = 0;
        curRec.addData(&mpdMStreamHdr, sizeof(mpdMStreamHdr));
        curRec.addData((frag->data.data()+sizeof(MStreamHeader)),
                                  mpdMStreamHdr.words32b*4);
        frag->clear();
    }

    // Send channals MStream blocks
    for (it=adcTrigInfo.fragments.constBegin(); it!=adcTrigInfo.fragments.constEnd(); ++it) {
        frag = it.value();
        diff = frag->fragmentId-minFr;
        if(!bigFragId) diff &= 0xFF;
        if(diff>maxDiff)
            maxDiff = diff;
        mpdMStreamHdr.subtype = frag->subType;
        mpdMStreamHdr.words32b = frag->curLen - (sizeof(MStreamHeader)-sizeof(MStreamFragmentHeader));
        mpdMStreamHdr.words32b /= 4;
        mpdMStreamHdr.usrDefBits = it.key(); // Channal Number
        curRec.addData(&mpdMStreamHdr, sizeof(mpdMStreamHdr));
        curRec.addData((frag->data.data()+sizeof(MStreamHeader)),
                                  mpdMStreamHdr.words32b*4);

        frag->clear();
    }
//#ifdef QT_DEBUG
//    if(lastSize!=mpdEventHdr.length){
//        qDebug()<<"Mpd Event size="<<mpdEventHdr.length<<"; previous="<<lastSize;
//        lastSize=mpdEventHdr.length;
//    }
//    qDebug()<<adcTrigInfo.evNum;
//#endif

    maxDiff =  minFr + maxDiff; // maxFr
    lastEventFragments.remove(0);
    lastEventFragments.push_back(QPair<quint16, quint16>(minFr, maxDiff) );
//    qInfo()<<"lastEventFragments.last:"<<minFr<<"-"<<maxDiff<<" evNum="<<adcTrigInfo.evNum;
    adcTrigInfo.clear();
    updateLastRecord();
}


void MStreamDump::closeHrb(BaseFragment &fragment, const quint8 exactTrigPos, bool keepEvNum)
{
    // fragment.curLen - length in bytes =2*32bit+Subtype 0 Payload
    mpdMStreamHdr.subtype=fragment.subType;
    mpdMStreamHdr.words32b=fragment.curLen-(sizeof(MStreamHeader)-sizeof(MStreamFragmentHeader));
    mpdMStreamHdr.words32b /= 4;
    mpdMStreamHdr.usrDefBits=exactTrigPos;

    mpdDeviceHdr.deviceId = fragment.deviceId;
    mpdDeviceHdr.deviceSerial = fragment.deviceSerial;
    mpdDeviceHdr.length = mpdMStreamHdr.words32b*4 + sizeof(MpdMStreamHdr);

    mpdEventHdr.length = mpdDeviceHdr.length + sizeof(MpdDeviceHdr);
    if(prevEvNum >= fragment.evNum)
        lastSpillEvNum += prevEvNum+1;
    prevEvNum = fragment.evNum;
    if(keepEvNum)
        mpdEventHdr.evNum = fragment.evNum;
    else
        mpdEventHdr.evNum = fragment.evNum + lastSpillEvNum;

    sendHeaders();
    curRec.addData(&mpdMStreamHdr, sizeof(mpdMStreamHdr));
    curRec.addData((fragment.data.data()+sizeof(MStreamHeader)),
              mpdMStreamHdr.words32b*4);
    fragment.clear();
    updateLastRecord();
}

void MStreamDump::closeTimeSlice(BaseFragment &fragment, const quint8 usrDefBits)
{
    static const quint32 reservedField = 0;

    // fragment.curLen - length in bytes =2*32bit+Subtype 0 Payload
    mpdMStreamHdr.subtype=fragment.subType;
    mpdMStreamHdr.words32b=fragment.curLen-(sizeof(MStreamHeader)-sizeof(MStreamFragmentHeader));
    mpdMStreamHdr.words32b /= 4;
    mpdMStreamHdr.usrDefBits=usrDefBits;

    mpdDeviceHdr.deviceId = fragment.deviceId;
    mpdDeviceHdr.deviceSerial = fragment.deviceSerial;
    mpdDeviceHdr.length = mpdMStreamHdr.words32b*4 + sizeof(MpdMStreamHdr);

    mpdHdr.sync = MPD_SYNC_EOB_TLV;
    mpdHdr.length = mpdDeviceHdr.length + sizeof(MpdDeviceHdr) + sizeof(reservedField);

    curRec.addData(&mpdHdr, sizeof(mpdHdr));
    curRec.addData(&reservedField, sizeof(reservedField));
    curRec.addData(&mpdDeviceHdr, sizeof(mpdDeviceHdr));
    ++stat.timeSliceBlocks;

    curRec.addData(&mpdMStreamHdr, sizeof(mpdMStreamHdr));

    curRec.addData(fragment.data.data()+sizeof(MStreamHeader),
              mpdMStreamHdr.words32b*4);
    fragment.clear();
    updateLastRecord();
}

void MStreamDump::sendHeaders()
{
    curRec.addData(&mpdEventHdr, sizeof(mpdEventHdr));
    curRec.addData(&mpdDeviceHdr, sizeof(mpdDeviceHdr));
    ++stat.evCount;
    stat.lastEvNum = mpdEventHdr.evNum;
}


void MStreamDump::flushData(const QByteArray &data)
{
    if (fw) {
        fw->write(data.data(), static_cast<size_t>(data.size()));
    }
    if (tcpServer) {
        tcpServer->addEvent(data.data(), data.size());
    }
}

void MStreamDump::addHistory(int fragId)
{
//    if(histPos==0)
//        printHistory();
    histFrags[histPos++] = fragId;
    if(histPos>=BUF_HISTORY_SIZE)
        histPos=0;
}


void MStreamDump::printHistory()
{
    QString histStr("fragment history:");
    quint16 minFrag = histFrags[histPos];
    quint16 curFrag = minFrag;
    quint16 fragRepeat=1;
    for(int pos=histPos+1;; ++pos){
        if(pos==BUF_HISTORY_SIZE)
            pos=0;
        if(pos==histPos)
            break;

        quint16 const &nextFrag = histFrags[pos];

        if(curFrag == nextFrag) {
            ++fragRepeat;
        } else {
            if(fragRepeat != 1) {
                histStr.append(QString("%1[%2] ").arg(curFrag).arg(fragRepeat));
                fragRepeat=1;
                minFrag = nextFrag;
            } else {
                if((curFrag+1) != nextFrag) {
                    if(minFrag == curFrag) {
                        if(fragRepeat==1) {
                            histStr.append(QString("%1 ").arg(curFrag));
                        }
                    } else
                        histStr.append(QString("%1->%2 ").arg(minFrag).arg(curFrag));
                    minFrag = nextFrag;
                }
            }
            curFrag = nextFrag;
        }
    }
    if(minFrag == curFrag)
        if(fragRepeat==1) {
            histStr.append(QString("%1 ").arg(curFrag));
        } else {
            histStr.append(QString("%1[%2] ").arg(curFrag).arg(fragRepeat));
        }
    else
        histStr.append(QString("%1->%2 ").arg(minFrag).arg(curFrag));

    print(histStr, true);
    print(QString("last Event fragments: %1 - %2")
          .arg(lastEventFragments.last().first)
          .arg(lastEventFragments.last().second));
}

void MStreamDump::printFragParts(quint16 fragId)
{
    print(QString("printFragParts for frId=%1:").arg(fragId));
    const BaseFragment &fr = fragments[fragId];
    if(fr.parts.isEmpty())
        print("empty");
    int checkAddr=0;
    foreach (int start, fr.parts.keys()) {
        if(checkAddr!=start)
            print(QString("Missing:%1-%2").arg(checkAddr).arg(start));
        checkAddr = fr.parts[start];
        print(QString("%1-%2").arg(start).arg(fr.parts[start]));
    }
}

void MStreamDump::sendMetrics()
{
    if (dumpAsApp)
        return;
    QMap<QString, QString> tags({{"serial", deviceIndex.getSerialStr()}});

    //    MetricSender::send("mstream.packets.total", stat.packageCount, tags);
    MetricSender::send("mstream.packets.valid", stat.goodFormatPackage, tags);
    MetricSender::send("mstream.packets.error", stat.badFormatPackage, tags);
    MetricSender::send("mstream.fragments.count", stat.fragTotal, tags);
    MetricSender::send("mstream.fragments.assemble.complete", stat.fragClosed, tags);
    MetricSender::send("mstream.fragments.assemble.dropped", stat.fragDroped, tags);
    MetricSender::send("mstream.fragments.assemble.missing", stat.fragMissed, tags);
    MetricSender::send("mstream.fragments.assemble.incompleted", stat.fragIncompled, tags);
    MetricSender::send("mstream.fragments.retransmit", stat.retransmitedFragParts, tags);
    MetricSender::send("mstream.data.events", stat.evCount, tags);
    MetricSender::send("mstream.data.bytes", stat.recvDataSize, tags);
//    prevStat = stat;
}

void MStreamDump::waitForData()
{
    if (!mr || hasDeleteFlag() || !waitForDataEn)
        return;
    QElapsedTimer t;
    t.start();
    int process_ev_time = 0;
    bool hasData;
    int i;
    int checkTimeCnt=0;
    const int DATA_SEM_TIMEOUT = 10;
    auto *mrDataSizeSem = sharedBuffer->getDataSizeSem();
    while(true) {
        hasData = mrDataSizeSem->tryAcquire(1, DATA_SEM_TIMEOUT);
        if(hasData) {
            int n = mrDataSizeSem->available();
            if(n)
                mrDataSizeSem->acquire(n);
            int task_size = n+1; // add one from tryAcquire
            for(i=0; i<task_size; ++i) {
                const PacketBuffer &pkbuf = sharedBuffer->getPendingTask(i);
                if (pkbuf.dataValid())
                    processData(pkbuf);
            }
            sharedBuffer->releaseDataSem(task_size);
            checkTimeCnt += task_size;
        } else {
            checkTimeCnt += DATA_SEM_TIMEOUT;
        }
        if(checkTimeCnt>100) {
            int time = t.elapsed();
            if(time > 1000)
                break;
            if(time-process_ev_time > 10) {
                QCoreApplication::processEvents(QEventLoop::AllEvents, 1);
                if (hasDeleteFlag() || !waitForDataEn)
                    return;
                process_ev_time = time;
            }
            checkTimeCnt=0;
        }
    }

    waitDataTimer->start();
}

void MStreamDump::msdump_reset(const DeviceIndex &index, int seqNum, QString fn, bool doStatClear)
{
    if(index != deviceIndex)
        return;
    preRunInit();

    const int frSize = bigFragId ? FRAGMENTS_SIZE_16_BIT : FRAGMENTS_SIZE_8_BIT;
    fragments.resize(frSize);
    for (int i=0;i<frSize; ++i) {
        fragments[i].clear();
        fragments[i].fragmentId = i;
    }

    adcTrigInfo.clear();

    lastEventFragments.clear();
    lastEventFragments.resize(20);

    histFrags.clear();
    histFrags.resize(BUF_HISTORY_SIZE);
    histPos=0;

    fragmentHdr= nullptr;
    lastClosedFragmentId = -1;
    initStage = true;

    lastSpillEvNum = 0;
    prevEvNum = 0;

    if(doStatClear)
        stat.clear();

    if(fn != fileName) {
        delete fw;
        fileName = fn;
        fw = fileName.isNull() ? nullptr : new MStreamFileWriter(fileName.toLocal8Bit());
    }

    emit reset(seqNum);
}

void MStreamDump::sendJsonBlock(const DeviceIndex &index, const QByteArray &data)
{
    if(index != deviceIndex)
        return;
    sendJsonBlockInternal(index, data);
}

void MStreamDump::sendCustomBlock(const DeviceIndex &index, const QByteArray &data)
{
    if(index != deviceIndex)
        return;
    static const quint32 reservedField = 0;

    // allign to 32-bit size
    QByteArray allinedData(data);
    if(int rest = (allinedData.size()%4))
        allinedData.append(4-rest, 0);

    mpdDeviceHdr.deviceId = index.getDevId();
    mpdDeviceHdr.deviceSerial = index.getSerial();
    mpdDeviceHdr.length = static_cast<quint32>(allinedData.size());

    mpdHdr.sync = MPD_SYNC_EOB_TLV;
    mpdHdr.length = mpdDeviceHdr.length + sizeof(MpdDeviceHdr) + sizeof(reservedField);

    curRec.addData(&mpdHdr, sizeof(mpdHdr));
    curRec.addData(&reservedField, sizeof(reservedField));
    curRec.addData(&mpdDeviceHdr, sizeof(mpdDeviceHdr));
    curRec.addData(allinedData.data(), mpdDeviceHdr.length);
    ++stat.customBlocks;
    updateLastRecord();
}

void MStreamDump::connectToDevice(const DeviceIndex &index)
{
    if (index != deviceIndex)
        return;
    emit connectToHardware();
}

void MStreamDump::disconnectFromDevice(const DeviceIndex &index)
{
    if (index != deviceIndex)
        return;
    emit disconnectFromHardware();
}


void MStreamDump::configSelectorChanged(const ConfigIndexName &cin)
{
    ClientIndex ci{Globals::getDaqModule(), cin.getIndex()};
    if(parentClientIndex == ci)
        return;

    parentClientIndex = ci;
    updateProgDescription();
}

void MStreamDump::requestLastRecord() const
{
    emit lastRecordUpdated(lastRec.data);
}

void MStreamDump::onPreRunCheck()
{
    if(initStage) {
        for(auto &f : fragments)
            f.clear();
    }
}

QString MStreamDump::formatStatText(int timeMs) const
{
    QString mess = QString("[stat]packages=%1;packLastSec=%8;retransmitCount=%2;"
                           "evCount=%10;evCountLastSec=%11;lastEv=%12;"
                           "fragAll=%5;fragDroped=%6;fragMissed=%7;"
                           "recvSize_B=%3;recvSizeLastSec_kBs=%4;"
                           "upTimeSec=%9")
                       .arg(stat.packageCount).arg(stat.retransmitedFragParts)
                       .arg(stat.recvDataSize).arg(1.*(stat.recvDataSize-prevStat.recvDataSize)/timeMs)
                       .arg(stat.fragTotal).arg(stat.fragDroped).arg(stat.fragMissed)
                       .arg((stat.packageCount-prevStat.packageCount)*1000/timeMs)
                       .arg(stat.upTime.elapsed()/1000)
                       .arg(stat.evCount).arg((stat.evCount-prevStat.evCount)*1000/timeMs)
                       .arg(stat.lastEvNum);

    if(stat.badFormatPackage)
        mess.append(QString(";badFormatPackages=%1").arg(stat.badFormatPackage));
    if(stat.missedEventsRanges) {
        mess.append(QString(";missedEvents=%1;missedEventsRanges=%2")
                        .arg(stat.missedEvents)
                        .arg(stat.missedEventsRanges));
    }
    if(stat.fragIncompled) {
        mess.append(QString(";fragIncompled=%1")
                        .arg(stat.fragIncompled));
    }
    if(tcpServer)
        mess.append(QString(";tcpOccupancy=%1").arg(tcpServer->getOccupancy(),0,'f',2));
    return mess;
}

void MStreamDump::printStat()
{
#ifdef Q_OS_UNIX
    if(dumpAsApp) {
        const pid_t pId = getppid();
        if (pId==1) {
            qWarning() << logPrefix() + "No parent process, exiting";
            exit(1);
        }
    }
#endif
//    if(statTimer->elapsed() > RUN_LOOP_PERIOD_MS){
        if(dumpAsApp){
            qint64 timeMs = qMax(statTime->restart(), 1);
            print(formatStatText(timeMs), true);
            std::cerr << std::endl;
        }
        stat.acquiredAtMs = QDateTime::currentMSecsSinceEpoch();
        emit stat_updated(stat);
        if(!lastRecSent) {
            emit lastRecordUpdated(lastRec.data);
            lastRecSent = true;
        }
        prevStat = stat;
        //    }
//        stat.time.start();
}

void MStreamDump::initialize()
{
    qInfo().noquote() << logPrefix()  << "initialization";
    initStage=false;

    quint16 minFragId = fragmentHdr->getFragId(bigFragId);
    quint16 frId=minFragId-1;
    quint16 fragLimit = minFragId-getAcceptedRange()-1;
    quint16 fragLimit2 = minFragId+getAcceptedRange()-1;
    if(!bigFragId) {
        frId &= 0xFF;
        fragLimit &= 0xFF;
    }
    for( ; frId!=fragLimit; ) {
        Q_ASSERT(frId != fragmentHdr->getFragId(bigFragId));
        if(!fragments[frId].free){
            minFragId = frId;
        }
        --frId;
        if(!bigFragId) frId &= 0xFF;
    }
    lastClosedFragmentId = minFragId-1;

    for( ; frId!=fragLimit2; ) {
        if(!fragments[frId].free){
            fragments[frId].clear();
        }
        --frId;
        if(!bigFragId) frId &= 0xFF;
    }

    if(!bigFragId) lastClosedFragmentId &= 0xFF;
/*
//    // Find out first eadge of fragment that would be lastClosedFragmentId
//    for(quint8 frId=fragmentHdr->fragmentId-1; frId!=fragmentHdr->fragmentId; --frId){
//        if(fragments[frId].free){
//            lastClosedFragmentId = frId;
//            break;
//        }
//    }

//    // All previous fragments will be removed
//    for(quint8 frId=lastClosedFragmentId-1; frId!=fragmentHdr->fragmentId; --frId){
//        if(!fragments[frId].free){
//            qDebug()<<"remove previous fragments fId="<<frId;
//            fragments[frId].clear();
//        }
//    }

    // Defines gap in intialization loop
//    QVectorIterator<BaseFragment> i(fragments);
//    QList<quint8> ids;
//    while(i.hasNext()){
//        const BaseFragment &f = i.next();
//        if (f.free)
//            continue;
//        ids.append(f.fragmentId);
//    }
//    quint8 min = ids[0];
//    quint8 max = ids[ids.size()-1];
//    quint8 halfLoop= -1;
//    halfLoop /=2;
//    if (max-min > halfLoop){
//        min = -1;
//        max = 0;
//        foreach (quint8 id, ids) {
//            if (id < halfLoop){
//                // candidate for max
//                if (max < id)
//                    max = id;
//            } else {
//                // candidate for min
//                if (min > id)
//                    min = id;
//            }
//        }
//    }
//    bool hasGap=false;
//    int id=min;
//    for (int i=0;i<HW_BUF_SIZE;++i){
//        if (!ids.contains(id)){
//            hasGap=true;
//            break;
//        }
//        ++id;
//    }
//    if (hasGap){
//        qDebug() << "There is gap "<<id<<" in fragmentIds:"<<ids;
//    } else {
//        qDebug() << "All first fragments are in range:"<<min<<"-"<<max;
//    }
//    lastClosedFragmentId = min-1;
*/
    qInfo().noquote() << logPrefix() +  QString(" set lastClosedFragmentId to %1 curFragId=%2").arg(lastClosedFragmentId).arg(fragmentHdr->getFragId(bigFragId));

    // Try to Close gained fragments
    quint16 nextFrId = lastClosedFragmentId+1;
    if(!bigFragId) nextFrId &= 0xFF;
    if (fragments[nextFrId].completed)
        closeFragment(nextFrId);
}

void MStreamDump::checkDeviceSerial(uint32_t serial)
{
    const auto &deviceSerial = deviceIndex.getSerial();
    if(serial==static_cast<uint32_t>(deviceSerial))
        return;

    if(serial==0) {
        qInfo().noquote() << logPrefix() + "Missing device serial number";
    } else if (deviceSerial) {
        qInfo().noquote() << logPrefix() << QString("Device serial number changed from 0x%1 to 0x%2")
            .arg(getSerialIdStr(true)).arg(serial, 8, 16,QChar('0'));
    } else {
        qInfo().noquote() << logPrefix() << QString("Device serial number set to 0x%1")
            .arg(getSerialIdStr(true));
    }
    deviceIndex = DeviceIndex(deviceIndex.getDevId(), serial);
}

void MStreamDump::debugPrintBufInHex(const char *buf, int size)
{
#ifndef QT_DEBUG
    return;
#endif
    qDebug().noquote().nospace()
        << "hex dump:\n"
        << HexDump::toStringList(buf, size).join("\n");
}

void MStreamDump::updateProgDescription()
{
    QString dev_serial = getSerialIdStr(true).prepend("0x").toLower();
    const auto &pi = ProgramIndex::fromString(dev_serial);
    progDescr.setClientIndex({DaqModule(DaqModule::Type::MStream), pi});
    progDescr.setParent(parentClientIndex);
    progDescr.ver_date = getAppDateStr();
    progDescr.ver_hash = getAppVersion();
    progDescr.options.clear();
    progDescr.options["dev_serial"] = dev_serial;
    progDescr.options["dev_id"] = QString("0x%1")
            .arg(deviceIndex.getDevId(), 2, 16, QChar('0'));
    if(!stat.devIdStr.isEmpty())
        progDescr.options["dev_id_str"] = stat.devIdStr;
    if(!deviceAddress.isNull()) {
        progDescr.options["dev_address"] = deviceAddress.toString();
        progDescr.options["out_port"] = QString::number(tcpPort);
    }
    if(mr)
        progDescr.options["myReceiverPort"] = QString::number(myReceiverPort);
    if(!fileName.isEmpty())
        progDescr.options["fileName"] = fileName;
    progDescr.interfaces.clear();
    const auto &connectionSt = tcpStatus.connectionSt;
    if(connectionSt.tcpPort) {
        ProgramInterface progIf;
        progIf.setType(ProgramInterfaceDataFlow);
        progIf.port = connectionSt.tcpPort;
        progIf.isFree = connectionSt.isFree;
        progIf.enabled = true;
        if(!connectionSt.isFree)
            progIf.addPeer(connectionSt.conHost, connectionSt.conPort);
        progDescr.interfaces.push_back(progIf);

        progIf.setType(ProgramInterfaceMStream);
        progIf.devIndex = deviceIndex;
        progIf.devTypeStr = stat.devIdStr;
        progDescr.interfaces.push_back(progIf);
    }

    emit updateProgramDescription(progDescr);
}

void MStreamDump::rxPortUpdated(quint16 port)
{
    myReceiverPort = port;
    updateProgDescription();
}

void MStreamDump::receiverDestroyed()
{
    myReceiverPort = 0;
    updateProgDescription();
}

void MStreamDump::peerConnected()
{
    emit connectionStateChanged(deviceIndex, true);
}

void MStreamDump::peerDisconnected()
{
    emit connectionStateChanged(deviceIndex, false);
}

void MStreamDump::createReciever()
{
    if(mr)
        destroyReciever();

    mr = new MlinkStreamReceiver(deviceIndex, sharedBuffer, bigFragId, hwBufSize, dumpAsApp);
    connect(mr, &MlinkStreamReceiver::rxPortUpdated, this, &MStreamDump::rxPortUpdated);
    connect(mr, &QObject::destroyed, this, &MStreamDump::receiverDestroyed);
    connect(this, &MStreamDump::hostDiscovered, mr, &MlinkStreamReceiver::hostDiscovered);
    connect(this, &MStreamDump::connectToHardware, mr, &MlinkStreamReceiver::connectToHardware);
    connect(this, &MStreamDump::disconnectFromHardware, mr, &MlinkStreamReceiver::disconnectFromHardware);
    connect(mr, &MlinkStreamReceiver::peerConnected, this, &MStreamDump::peerConnected);
    connect(mr, &MlinkStreamReceiver::peerDisconnected, this, &MStreamDump::peerDisconnected);
#if (! SEM_INTERACT)
    connect(this, &MStreamDump::releaseData, mr, &MlinkStreamReceiver::releaseData);
    connect(mr, &MlinkStreamReceiver::gotData, this, &MStreamDump::gotData);
#endif
    connect(mr, &MlinkStreamReceiver::stat_from_receiver_updated, this, &MStreamDump::stat_from_receiver_updated);
    new ThreadController(mr, QThread::HighPriority, this);

#if SEM_INTERACT
    waitDataTimer->start();
#endif

    if(!deviceAddress.isNull())
        emit hostDiscovered(deviceAddress);
    waitForDataEn=true;
}

void MStreamDump::destroyReciever()
{
    waitForDataEn=false;
    if(!mr)
        return;

    mr_destroy = true;
    if (mr)
        mr->deleteReceiver();
    waitDataTimer->stop();
    mr = nullptr;
    mr_destroy = false;
}

void MStreamDump::print(QString s, bool noLog)
{
#ifdef _WIN32
    std::cerr<<s.toStdString();
#else
#if QT_VERSION < QT_VERSION_CHECK(5, 5, 0)
    qDebug()<<s;
#else
    if(noLog)
        std::cerr<<s.toStdString();
    else
        qInfo()<<s;
#endif // QT_VERSION
#endif // _WIN32
}

/*
 * = HRB =
 * == Recive ==
 * [MStreamHeader]
 *  [MStreamFragmentHeader]
 *   fragId+fragOffset
 *   devId+flags+subType+fragLen
 *  [/MStreamFragmentHeader]
 *  evNum+userDefBits
 *  devSerial
 * [/MStreamHeader]
 *
 * [RawData]
 *  [TS]
 *   tsSec
 *   tsNSec
 *  [/TS]
 *  [Samples/]
 * [/RawData]
 *
 * /// size(RawData)=fragLen-2
 *
 * == Send ==
 * [MpdEventHdr/]
 * [MpdDeviceHdr/]
 * [MpdMStreamHdr/]
 * [frag]
 *  trigPos+Len+SubType
 *  [RawData]
 *   [TS]
 *    tsSec
 *    tsNSec
 *   [/TS]
 *   [Samples/]
 *  [/RawData]
 * [/frag]
 *
 *
 *
 * = ADC64 =
 * == Recive ==
 * [MStreamHeader]
 *  [MStreamFragmentHeader]
 *   fragId+fragOffset
 *   devId+flags+subType+fragLen
 *  [/MStreamFragmentHeader]
 *  evNum+userDefBits
 *  devSerial
 * [/MStreamHeader]
 *
 * [RawData/]
 *
 *
 * fragLen=2+size(RawData)
 *
 * == Send ==
 * [MpdEventHdr/]
 * [MpdDeviceHdr/]
 * [MpdMStreamHdr/]
 * [frag]
 *  trigPos+Len+SubType
 *  [RawData/]
 * [/frag]
 *
 */
