//
//    Copyright 2012 AFI Electronics
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "TcpServer.h"

#include <QDebug>
#include <QLoggingCategory>
#include <QTcpServer>
#include <QTimer>
#include <QtNetwork>

#include <unistd.h>
#include <stdexcept>

//#include "mpd-common/PrintInfo.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_mstream_tcp)
Q_LOGGING_CATEGORY(lc_mstream_tcp, "afi.mstream.tcp")
}

//const int TCP_SERVER_BUFFER_SIZE = 10000;
// EventSize(s)=40+4*4*s; EventSize(200)=3240; 10Mb for 3236 Events
namespace {
const int TCP_SERVER_BUFFER_SIZE = 10*1024*1024;
const int SOCKET_PENDING_DATA_LIMIT = TCP_SERVER_BUFFER_SIZE;
const int SEND_TIMEOUT_MS = 10;
const int UPDATE_TIMEOUT_MS = 1000;
}

TcpServer::TcpServer(int port, QObject *parent, QString _logPrefix) :
    QObject(parent),
    logPrefix(_logPrefix),
    tcpPort(port),
    qTcpServer(new QTcpServer(this)),
    tcpSocket(nullptr),
    sendTimer(new QTimer(this)),
    updateTimer(new QTimer(this)),
    slow_connection_timer(new QElapsedTimer())
{
    reset(0);
    qRegisterMetaType<TcpServerStatus>();

    logPrefix = QString("[%1]: ").arg(logPrefix);

    while(!qTcpServer->listen(QHostAddress::Any, tcpPort)) {
        QAbstractSocket::SocketError errNum = qTcpServer->serverError();
        if(errNum==QAbstractSocket::AddressInUseError){
            qCritical(lc_mstream_tcp).noquote()
                << logPrefix + tr("Unable to start server on port %1: %2, exiting...").arg(tcpPort).arg(qTcpServer->errorString());
            // sleep(5);
            exit(-1);
            continue;
        }
        QString s = tr("Unable to start server on port %1: %2").arg(tcpPort).arg(qTcpServer->errorString());
        qCritical(lc_mstream_tcp).noquote() << logPrefix + s;
        throw std::runtime_error(s.toStdString());
    }
    tcpPort = qTcpServer->serverPort();
    status.connectionSt.tcpPort = tcpPort;
    connect(sendTimer, &QTimer::timeout, this, &TcpServer::on_sendTimer_timeout);
    sendTimer->setSingleShot(true);
    sendTimer->setInterval(SEND_TIMEOUT_MS);

    connect(updateTimer, &QTimer::timeout, this, &TcpServer::updateStatus);
    updateTimer->setInterval(UPDATE_TIMEOUT_MS);
    updateTimer->start();

    connect(qTcpServer, &QTcpServer::newConnection,
            this, &TcpServer::updateConnections);

    qDebug(lc_mstream_tcp).noquote()
        << logPrefix + QString("TcpServer created on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);

    QTimer::singleShot(0, this, &TcpServer::deferredInit);
}

TcpServer::~TcpServer()
{
    qint64 bytes = getDataSize();
    if (tcpSocket)
        bytes += tcpSocket->bytesToWrite();
    if (bytes)
        qWarning(lc_mstream_tcp).noquote()
            << logPrefix + QString("TcpServer not empty and destroyed, %1 bytes lost").arg(bytes);
    delete slow_connection_timer;
}

void TcpServer::deferredInit()
{
    qInfo(lc_mstream_tcp).noquote()
        << logPrefix + QString("TcpServer listening on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
}

void TcpServer::handleDisconnect()
{
    if (!m_connectionName.isEmpty())
        qInfo(lc_mstream_tcp).noquote()
            << logPrefix + QString("Closed connection from %1").arg(m_connectionName);
    deleteSocket();
    updateStatus();
}

void TcpServer::checkPendingData()
{
    if(!requestFinish)
        return;
    if((getDataSize()==0) && tcpSocket && (tcpSocket->bytesToWrite()==0))
        emit workCompleted();
}

void TcpServer::updateStatus()
{
    emit statusUpdated(status);
}

void TcpServer::reset(int seqNum)
{
    writePos = 0;
    readPos = 0;
//    freeSpace.release(TCP_SERVER_BUFFER_SIZE-freeSpace.available());
    slowConnection = false;
    numberBytesSent = 0;
    numberBytesDropped = 0;
    dropOnFree = true;
    requestFinish = false;

    data.clear();
    data.resize(TCP_SERVER_BUFFER_SIZE);

    QTimer::singleShot(0, this, &TcpServer::updateStatus);
    emit reseted(seqNum);
}

void TcpServer::deleteSocket()
{
    tcpSocket->deleteLater();
    tcpSocket = nullptr;
    status.connectionSt.conHost.clear();
    status.connectionSt.conPort = 0;
    status.connectionSt.isFree = isFree();
    numberBytesSent = 0;
    numberBytesDropped = 0;
    m_connectionName = QString();
}

bool TcpServer::isEnoughSpace(int size) const
{
    int freeWriteSize = getFreeSize();
    if(freeWriteSize < size)
        return false;

    double freeSizeRate = 1.*freeWriteSize/TCP_SERVER_BUFFER_SIZE;
    if(freeSizeRate<0.5){
        // Enable porosity
        double chance = 1.*qrand()/RAND_MAX;
        if(chance>freeSizeRate)
            //Skipp this event
            return false;
    }
    return true;
}

double TcpServer::getOccupancy() const
{
    return 1.*getDataSize()/TCP_SERVER_BUFFER_SIZE;
}

QString TcpServer::connectionName() const
{
    Q_ASSERT(tcpSocket);
    return connectionName(*tcpSocket);
}

QString TcpServer::connectionName(const QTcpSocket &socket) const
{
    return QString("%1:%2").arg(socket.peerAddress().toString()).arg(socket.peerPort());
}

void TcpServer::on_sendTimer_timeout()
{
    sendData();
    if (tcpSocket)
        sendTimer->start(SEND_TIMEOUT_MS);
}

void TcpServer::sendData()
{
    //wait data available
    int sentSize = 0;
    while(int ds = getContiniousDataSize()) {
        Q_ASSERT(ds>0);
        //Check connetion
        if(!tcpSocket) {
            numberBytesDropped += ds;
            readPos = (readPos+ds) % TCP_SERVER_BUFFER_SIZE;
            continue;
        }
        if(tcpSocket->bytesToWrite()>SOCKET_PENDING_DATA_LIMIT){
            if(!slowConnection){
                slowConnection = true;
                slow_connection_timer->start();
            }
            break;
        }
        if(slowConnection) {
            slowConnection = false;
            if(slow_connection_timer->isValid()) {
                status.slow_connection_duration += slow_connection_timer->restart();
            }
        }
        if(!tcpSocket->isValid() && tcpSocket->state() != QTcpSocket::ConnectedState){
            deleteSocket();
            updateStatus();
            sendTimer->stop();
            continue;
        }

        sentSize = tcpSocket->write(data.constData()+readPos, ds);
        Q_ASSERT(sentSize != -1);
        if(sentSize > 0) {
            readPos = (readPos+sentSize)%TCP_SERVER_BUFFER_SIZE; // TCP_SERVER_BUFFER_SIZE-1
//                freeSpace.release(tailSize);
            numberBytesSent += sentSize;
            emit bytesSent(numberBytesSent);
       } else {
            qWarning().noquote() << logPrefix + "Can't sent data to socket:"  << tcpSocket->errorString();
            break;
        }
    }

    writerWC.wakeAll();
    if(requestFinish)
        checkPendingData();
}


void TcpServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()){
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)) {
            tcpSocket->abort();
            deleteSocket();
        }
        if(tcpSocket==nullptr) {
            tcpSocket = pendingConnection;
            status.connectionSt.conHost = tcpSocket->peerAddress();
            status.connectionSt.conPort = tcpSocket->peerPort();
            status.connectionSt.isFree = isFree();
            if(requestFinish)
                connect(tcpSocket, &QTcpSocket::bytesWritten,
                        this, &TcpServer::checkPendingData);
            connect(tcpSocket, &QTcpSocket::disconnected,
                    this, &TcpServer::handleDisconnect);
            m_connectionName = connectionName();
            qInfo(lc_mstream_tcp).noquote()
                << logPrefix + " " + QString("Accepted connection from %1").arg(m_connectionName);
            sendTimer->start(0);
            updateStatus();
        } else {
            qInfo(lc_mstream_tcp).noquote()
                << logPrefix + QString("Refused connection from %1").arg(connectionName(*pendingConnection));
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
//    if(tcpSocket==nullptr){
//        qDebug()<<"No connections";
//    } else {
//        qDebug() << QString("Current connection: %1").arg(connectionName());
//    }
}


bool TcpServer::addEvent(const char *event, int len)
{
//    qDebug() << QTime::currentTime().toString("hh:mm:ss.zzz") << QString("addEvent: %1 bytes").arg(len);
    if(dropOnFree && tcpSocket==nullptr){
        numberBytesDropped += len;
        return true;
    }

    Q_ASSERT(len < TCP_SERVER_BUFFER_SIZE); //todo complex partial writing
    int freeWriteSize = getFreeSize();

    while(freeWriteSize<len){
        //wait client read
        if(!bufFullSuppressTimer.isValid() || bufFullSuppressTimer.elapsed()>6000) {
            qWarning().noquote() << logPrefix + "Buffer is full, waiting client";
        }
        bufFullSuppressTimer.restart();
        writerMutex.lock();
        bool v = writerWC.wait(&writerMutex, 5000);
        auto freeWriteSizeNew = getFreeSize();
        if(!v) {
            qInfo() << __PRETTY_FUNCTION__
                    << len
                    << freeWriteSize
                    << freeWriteSizeNew;
        }
        freeWriteSize = getFreeSize();
        writerMutex.unlock();
    }
    if(requestFinish) {
        requestFinish = false;
        disconnect(tcpSocket, &QTcpSocket::bytesWritten,
                   this, &TcpServer::checkPendingData);
    }
//    if(freeSpace.tryAcquire(len)){
//        qWarning()<<"Buffer is full, waiting client";
//        freeSpace.acquire(len);
//    }

    int writtenSize=0;
    do {
        int tailSize = getContiniousFreeSize();
        tailSize = std::min(tailSize, len-writtenSize);
        Q_ASSERT(tailSize > 0);
        memcpy(data.data()+writePos, event+writtenSize, tailSize);
        writtenSize+=tailSize;
        writePos = (writePos+tailSize) % TCP_SERVER_BUFFER_SIZE;
    } while(len > writtenSize);
    return true;
}

void TcpServer::requestWorkCompleted()
{
    requestFinish = true;
    if(tcpSocket)
        connect(tcpSocket, &QTcpSocket::bytesWritten,
                this, &TcpServer::checkPendingData);
    checkPendingData();
}

int TcpServer::getFreeSize() const
{
    return (readPos+TCP_SERVER_BUFFER_SIZE-writePos-1)%TCP_SERVER_BUFFER_SIZE;
}

int TcpServer::getDataSize() const
{
    return (writePos+TCP_SERVER_BUFFER_SIZE-readPos)%TCP_SERVER_BUFFER_SIZE;
}

int TcpServer::getContiniousFreeSize() const
{
    return getContiniousSize(true);
}

int TcpServer::getContiniousDataSize() const
{
    return getContiniousSize(false);
}

int TcpServer::getContiniousSize(bool freeSize) const
{
    int size = freeSize ? getFreeSize() : getDataSize();
    if(size == 0)
        return 0;

    int tailSize = TCP_SERVER_BUFFER_SIZE-(freeSize ? writePos : readPos);
    if(tailSize <= 0)
        qCritical(lc_mstream_tcp).noquote()
                << QString("[%1] %2 tailSize<0: tailSize=%3; readPos=%4;"
                           " writePos=%5; TCP_SERVER_BUFFER_SIZE=%6")
                   .arg(logPrefix).arg(freeSize ? "free" : "data")
                   .arg(tailSize)
                   .arg(readPos).arg(writePos)
                   .arg(TCP_SERVER_BUFFER_SIZE);

    return std::min(size, tailSize);
}
