#include "PNPMessage.h"

#include <QStringList>
#include <QDomDocument>
#include <QDomNodeList>

#include "base/DaqModule.h"
#include "base/ProgramIndex.h"

#include "PNPServerUtil.h"
#include "PnpXmlProtocol.h"

namespace {
const QMap<ProgramInterfaceType, QString> ProgramInterfaceTypeNameMap
    {
        {ProgramInterfaceUnknown, "unknown"},
        {ProgramInterfaceDataFlow, "data flow"},
        {ProgramInterfaceMStream, "MStream"},
        {ProgramInterfaceRemoteControl, "RemoteControl"},
        {ProgramInterfaceMonitorData, "Monitor output data flow"},
        { ProgramInterfaceConfigTransfer, "Configuration transfer"}
    };
} // namespace

QString PNPMessage::toString() const
{
    QString res;
    switch (type) {
    case PNPMessage_Type_Search: {
        res = QString("Discover_request for: %1").arg(getSearchString(searchTargets));
        break;
    }
    case PNPMessage_Type_Discover:
        res = QString("Program announcement: clientIndex='%1' uuid='%2'")
                .arg(progDesc.getName())
                .arg(progDesc.uuid.toString());
        for (auto it : progDesc.interfaces) {
            QStringList peerStr;
            for (auto peer : it.peers)
                peerStr << peer.toString();
            res.append(QString("\ninterface %1 %5 at %2:%3 %4")
                       .arg(it.type).arg(it.host.toString()).arg(it.port)
                       .arg(it.isFree?"free":QString("busy (%1)").arg(peerStr.join(";")))
                       .arg(it.enabled?"enabled":"disabled"));
        }
        for (auto opKey : progDesc.options.keys()) {
            res.append(QString("\n[%1]: %2").arg(opKey).arg(progDesc.options[opKey]));
        }
        break;
    case PNPMessage_Type_Close:
        res = QString("Program closing: clientIndex='%1' uuid='%2'")
                .arg(progDesc.getName())
                .arg(progDesc.uuid.toString());
        break;
    default:
        break;
    }
    return res;
}

ProgramDescription::ProgramDescription()
    : uuid(QUuid::createUuid())
    , hostName(QHostInfo::localHostName())
{
    discoverTimer.start();
}

bool ProgramDescription::isOutOfDate(const QElapsedTimer &other) const
{
    return !discoverTimer.isValid() || discoverTimer.msecsTo(other) > 1000;
}

ProgramInterface& ProgramDescription::getProgramInterface(ProgramInterfaceType type, bool *ok, int id)
{
    for(int i=0; i<interfaces.size(); ++i) {
        if(id>=0 && interfaces[i].id!=id)
            continue;
        if(interfaces[i].type == type) {
            if(ok)
                *ok = true;
            return interfaces[i];
        }
    }
    if(ok)
        *ok = false;
    return dummyInterface;
}

const ProgramInterface& ProgramDescription::getProgramInterface(ProgramInterfaceType type, bool *ok) const
{
    for(const ProgramInterface &i : interfaces) {
        if(i.type == type) {
            if(ok)
                *ok = true;
            return i;
        }
    }
    if(ok)
        *ok = false;
    return dummyInterface;
}

bool ProgramDescription::containsProgramInterface(ProgramInterfaceType type, int id)
{
    for (const ProgramInterface &i : interfaces) {
        if(i.type == type && i.id ==id){
            return true;
        }
    }
    return false;
}

bool ProgramDescription::updateProgramInterface(ProgramInterfaceType type, int id, int port, bool isListening, bool isFree, const QHostAddress &peerHost, int peerPort)
{
    bool hasOld;
    ProgramInterface oldInterface = getProgramInterface(type, &hasOld, id);

    if(!hasOld) {
        ProgramInterface new_interface;
        new_interface.setType(type);
        new_interface.id = id;
        interfaces.push_back(new_interface);
    }
    ProgramInterface &interface = getProgramInterface(type, nullptr, id);
    interface.port = port;
    interface.enabled = isListening;
    interface.isFree = isFree;

    if(isFree) {
        interface.peers.clear();
    } else {
        interface.setPeer(peerHost, peerPort);
    }

    return !hasOld || oldInterface!= interface;
}

QString ProgramDescription::getAnnounceStr() const
{
    return PnpXml::encodeDiscover(*this);
}

QString ProgramDescription::toString() const
{
    QStringList s;
    s << QString("Uuid:\t%1").arg(uuid.toString());
    s << QString("Seq:\t%1").arg(QString::number(seq));
    s << QString("Type:\t%1").arg(getDaqModule().toString());
    s << QString("Index:\t%1").arg(getProgramIndex().toString());
    if(!getParent().getProgramType().isEmpty())
        s << QString("Parent:\t%1").arg(getParent().toString());
    s << QString("host:\t%1").arg(host.toString());
    s << QString("hostName:\t%1").arg(hostName);
    s << QString("version hash:\t%1").arg(ver_hash);
    s << QString("version date:\t%1").arg(ver_date);
    if(interfaces.size()) {
        int count=1;
        for (const ProgramInterface &i : interfaces) {
            s << QString("Interface %1:%2").arg(count++)
                 .arg(i.enabled ? (i.isFree?" (free)":" (busy)") : " (disabled)");
            s << QString("\ttype:\t%1 #%2").arg(i.getTypeName()).arg(i.id);
            s << QString("\tport:\t%1").arg(i.port);
        }
    }
    if(options.size()) {
        s << QString("Options:");
        for (QString opKey : options.keys())
            s << QString("\t%1: %2").arg(opKey).arg(options[opKey]);
    }

    return s.join("\n");
}

void ProgramDescription::clear()
{
    options.clear();
    interfaces.clear();
}

quint32 ProgramDescription::getHostIp() const
{
    return host.toIPv4Address();
}

QString ProgramDescription::getHostName() const
{
    if(!hostName.isEmpty())
        return hostName;
    else
        return host.toString();
}

QString ProgramInterface::getTypeName() const
{
    return getTypeName(type);
}

QString ProgramInterface::getTypeName(ProgramInterfaceType type)
{
    return ProgramInterfaceTypeNameMap[type];
}

ProgramInterfaceType ProgramInterface::getTypeByName(QString str)
{
    return ProgramInterfaceTypeNameMap.key(str, ProgramInterfaceUnknown);
}

void ProgramInterface::setType(ProgramInterfaceType t)
{
    type = t;
}

void ProgramInterface::setPeer(const QHostAddress &h, quint32 p)
{
    setPeer({h,p});
}

void ProgramInterface::setPeer(const ProgramInterfacePeer &p)
{
    peers.clear();
    addPeer(p);
}

void ProgramInterface::addPeer(const QHostAddress &h, quint32 p) {
    addPeer({h,p});
}

void ProgramInterface::addPeer(const ProgramInterfacePeer &p)
{
    peers << p;
}
