#ifndef PNPMESSAGE_H
#define PNPMESSAGE_H

class QByteArray;
class QDomElement;
#include <QHostAddress>
#include <QHostInfo>
#include <QStringList>
#include <QMetaType>
#include <QUuid>
#include <QElapsedTimer>

#include "base/ClientIndex.h"
#include "base/DeviceIndex.h"

class DaqModule;
class ProgramIndex;

enum ProgramInterfaceType { ProgramInterfaceUnknown = 0x0, // unused
                            ProgramInterfaceRemoteControl = 0x1,
                            ProgramInterfaceMStream = 0x2,
                            ProgramInterfaceDataFlow = 0x3,
                            ProgramInterfaceMonitorData = 0x4,
                            ProgramInterfaceConfigTransfer = 0x5 // unused
                          };

struct ProgramInterfacePeer
{
    ProgramInterfacePeer() = default;
    ProgramInterfacePeer(const QHostAddress &h, quint32 p) : host(h), port(static_cast<quint16>(p))
    {}
    friend bool operator==(const ProgramInterfacePeer &, const ProgramInterfacePeer &) = default;

    friend uint qHash(const ProgramInterfacePeer &r, uint seed = 0)
    {
        return qHash(r.host) ^ qHash(r.port) ^ seed;
    }

    QString toString() const { return QString("%1:%2").arg(host.toString()).arg(port); }

    QHostAddress host;
    quint16 port = 0;
};
Q_DECLARE_METATYPE(ProgramInterfacePeer)

class ProgramInterface
{
public:
    ProgramInterface() = default;
    ProgramInterface(const QHostAddress &h, const quint32 p, const ProgramInterfaceType &t, bool free = true, int i = 0) :
        host(h), port(static_cast<quint16>(p)), type(t), enabled(true), isFree(free), id(i)
    {}

    friend bool operator==(const ProgramInterface &, const ProgramInterface &) = default;

    QString getTypeName() const;
    static QString getTypeName(ProgramInterfaceType type);
    static ProgramInterfaceType getTypeByName(QString str);
    void setType(ProgramInterfaceType t);
    bool isEmpty() const { return type == ProgramInterfaceUnknown; }

    void setPeer(const QHostAddress &h, quint32 p);
    void setPeer(const ProgramInterfacePeer &p);
    void addPeer(const QHostAddress &h, quint32 p);
    void addPeer(const ProgramInterfacePeer &p);

    QHostAddress host;
    quint16 port = 0;
    ProgramInterfaceType type = ProgramInterfaceUnknown;
    DeviceIndex devIndex;
    QString devTypeStr;
    QSet<ProgramInterfacePeer> peers;
    bool enabled = false;
    bool isFree = true;
    int id = 0;
};
Q_DECLARE_METATYPE(ProgramInterface)

struct ProgramDescription
{
    ProgramDescription();

    QUuid uuid;
    int ttl = -1; // rest of ttl of incoming package
    quint32 seq = 0;
    ClientIndex clientIndex;
    ClientIndex parentClientIndex;
    QHostAddress host;
    QString hostName;
    QVector<ProgramInterface> interfaces;
    ProgramInterface dummyInterface;
    QMap<QString, QString> options;
    QString ver_hash;
    QString ver_date;

    ProgramInterface &getProgramInterface(ProgramInterfaceType type, bool *ok=nullptr, int id=-1);
    const ProgramInterface &getProgramInterface(ProgramInterfaceType type, bool *ok=nullptr) const;
    bool containsProgramInterface(ProgramInterfaceType type, int id);
    bool updateProgramInterface(ProgramInterfaceType type, int id, int port, bool isListening, bool isFree, const QHostAddress &peerHost, int peerPort);

    bool isOutOfDate(const QElapsedTimer &other) const;
    QString getAnnounceStr() const;
    QString toString() const;
    void clear();
    quint32 getHostIp() const;
    QString getHostName() const;
    DaqModule getDaqModule() const { return clientIndex.getProgramType(); }
    ProgramIndex getProgramIndex() const { return clientIndex.getProgramIndex(); }
    ClientIndex getClientIndex() const { return clientIndex; }
    void setClientIndex(const ClientIndex &i) { clientIndex = i; }
    ClientIndex getParent() const { return parentClientIndex; }
    void setParent(const ClientIndex &i) { parentClientIndex = i; }
    QElapsedTimer getDiscoverTimer() const { return discoverTimer; }

    void setProgramIndex(const ProgramIndex &i) { clientIndex.program_index = i; }
    QString getName() const { return clientIndex.toString(); }
    bool isEmpty() const { return clientIndex.program_type.isEmpty()
                || clientIndex.program_index.isDefault(); }
    bool sameProgram(const ProgramDescription &descr) const {
        return clientIndex == descr.getClientIndex();
    }
    friend bool operator <(const ProgramDescription &pd1, const ProgramDescription &pd2){
        return pd1.getName()<pd2.getName() || (pd1.getName()==pd2.getName() && pd1.uuid<pd2.uuid);
    }
    friend bool operator==(const ProgramDescription &, const ProgramDescription &) = default;

private:
    QElapsedTimer discoverTimer;
};
Q_DECLARE_METATYPE(ProgramDescription)

class PnpInfoCache: public QHash<QUuid, ProgramDescription> {};
Q_DECLARE_METATYPE(PnpInfoCache)

enum PNPMessage_Type {
    PNPMessage_Type_Discover = 1,
    PNPMessage_Type_Close = 2,
    PNPMessage_Type_Search = 3
};

struct PNPMessage
{
    QString toString() const;

    PNPMessage_Type type;
    DaqModuleSet searchTargets; // list program types

    ProgramDescription progDesc;
};

#endif // PNPMESSAGE_H
