// QString s = "<!DOCTYPE pnp_message>"
//             "<discover_request/>"

// QString s = "<!DOCTYPE pnp_message>"
//             "<program name=\"pnp_search\">"
//             " <options>"
//             "  <option name=\"targets\" value=\"all\"/>"
//             " <options>"
//             "</program>";

//    QString s = "<!DOCTYPE pnp_message><program name=\"mstream\">"
//            "<options><option name=\"address\" value=\"10.93.120.171\"/><option name=\"port\" value=\"1222\"/></options>"
//            "<interfaces><interface type=\"data downflow\" host=\"0.0.0.0\" port=\"33511\"/></interfaces></program>";

#include "PNPServer.h"

#include <QCoreApplication>
#include <QDebug>
#include <QNetworkInterface>
#include <QTimer>
#include <QSettings>
#include <QUdpSocket>
#include <arpa/inet.h>
#include <cmath>
#ifdef QT_DEBUG
#include <QTime>
#endif

#include "base/Globals.h"
#include "multicast-listener/MultiCastListener.h"

#include "PNPServerUtil.h"
#include "PnpRawProtocol.h"
#include "PnpXmlProtocol.h"

namespace {
constexpr const int SEARCH_INTERVAL_MS = 5000;
constexpr const quint16 PNP_DISCOVER_PORT = 33304;
constexpr const auto PNP_DISCOVER_IP_ADDR = "239.192.1.2";
constexpr const auto k_pnp = "pnp";
constexpr const auto k_pnp_port = "port";
constexpr const auto k_pnp_sendXml = "sendXml";
constexpr const auto k_pnp_sendRaw = "sendRaw";

void refreshSendEnv(bool *sendType, const char *key)
{
    const auto &settings = Globals::getSettings();

    QString k = Globals::getEnvironment() + "/" + k_pnp + "_" + key;
    if (settings->contains(k))
        *sendType = settings->value(k).toBool();
}
} // namespace

PNPServer::PNPServer(QObject *parent)
    : QObject(parent)
    , searchTimer(new QTimer(this))
    , multiCastListener(new MultiCastListener(QHostAddress(PNP_DISCOVER_IP_ADDR), getPnpPort(), this))
    , multiCastSender(new QUdpSocket(this))
    , deferredSearchResponceTimer(new QTimer(this))
{
    setObjectName("PnP Server");
    searchTimer->setInterval(SEARCH_INTERVAL_MS);
    connect(searchTimer, &QTimer::timeout, this, &PNPServer::sendSearchRequest);
    connect(multiCastListener, &MultiCastListener::received, this, &PNPServer::decode);
    connect(multiCastListener, &MultiCastListener::multicastInterfacesUpdated,
            this, &PNPServer::sendSearchRequest);
    qRegisterMetaType<ProgramDescription>();
    if(!multiCastSender->bind())
        qFatal("Can't bind UDP socket for PNP multicast");
    u_char ttl=15;
    if(setsockopt(static_cast<int>(multiCastSender->socketDescriptor()),
                  IPPROTO_IP, IP_MULTICAST_TTL, &ttl,sizeof(ttl)) != 0) {
        qCritical("Can't set TTL to %d for PNP multicast", ttl);
    }
    connect(qApp, &QCoreApplication::aboutToQuit,
            this, &QObject::deleteLater);
    connect(deferredSearchResponceTimer, &QTimer::timeout, this, &PNPServer::deferredSearchResponce);
    deferredSearchResponceTimer->setSingleShot(true);
    deferredSearchResponceTimer->setInterval(500);

    refreshSendEnv(&sendXml, k_pnp_sendXml);
    refreshSendEnv(&sendRaw, k_pnp_sendRaw);
}

PNPServer::~PNPServer()
{
    if(searchEnabled)
        stopSearch();
    while(!progDesc.isEmpty()) {
        closeProgram(progDesc.first());
    }
    multiCastSender->flush();
}

bool PNPServer::decode(const QByteArray &data, const QHostAddress &host, int ttl)
{
    PNPMessage msg;
    bool ok = false;

    if(PnpRaw::canDecodeRaw(data)) {
        msg = PnpRaw::decodeRaw(data, host, &ok);
    } else {
        msg = PnpXml::decodeXml(data, host, &ok);
    }
    if(ok) {
        switch (msg.type) {
        case PNPMessage_Type_Search:
        {
            bool instantSend = !deferredSearchResponceTimer->isActive() &&
                (!lastSearchTime.isValid() || lastSearchTime.elapsed()>500);
            bool wasSent = false;
            for (auto it=progDesc.cbegin(); it!=progDesc.cend(); ++it) {
                if(msg.searchTargets.isEmpty() || msg.searchTargets.contains(it->getDaqModule())){
                    if(instantSend) {
                        sendProgDescr(*it);
                        wasSent = true;
                    } else
                        searchRequestSet << it->uuid;
                }
            }
            if(wasSent)
                lastSearchTime.restart();
            if(!searchRequestSet.isEmpty() && !deferredSearchResponceTimer->isActive())
                deferredSearchResponceTimer->start();
            break;
        }
        case PNPMessage_Type_Discover:
        case PNPMessage_Type_Close:
            if(progsToDebugPrint.remove(msg.progDesc.uuid)) {
                auto data2print = PnpRaw::canDecodeRaw(data) ? data.toHex().toUpper() : data;
                qInfo() << __PRETTY_FUNCTION__  << '\n'
                        << msg.progDesc.uuid.toString()
                        << msg.progDesc.clientIndex.toString()
                        << host.toString() << '\n'
                        << data2print;
            }
            if(searchEnabled) {
                if(searchTargets.isEmpty() || searchTargets.contains(msg.progDesc.getDaqModule())){
                    bool isMine=false;
                    for (const auto &pd : qAsConst(progDesc)) {
                        if(msg.progDesc.uuid == pd.uuid){
                            isMine = true;
                            break;
                        }
                    }
                    if(!isMine){
//                        print("==========");
//                        print(msg.toString());
                        if(msg.type==PNPMessage_Type_Discover) {
                            msg.progDesc.ttl = ttl;
                            emit programDiscovered(msg.progDesc, data.size());
                        } else
                            emit programClosed(msg.progDesc);
                    }
                }
            }
            break;
        default:;
        }
    } else {
        static QMap<quint32, QPair<QElapsedTimer, int> > hostWarnTimeout;
        auto ip = host.toIPv4Address();
        auto pair = hostWarnTimeout[ip];
        if(!pair.first.isValid() || pair.first.elapsed()>500*(pow(1.5,pair.second))) {
            if(!pair.first.isValid())
                pair.second = 1;
            else
                ++pair.second;
            pair.first.start();
            qWarning() << QString("Failed decode msg from host '%1':")
                              .arg(host.toString())
                       << data;
        }
    }
    return ok;
}

void PNPServer::sendSearchRequest()
{
    if(!searchEnabled)
        return;
    if(sendXml) {
        const QString searchStr = PnpXml::encodeSearch(searchTargets);
        sendMulticastMsg(searchStr);
    }
    if(sendRaw) {
        sendMulticastMsg(PnpRaw::encodeSearch(searchTargets));
    }
}

void PNPServer::deferredSearchResponce()
{
    auto cachedSet = searchRequestSet; // searchRequestSet will be modified by sendProgDescr(...)
    for(auto i : cachedSet) {
        for(auto pd : progDesc) {
            if(pd.uuid == i) {
                sendProgDescr(pd);
                break;
            }
        }
    }
    searchRequestSet.clear();
    lastSearchTime.restart();
}

void PNPServer::changeProgramDescription(const ProgramDescription &pd)
{
    progDesc.clear();
    if(pd.isEmpty())
        return;
    progDesc.append(pd);
    sendProgDescr(pd);
}

void PNPServer::updateProgramDescription(const ProgramDescription &pd)
{
    if(pd.isEmpty())
        return;

    bool found = false;
    for(int i=0; i<progDesc.size(); ++i) {
        ProgramDescription &oldPd = progDesc[i];
        if(oldPd.sameProgram(pd)){
            if(progDesc[i] == pd)
                return;
            progDesc[i] = pd;
            found = true;
            break;
        }
    }

    if(!found)
        progDesc.append(pd);

    sendProgDescr(pd);
}

void PNPServer::closeProgram(const ProgramDescription &pd)
{
    if(sendXml)
        sendMulticastMsg(PnpXml::encodeClose(pd));
    if(sendRaw)
        sendMulticastMsg(PnpRaw::encodePD(PNPMessage_Type_Close, pd));

    // Be aware of using direct pd, after first removing pd will be reseted
    ClientIndex pdCi = pd.getClientIndex();
    for(int i=0; i<progDesc.size();) {
        if(progDesc[i].getClientIndex() == pdCi){
            progDesc.removeAt(i);
        } else
            ++i;
    }
}

void PNPServer::addProgToDebugPrint(const QUuid &id)
{
    progsToDebugPrint << id;
}

void PNPServer::sendProgDescr(ProgramDescription pd)
{
    searchRequestSet.remove(pd.uuid);
    pd.seq = ++(seqMap[pd.uuid]);
    if(sendXml)
        sendMulticastMsg(PnpXml::encodeDiscover(pd));
    if(sendRaw)
        sendMulticastMsg(PnpRaw::encodePD(PNPMessage_Type_Discover, pd));
}

void PNPServer::sendMulticastMsg(const QString &msg)
{
    const QByteArray data = msg.toLatin1();
    sendMulticastMsg(data);
}

void PNPServer::sendMulticastMsg(const QByteArray &data)
{
    qint64 nwr = multiCastSender->writeDatagram(data, QHostAddress(PNP_DISCOVER_IP_ADDR), getPnpPort());
    if (nwr != data.size()) {
        qCritical() << "[PNPServer::sendMulticastMsg] "
                    << multiCastSender->errorString();
    }
}

void PNPServer::print(QString s) const
{
#if QT_VERSION < QT_VERSION_CHECK(5, 5, 0)
    qDebug()<<s;
#else
    qInfo()<<s;
#endif
}

quint16 PNPServer::getPnpPort()
{
    const auto &settings = Globals::getSettings();
    QString key = Globals::getEnvironment() + "/" + k_pnp + "_" + k_pnp_port;
    quint16 port = 0;
    if (settings->contains(key)) {
        port = settings->value(key).toInt();
    }

    return port ? port : PNP_DISCOVER_PORT;
}

void PNPServer::startSearch(DaqModuleSet targets, bool single)
{
    searchTargets = targets;
    searchEnabled = true;
    sendSearchRequest();
    if(single) {
//        print(QString("Pnp_search single request for targets:%1")
//              .arg(getSearchString(searchTargets)));
    } else {
        searchTimer->start();
        print(QString("Pnp_search started with %1ms interval, targets:%2")
                  .arg(SEARCH_INTERVAL_MS).arg(getSearchString(searchTargets))
              );
    }
}

void PNPServer::stopSearch()
{
    searchEnabled = false;
    searchTimer->stop();
    print("Pnp_search stopped");
}

QVector<QHostAddress> PNPServer::getLocalHostAddresses()
{
    QVector<QHostAddress> addresses;
    const auto &allInterfaces = QNetworkInterface::allInterfaces();
    for (const auto &interface : allInterfaces) {
        if ( !(interface.flags() & QNetworkInterface::IsUp) ) continue;
        if ( !(interface.flags() & QNetworkInterface::IsRunning) ) continue;
        if (  (interface.flags() & QNetworkInterface::IsLoopBack) ) continue;
        const auto &allAddresses = interface.allAddresses();
        for (const auto &addr : allAddresses) {
            if(addr.isNull() || addr.protocol() != QAbstractSocket::IPv4Protocol)
                continue;
            if(addr.isLoopback())
                continue;
            addresses.push_back(addr);
        }
    }
    return addresses;
}

void PNPServer::enableSearch()
{
    searchEnabled = true;
    sendSearchRequest();
    searchTimer->start();
    print(QString("Pnp_search started with %1ms interval, targets:%2")
              .arg(SEARCH_INTERVAL_MS).arg(getSearchString(searchTargets))
          );
}
