#include "PnpRawProtocol.h"

#include "PNPMessage.h"
#include <QDateTime>
#include <QDataStream>

namespace {
constexpr const quint32 CUR_PNP_RAW_VER = 1; // at this moment one for all types
constexpr const quint32 PnpRawSyncWord = 0x5F506E50; // _PnP
constexpr const quint16 k_pnp_uuid = 0x01; // 6 bytes
constexpr const quint16 k_pnp_seq = 0x02; // 4 bytes
constexpr const quint16 k_pnp_host = 0x03; // 6 bytes
constexpr const quint16 k_pnp_hostName = 0x04; // 6 bytes
constexpr const quint16 k_pnp_verHash = 0x05; // 6 bytes
constexpr const quint16 k_pnp_verDate = 0x06; // 6 bytes
constexpr const quint16 k_pnp_clIndex = 0x07; // 6 bytes
constexpr const quint16 k_pnp_parClIndex = 0x08; // 6 bytes
constexpr const quint16 k_pnp_options = 0x09; // 6 bytes
constexpr const quint16 k_pnp_interfaces = 0x0a; // 6 bytes

constexpr const quint16 k_pnp_ci_type = 0x01; // 3-15 bytes at this moment
constexpr const quint16 k_pnp_ci_domain = 0x02; // skiped, 6 bytes
constexpr const quint16 k_pnp_ci_index = 0x03; // up to 16 bytes at this moment
const static constexpr char k_pnp_opt_sep(0x1E); // RS 0x1E record separator

struct afi_pnp_main_hdr_t {
    uint32_t sync;
    uint8_t type;
    uint8_t ver;
    uint16_t len;
} __attribute__((packed));


typedef union __attribute__((packed)) PnpTypeLen {
    uint16_t all;
    struct
    {
        uint16_t type : 4; // LSB
        uint16_t len : 10; // MSB
    } bit;
} PnpTypeLen;

typedef union __attribute__((packed)) ProgramInterfaceMix {
    uint16_t all;
    struct
    {
        uint16_t unused : 5;
        uint16_t id : 4;
        uint16_t type : 4;
        uint16_t en : 1;
        uint16_t isFree : 1;
        uint16_t hostDiffer : 1;
    } bit;
} ProgramInterfaceMix;

// ========= Decode part =========
ClientIndex decodeRawClIndex(const QByteArray &data, bool *ok=nullptr)
{
    quint16 payloadSize = data.size();
    QString typeStr, domainStr, indexStr;
    QString *strPtr;

    PnpTypeLen ptl;
    for(size_t pos = 0; pos+2<=payloadSize;) {
        ptl.all = *(quint16*)(data.data()+pos);

        switch (ptl.bit.type) {
        case k_pnp_ci_type: strPtr = &typeStr; break;
        case k_pnp_ci_domain: strPtr = &domainStr; break;
        case k_pnp_ci_index: strPtr = &indexStr; break;
        default:
            *ok = false;
            return {};
        }
        *strPtr = QString::fromLatin1(data.data()+pos+2, ptl.bit.len);
        pos += 2+ptl.bit.len;
    }

    *ok = true;
    return ClientIndex(DaqModule::fromString(typeStr),
                       ProgramIndex(domainStr, indexStr));
}

QPair<DeviceIndex, QString> decodeRawDevIndex(const QByteArray &data, bool *ok)
{
    quint16 payloadSize = data.size();
    if(payloadSize < 9) {
        *ok = false;
        return {};
    }

    uint8_t id = *(const uint8_t*)data.data();
    quint64 ser = *(const quint64*)(data.data()+1);
    QString typeStr;
    if(payloadSize > 9)
        typeStr = QString::fromLatin1(data.mid(1+8));

    *ok = true;
    return {DeviceIndex(DeviceId(id), ser), typeStr};
}

ProgramInterface decodeRawInterface(const QByteArray &data, bool *ok)
{
    quint16 payloadSize = data.size();
    ProgramInterface i;
    if(payloadSize < 4) {
        *ok = false;
        return {};
    }
    i.port = *(quint16*)(data.data());
    int pos = 2;
    const ProgramInterfaceMix *mix = (const ProgramInterfaceMix*)(data.data()+pos);
    i.enabled = mix->bit.en;
    i.isFree = mix->bit.isFree;
    i.id = mix->bit.id;
    i.setType(static_cast<ProgramInterfaceType>(mix->bit.type));
    pos += 2;

    if(mix->bit.hostDiffer) {
        i.host.setAddress(*(const quint32*)(data.data()+pos));
        pos += 4;
    }

    if(i.type == ProgramInterfaceMStream) {
        uint8_t len = *(const uint8_t*)(data.data()+pos);
        auto p = decodeRawDevIndex(data.mid(pos+1, len), ok);
        if(!*ok)
            return {};
        i.devIndex = p.first;
        i.devTypeStr = p.second;
        pos += 1+len;
    }
    while(pos < payloadSize) {
        Q_ASSERT(payloadSize-pos>=6);
        if(payloadSize-pos<6) {
            *ok = false;
            return {};
        }
        ProgramInterfacePeer peer;
        peer.host.setAddress(*(const quint32*)(data.data()+pos));
        pos += 4;
        peer.port = *(const quint16*)(data.data()+pos);
        pos += 2;
        i.peers << peer;
    }

    *ok = true;
    return i;
}

QVector<ProgramInterface> decodeRawInterfaces(const QByteArray &data, bool *ok=nullptr)
{
    quint16 payloadSize = data.size();
    quint8 len;
    QVector<ProgramInterface> iVec;

    *ok = true;
    for(size_t pos = 0; pos+1<=payloadSize;) {
        len = *(quint8*)(data.data()+pos);
        iVec.push_back(decodeRawInterface(data.mid(pos+1, len), ok));
        if(!*ok)
            return {};
        pos += 1+len;
    }

    *ok = true;
    return iVec;
}

QMap<QString, QString> decodeRawOptions(const QByteArray &data, bool *ok=nullptr)
{
    QMap<QString, QString> map;
    int pos = 0;
    int sepPos = data.indexOf(k_pnp_opt_sep, pos);
    QString key;
    while(sepPos != -1) {
        key = data.mid(pos, sepPos-pos);

        pos = sepPos + 1;
        sepPos = data.indexOf(k_pnp_opt_sep, pos);

        map[key] = data.mid(pos, sepPos-pos);
        if(sepPos == -1)
            break;
        pos = sepPos + 1;
        sepPos = data.indexOf(k_pnp_opt_sep, pos);
    }
    *ok = true;
    return map;
}

PNPMessage decodeRawSearch(const QByteArray &data, const QHostAddress &host, bool *ok)
{
    Q_ASSERT(ok);
    PNPMessage pnp;
    auto &searchTargets = pnp.searchTargets;

    uint fullSize = data.size();
    auto rawHdr  = reinterpret_cast<const afi_pnp_main_hdr_t*>(data.data());
    pnp.type = static_cast<PNPMessage_Type>(rawHdr->type);
    quint16 payloadSize = rawHdr->len;
    if(fullSize < sizeof(afi_pnp_main_hdr_t)+payloadSize) {
        *ok = false;
        return {};
    }

    // pos - is current position of payload relative to payload beginig
    for(size_t pos = 0; pos+1<=payloadSize;) {
        quint8 len = *(quint8*)(data.data()+pos);
        auto str = QString::fromLatin1(data.data()+pos+1, len);
        auto dm = DaqModule::fromString(str);
        if(dm.isEmpty()) {
            static bool singleWarn = true;
            if(singleWarn) {
                singleWarn = false;
                qWarning() << QString("Got unknown/empty DaqModule from host:")
                                  .arg(host.toString());
            }
        } else {
            searchTargets << dm;
        }
        pos += sizeof(len)+len;
    }

    *ok = true;
    return pnp;
}

PNPMessage decodeRawDiscover(const QByteArray &data, const QHostAddress &host, bool *ok)
{
    PNPMessage pnp;
    ProgramDescription &progDesc = pnp.progDesc;
    bool local_ok=true;

    uint fullSize = data.size();
    auto rawHdr  = reinterpret_cast<const afi_pnp_main_hdr_t*>(data.data());
    quint16 payloadSize = rawHdr->len;
    pnp.type = static_cast<PNPMessage_Type>(rawHdr->type);

    if(fullSize < sizeof(afi_pnp_main_hdr_t)+payloadSize) {
        *ok = false;
        return {};
    }
    PnpTypeLen ptl;
    progDesc.host = host;

    // blockPos - is current position of payload
    size_t blockPos = sizeof(afi_pnp_main_hdr_t);
    while(blockPos+sizeof(PnpTypeLen)<=sizeof(afi_pnp_main_hdr_t)+payloadSize) {
        ptl.all = *(quint16*)(data.data()+blockPos);
        const auto &len = ptl.bit.len;
        if(blockPos+sizeof(ptl)+len > sizeof(afi_pnp_main_hdr_t)+payloadSize) {
            *ok = false;
            return {};
        }
        blockPos += sizeof(PnpTypeLen);

        switch (ptl.bit.type) {
        case k_pnp_uuid:
            if(len != 16) {
                *ok = false;
                return {};
            }
            progDesc.uuid = QUuid::fromRfc4122(data.mid(blockPos,len));
            break;
        case k_pnp_seq:
            if(len != 4) {
                *ok = false;
                return {};
            }
            progDesc.seq= *(quint32*)(data.data()+blockPos);
            break;
        case k_pnp_host:
            if(len != sizeof(quint32)) {
                *ok = false;
                return {};
            }
            progDesc.host.setAddress(*(quint32*)(data.data()+blockPos));
            break;
        case k_pnp_hostName:
            progDesc.hostName = QString::fromLatin1(data.data()+blockPos, len);
            break;
        case k_pnp_verHash:
            progDesc.ver_hash = QString::fromLatin1(data.data()+blockPos, len);
            break;
        case k_pnp_verDate:
            progDesc.ver_date =
                QDateTime::fromSecsSinceEpoch(*(quint32*)(data.data()+blockPos))
                    .toString(Qt::ISODate);
            break;
        case k_pnp_clIndex:
        {
            const auto &ci = decodeRawClIndex(data.mid(blockPos,len), &local_ok);
            progDesc.setClientIndex(ci);
            break;
        }
        case k_pnp_parClIndex:
        {
            const auto &ci = decodeRawClIndex(data.mid(blockPos,len), &local_ok);
            progDesc.setParent(ci);
            break;
        }
        case k_pnp_options:
            progDesc.options = decodeRawOptions(data.mid(blockPos,len),
                                                &local_ok);
            break;
        case k_pnp_interfaces:
            progDesc.interfaces = decodeRawInterfaces(data.mid(blockPos,len),
                                                      &local_ok);
            break;
        default:
            // unknown PTL type
            *ok = false;
            return {};
        }
        if(!local_ok) {
            *ok = false;
            return {};
        }
        blockPos += len;
    }

    if(progDesc.uuid.isNull() || progDesc.clientIndex.isEmpty()) {
        *ok = false;
        return {};
    }
    for(ProgramInterface &i : progDesc.interfaces) {
        if(i.host.isNull())
            i.host = progDesc.host;
    }

    *ok = true;
    return pnp;
}

// ========= Encode part =========

void encodeClientIndex(QDataStream &out, quint16 key,
                               const ClientIndex &clientIndex)
{
    if(clientIndex.isEmpty())
        return;
    QByteArray d;
    QDataStream stream(&d, QIODevice::WriteOnly);
    stream.setByteOrder(QDataStream::LittleEndian);
    PnpTypeLen ptl;
    const auto &type = clientIndex.getProgramType();
    if(!type.isEmpty()) {
        const auto &d = type.toString().toLatin1();
        ptl.bit.type = k_pnp_ci_type;
        ptl.bit.len = d.size();
        stream << ptl.all;
        stream.writeRawData(d.data(), d.size());
    }

    const auto &index = clientIndex.getProgramIndex();
    if(!index.isDomainDefault()) {
        const auto &d = index.getDomain().toLatin1();
        ptl.bit.type = k_pnp_ci_domain;
        ptl.bit.len = d.size();
        stream << ptl.all;
        stream.writeRawData(d.data(), d.size());
    }
    if(!index.isIndexDefault()) {
        const auto &d = index.getIndex().toLatin1();
        ptl.bit.type = k_pnp_ci_index;
        ptl.bit.len = d.size();
        stream << ptl.all;
        stream.writeRawData(d.data(), d.size());
    }

    ptl.bit.type = key;
    ptl.bit.len = d.size();
    out << ptl.all;
    out.writeRawData(d.data(), d.size());
}

void encodeInterfaces(QDataStream &out,
                      const QVector<ProgramInterface> &interfaces,
                      const QHostAddress &host)
{
    if(interfaces.isEmpty())
        return;

    QByteArray dataIntAll;
    QDataStream streamAll(&dataIntAll, QIODevice::WriteOnly);
    streamAll.setByteOrder(QDataStream::LittleEndian);
    for(auto it=interfaces.cbegin(); it!=interfaces.cend(); ++it) {
        QByteArray dataIntOne;
        QDataStream streamOne(&dataIntOne, QIODevice::WriteOnly);
        streamOne.setByteOrder(QDataStream::LittleEndian);
        streamOne << it->port;
        ProgramInterfaceMix mix;
        mix.all = 0;
        mix.bit.id = it->id;
        mix.bit.type = it->type;
        mix.bit.en = it->enabled;
        mix.bit.isFree = it->isFree;
        mix.bit.hostDiffer = (!it->host.isNull()) && (it->host != host);
        streamOne << mix.all;

        if(mix.bit.hostDiffer) {
            streamOne << it->host.toIPv4Address();
        }
        if(mix.bit.type == ProgramInterfaceMStream) {
            QByteArray dataDevIndex;
            QDataStream streamDevIndex(&dataDevIndex, QIODevice::WriteOnly);
            streamDevIndex.setByteOrder(QDataStream::LittleEndian);

            streamDevIndex << it->devIndex.getDevId(); // uint8_t
            streamDevIndex << it->devIndex.getSerial();
            dataDevIndex.append(it->devTypeStr.toLatin1());

            streamOne << quint8(dataDevIndex.size());
            streamOne.writeRawData(dataDevIndex.data(), dataDevIndex.size());
            // streamOne << dataDevIndex;
        }
        for (auto itPeer : it->peers) {
            streamOne << itPeer.host.toIPv4Address();
            streamOne << itPeer.port;
        }

        streamAll << quint8(dataIntOne.size());
        // streamAll << dataIntOne;
        streamAll.writeRawData(dataIntOne.data(), dataIntOne.size());
    }

    PnpTypeLen ptl;
    ptl.bit.type = k_pnp_interfaces;
    ptl.bit.len = dataIntAll.size();

    out << ptl.all;
    out.writeRawData(dataIntAll.data(), dataIntAll.size());
}

void encodeOptions(QDataStream &out,
                      const QMap<QString, QString> &options)
{
    if(options.isEmpty())
        return;

    QByteArray d;
    for(auto it=options.cbegin(); it!=options.cend(); ++it) {
        if(!d.isEmpty())
            d.append(k_pnp_opt_sep);
        d.append(it.key().toLatin1());
        d.append(k_pnp_opt_sep);
        d.append(it->toLatin1());
    }

    PnpTypeLen ptl{0};
    ptl.bit.type = k_pnp_options;
    ptl.bit.len = d.size();
    out << ptl.all;
    out.writeRawData(d.data(), d.size());
}
} // namespace

bool PnpRaw::canDecodeRaw(const QByteArray &data)
{
    uint fullSize = data.size();
    if(fullSize < sizeof(afi_pnp_main_hdr_t))
        return false;
    auto rawHdr  = reinterpret_cast<const afi_pnp_main_hdr_t*>(data.data());
    if(rawHdr->sync == PnpRawSyncWord)
        return true;
    return false;
}

PNPMessage PnpRaw::decodeRaw(const QByteArray &data, const QHostAddress &host, bool *ok)
{
    auto rawHdr  = reinterpret_cast<const afi_pnp_main_hdr_t*>(data.data());
    switch (rawHdr->type) {
    case PNPMessage_Type_Discover:
        return decodeRawDiscover(data, host, ok);
    case PNPMessage_Type_Close:
        return decodeRawDiscover(data, host, ok);
    case PNPMessage_Type_Search:
        return decodeRawSearch(data, host, ok);
    default:
    {
        static QSet<quint16> unknownTypes;
        if(unknownTypes.size() < 10 && !unknownTypes.contains(rawHdr->type)) {
            unknownTypes << rawHdr->type;
            qWarning() << QString("Got unknown msgType:%1 from host:%2")
                              .arg(rawHdr->type).arg(host.toString());
        }
        break;
    }
    }

    *ok = false;
    return {};
}

QByteArray PnpRaw::encodePD(const quint8 &type, const ProgramDescription &pd)
{
    QByteArray data;
    QDataStream stream(&data, QIODevice::WriteOnly);
    stream.setByteOrder(QDataStream::LittleEndian);
    PnpTypeLen ptl;
    ptl.all = 0;
    if(!pd.uuid.isNull()) {
        ptl.bit.type = k_pnp_uuid;
        ptl.bit.len = 16;
        stream << ptl.all;
        const auto &uuidData = pd.uuid.toRfc4122();
        stream.writeRawData(uuidData.data(), uuidData.size());
    }

    ptl.all = 0;
    ptl.bit.type = k_pnp_seq;
    ptl.bit.len = 4;
    stream << ptl.all;
    stream << pd.seq;

    if(!pd.host.isNull()) {
        ptl.bit.type = k_pnp_host;
        ptl.bit.len = 4;
        stream << ptl.all;
        stream << pd.host.toIPv4Address();
    }

    if(!pd.hostName.isEmpty()) {
        const auto &d = pd.hostName.toLatin1();
        ptl.bit.type = k_pnp_hostName;
        ptl.bit.len = d.size();
        stream << ptl.all;
        stream.writeRawData(d.data(), d.size());
    }
    if(!pd.ver_hash.isEmpty()) {
        const auto &d = pd.ver_hash.toLatin1();
        ptl.bit.type = k_pnp_verHash;
        ptl.bit.len = d.size();
        stream << ptl.all;
        stream.writeRawData(d.data(), d.size());
    }
    if(!pd.ver_date.isEmpty()) {
        const auto &d = QDateTime::fromString(pd.ver_date, Qt::ISODate);
        if(!d.isNull()) {
            quint32 sec = d.toSecsSinceEpoch();
            ptl.bit.type = k_pnp_verDate;
            ptl.bit.len = 4;
            stream << ptl.all;
            stream << sec;
        } else {
            qWarning() << __PRETTY_FUNCTION__ << "errrrrrrrr";
        }
    }
    encodeClientIndex(stream, k_pnp_clIndex, pd.getClientIndex());
    encodeClientIndex(stream, k_pnp_parClIndex, pd.getParent());

    encodeOptions(stream, pd.options);
    encodeInterfaces(stream, pd.interfaces, pd.host);

    afi_pnp_main_hdr_t hdr;
    hdr.sync = PnpRawSyncWord;
    hdr.type = type;
    hdr.ver = CUR_PNP_RAW_VER;
    hdr.len = data.size();
    data.prepend(reinterpret_cast<char*>(&hdr), sizeof(hdr));

    return data;
}

QByteArray PnpRaw::encodeSearch(const DaqModuleSet &searchTargets)
{
    QByteArray data;
    QDataStream stream(&data, QIODevice::WriteOnly);
    stream.setByteOrder(QDataStream::LittleEndian);
    for(auto it=searchTargets.cbegin(); it!=searchTargets.cend(); ++it) {
        const auto &d = it->toString().toLatin1();
        stream << quint8(d.size());
        stream.writeRawData(d.data(), d.size());
    }

    afi_pnp_main_hdr_t hdr;
    hdr.sync = PnpRawSyncWord;
    hdr.type = PNPMessage_Type_Search;
    hdr.ver = CUR_PNP_RAW_VER;
    hdr.len = data.size();
    data.prepend(reinterpret_cast<char*>(&hdr), sizeof(hdr));

    return data;
}
