#include "PnpXmlProtocol.h"

#include <QDebug>
#include <QDomDocument>
#include <QDomNodeList>

#include "base/DaqModule.h"

#include "PNPMessage.h"
#include "PnpXmlProtocolTags.h"

using namespace PnpXml;

namespace {
QMap<QString, QString> decodeOptions(const QDomElement &domEl, const QString &hostName, bool *ok)
{
    QMap<QString, QString> options;

    if(domEl.tagName() != TAG_OPTIONS_SECTION) {
        *ok = false;
        return {};
    }
    for(int i=0; i < domEl.childNodes().size(); ++i) {
        const QDomElement optionEl = domEl.childNodes().at(i).toElement();
        if (optionEl.tagName() != TAG_OPTION_SECTION) {
            qWarning()<<QString("Unexpected tag name:%1 from:%2")
            .arg(optionEl.tagName()).arg(hostName);
            *ok = false;
            return {};
        }
        if (!optionEl.hasAttribute(TAG_OPTION_NAME)) {
            qWarning()<<QString("There is no attribute :%1 from:%2")
            .arg(TAG_OPTION_NAME).arg(hostName);
            *ok = false;
            return {};
        }
        if (!optionEl.hasAttribute(TAG_OPTION_VALUE)) {
            qWarning()<<QString("There is no attribute :%1 from:%2")
            .arg(TAG_OPTION_VALUE).arg(hostName);
            *ok = false;
            return {};
        }
        //        if (optionEl.tagName() != TAG_OPTION_SECTION
        //                || !optionEl.hasAttribute(TAG_OPTION_NAME)
        //                || !optionEl.hasAttribute(TAG_OPTION_VALUE) )
        //            return false;
        QString name = optionEl.attribute(TAG_OPTION_NAME);
        if(!name.isEmpty())
            options[name] = optionEl.attribute(TAG_OPTION_VALUE);
    }
    *ok = true;
    return options;
}

void decodeInterfaces(ProgramDescription *progDesc, const QDomElement &domEl, const QHostAddress &host, bool *ok)
{
    if(domEl.tagName() != TAG_INTERFACES_SECTION) {
        *ok = false;
        return;
    }
    for(int i=0; i < domEl.childNodes().size(); ++i) {
        const QDomElement interfaceEl = domEl.childNodes().at(i).toElement();
        if (interfaceEl.tagName() != TAG_INTERFACE_SECTION){
            qWarning() << QString("Unknow tag (%1) in interface section from %2")
            .arg(interfaceEl.tagName()).arg(progDesc->getName());
            *ok = false;
            return;
        }
        if(!interfaceEl.hasAttribute(TAG_INTERFACE_TYPE) ){
            qWarning() << QString("There is no attribute  (%1) in interface section from %2")
            .arg(TAG_INTERFACE_TYPE).arg(progDesc->getName());
            *ok = false;
            return;
        }
        if(!interfaceEl.hasAttribute(TAG_INTERFACE_PORT) ){
            qWarning() << QString("There is no attribute (%1) in interface section from %2")
            .arg(TAG_INTERFACE_PORT).arg(progDesc->getName());
            *ok = false;
            return;
        }
        if(!interfaceEl.hasAttribute(TAG_INTERFACE_ISFREE) ){
            qWarning() << QString("There is no attribute (%1) in interface section from '%2'")
            .arg(TAG_INTERFACE_ISFREE).arg(progDesc->getName());
            *ok = false;
            return;
        }
        Q_ASSERT(!host.isNull());
        ProgramInterface interface;
        if(interfaceEl.hasAttribute(TAG_INTERFACE_HOST))
            interface.host = QHostAddress(interfaceEl.attribute(TAG_INTERFACE_HOST));
        else
            interface.host = host;
        auto typeStr = interfaceEl.attribute(TAG_INTERFACE_TYPE);
        interface.setType(ProgramInterface::getTypeByName(typeStr));
        if(interface.type == ProgramInterfaceMStream) {
            if(!interfaceEl.hasAttribute(TAG_INTERFACE_DEV_INDEX)) {
                qWarning() << QString("There is no attribute (%1) in interface section from '%2'")
                .arg(TAG_INTERFACE_DEV_INDEX).arg(progDesc->getName());
                *ok = false;
                return;
            }

            interface.devIndex = DeviceIndex::fromString(interfaceEl.attribute(TAG_INTERFACE_DEV_INDEX));
            interface.devTypeStr = interfaceEl.attribute(TAG_INTERFACE_DEV_TYPE_STR);
            if(interface.devIndex.isNull()) {
                qWarning() << QString("Got empty devIndex from attribute (%1: %2) in interface section from '%3'")
                                  .arg(TAG_INTERFACE_DEV_INDEX)
                                  .arg(interfaceEl.attribute(TAG_INTERFACE_DEV_INDEX))
                                  .arg(progDesc->getName());
                *ok = false;
                return;
            }
        }
        bool ok;
        bool enabled = interfaceEl.attribute(TAG_INTERFACE_ENABLED).toInt(&ok)==1;
        if(ok)
            interface.enabled = enabled;
        bool isFree = interfaceEl.attribute(TAG_INTERFACE_ISFREE).toInt(&ok)==1;
        if(ok)
            interface.isFree = isFree;
        int id = interfaceEl.attribute(TAG_INTERFACE_ID).toInt(&ok);
        if(ok)
            interface.id = id;
        interface.port = interfaceEl.attribute(TAG_INTERFACE_PORT).toInt();
        for(int peerId=0; peerId < interfaceEl.childNodes().size(); ++peerId) {
            const QDomElement peerEl = interfaceEl.childNodes().at(peerId).toElement();
            if(peerEl.tagName() != TAG_PEER_SECTION) {
                qWarning() << QString("Unknow tag (%1) in peer section from '%2'")
                .arg(peerEl.tagName()).arg(progDesc->getName());
                continue;
            }
            ProgramInterfacePeer itPeer;
            if(peerEl.hasAttribute(TAG_PEER_HOST)) {
                QString conHostStr = peerEl.attribute(TAG_PEER_HOST);
                if(!conHostStr.isEmpty()) {
                    QHostAddress conHost(conHostStr);
                    if(!conHost.isNull()) {
                        itPeer.host = conHost;
                    }
                }
            }
            if(peerEl.hasAttribute(TAG_PEER_PORT)) {
                itPeer.port = peerEl.attribute(TAG_PEER_PORT).toUInt(&ok);
                if(!ok && itPeer.port==0) {
                    qWarning() << QString("Bad attribute  (%1) in peer section from '%2'")
                    .arg(peerEl.attribute(TAG_PEER_PORT)).arg(progDesc->getName());
                    continue;
                }
            } else {
                qWarning() << QString("There is no attribute (%1) in peer section from '%2'")
                .arg(TAG_PEER_PORT).arg(progDesc->getName());
                continue;
            }
            interface.peers << itPeer;
        }

        if(interface.port==0 && interface.enabled){
            qWarning() << QString("Empty port at enabled interface from '%1'")
                              .arg(progDesc->getName());
            ok = false;
            return ;
        }

        progDesc->interfaces.push_back(interface);
    }

    *ok = true;
}

} // namespace

PNPMessage PnpXml::decodeXml(const QByteArray &data, const QHostAddress &host, bool *ok)
{
    PNPMessage pnp;
    auto &progDesc = pnp.progDesc;
    QDomDocument mDocument;
    if(!mDocument.setContent(data) || mDocument.doctype().name() != DOC_TYPE) {
        *ok = false;
        return {};
    }
    bool local_ok=true;
    QDomElement  rootElement;
    rootElement = mDocument.documentElement();

    if(rootElement.tagName() == TAG_PROGRAM_ANNOUNCE ||
        rootElement.tagName() == TAG_PROGRAM_CLOSE) {
        QString daqTypeStr, programIndexStr;
        if(rootElement.tagName() == TAG_PROGRAM_ANNOUNCE)
            pnp.type = PNPMessage_Type_Discover;
        else
            pnp.type = PNPMessage_Type_Close;
        if(rootElement.hasAttribute(TAG_PROGRAM_HOST_NAME)) {
            progDesc.hostName = rootElement.attribute(TAG_PROGRAM_HOST_NAME);
        } else {
            progDesc.hostName = host.toString();
            qWarning() << QString("Program announce from %1 has no attribute %2")
                              .arg(progDesc.hostName)
                              .arg(TAG_PROGRAM_HOST_NAME);
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_TYPE)) {
            daqTypeStr = rootElement.attribute(TAG_PROGRAM_TYPE);
        } else {
            qWarning() << QString("Program announce from %1 has no attribute %2")
            .arg(progDesc.hostName)
                .arg(TAG_PROGRAM_TYPE);
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_INDEX)) {
            programIndexStr = rootElement.attribute(TAG_PROGRAM_INDEX);
        } else {
            qWarning() << QString("Program announce from %1 has no attribute %2")
            .arg(progDesc.hostName)
                .arg(TAG_PROGRAM_INDEX);
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_PARENT)) {
            progDesc.parentClientIndex = ClientIndex::fromString(
                rootElement.attribute(TAG_PROGRAM_PARENT));
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_UUID)) {
            progDesc.uuid = QUuid(rootElement.attribute(TAG_PROGRAM_UUID));
        } else {
            qWarning() << QString("Program announce from '%1' has no attribute %2")
            .arg(progDesc.getName())
                .arg(TAG_PROGRAM_UUID);
            local_ok = false;
        }
        if(rootElement.hasAttribute(TAG_PROGRAM_SEQ)) {
            bool seqOk;
            progDesc.seq = rootElement.attribute(TAG_PROGRAM_SEQ).toUInt(&seqOk);
            if(!seqOk){
                progDesc.seq = 0;
                //                qWarning() << QString("Program announce from '%1' has wrong attribute %2: %3")
                //                              .arg(progDesc.getName())
                //                              .arg(TAG_PROGRAM_SEQ)
                //                              .arg(rootElement.attribute(TAG_PROGRAM_SEQ));
                //                local_ok = false;
            }
        } else {
            //            qWarning() << QString("Program announce from '%1' has no attribute %2")
            //                          .arg(progDesc.getName())
            //                          .arg(TAG_PROGRAM_SEQ);
            //            local_ok = false;
        }
        if(rootElement.hasAttribute(TAG_VER_HASH)) {
            progDesc.ver_hash = rootElement.attribute(TAG_VER_HASH);
        }
        if(rootElement.hasAttribute(TAG_VER_DATE)) {
            progDesc.ver_date = rootElement.attribute(TAG_VER_DATE);
        }

        progDesc.host = rootElement.hasAttribute(TAG_PROGRAM_HOST)
            ? QHostAddress(rootElement.attribute(TAG_PROGRAM_HOST))
            : host;

        for(int i=0; (i < rootElement.childNodes().size()) && local_ok; ++i) {
            const QDomElement domEl = rootElement.childNodes().at(i).toElement();
            if(domEl.tagName() == TAG_PROGRAM_TYPE){
                daqTypeStr = domEl.text();
            } else if(domEl.tagName() == TAG_PROGRAM_INDEX){
                programIndexStr = domEl.text();
            } else if(domEl.tagName() == TAG_PROGRAM_HOST_NAME){
                progDesc.hostName = domEl.text();
            } else if(domEl.tagName() == TAG_PROGRAM_UUID){
                progDesc.uuid = QUuid(domEl.text());
            } else if(domEl.tagName() == TAG_PROGRAM_SEQ){
                progDesc.seq = domEl.text().toUInt();
            } else if(domEl.tagName() == TAG_VER_HASH){
                progDesc.ver_hash = domEl.text();
            } else if(domEl.tagName() == TAG_VER_DATE){
                progDesc.ver_date = domEl.text();
            } else if(domEl.tagName() == TAG_OPTIONS_SECTION){
                progDesc.options = decodeOptions(domEl, progDesc.hostName, &local_ok);
            } else if(domEl.tagName() == TAG_INTERFACES_SECTION){
                decodeInterfaces(&progDesc, domEl, progDesc.host, &local_ok);
            }  else {
                qWarning() << QString("Unknown tag:%1 from: %2")
                .arg(domEl.tagName()).arg(host.toString());
                *ok = false;
                return {};
            }
        }
        const auto &daqModule = DaqModule::fromString(daqTypeStr);
        const auto &programIndex = ProgramIndex::fromString(programIndexStr);
        progDesc.setClientIndex({daqModule, programIndex});
        if(pnp.type == PNPMessage_Type_Discover)
            local_ok &= !progDesc.isEmpty();
    } else if(rootElement.tagName() == TAG_SEARCH) {
        pnp.type = PNPMessage_Type_Search;
        for(int i=0; (i < rootElement.childNodes().size()) && local_ok; ++i) {
            const QDomElement domEl = rootElement.childNodes().at(i).toElement();
            if(domEl.tagName() == TAG_SEARCH_TARGET) {
                const auto &searchTarg = DaqModule::fromString(domEl.text());
                if(searchTarg.isEmpty()) {
                    static bool showWarn = true;
                    if(showWarn) {
                        qWarning() << "Can't detect search target:" << domEl.text();
                        qWarning() << "Whole pnp msg is:\n" << data;
                        showWarn = false;
                    }
                } else {
                    pnp.searchTargets << searchTarg;
                }
            } else
                local_ok = false;
        }
    } else
        local_ok = false;
    if(ok)
        *ok = local_ok;
    return pnp;
}

QByteArray PnpXml::encodePD(const quint8 &type, const ProgramDescription &pd)
{
    return {};
}

QString PnpXml::encodeSearch(const DaqModuleSet &searchTargets)
{
    QDomDocument doc(DOC_TYPE);
    QDomElement root = doc.createElement(TAG_SEARCH);
    doc.appendChild(root);
    if(!searchTargets.isEmpty()){
        for (auto daqMod : searchTargets) {
            if(daqMod.isEmpty()) {
                static bool showWarn = true;
                if(showWarn) {
                    qWarning() << __PRETTY_FUNCTION__ << "searchTargets contains empty DaqModule";
                    showWarn = false;
                }
                continue;
            }
            QDomElement target = doc.createElement(TAG_SEARCH_TARGET);
            root.appendChild(target);
            QDomText text = doc.createTextNode(daqMod.toString());
            target.appendChild(text);
        }
    }
    return doc.toString();
}

QString PnpXml::encodeClose(const ProgramDescription &pd)
{
    QDomDocument doc(DOC_TYPE);
    QDomElement root = doc.createElement(TAG_PROGRAM_CLOSE);
    root.setAttribute(TAG_PROGRAM_TYPE, pd.getDaqModule().toString());
    root.setAttribute(TAG_PROGRAM_INDEX, pd.getProgramIndex().toString());
    if(!pd.getParent().getProgramType().isEmpty())
        root.setAttribute(TAG_PROGRAM_PARENT, pd.getParent().toString());
    root.setAttribute(TAG_PROGRAM_HOST_NAME, pd.hostName);
    root.setAttribute(TAG_PROGRAM_UUID, pd.uuid.toString());
    root.setAttribute(TAG_PROGRAM_SEQ, QString::number(pd.seq));
    root.setAttribute(TAG_VER_HASH, pd.ver_hash);
    root.setAttribute(TAG_VER_DATE, pd.ver_date);
    doc.appendChild(root);
    //    qDebug() << doc.toString();
    return doc.toString();
}

QString PnpXml::encodeDiscover(const ProgramDescription &pd)
{
    QDomDocument doc(DOC_TYPE);
    QDomElement root = doc.createElement(TAG_PROGRAM_ANNOUNCE);
    root.setAttribute(TAG_PROGRAM_TYPE, pd.getDaqModule().toString());
    root.setAttribute(TAG_PROGRAM_INDEX, pd.getProgramIndex().toString());
    if(!pd.getParent().getProgramType().isEmpty())
        root.setAttribute(TAG_PROGRAM_PARENT, pd.getParent().toString());
    root.setAttribute(TAG_PROGRAM_HOST_NAME, pd.hostName);
    if(!pd.host.isNull())
        root.setAttribute(TAG_PROGRAM_HOST, pd.host.toString());
    root.setAttribute(TAG_PROGRAM_UUID, pd.uuid.toString());
    root.setAttribute(TAG_PROGRAM_SEQ, QString::number(pd.seq));
    if(!pd.ver_hash.isEmpty())
        root.setAttribute(TAG_VER_HASH, pd.ver_hash);
    if(!pd.ver_date.isEmpty())
        root.setAttribute(TAG_VER_DATE, pd.ver_date);
    doc.appendChild(root);
    if(!pd.options.isEmpty()){
        QDomElement optionsSection = doc.createElement(TAG_OPTIONS_SECTION);
        root.appendChild(optionsSection);
        for (auto op : pd.options.keys()) {
            QDomElement optionSection = doc.createElement(TAG_OPTION_SECTION);
            optionsSection.appendChild(optionSection);
            optionSection.setAttribute(TAG_OPTION_NAME, op);
            optionSection.setAttribute(TAG_OPTION_VALUE, pd.options[op]);
        }
    }

    if(!pd.interfaces.isEmpty()){
        QDomElement interfacesSection = doc.createElement(TAG_INTERFACES_SECTION);
        root.appendChild(interfacesSection);
        for (const ProgramInterface &i : pd.interfaces) {
            QDomElement interfaceSection = doc.createElement(TAG_INTERFACE_SECTION);
            interfacesSection.appendChild(interfaceSection);
            if(!i.host.isNull() && i.host != pd.host)
                interfaceSection.setAttribute(TAG_INTERFACE_HOST, i.host.toString());
            interfaceSection.setAttribute(TAG_INTERFACE_PORT, i.port);
            interfaceSection.setAttribute(TAG_INTERFACE_ENABLED, i.enabled?1:0);
            interfaceSection.setAttribute(TAG_INTERFACE_ISFREE, i.isFree?1:0);
            interfaceSection.setAttribute(TAG_INTERFACE_ID, i.id);
            for (auto itPeer : i.peers) {
                QDomElement peerSection = doc.createElement(TAG_PEER_SECTION);
                peerSection.setAttribute(TAG_PEER_HOST, itPeer.host.toString());
                peerSection.setAttribute(TAG_PEER_PORT, itPeer.port);
                interfaceSection.appendChild(peerSection);
            }

            interfaceSection.setAttribute(TAG_INTERFACE_TYPE, i.getTypeName());
            if(i.type == ProgramInterfaceMStream) {
                interfaceSection.setAttribute(TAG_INTERFACE_DEV_INDEX, i.devIndex.toString());
                interfaceSection.setAttribute(TAG_INTERFACE_DEV_TYPE_STR, i.devTypeStr);
            }
        }
    }

    return doc.toString();
}
