#include <QHostAddress>

#include "ProgramDescriptionConverter.h"
#include "base/ConfigConverterUtil.h"
#include "base/DaqModuleConverter.h"
#include "base/ProgramIndexConverter.h"
#include "pnp-server/PNPMessage.h"
#include "pnp-server/PnpDiscovery.h"

namespace {
    constexpr const char *k_host = "host";
    constexpr const char *k_port = "port";
    constexpr const char *k_type = "type";
    constexpr const char *k_peers = "peers";
    constexpr const char *k_enabled = "enabled";
    constexpr const char *k_isFree = "isFree";
    constexpr const char *k_id = "id";
    constexpr const char *k_uuid = "uuid";
    constexpr const char *k_seq = "seq";
    constexpr const char *k_index = "index";
    constexpr const char *k_hostName = "hostName";
    constexpr const char *k_interfaces = "interfaces";
    constexpr const char *k_dummyInterface = "dummyInterface";
    constexpr const char *k_options = "options";
    constexpr const char *k_ver_hash = "ver_hash";
    constexpr const char *k_ver_date = "ver_date";
}

// --- ProgramInterfacePeer ---
template<>
bool AbstractConfigConverter<ProgramInterfacePeer>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<ProgramInterfacePeer>::toJsonObject(const ProgramInterfacePeer &in)
{
    QVariantMap m;
    m[k_host] = QVariant::fromValue(in.host.toString());
    m[k_port] = QVariant::fromValue(in.port);
    return fromVariantMap(m);
}

template<>
ProgramInterfacePeer AbstractConfigConverter<ProgramInterfacePeer>::fromJsonObject(const QJsonObject &m)
{
    ProgramInterfacePeer out;
    out.host =  QHostAddress(QVariant::fromValue(m[k_host]).value<QString>());
    out.port = QVariant::fromValue(m[k_port]).value<quint16>();
    return out;
}

template<>
QJsonValue AbstractConfigConverter<ProgramInterfacePeer>::toJsonValue(const ProgramInterfacePeer &in)
{
    return toJsonObject(in);
}

template<>
ProgramInterfacePeer AbstractConfigConverter<ProgramInterfacePeer>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<ProgramInterfacePeer>::registerSubTypes()
{}

// --- ProgramInterface ---
template<>
bool AbstractConfigConverter<ProgramInterface>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<ProgramInterface>::toJsonObject(const ProgramInterface &in)
{
    QVariantMap m;
    m[k_host] = QVariant::fromValue(in.host.toString());
    m[k_port] = QVariant::fromValue(in.port);
    m[k_type] = QVariant::fromValue(in.getTypeName());
    m[k_peers] = ConfigConverterUtil::toJsonValue(in.peers.values());
    m[k_enabled] = QVariant::fromValue(in.enabled);
    m[k_isFree] = QVariant::fromValue(in.isFree);
    m[k_id] = QVariant::fromValue(in.id);
    return fromVariantMap(m);
}

template<>
ProgramInterface AbstractConfigConverter<ProgramInterface>::fromJsonObject(const QJsonObject &m)
{
    ProgramInterface out;
    out.host =  QHostAddress(QVariant::fromValue(m[k_host]).value<QString>());
    out.port = QVariant::fromValue(m[k_port]).value<quint16>();
    auto typeStr = QVariant::fromValue(m[k_type]).value<QString>();
    out.setType(ProgramInterface::getTypeByName(typeStr));
    auto peers = ConfigConverterUtil::toList<ProgramInterfacePeer>(m[k_peers]);
    out.peers = QSet<ProgramInterfacePeer>(peers.cbegin(), peers.cend());
    out.enabled = QVariant::fromValue(m[k_enabled]).value<bool>();
    out.isFree = QVariant::fromValue(m[k_isFree]).value<bool>();
    out.id = QVariant::fromValue(m[k_id]).value<int>();
    return out;
}

template<>
QJsonValue AbstractConfigConverter<ProgramInterface>::toJsonValue(const ProgramInterface &in)
{
    return toJsonObject(in);
}

template<>
ProgramInterface AbstractConfigConverter<ProgramInterface>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<ProgramInterface>::registerSubTypes()
{
    AbstractConfigConverter<ProgramInterfacePeer>::registerTypes();
}

//----ProgramDescription ----
template<>
bool AbstractConfigConverter<ProgramDescription>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<ProgramDescription>::toJsonObject(const ProgramDescription &in)
{
    QVariantMap m;
    m[k_uuid] = QVariant::fromValue(in.uuid.toString());
    m[k_seq] = QVariant::fromValue(in.seq);
    m[k_type] = QVariant::fromValue(in.getDaqModule());
    m[k_index] = QVariant::fromValue(in.getProgramIndex());
    m[k_host] = QVariant::fromValue(in.host.toString());
    m[k_hostName] = QVariant::fromValue(in.hostName);
    m[k_interfaces] = ConfigConverterUtil::toJsonValue(in.interfaces);
    m[k_dummyInterface] = QVariant::fromValue(in.dummyInterface);
    m[k_options] = ConfigConverterUtil::toJsonValue(in.options);
    m[k_ver_hash] = QVariant::fromValue(in.ver_hash);
    m[k_ver_date] = QVariant::fromValue(in.ver_date);
    return fromVariantMap(m);
}

template<>
ProgramDescription AbstractConfigConverter<ProgramDescription>::fromJsonObject(const QJsonObject &m)
{
    ProgramDescription out;
    out.uuid =  QUuid(QVariant::fromValue(m[k_uuid]).value<QString>());
    out.seq = QVariant::fromValue(m[k_seq]).value<quint32>();
    out.setClientIndex({QVariant::fromValue(m[k_type]).value<DaqModule>(),
                        QVariant::fromValue(m[k_index]).value<ProgramIndex>()});
    out.host = QHostAddress(QVariant::fromValue(m[k_host]).value<QString>());
    out.hostName = QVariant::fromValue(m[k_hostName]).value<QString>();
    out.interfaces = ConfigConverterUtil::toVector<ProgramInterface>(m[k_interfaces]);
    out.dummyInterface = QVariant::fromValue(m[k_dummyInterface]).value<ProgramInterface>();
    out.options = ConfigConverterUtil::toMap<QString, QString>(m[k_options]);
    out.ver_hash = QVariant::fromValue(m[k_ver_hash]).value<QString>();
    out.ver_date = QVariant::fromValue(m[k_ver_date]).value<QString>();
    return out;
}


template<>
QJsonValue AbstractConfigConverter<ProgramDescription>::toJsonValue(const ProgramDescription &in)
{
    return toJsonObject(in);
}

template<>
ProgramDescription AbstractConfigConverter<ProgramDescription>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<ProgramDescription>::registerSubTypes()
{
    AbstractConfigConverter<ProgramInterfacePeer>::registerTypes();
    AbstractConfigConverter<ProgramInterface>::registerTypes();
    AbstractConfigConverter<DaqModule>::registerTypes();
    AbstractConfigConverter<ProgramIndex>::registerTypes();
}

//----PnpInfoCache ----
template<>
bool AbstractConfigConverter<PnpInfoCache>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<PnpInfoCache>::toJsonObject(const PnpInfoCache &in)
{
    return ConfigConverterUtil::toJsonValue(in).toObject();
}

template<>
PnpInfoCache AbstractConfigConverter<PnpInfoCache>::fromJsonObject(const QJsonObject &m)
{
    QMap<QUuid, ProgramDescription> res = ConfigConverterUtil::toMap<QUuid, ProgramDescription>(m);
    PnpInfoCache out ;
    auto it = res.cbegin();
    while (it != res.cend()) {
        out.insert(it.key(), it.value());
    }
    return out;
}

template<>
QJsonValue AbstractConfigConverter<PnpInfoCache>::toJsonValue(const PnpInfoCache &in)
{
    return toJsonObject(in);
}

template<>
PnpInfoCache AbstractConfigConverter<PnpInfoCache>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<PnpInfoCache>::registerSubTypes()
{
    AbstractConfigConverter<ProgramInterfacePeer>::registerTypes();
    AbstractConfigConverter<ProgramInterface>::registerTypes();
    AbstractConfigConverter<ProgramDescription>::registerTypes();
}
